
/*
 * FILE:   controller.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.66 $
 * $Date: 2007/01/08 22:22:22 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "version.h"
#include "environment.h"
#include "trace_ug.h"
#include "trace.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/rtp_callback.h"
#include "rtp/pbuf.h"
#include "rtp/sbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "pdb.h"
#include "tv.h"
#include "transmit.h"
#include "receive.h"
#include "controller.h"

#define RTCP_BW		(5*1024*1024)
#define UDP_BUFSIZE	(512*1024)	/* 500Kb of buffer, for the network device */

typedef enum {
	t_trans,
	t_recep,
} t_subsystem;

ctrlr_err_t	ctrlr_cleanup( t_subsystem ); 

/*
 * Controller state
 */
static struct ctrlr_state_t {
	struct timeval		 base_time;		/* Network base time */

	struct rtp		*network_device;	
	
	struct vidcap		*capture_device;

	struct display		*display_device;
	struct pdb		*participants;
	
	struct video_rx		*rx;
	struct video_tx		*tx;

	struct cc_rcvr_t	*rx_cc;		/* CC state for the receiver	*/
	struct cc_sndr_t	*tx_cc;		/* ...and for the sender	*/

	unsigned int		 encoder_id;	/* Encoder id for sent frames	*/
	unsigned int		 decoder_id;	/*... and for grabbed frames	*/
	
	
	unsigned int		 working;
} ctrlr;



/*******************************************************************************
 * Video capture & display initialization
 ******************************************************************************/
/* Initialize the video codecs */
int
ctrlr_init_video_codecs (struct run_env *env)
{
	unsigned int		 nc;
	unsigned int		 i;
	const char		*codec_name	= NULL;
	
	assert( env        != NULL );
	assert( env->codec != NULL );

	vcodec_init();
	nc = vcodec_get_num_codecs();

	/* Get the list, looking for known codecs... */ 
	printf ("Video codecs:\n");
	for (i = 0; i < nc; i++) {
		if (vcodec_can_encode(i)) {
			printf ("  Encoder : %s\n", vcodec_get_description(i));
		}
		if (vcodec_can_decode(i)) {
			printf ("  Decoder : %s\n", vcodec_get_description(i));
		}

		codec_name = vcodec_get_name (i);
		if (codec_name != NULL) {
			/* Map static and "well-known" dynamic payload types */
			if (strcmp (codec_name, "uv_yuv422") == 0) {
				vcodec_map_payload (30, i);	/* 30 : Uncompressed YUV 4:2:2 8 bit */
			}

			if (strcmp (codec_name, "dv") == 0) {
				vcodec_map_payload (60, i);	/* Unassigned PT for DV */
			}

			if (strcmp (codec_name, "jpeg") == 0) {
				vcodec_map_payload (61, i);	/* Unassigned PT for JPEG */
			}

			if (strcmp (codec_name, "mpeg1") == 0) {
				vcodec_map_payload (62, i);	/* Unassigned PT for MPEG1 */
			}
		}
	}
		
	/* Initilize the encoder that is going to be used */
	ctrlr.encoder_id = vcodec_get_by_name (env->codec);
	
	if (ctrlr.encoder_id == VCODEC_INVALID_ID) {
		fprintf (stderr, "Error: '%s' codec not found or invalid.\n", env->codec);
		return FALSE;
	}
	
	return TRUE;
}

/* Initialize the video display subsystem */ 
struct display*
ctrlr_init_video_display (struct run_env* env)
{
	unsigned int             i;
	display_type_t		*dt;
	display_format_t	*format	= NULL;
	struct display		*res	= NULL;
	
	assert (env		!= NULL);	
	assert (env->display	!= NULL);	
	
	ctrlr.display_device = NULL;
	
	/* We can force a specific format */
	display_init_devices (format, env->display_args);
	
	/* For every display available, check if it is the display we want... */
	for (i = 0; i < display_get_device_count(); i++) {
		dt = display_get_device_details (i);
		
		assert(dt->name != NULL);
		
		/* Found -> initialize it */
		if (strcmp (env->display, dt->name) == 0) {
			res = display_init (dt->id, dt->formats, env->display_args, NULL);
			break;
		}
	}
	display_free_devices();

	return (ctrlr.display_device = res);
}

/* Initialize the video capture subsystem */ 
struct vidcap *
ctrlr_init_video_capture (struct run_env *env)
{
	unsigned int		 i;
	struct vidcap_type	*vt;
	capture_format_t	 format;
	struct vidcap		*res	= NULL;
	
	assert (env		!= NULL);
	assert (env->capture	!= NULL);
	
	ctrlr.capture_device = NULL;

	vidcap_init_devices (env);
	for (i = 0; i < vidcap_get_device_count(); i++) {
		vt     = vidcap_get_device_details(i);

		assert (vt->name != NULL);

		/* FIXME: we use the first capture format as the default... */
		format.colour_mode	= vt->formats[0].colour_mode;
		format.encoding		= vt->formats[0].encoding;
		format.size		= vt->formats[0].size;
		format.fps		= env->fps;

		if (strcmp (vt->name, env->capture) == 0) {
			res = vidcap_init (vt->id, &format, env->capture_args);
			ctrlr.decoder_id = vcodec_get_by_enc (format.encoding);
			break;
		}
	}
	vidcap_free_devices();

	return (ctrlr.capture_device = res);
}


/*******************************************************************************
 * Network initialization
 ******************************************************************************/

/* Initialize the network (congestion control, RTP, RTCP...) */
struct rtp *
ctrlr_init_network (struct run_env * env)
{
	unsigned int	bytes_per_sec;
	cc_algo_t	cc_algo;
	struct rtp     *res		= NULL;
	double          rtcp_bw		= RTCP_BW;	/* FIXME */
	unsigned int	udpbufsize	= UDP_BUFSIZE;

	assert (env			!= 0);
	assert (env->fps		> 0);
	assert (ctrlr.encoder_id	!= VCODEC_INVALID_ID);
	
	/* Set a common base time for RTP and the congestion control system */
	gettimeofday (&ctrlr.base_time, NULL);

	cc_algo	= cc_get_by_name (env->cc_algo);
	debug_msg ("Using '%s' congestion control.\n", cc_get_name (cc_algo));

	/* Expected bitrate of the sender (bits/sec)*/
	bytes_per_sec = vcodec_encoder_get_framelen (ctrlr.encoder_id) * env->fps;
	assert (bytes_per_sec > 0);
	
	/* Initialize the congestion control system */
	ctrlr.tx_cc = cc_sndr_init (cc_algo, 0, bytes_per_sec, env->mtu, 0);
	ctrlr.rx_cc = cc_rcvr_init (cc_algo, 0, env->mtu, 0);

	ctrlr.participants = pdb_init();
	
	if( ctrlr.participants != NULL ) {
		res = rtp_init (env->addr, env->rx_port, env->tx_port, 255,
				rtcp_bw, udpbufsize,
				&ctrlr.base_time,
				rtp_recv_callback, (void*) ctrlr.participants,
				ctrlr.tx_cc, ctrlr.rx_cc);
		
		if (res != NULL) {
			pdb_add (ctrlr.participants, rtp_my_ssrc (res));
			rtp_set_option (res, RTP_OPT_WEAK_VALIDATION, TRUE);
			rtp_set_sdes (res, rtp_my_ssrc (res), RTCP_SDES_TOOL, ULTRAGRID_VERSION, strlen(ULTRAGRID_VERSION));
		}
	}
	
	return (ctrlr.network_device = res);
}

/* Initialize the transmitter */
struct video_tx *
ctrlr_init_transmit (struct run_env * env)
{
	struct tx_ctx	trans_ctx;
	
	assert (env != NULL);
	
	/* Create a transmission context and start it */
	trans_ctx.fps		= env->fps;
	trans_ctx.mtu		= env->mtu;
	trans_ctx.max_framelen	= vcodec_encoder_get_framelen (ctrlr.encoder_id);

	return (ctrlr.tx = tx_init (&trans_ctx,
				    ctrlr.network_device,
				    ctrlr.capture_device,
				    ctrlr.encoder_id, ctrlr.decoder_id,
				    &ctrlr.base_time,
				    ctrlr.tx_cc,
				    &ctrlr.working));
}

/* Initialize the receiver */
struct video_rx*
ctrlr_init_receive (struct run_env * env)
{
	struct rx_ctx	ctx;
	
	assert (env != NULL);
	assert (env->fps > 0);
	
	return (ctrlr.rx = rx_init (&ctx,
				    ctrlr.display_device,
				    ctrlr.participants,
				    ctrlr.network_device,
				    &ctrlr.base_time,
				    ctrlr.rx_cc,
				    &ctrlr.working));
}

/*******************************************************************************
 * Control of subsystems: start, stop...
 ******************************************************************************/
/* Initialize all subsystems */
ctrlr_err_t
ctrlr_init_all (struct run_env * env)
{
	assert (env != NULL);
	
	if (!ctrlr_init_video_codecs (env))	return ERR_CODEC;
	if (!ctrlr_init_video_display (env))	return ERR_DISPLAY;
	if (!ctrlr_init_video_capture (env))	return ERR_CAPTURE;	
	if (!ctrlr_init_network (env))		return ERR_NET;
	if (!ctrlr_init_receive (env))		return ERR_RECV;
	if (!ctrlr_init_transmit (env))		return ERR_TRANS;
	
	return ERR_OK;
}

/* Starts the transmission */
ctrlr_err_t
ctrlr_start (struct run_env* env)
{
	UNUSED (env);
	
	ctrlr.working	= TRUE;
	
	/* Trace some info */
	//trace_val (te_env_fps, env->fps);
	//trace_val (te_env_mtu, env->mtu);

	/* Start subsystems */
	rx_start (ctrlr.rx);			/* start reception		*/
	while (!rx_working (ctrlr.rx));		/*... wait for it		*/
	tx_start (ctrlr.tx);			/*... and start transmission	*/
	
	return ERR_OK;
}

/* Stop everything, waiting for threads and cleaning the state */
ctrlr_err_t
ctrlr_done (void)
{	
	struct timeval	 curr_time;
	struct pdb_e	*cp;
	uint32_t	 ssrc;
	
	assert (ctrlr.rx != NULL);
	assert (ctrlr.tx != NULL);
	
	/* Finish all the subsystems */
	debug_msg ("Stopping all subsystems...\n");
	
	ctrlr.working = FALSE;

	tx_done (ctrlr.tx);
	cc_sndr_done (ctrlr.tx_cc);	
	
	rx_done (ctrlr.rx);	
	cc_rcvr_done (ctrlr.rx_cc);

	rtp_done (ctrlr.network_device);
	vidcap_done (ctrlr.capture_device);
	display_done (ctrlr.display_device);
	vcodec_done ();
	
	/*
	 * Pause to allow playout time for any queued data to pass,
	 * then empty the playout buffers...
	 */ 
	sleep (1);
	gettimeofday (&curr_time, NULL);
	
	cp = pdb_iter_init (ctrlr.participants);
	while (cp != NULL) {
		pbuf_remove (cp->playout_buffer, curr_time);
		pbuf_done (cp->playout_buffer);
		ssrc = cp->ssrc;
		cp = pdb_iter_next (ctrlr.participants);
		pdb_remove (ctrlr.participants, ssrc, NULL);
	}
	
	pdb_iter_done (ctrlr.participants);
	pdb_destroy (&ctrlr.participants);
	
	return ERR_OK;
}

/* Returns TRUE if the system is working */
int
ctrlr_working (void)
{
	return ctrlr.working;
}
