/*
 * FILE:   environment.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.27 $
 * $Date: 2006/02/06 13:34:58 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "controller.h"
#include "version.h"

/* The following are needed by video_codec.h */
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_types.h"
#include "video_capture.h"

#include "video_codec.h"

#include "trace.h"
#include "trace_ug.h"

static void 
usage(void) 
{
	int	 i, nc;

        printf("Usage: uv [OPTION]... address[/rx_port[/tx_port]]...\n");
	printf("\n");
	printf("Options:\n");
	printf(" Display:\n");	
	printf("  -d <display_device>     Display device\n");
	printf("  -D <display_args>       Optional arguments\n");
	printf("\n");
	printf(" Capture:\n");
	printf("  -t <capture_device>     Capture device\n");
	printf("  -T <capture_args>       Optional arguments for the capture device\n");
	printf("  -c <codec_name>         Codec to use for transmission:\n");

	vcodec_init();
	nc = vcodec_get_num_codecs();
	for (i = 0; i < nc; i++) {
		if (vcodec_can_encode(i)) {
			printf("     %-20s    %s\n", vcodec_get_name(i), vcodec_get_description(i));
		}
	}
	vcodec_done();
	
	printf("\n");
	printf(" Transmission:\n");
	printf("  -m <mtu>                Path MTU of the network\n");
	printf("  -f <framerate>          Frames per second\n");
	printf("  -k <cc_algo>            Congestion Control algorithm\n");
	printf("\n");
	printf(" Other:\n");
	printf("  -g <filename>           Trace file name\n");
	printf("  -G <event>[,<event>...] Trace events list\n");	
	printf("                          (use 'list' as event name for a complete list)\n");	
	printf("  -j                      Trace file uses absolute times.\n");
}

/* Dump the events list */
static
void usage_events (void)
{
	unsigned int i;
	
#ifdef DEBUG
	printf ("%-25s %-25s\n", "[Event name]", "[Description]");
	for (i = 0; i < TRACE_EVENT_UG_NUM; i++) {
		printf ("%-25s %-25s\n", trace_names_ug [i], trace_descs_ug[i]);		
	}
	
	printf ("\n");
	printf ("Some of these events may be unavailable for tracing, "
		"depending on the current configuration. Refer to the UltraGrid "
		"documentation for details on the tracefile format.\n");
#endif
}

/*
 * Creation of a running environment, setting some defaults.
 */
struct run_env*
env_init (void)
{
	struct run_env* env;
	
	env = (struct run_env*)malloc(sizeof(struct run_env));
	if(env == NULL)
		fprintf(stderr, "Not enought memory for run_env allocation.\n");
	else {
		env->addr           = NULL;
		env->fps            = 60;
		env->rx_port        = 5004;
		env->tx_port        = 5004;
		env->mtu            = 1500;
		env->codec          = "uv_yuv422";
		env->capture        = "none";
		env->capture_args   = NULL;
		env->capture_height = 0;	/* Not used */
		env->capture_width  = 0;	/* Not used */
		env->display        = "none";
		env->display_args   = NULL;
		env->trace_file     = NULL;
		env->trace_events   = NULL;
		env->trace_t_abs    = FALSE;
		env->cc_algo        = NULL;
        }
	
        return env;
}

/*
 * Modifies the given running environment, setting the values 
 * specified with the command line options.
 * 
 * FIXME: every modification to argc/argv must be available when
 *        we return, as TCL initialization depends on that...
 */ 
int
env_load_cmdline(int* argc, char *argv[], struct run_env* env)
{
        int      ch;
        char    *token;
        int      port;       
        
        while ((ch = getopt(*argc, argv, "d:D:c:t:k:T:m:f:G:g:jv")) != -1) {
                switch (ch) {
                case 'd' :
                        env->display      = optarg;
                        break;
                case 'D' :
                        env->display_args = optarg;
                        break;
		case 'k' :
			env->cc_algo	  = optarg;
			break;
		case 'c' :
			env->codec	  = optarg;
			break;
                case 't' :
                        env->capture      = optarg;
                        break;
                case 'T' :
                        env->capture_args = optarg;
                        break;
                case 'm' :
                        env->mtu          = atoi(optarg);
                        break;
                case 'f' :
                        env->fps          = atoi(optarg);
                        break;
                case 'g' :
                        env->trace_file   = optarg;
                        break;
                case 'G' :
                        env->trace_events = optarg;
                        break;
		case 'j':
			env->trace_t_abs = TRUE;
			break;
                case 'v' :
                        printf("%s\n", ULTRAGRID_VERSION);
                        return ERR_OK;
                case '?' :
                        usage();
                        return ERR_USAGE;
                }
        }
	
        *argc -= optind;
        argv += optind;

	/* Check if we should dump the list of events */
	if (env->trace_events != NULL) {
		if (strcmp (env->trace_events, "list") == 0) {
			usage_events ();
			return ERR_USAGE;		
		}
	}
        
	if (*argc != 1) {
                usage();
                return ERR_USAGE;
        }
        
        env->addr = (char*)strtok(argv[0], "/");

        token = strtok(NULL, "/");
        if (token) {
                port         = atoi(token);
                port        &= ~1;
                env->rx_port = port;
        }

        token = strtok(NULL, "/");
        if (token) {
                port         = atoi(token);
                port        &= ~1;
                env->tx_port = port;
        }
        
        return ERR_OK;
}

void
env_done(struct run_env* env) {
	free( env );
}
