/*
 * FILE:   receive.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.17 $
 * $Date: 2006/05/12 15:08:20 $
 */



#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace_ug.h"
#include "trace.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"	/* For struct coded_data */
#include "pdb.h"
#include "tv.h"
#include "video_capture.h"
#include "video_codec.h"
#include "receive.h"


/* Receiver constants */
#define RECEIVE_MAGIC		0x8a0ab25f

#define RX_MAX_RTP_UPDATE	900000		/* Max time between rtp_update's */

struct video_rx {
	uint32_t		 magic;
	
	struct timeval		*base_time;	/* Pointer to the controller's start_time */

	struct display		*display_device;
	struct pdb		*participants;
	struct rtp		*network_device;

	struct cc_rcvr_t	*cc_rcvr;
	
	struct timeval		 recv_timeout;
	
	pthread_t		 rx_receive_thread,
				 rx_render_thread;	

	unsigned int		 rx_working;
	unsigned int		*all_working;
};

static void	rx_stop_all (struct video_rx*);

/*******************************************************************************
 * Threads
 ******************************************************************************/

static void*
rx_render_thread (void *arg)
{
	struct video_rx		*rx		= (struct video_rx*) arg;	
	struct pdb_e		*cp		= NULL;
	display_result_t	 display_result	= DISPLAY_OK;
	struct timeval		 curr_time;

	assert (rx != NULL);
	assert (rx->magic == RECEIVE_MAGIC);

	while (rx_working (rx)) {

		cp = pdb_iter_init (rx->participants);
		if (cp != NULL) {
			/* Decode and render for each participant in the conference... */
			while (cp != NULL && display_result == DISPLAY_OK) {
				/* Decode and render video... */
				gettimeofday (&curr_time, NULL);
				display_result = pbuf_decode (cp->playout_buffer,
							      curr_time,
							      rx->display_device);
				pbuf_remove (cp->playout_buffer, curr_time);
				cp = pdb_iter_next (rx->participants);
			}
			
			if (display_result == DISPLAY_EXIT_REQ) {
				debug_msg ("Exit requested in display...\n");
				break;
			}
			
			if (display_result == DISPLAY_ERR) {
				debug_msg ("Display error...\n");
				break;
			}			
		} 
		
		/* Sleep for some time... */
		tv_sleep_for (TV_MIN_SLEEP);			

		pdb_iter_done (rx->participants);
	}
	
	rx_stop_all (rx);

	return NULL;
}

/*
 * Update the reception timeout.
 * This function determines the amount of time we are going to wait for a data
 * packet. For some CC algorithms, this can be a long time, but for other like
 * TFRC, where feedbacks must be periodically generated, this must be equal to
 * the reports period...
 */
static uint32_t
rx_receive_update_timeout (struct timeval	*timeout,
			   uint32_t		 now_usecs,
			   struct cc_rcvr_t	*cc_rcvr)
{
	uint32_t	sendtime_usecs	= 0,
			timeout_usecs	= 0;
	
	assert (timeout != NULL);
	
	/* There is an upper limit to the timeout: the RTP update frequency */
	timeout_usecs = RX_MAX_RTP_UPDATE;
	
	/*
	 * This could do some fancy things like, for example, keeping track
	 * of the udp_select() error...
	 */
	if (cc_rcvr_get_algo (cc_rcvr) == tfrc_e) {
		
		/* If we are using TFRC, use the next sending time as the timeout. */
		sendtime_usecs = cc_rcvr_get_sendtime (cc_rcvr);
		
		if (sendtime_usecs > now_usecs) {
			/* Substract the current time */
			timeout_usecs = sendtime_usecs - now_usecs;
		} else {
			timeout_usecs = 0;
		}
	}

	usec2tv (timeout, MIN (timeout_usecs, RX_MAX_RTP_UPDATE));

	return timeout_usecs;
}

/*
 * Reception thread.
 * Receives and enqueues packets from the network, and launches the rendering
 * of completed frames.
 */
static void*
rx_receive_thread (void *arg)
{
	struct video_rx		*rx		= (struct video_rx*) arg;	
	struct timeval		 tv_now,
				 timeout;
	uint32_t		 usecs_now	= 0,
				 usecs_now_mt	= 0,
				 usecs_prev	= 0,
				 timeout_usecs	= 0;
	
	assert (rx->magic          == RECEIVE_MAGIC);
	assert (rx->participants   != NULL );
	assert (rx->network_device != NULL );
	
	debug_msg ("Started receiver thread\n");

	while (rx_working (rx)) {

		usecs_now	= tv_usecs_since (rx->base_time, NULL);
		usecs_now_mt	= usecs_now * 90000;
		
		/* Update the timeout allowed */
		timeout_usecs = rx_receive_update_timeout (&timeout, usecs_now, rx->cc_rcvr);

		/* Receive packets from the network... */		
		rtp_recv (rx->network_device, &timeout, usecs_now_mt);

		/*
		 * Housekeeping and RTCP...
		 * Update the receiver status and send any control packet (if needed).
		 */
		usecs_now	= tv_usecs_since (rx->base_time, &tv_now);
		usecs_now_mt	= usecs_now * 90000;

		rtp_update    (rx->network_device, tv_now);
		rtp_send_ctrl (rx->network_device, usecs_now_mt, NULL, tv_now);
		
		/* Trace some events */
		trace_val_at (&tv_now, te_rx_sleep, usecs_now - usecs_prev);
		
		usecs_prev	= usecs_now;
	}

	debug_msg("Stopped receiver thread\n");
	
	rx_stop_all (rx);
	
	return NULL;
}

/*******************************************************************************
 * Public API
 ******************************************************************************/

/* Start the reception system */
struct video_rx*
rx_init (struct rx_ctx		*ctx,
	 struct display		*display_device,
	 struct pdb		*participants,
	 struct rtp		*network_device,
	 struct timeval		*base_time,
	 struct cc_rcvr_t	*cc_rcvr,
	 unsigned int		*working)
{
	struct video_rx	*res;

	UNUSED (ctx);

	res = (struct video_rx*) malloc (sizeof(struct video_rx));
	if (res != NULL) {
		res->magic		= RECEIVE_MAGIC;

		res->display_device	= display_device;
		res->participants	= participants;
		res->network_device	= network_device; 
		res->cc_rcvr		= cc_rcvr;

		res->rx_working		= FALSE;
		res->all_working	= working;

		res->base_time		= base_time;

	} else {
		fprintf (stderr, "Error; allocating memory for 'struct video_rx'\n");
	}

	return res;
}

/* Start threads... */
int
rx_start (struct video_rx *rx)
{
	assert (rx != NULL);
	assert (rx->magic == RECEIVE_MAGIC);

	rx->rx_working = TRUE;

	/* Start the transmit thread */
	if (pthread_create (&rx->rx_render_thread, NULL, rx_render_thread, rx) != 0) {
		fprintf (stderr, "Error: creating the rx control thread.\n");			
		return 1;
	}
	
	/* Start the grab & enqueue thread */
	if (pthread_create (&rx->rx_receive_thread, NULL, rx_receive_thread, rx) != 0) {
		fprintf (stderr, "Error: creating the rx receive thread.\n");			
		return 1;
	}

	return 0;
}

void
rx_stop (struct video_rx *rx)
{
	assert (rx != NULL);

	rx->rx_working = FALSE;
}

static void
rx_stop_all (struct video_rx *rx)
{
	assert (rx != NULL);

	rx_stop (rx);
	*(rx->all_working) = FALSE;
}


/* Stop everything */
void
rx_done	(struct video_rx *rx)
{
	assert (rx != NULL);
	assert (rx->magic == RECEIVE_MAGIC);
	
	rx_stop_all (rx);
	
	/* Waits for the transmit thread */	
	if (pthread_join (rx->rx_receive_thread, NULL) != 0) {
		fprintf (stderr, "Error: joining the rx receive thread.\n");
	}
	
	if (pthread_join (rx->rx_render_thread, NULL) != 0) {
		fprintf (stderr, "Error: joining the rx control thread.\n");			
	}
		
	free (rx);
}

unsigned int
rx_working (struct video_rx *rx)
{
	assert (rx != NULL);
	
	return (*(rx->all_working) && rx->rx_working);
}

