/*
 * FILE:   rtp/cc.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.16 $
 * $Date: 2007/01/08 22:22:22 $
 *
 */


#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "version.h"

#ifdef HAVE_TFRC	
#include "tfrc_rcvr.h"
#include "tfrc_sndr.h"
#endif

#include "rtp/cc.h"
#include "tv.h"

/* Switches */
#define USE_CC_LOCK

#define CC_NONE_Q	0.9	/* RTT weighting factor for the 'null' cc system */

#ifndef TFRC_NULL_TIME 
#define TFRC_NULL_TIME CC_NULL_TIME
#endif

/* Utility macros */
#ifdef USE_CC_LOCK
#define cc_lock_init(x)		pthread_mutex_init(&(x->lock), NULL)
#define cc_lock(x)		pthread_mutex_lock(&(x->lock))
#define cc_unlock(x)		pthread_mutex_unlock(&(x->lock))
#else
#define cc_lock_init(x)
#define cc_lock(x)
#define cc_unlock(x)
#endif

struct cc_descr_t {
	char *name;
};

/* Names of the CC algorithms (in the same order as cc_algo_t) */
static struct cc_descr_t cc_algos[] = {
	{ "none" },
	{ "tfrc" },
};

#define NUM_CC_ALGOS	(sizeof (cc_algos) / sizeof(struct cc_descr_t))


/*******************************************************************************
 * Sender
 ******************************************************************************/

/* CC_none: CC state when there is no CC */
struct cc_none_sndr_t {
	uint32_t	last_send_time;
	uint16_t	s;		/* Packet size (in bytes)		*/
	uint32_t	ipi;		/* Inter-packet interval (in usecs)	*/
	uint32_t	bytes_per_sec;	/* Send rate (bytes/sec)		*/
	uint32_t	rtt;		/* RTT					*/
};

struct cc_sndr_t {
	cc_algo_t		 alg;
	void			*state;
	pthread_mutex_t		 lock;
};

struct cc_sndr_t*
cc_sndr_init (cc_algo_t	alg, 
	      uint32_t	now,
	      uint32_t	bytes_per_sec,
	      uint16_t	s,
	      uint32_t	rtt)
{
	struct cc_sndr_t	*ret;
	
	ret = (struct cc_sndr_t*) malloc (sizeof(struct cc_sndr_t));
	if (ret != NULL) {
		ret->alg = alg;
		
		cc_lock_init (ret);
		
		if (alg == none_e) {
			ret->state = (struct cc_none_sndr_t*) malloc (sizeof (struct cc_none_sndr_t));
			if (ret->state != NULL) {
				double			 pcks_per_sec;
				struct cc_none_sndr_t	*state = ret->state;
				
				state->bytes_per_sec	= bytes_per_sec;
				state->s		= s;
				state->last_send_time	= now;
				state->rtt		= CC_NULL_RTT;
				
				pcks_per_sec	= ((double)bytes_per_sec) / ((double)s);
				state->ipi	= 1000000.0/pcks_per_sec;
				
				assert (state->ipi > 0);
				
				debug_msg ("IPI: %d usecs\n", state->ipi);
				
			} else {
				fprintf (stderr, "Error: allocating memory for cc_none_sndr_t.\n");
				free (ret);
				ret = NULL;
			}
#ifdef HAVE_TFRC	
		} else if (alg == tfrc_e) {

			assert (TFRC_NULL_TIME == CC_NULL_TIME);
			
			ret->state = (void*) tfrc_sndr_init (now, s, rtt);
#else
			UNUSED (rtt);
#endif			
		} else {
			fprintf (stderr, "Invalid CC algorithm.\n");
			free (ret);
			ret = NULL;
		}
	}
	
	return ret;
}

void
cc_sndr_done (struct cc_sndr_t *cc)
{
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		struct cc_none_sndr_t	*state = (struct cc_none_sndr_t*)(cc->state);
		
		if (state != NULL) {
			cc_lock (cc);
			free (state);
			cc_unlock (cc);
		}

		free (cc);

#ifdef HAVE_TFRC			
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr	*state = (struct tfrc_sndr*)(cc->state);
		
		if (state != NULL) {
			cc_lock (cc);
			tfrc_sndr_done ((struct tfrc_sndr*)(cc->state));
			cc_unlock (cc);
		}
		
		free (cc);
#endif		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
}

/* Update the internal state */
uint32_t
cc_sndr_update (struct cc_sndr_t *cc, uint32_t now)
{
	uint32_t next_sendtime	= 0;

	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Nothing to do. Just return next sending time */
		struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		next_sendtime = state->last_send_time + state->ipi;
		cc_unlock (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		next_sendtime = tfrc_sndr_update (state, now);
		cc_unlock (cc);
#else
		UNUSED (cc);
		UNUSED (now);
#endif	
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return next_sendtime;
}

/* Process information coming from the receiver */
uint32_t
cc_sndr_process	(struct cc_sndr_t *cc, uint32_t now, cc_rcvr_info_t *info)
{
	uint32_t next_sendtime	= 0;

	assert (cc != NULL);	
	
	if (cc->alg == none_e) {
		struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		state->rtt	= CC_NONE_Q*state->rtt + (1-CC_NONE_Q)*info->rtt;
		next_sendtime	= state->last_send_time + state->ipi;
		cc_unlock (cc);
		
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		next_sendtime = tfrc_sndr_process_feedback(state, now,
							   info->tdelay, info->x_recv,
							   info->ts, info->p);
		cc_unlock (cc);
#else
		UNUSED (info);
		UNUSED (now);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}

	return next_sendtime;
}
 
/* Prepare the info and format a 'cc_sndr_info_t' */
int
cc_sndr_format (struct cc_sndr_t *cc, uint32_t now, uint16_t seq, cc_sndr_info_t *info)
{
	assert (cc   != NULL);
	assert (info != NULL);

	if (cc->alg == none_e) {
		/* Nothing to do: no CC info transmitted */
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		/* Format the data needed by the CC system and attached in a PCK. */
		cc_lock (cc);
		info->rtt	= tfrc_sndr_get_avg_rtt (state);
		info->send_ts	= now;
		info->seq	= seq;
		cc_unlock (cc);
#else
		UNUSED (now);
		UNUSED (seq);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
		return FALSE;
	}
	
	return TRUE;
}

/* Return TRUE if a data packet can be sent */ 
int
cc_sndr_send_ok (struct cc_sndr_t *cc, uint32_t now)
{
	unsigned int	send_ok		= FALSE;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		send_ok=TRUE;

		/* this stuff is not working with HD - more experimentation needed - LG */
		//struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state); 
		//assert (state != NULL);
		
		//cc_lock (cc);
		//send_ok = (now >= (state->last_send_time + state->ipi));
		//cc_unlock (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state =(struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		send_ok = tfrc_sndr_send_ok (state, now);
		cc_unlock (cc);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return send_ok;
}
 
/* Register the delivery of a data packet */ 
uint32_t
cc_sndr_send_done (struct cc_sndr_t *cc, uint32_t now)
{
	uint32_t	next_sendtime	= CC_NULL_TIME;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		state->last_send_time	= now;
		next_sendtime		= state->last_send_time + state->ipi;
		cc_unlock (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		next_sendtime = tfrc_sndr_send_done (state, now);
		cc_unlock (cc);
#else
		UNUSED (cc);
		UNUSED (now);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return next_sendtime;
}
 
/* Get the CC algorithm used */
cc_algo_t
cc_sndr_get_algo (struct cc_sndr_t *cc)
{
	cc_algo_t	ret	= none_e;
	
	if (cc != NULL) {
		ret = cc->alg;
	}
	
	return ret;
}

/* Return the current RTT (usecs) */ 
uint32_t
cc_sndr_get_avg_rtt (struct cc_sndr_t *cc)
{
	uint32_t	rtt	= CC_NULL_RTT;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		struct cc_none_sndr_t* state = (struct cc_none_sndr_t*)(cc->state);
		
		cc_lock (cc);
		rtt = state->rtt;
		cc_unlock (cc);
		
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		rtt = tfrc_sndr_get_avg_rtt (state);
		cc_unlock (cc);
#else
		UNUSED (cc);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return rtt;
}

/* Return the current sending rate (bytes/sec) */ 
double
cc_sndr_get_sendrate (struct cc_sndr_t *cc, uint32_t now)
{
	double	sendrate = 0;
	
	assert (cc != NULL);

	if (cc->alg == none_e) {
		struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state);

		cc_lock (cc);
		sendrate = state->bytes_per_sec;	/* Convert to bytes/sec */
		cc_unlock (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		sendrate = tfrc_sndr_get_sendrate (state, now);
		cc_unlock (cc);
#else
		UNUSED (now);
#endif	
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return sendrate;
}

/* Return the next sending time (usecs since start_time) */ 
uint32_t
cc_sndr_get_sendtime (struct cc_sndr_t *cc)
{
	uint32_t	sendtime	= 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		state = cc->state;
		sendtime = (state->last_send_time + state->ipi);
		cc_unlock (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		sendtime = tfrc_sndr_get_sendtime (state);
		cc_unlock (cc);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return sendtime;
}

/* Return the current IPI */ 
uint32_t
cc_sndr_get_ipi	(struct cc_sndr_t *cc)
{
	uint32_t	ipi	= 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		struct cc_none_sndr_t *state = (struct cc_none_sndr_t*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		state = cc->state;
		ipi = state->ipi;
		cc_unlock (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		ipi = tfrc_sndr_get_ipi (state);
		cc_unlock (cc);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return ipi;
}

/* Return the current loss event rate */ 
double
cc_sndr_get_p	(struct cc_sndr_t *cc)
{
	double	p = 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Nothing to do */
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_sndr *state = (struct tfrc_sndr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		p = tfrc_sndr_get_p (state);
		cc_unlock (cc);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return p;
}


/*******************************************************************************
 * Receiver
 ******************************************************************************/

struct cc_rcvr_t {
	cc_algo_t	 alg;
	void		*state;
	pthread_mutex_t	 lock;
};

struct cc_none_rcvr_t {
	uint32_t	rtt;
};

/* Initialize the receiver */
struct cc_rcvr_t*
cc_rcvr_init (cc_algo_t alg, uint32_t now, uint16_t s, uint32_t rtt)
{
	struct cc_rcvr_t	*ret;
	
	ret = (struct cc_rcvr_t*) malloc (sizeof (struct cc_rcvr_t));
	if (ret != NULL) {
		ret->alg = alg;
		
		cc_lock_init (ret);
		
		if (ret->alg == none_e) {
			ret->state = (struct cc_none_rcvr_t*) malloc (sizeof (struct cc_none_rcvr_t));
			if (ret->state != NULL) {
				struct cc_none_rcvr_t	*state = ret->state;				
				state->rtt		= CC_NULL_RTT;
			} else {
				fprintf (stderr, "Error: allocating memory for cc_none_rcvr_t.\n");
				free (ret);
				ret = NULL;
			}
			
#ifdef HAVE_TFRC	
		} else if (ret->alg == tfrc_e) {
			
			assert (TFRC_NULL_TIME == CC_NULL_TIME);

			ret->state = (void*) tfrc_rcvr_init (now, s, rtt);
#else
			UNUSED (now);
			UNUSED (s);
			UNUSED (rtt);
#endif
		} else {
			fprintf (stderr, "Invalid CC algorithm.\n");
			free (ret);
			ret = NULL;
		}
	}
	
	return ret;
}

/* Free the receiver memory and state */
void
cc_rcvr_done (struct cc_rcvr_t *cc)
{
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		free (cc);

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		tfrc_rcvr_done (state);
		cc_lock (cc);

		free (cc);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
}

/* Update the state of the receiver */
uint32_t
cc_rcvr_update (struct cc_rcvr_t *cc, uint32_t now)
{
	uint32_t	next_sendtime	= 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);
		
		UNUSED (state);
#endif
		UNUSED (now);
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return next_sendtime;
}

/* Process the arrival of some data */
void
cc_rcvr_process	(struct cc_rcvr_t *cc, uint32_t now, cc_sndr_info_t *info)
{	
	assert (cc != NULL);
	
	
	
	if (cc->alg == none_e) {
		struct cc_none_rcvr_t *state = (struct cc_none_rcvr_t*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		/* RTT computation for the null CC. We need it for other parts of the system... */
		state->rtt	= CC_NONE_Q*state->rtt + (1-CC_NONE_Q)*info->rtt;
		cc_unlock (cc);	
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);

		if (info->rtt > 0) {
			tfrc_rcvr_process_rtt_sample (state, now, info->rtt);
		}
		
		tfrc_rcvr_process_arrival (state, now, info->seq, info->send_ts);
		cc_unlock (cc);
#else
		UNUSED (now);
		UNUSED (info);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
}

/* Prepare and get a 'cc_rcvr_info_t' structure for seending */
int
cc_rcvr_format	(struct cc_rcvr_t *cc, uint32_t now, cc_rcvr_info_t *info)
{
	unsigned int	 res	= FALSE;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */
		
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		uint32_t	 t_delay,
				 x_recv,
				 send_ts,
				 p;
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		
		assert (state != NULL);

		cc_lock (cc);
		tfrc_rcvr_format_feedback (state, now, &t_delay, &x_recv, &send_ts, &p);
		cc_unlock (cc);
		
		info->ts	= send_ts;
		info->tdelay	= t_delay;
		info->x_recv	= x_recv;
		info->p		= p;
		
		res = TRUE;
#else
		UNUSED (now);
		UNUSED (info);
#endif
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return res;
}

/* Return TRUE if the receiver can send */
int
cc_rcvr_send_ok (struct cc_rcvr_t *cc, uint32_t now)
{
	unsigned int	res	= FALSE;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		res = tfrc_rcvr_send_ok (state, now);
		cc_unlock (cc);
#else
		UNUSED (now);
#endif
		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}

	return res;
}

/* Register the delivery of receiver information */
uint32_t
cc_rcvr_send_done (struct cc_rcvr_t *cc, uint32_t now )
{
	uint32_t	next_sendtime	= 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);
		
		cc_lock (cc);
		next_sendtime = tfrc_rcvr_send_done (state, now);
		cc_unlock (cc);
#else
		UNUSED (now);
#endif
		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return next_sendtime;
}
 
 
/* Get the CC algorithm used */
cc_algo_t
cc_rcvr_get_algo (struct cc_rcvr_t *cc)
{
	cc_algo_t	ret	= none_e;
	
	if (cc != NULL) {
		ret = cc->alg;
	}
	
	return ret;
}

/* Get the current RTT */
uint32_t
cc_rcvr_get_avg_rtt (struct cc_rcvr_t *cc)
{
	uint32_t	rtt	= CC_NULL_RTT;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		struct cc_none_rcvr_t* state = (struct cc_none_rcvr_t*)(cc->state);
		
		cc_lock (cc);
		rtt = state->rtt;
		cc_unlock (cc);
		
#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		rtt = tfrc_rcvr_get_avg_rtt (state);
		cc_unlock (cc);
#endif
		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}

	return rtt;
}

/* Get the current loss event rate */
double
cc_rcvr_get_p (struct cc_rcvr_t *cc)
{
	double p	= 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		p = tfrc_rcvr_get_p (state);
		cc_unlock (cc);
#endif
		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}

	return p;
}

/* Get the next send time */
uint32_t
cc_rcvr_get_sendtime (struct cc_rcvr_t *cc)
{
	uint32_t	sendtime	= 0;

	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		sendtime = tfrc_rcvr_get_sendtime (state);
		cc_unlock (cc);
#endif
		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}
	
	return sendtime;
}

/* Get the current send rate known by the receiver */
double
cc_rcvr_get_sendrate (struct cc_rcvr_t *cc)
{
	double	sendrate	= 0;
	
	assert (cc != NULL);
	
	if (cc->alg == none_e) {
		/* Do nothing */

#ifdef HAVE_TFRC	
	} else if (cc->alg == tfrc_e) {
		struct tfrc_rcvr *state = (struct tfrc_rcvr*)(cc->state);
		assert (state != NULL);

		cc_lock (cc);
		sendrate = tfrc_rcvr_get_x_recv (state);
		cc_unlock (cc);
#endif
		
	} else {
		fprintf (stderr, "Invalid CC algorithm.\n");
	}

	return sendrate;
}


/*******************************************************************************
 * Utilities
 ******************************************************************************/
/* Get the printable name of the algorithm provided */
char*
cc_get_name	(cc_algo_t algo)
{
	unsigned int	algo_num = (unsigned int) algo;
	
	if (algo_num < NUM_CC_ALGOS)
		return cc_algos[algo_num].name;
	else
		return NULL;	
}

/* Retrieve an  algorithm by name */
cc_algo_t
cc_get_by_name (const char* name)
{
	unsigned int i	= 0;
	
	if (name != NULL) {
		while (i < NUM_CC_ALGOS) {
			if (strcmp (cc_algos[i].name, name) == 0) {
				return (cc_algo_t)i;
			}
			i++;
		}
	}
	
	return (cc_algo_t)0;
}

