/*
 * FILE:   rtp/cc.h
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2005/12/14 17:15:40 $
 *
 */

#ifndef __CC_H
#define __CC_H

/*
 * Congestion control algorithms.
 * So far, the only CC algorithm are the null and the TFRC congestion controls.
 */
typedef enum {
	none_e,
	tfrc_e
} cc_algo_t;

/* Utilities */
char		*cc_get_name	(cc_algo_t algo);
cc_algo_t	 cc_get_by_name (const char* name);

/* Some constants */
#ifdef HAVE_TFRC	
#define TFRC_NUM_INTERVALS	8		/* Number of intervals */
#define TFRC_MAX_DROPOUT	1000
#endif

#define CC_NULL_TIME		0
#define CC_NULL_RTT		CC_NULL_TIME


/*******************************************************************************
 * Interchange of information
 ******************************************************************************/

/* Info provided by the receiver to the sender */
typedef struct {
	uint32_t	ts;
	uint32_t	tdelay;
	uint32_t	x_recv;
	uint32_t	p;		/* Loss event rate */
	uint32_t	rtt;		/* RTT, if available */
} cc_rcvr_info_t;

/* Info provided by the sender to the receiver */
typedef struct {
	uint16_t	seq;
	uint32_t	send_ts;
	uint32_t	rtt;
} cc_sndr_info_t;

/*******************************************************************************
 * Sender
 ******************************************************************************/

struct cc_sndr_t;

/* Initialize the CC sender */
struct cc_sndr_t *cc_sndr_init			(cc_algo_t	alg,
						 uint32_t	now,
						 uint32_t	bytes_per_sec,
						 uint16_t	s,
						 uint32_t	rtt); /* 0 = unknown */

/* Finalize the CC sender */
void		 cc_sndr_done			(struct cc_sndr_t	*cc);
		
/* Update the sender */
uint32_t	 cc_sndr_update			(struct cc_sndr_t	*cc, 
						 uint32_t		 now);

/* Process a feedback */
uint32_t	 cc_sndr_process		(struct cc_sndr_t	*cc, 
						 uint32_t		 now,
						 cc_rcvr_info_t		*info);

/* Prepare and return the CC info */
int		 cc_sndr_format			(struct cc_sndr_t	*cc,
						 uint32_t		 now,
						 uint16_t		 seq,
						 cc_sndr_info_t		*info);

/* Check if a data packet can be sent */
int		 cc_sndr_send_ok		(struct cc_sndr_t	*cc,
						 uint32_t		 now);
				
/* Register the delivery of a data packet */
uint32_t	 cc_sndr_send_done		(struct cc_sndr_t	*cc, 
						 uint32_t		 now);

/* State info */
cc_algo_t	 cc_sndr_get_algo		(struct cc_sndr_t	*cc);
uint32_t	 cc_sndr_get_avg_rtt		(struct cc_sndr_t	*cc);
double		 cc_sndr_get_sendrate		(struct cc_sndr_t	*cc,
						uint32_t		 now);
uint32_t	 cc_sndr_get_sendtime		(struct cc_sndr_t	*cc);
uint32_t	 cc_sndr_get_ipi		(struct cc_sndr_t	*cc);
double		 cc_sndr_get_p			(struct cc_sndr_t	*cc);

						 
/*******************************************************************************
 * Receiver
 ******************************************************************************/

struct cc_rcvr_t;
						 
/* Initialize the CC receiver */
struct cc_rcvr_t *cc_rcvr_init			(cc_algo_t		 alg,
						 uint32_t		 now, 
						 uint16_t		 s,
						 uint32_t		 rtt);

/* Finalize the CC receiver */
void             cc_rcvr_done			(struct cc_rcvr_t	*cc);

/* Update the receiver */
uint32_t	 cc_rcvr_update			(struct cc_rcvr_t	*cc,
						 uint32_t		 now);

/* Process a data arrival */
void		 cc_rcvr_process		(struct cc_rcvr_t	*cc, 
						 uint32_t		 now,
						 cc_sndr_info_t		*info);

/* Prepare and format the feedback info */						 
int		 cc_rcvr_format			(struct cc_rcvr_t	*cc, 
						 uint32_t		 now,
						 cc_rcvr_info_t		*info);

/* Check if a feedback can be sent */
int		 cc_rcvr_send_ok		(struct cc_rcvr_t	*cc,
						 uint32_t		 now );
						 
/* Register a feedback as sent */						 
uint32_t         cc_rcvr_send_done		(struct cc_rcvr_t	*cc,
						 uint32_t		 now);

/* State info */
cc_algo_t	 cc_rcvr_get_algo		(struct cc_rcvr_t	*cc);
uint32_t	 cc_rcvr_get_avg_rtt		(struct cc_rcvr_t	*cc);
double		 cc_rcvr_get_p			(struct cc_rcvr_t	*cc);
uint32_t	 cc_rcvr_get_sendtime		(struct cc_rcvr_t	*cc);
double		 cc_rcvr_get_sendrate		(struct cc_rcvr_t	*cc);
		

#endif

