/*
 * FILE:    rtp/pbuf.c
 * AUTHORS: N.Cihan Tas   <ctas@cs.umd.edu>
 *          Ladan Gharai  <ladan@isi.edu>
 *          Colin Perkins <csp@csperkins.org>
 * 
 * This file implements a linked list for the playout buffer.
 *
 * Copyright (c) 2003-2005 University of Southern California
 * Copyright (c) 2003-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.108 $
 * $Date: 2006/05/12 15:13:05 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "tv.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "video_convert.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_codec.h"
#include "trace.h"
#include "trace_ug.h"

struct pbuf_node {
        struct pbuf_node  *next;
        struct pbuf_node  *prev;
        uint32_t           rtp_timestamp;	/* RTP timestamp for the frame           */
        struct timeval     arrival_time;	/* Arrival time of first packet in frame */
        struct timeval     playout_time;	/* Playout time for the frame            */
        struct timeval     discard_time;	/* Discard time for the frame            */
        struct coded_data *cdata;		/*					 */
	unsigned           cdata_count;
	uint16_t           seqno_min;
	uint16_t           seqno_max;
	int		   decoded;		/* Non-zero if we've decoded this frame  */
	int     	   mbit;		/* Set if marker bit of frame seen       */
	uint8_t            pt;			/* RTP payload type for the frame        */
};

struct pbuf {
	struct pbuf_node	*frst;
	struct pbuf_node	*last;
	unsigned		 size;
	void			*decoder_state;
	unsigned		 decoder_id;
	struct timeval		 last_playout_time;
};

/*********************************************************************************/

int
seq_lt(uint16_t s1, uint16_t s2)
{
	/* Return 1 if s1 < s2 modulo 65535; 0 otherwise     */
	/* Typically used for comparing RTP sequence numbers */
	/* and assumes s1 and s2 are within half the space.  */
	uint16_t   diff = s2 - s1;

	if (diff < 0x8000u) {
		return 1;
	}
	return 0;
}

static void
pbuf_validate(struct pbuf *playout_buf)
{
	/* Run through the entire playout buffer, checking pointers, etc.  */
	/* Only used in debugging mode, since it's a lot of overhead [csp] */
#ifdef DEBUG_PBUF
	struct pbuf_node	*cpb, *ppb;
	struct coded_data	*ccd, *pcd;
	unsigned		 cdc,  pbc;

	cpb = playout_buf->frst;
	ppb = NULL;
	pbc = 0;
	while (cpb != NULL) {
		assert(cpb->prev == ppb);

		if (cpb->prev != NULL) {
			assert(cpb->prev->next == cpb);
			assert(cpb->rtp_timestamp > ppb->rtp_timestamp);     /* Nodes stored in RTP timestamp order */
			/* FIXME: previous assert will fail when timestamp wraps */
			assert(tv_gt(cpb->playout_time, ppb->playout_time)); /* Nodes stored in playout time order  */
		}

		if (cpb->next != NULL) {
			assert(cpb->next->prev == cpb);
		} else {
			assert(cpb = playout_buf->last);
		}

		if (cpb->cdata != NULL) {
			/* We have coded data... check all the pointers on that list too... */
			cdc = 0;
			ccd = cpb->cdata;
			pcd = NULL;
			while (ccd != NULL) {
				cdc++;
				assert(ccd->data != NULL);
				assert(ccd->prev == pcd);
				if (ccd->prev != NULL) {
					assert(ccd->prev->next == ccd);
					assert(seq_lt(ccd->seqno, pcd->seqno)); /* Coded data stored in descending order */
				}
				if (ccd->next != NULL) {
					assert(ccd->next->prev == ccd);
				}
				pcd = ccd;
				ccd = ccd->next;
			}
			assert(cpb->cdata_count == cdc);
		}

		/* Advance to next node in the playout buffer... */
		ppb = cpb;
		cpb = cpb->next;
		pbc++;
	}
	assert(pbc == playout_buf->size);
#else
	UNUSED(playout_buf);
#endif
}

struct pbuf *
pbuf_init(void)
{  
	struct pbuf	*playout_buf = NULL;

	playout_buf = malloc(sizeof(struct pbuf));
	if (playout_buf != NULL) {
		playout_buf->frst=NULL;
		playout_buf->last=NULL;
	} else {
		debug_msg("Failed to allocate memory for playout buffer\n");
	}
	playout_buf->size          = 0;
	playout_buf->decoder_state = NULL;
	playout_buf->decoder_id    = VCODEC_INVALID_ID;
	
	gettimeofday(&(playout_buf->last_playout_time), NULL);

	return playout_buf;
}

void
pbuf_done(struct pbuf *playout_buf)
{
	if (playout_buf->size != 0) {
		printf("ERROR: Destroying non-empty playout buffer\n");
		abort();
	}
	free(playout_buf);
}

static void
add_coded_unit(struct pbuf_node *node, rtp_packet *pkt) 
{
	/* Add "pkt" to the frame represented by "node". The "node" has    */
	/* previously been created, and has some coded data already...     */
	/* Packets are stored in descending order of RTP sequence number.  */

	struct coded_data *new_cdata, *curr; 

	assert(node->rtp_timestamp == pkt->fields.ts);
	assert(node->cdata != NULL);
	
	new_cdata = malloc(sizeof(struct coded_data));
	if (new_cdata == NULL) {
		/* Out of memory, drop the packet... */
		free(pkt);
		return;
	}

	if (node->pt != pkt->fields.pt) {
		debug_msg("Payload type changed within frame - packet discarded\n");
		free(pkt);
		return;
	}

	node->mbit |= pkt->fields.m;
	node->cdata_count++;

	if (seq_lt(pkt->fields.seq, node->seqno_min)) node->seqno_min = pkt->fields.seq;
	if (seq_lt(node->seqno_max, pkt->fields.seq)) node->seqno_max = pkt->fields.seq;

	new_cdata->seqno = pkt->fields.seq;
	new_cdata->data  = pkt;
	new_cdata->prev  = NULL;
	new_cdata->next  = NULL;

	curr = node->cdata;
	while (curr != NULL) {
		if (seq_lt(curr->seqno, new_cdata->seqno)) {
			/* Insert into list before "curr".  In the usual case, when packets  */
			/* arrive in order, this inserts the packet at the head of the list. */
			new_cdata->next = curr;
			new_cdata->prev = curr->prev;
			if (curr->prev != NULL) {
				curr->prev->next = new_cdata;
			}
			curr->prev = new_cdata;
			if (curr == node->cdata) {
				node->cdata = new_cdata;
			}
			return;
		}
		if (curr->next == NULL) {
			/* Got to the end of the list without inserting the newly arrived */
			/* packet. This can occur due to reordering. Add at end of list.  */
			curr->next = new_cdata;
			new_cdata->prev = curr;
			return;
		}
		curr = curr->next;
	}
}

static struct pbuf_node * 
create_new_pnode(rtp_packet *pkt) 
{
	struct pbuf_node *tmp;

	tmp = malloc(sizeof(struct pbuf_node));
	if (tmp != NULL) {
		tmp->next          = NULL;
		tmp->prev          = NULL;
		tmp->decoded       = 0;
		tmp->rtp_timestamp = pkt->fields.ts;
		tmp->mbit	   = pkt->fields.m; 
		tmp->pt            = pkt->fields.pt; /* FIX ME: should we always and? - LG */
		tmp->cdata_count   = 1;
		tmp->seqno_min     = pkt->fields.seq;
		tmp->seqno_max     = pkt->fields.seq;
		gettimeofday(&(tmp->arrival_time), NULL);
		gettimeofday(&(tmp->playout_time), NULL);	
		gettimeofday(&(tmp->discard_time), NULL);
		
		/* Playout delay... should really be adaptive, based on the */
		/* jitter, but we use a fixed 50ms delay (assuming a frame  */
		/* rate of at least 20 fps)                                 */
		tv_add (&(tmp->playout_time), PBUF_PLAY_DELAY/1000000.0);
		
		/* MUST be > playout time + maximum network queuing delay */
		tv_add (&(tmp->discard_time),
			(PBUF_PLAY_DELAY + PBUF_PLAY_DISCARD)/1000000.0);

		tmp->cdata = malloc(sizeof(struct coded_data));
		if (tmp->cdata != NULL) {
			tmp->cdata->next   = NULL;
			tmp->cdata->prev   = NULL;
			tmp->cdata->seqno = pkt->fields.seq;
			tmp->cdata->data  = pkt;
		} else {
			free(pkt); 	
			free(tmp);
			return NULL;
		}
	} else {
		free(pkt);
	}
	return tmp;
}


void
pbuf_insert(struct pbuf *playout_buf, rtp_packet *pkt) 
{
	struct pbuf_node	*tmp;

	pbuf_validate(playout_buf);

	/* Check if this is a valid packet */
	if (pkt->fields.pt == PBUF_DUMMY_PT) {
		free (pkt);
		return;		
	}	
	
	if (playout_buf->frst==NULL && playout_buf->last==NULL) {
		/* Playout buffer is empty - add the new frame */
		playout_buf->frst = create_new_pnode(pkt);
		playout_buf->last = playout_buf->frst;
		playout_buf->size++;
		return;
	} 

	if (playout_buf->last->rtp_timestamp == pkt->fields.ts) {
		/* Packet belongs to last frame stored in the playout buffer. */
		/* This is the most likely scenario for video formats, which  */
		/* split each frame into multiple RTP packets.                */
		add_coded_unit(playout_buf->last, pkt);
	} else if (playout_buf->last->rtp_timestamp < pkt->fields.ts) {
		/* Packet belongs to a new frame. This is the most likely     */
		/* scenario for audio formats, where each frame is a single   */
		/* packet. Add a new node to the end of the playout buffer.   */
		tmp = create_new_pnode(pkt);
		playout_buf->last->next = tmp;
		tmp->prev = playout_buf->last;
		playout_buf->last = tmp;
		playout_buf->size++;
	} else {
		/* Packet belongs to a previous frame. This is unlikely, and  */
		/* should only occur if there has been packet reordering.     */
		if (playout_buf->frst->rtp_timestamp > pkt->fields.ts) {
			debug_msg("A very old packet - discarded\n");
		} else {
			debug_msg("A packet for a previous frame, but might still be useful\n");
			/* FIXME: should insert the packet into the playout buffer... */
		}
		free(pkt);
	} 
	pbuf_validate(playout_buf);
}

static void
free_cdata(struct coded_data *head)
{
	struct coded_data *tmp;

	while (head != NULL) {
		free(head->data);
		tmp  = head;
		head = head->next;
		free(tmp);
	}
}

void
pbuf_remove(struct pbuf *playout_buf, struct timeval curr_time)
{
	/* Remove previously decoded frames that have passed their playout  */
	/* time from the playout buffer. Incomplete frames that have passed */
	/* their playout time are also discarded.                           */

	struct pbuf_node *curr, *temp;

	pbuf_validate(playout_buf);

        curr=playout_buf->frst;
        while (curr != NULL) {
		temp = curr->next;
		if (tv_gt(curr_time, curr->discard_time)) {
			if (curr == playout_buf->frst) {
				playout_buf->frst = curr->next;
			}
			if (curr == playout_buf->last) {
				playout_buf->last = curr->prev;
			}
			if (curr->next != NULL) {
				curr->next->prev = curr->prev;
			}
			if (curr->prev != NULL) {
				curr->prev->next = curr->next;
			}
			free_cdata(curr->cdata);
			free (curr);
			assert(playout_buf->size > 0);
			playout_buf->size--;
		} else {
			/* The playout buffer is stored in order, so once  */
			/* we see one frame that has not yet reached it's  */
			/* playout time, we can be sure none of the others */
			/* will have done so...                            */
			break;
		}
		curr = temp;
        }

	pbuf_validate(playout_buf);
	return;
}

static int
frame_complete(struct pbuf_node *frame)
{
        /* Return non-zero if the list of coded_data represents a     */
	/* complete frame of video.                                   */
	/* A video-specific test: if we've received a packet with the */
	/* marker bit set or if we've started to received packets for */
	/* the next frame, we might have a complete packet. Otherwise */
	/* assume some data is missing.                               */

	struct coded_data	*curr, *prev;

	if (frame->mbit || (frame->next != NULL)) {
		/* Does the number of packets in this frame match the gap between the */
		/* last sequence number of the previous frame and the first sequence  */
		/* number of the next frame? If so, we know we have a complete frame. */
		if ((frame->prev != NULL) && (frame->next != NULL)) {
			unsigned gap = frame->next->seqno_min - frame->prev->seqno_max - 1;
			if (gap == frame->cdata_count) {
				return 1;
			}
		}

		/* Check for a complete frame by walking the list of packets... This  */
		/* is expensive, but only called if there's been reordering or packet */
		/* loss.                                                              */
		curr = frame->cdata;
		prev = NULL;
		while (curr != NULL) {
			if ((prev != NULL) && (curr->seqno != prev->seqno - 1u)) {
				return 0;  /* Frame is not complete... */
			}
			prev = curr;
			curr = curr->next;
		}
		/* Frame complete but missing M bit... */
		return 1;
	}

	/* Frame is not complete... */
	return 0;
}

int
pbuf_decode(struct pbuf *playout_buf, struct timeval curr_time, struct display *display)
{
	/* Find the first complete frame that has reached it's playout */
	/* time, and decode it into the framebuffer. Mark the frame as */
	/* decoded, but otherwise leave it in the playout buffer.      */
	struct pbuf_node	*curr;
	struct video_frame	*dispframe, *convframe;
	unsigned 		 curr_decoder_id;
	display_size_t		 display_size;
	video_colour_mode_t	 display_cm;
	display_result_t	 display_result = DISPLAY_OK;
	int			 res;

	pbuf_validate (playout_buf);

        curr = playout_buf->frst;
        while (curr != NULL) {
		if (!curr->decoded && tv_gt (curr_time, curr->playout_time)) {

			assert (tv_gt (curr->playout_time, playout_buf->last_playout_time));
			playout_buf->last_playout_time = curr->playout_time;
			
			/* Adjust the decoder for the playout buffer */
			curr_decoder_id = vcodec_get_by_payload(curr->pt);
			if (curr_decoder_id != playout_buf->decoder_id) {
				if (playout_buf->decoder_state != NULL) {
					vcodec_decoder_destroy (playout_buf->decoder_id,
								playout_buf->decoder_state);
				}
				if (curr_decoder_id != VCODEC_INVALID_ID) {
					playout_buf->decoder_state = vcodec_decoder_create (curr_decoder_id);
				}
				playout_buf->decoder_id    = curr_decoder_id;
			}

			if (playout_buf->decoder_id != VCODEC_INVALID_ID) {

				/* Decode and display the frame */
				display_size = vcodec_decoder_get_size(playout_buf->decoder_id,
								       playout_buf->decoder_state);
				display_cm   = vcodec_decoder_get_cm  (playout_buf->decoder_id,
								       playout_buf->decoder_state);

				dispframe    = display_get_frame      (display, display_size, display_cm);
				
				/*
				 * Some displays return a NULL frame (ie, the null
				 * display). Then, avoid all the decode/convert/display
				 * stuff.
				 */
				display_result = DISPLAY_OK;
				
				if (dispframe != NULL) {
					
					convframe = vidconv_prepare (dispframe, display_cm);
					
					assert (curr->decoded == FALSE);
					
					if (frame_complete (curr)) {
						res = vcodec_decode (playout_buf->decoder_id, 
								     playout_buf->decoder_state,
								     curr->cdata, convframe);
					} else {
						res = vcodec_decode_partial (playout_buf->decoder_id,
									     playout_buf->decoder_state,
									     curr->cdata, convframe);
					}
					
					/*
					 * The decoding process can return FALSE
					 * if it was unable to decode the frame. Then,
					 * skip the display.
					 */
					if (res) {
						vidconv_convert (convframe, dispframe);
					
						trace (te_pbuf_put_frame);
						
						display_result = display_put_frame (display, dispframe);
					} else {
						debug_msg ("Skiping invalid frame...\n");
					}
				}
				
				curr->decoded = TRUE;

			} else {
				debug_msg ("No decoder available for PT=%d.\n", curr->pt);
			}

			return display_result; 
		}
                curr = curr->next;
        }
	
	return DISPLAY_OK;
}

unsigned int
pbuf_length (struct pbuf *playout_buf)
{
	assert (playout_buf);
	
	return playout_buf->size;
}
