/*
 * FILE:     rtp_crypto.c
 * AUTHOR:   Colin Perkins   <csp@csperkins.org>
 * MODIFIED: Orion Hodson    <o.hodson@cs.ucl.ac.uk>
 *           Markus Germeier <mager@tzi.de>
 *           Bill Fenner     <fenner@research.att.com>
 *           Timur Friedman  <timur@research.att.com>
 *           Ladan Gharai    <ladan@isi.edu>
 *           Steve Smith     <ssmith@vislab.usyd.edu.au>
 *
 * The routines in this file implement the Real-time Transport Protocol,
 * RTP, as specified in RFC 3550. Portions of the code are derived from 
 * the algorithms published in that specification.
 *
 * Copyright (c) 1998-2001 University College London
 * Copyright (c) 2001-2004 University of Southern California
 * Copyright (c) 2003-2005 University of Glasgow
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions 
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Science
 *      Department at University College London and by the University of
 *      Southern California Information Sciences Institute.
 * 4. Neither the name of the University nor of the Department may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Revision: 1.4 $		[Derived revision 1.298 of rtp.c]
 * $Date: 2006/02/22 00:54:45 $
 * 
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "crypto/crypt_des.h"
#include "crypto/crypt_aes.h"
#include "rtp/rtp_crypto.h"

struct des_state {
	unsigned char  encryption_key[8];
};

struct aes_state {
	keyInstance keyInstEncrypt;
	keyInstance keyInstDecrypt;
	cipherInstance cipherInst;
};

void *
des_initialize(unsigned char *hash, int hashlen, int *pad_len)
{
	int	          i, j, k;
	struct des_state *s;
	
	UNUSED(hashlen);

	*pad_len = 8;
	
	s = malloc(sizeof(struct des_state));

	/* Step 3: take first 56 bits of the MD5 hash */
	s->encryption_key[0] = hash[0];
	s->encryption_key[1] = hash[0] << 7 | hash[1] >> 1;
	s->encryption_key[2] = hash[1] << 6 | hash[2] >> 2;
	s->encryption_key[3] = hash[2] << 5 | hash[3] >> 3;
	s->encryption_key[4] = hash[3] << 4 | hash[4] >> 4;
	s->encryption_key[5] = hash[4] << 3 | hash[5] >> 5;
	s->encryption_key[6] = hash[5] << 2 | hash[6] >> 6;
	s->encryption_key[7] = hash[6] << 1;

	/* Step 4: add parity bits */
	for (i = 0; i < 8; ++i) {
		k = s->encryption_key[i] & 0xfe;
		j = k;
		j ^= j >> 4;
		j ^= j >> 2;
		j ^= j >> 1;
		j = (j & 1) ^ 1;
		s->encryption_key[i] = k | j;
	}

	return (void *) s;
}

int 
des_encrypt(void *state, unsigned char *data, unsigned int size, unsigned char *initVec)
{
	struct des_state *s = (struct des_state *) state;
	qfDES_CBC_e(s->encryption_key, data, size, initVec);
	return TRUE;
}

int 
des_decrypt(void *state, unsigned char *data, unsigned int size, unsigned char *initVec)
{
	struct des_state *s = (struct des_state *) state;
	qfDES_CBC_d(s->encryption_key, data, size, initVec);
	return TRUE;
}

void *
rijndael_initialize(unsigned char *hash, int hash_len, int *pad_len)
{
	int               rc;
	struct aes_state *s = malloc(sizeof(struct aes_state));
	
	*pad_len = 16;

	rc = makeKey(&s->keyInstEncrypt, DIR_ENCRYPT, hash_len * 8, hash);
	if (rc < 0) {
		debug_msg("makeKey failed: %d\n", rc);
		return FALSE;
	}

	rc = makeKey(&s->keyInstDecrypt, DIR_DECRYPT, hash_len * 8, hash);
	if (rc < 0) {
		debug_msg("makeKey failed: %d\n", rc);
		return FALSE;
	}

	rc = cipherInit(&s->cipherInst, MODE_ECB, NULL);
	if (rc < 0) {
		debug_msg("cipherInst failed: %d\n", rc);
		return FALSE;
	}
	return (void *) s;
}

int 
rijndael_encrypt(void *state, unsigned char *data, unsigned int size, unsigned char *initVec)
{
	struct aes_state *s = (struct aes_state *) state;
	int               rc;

	UNUSED(initVec);

	/*
	 * Try doing this in place. If it doesn't work that way,
	 * we'll have to allocate a buffer and copy back.
	 */
	rc = blockEncrypt(&s->cipherInst, &s->keyInstEncrypt, data, size * 8, data);
	return rc;
}

int 
rijndael_decrypt(void *state, unsigned char *data, unsigned int size, unsigned char *initVec)
{
	struct aes_state *s = (struct aes_state *) state;
	int               rc;

	UNUSED(initVec);

	/*
	 * Try doing this in place. If it doesn't work that way,
	 * we'll have to allocate a buffer and copy back.
	 */
	rc = blockDecrypt(&s->cipherInst, &s->keyInstDecrypt, data, size * 8, data);
	return rc;
}

