/*
 * FILE:     rtp_db.c
 * AUTHOR:   Colin Perkins   <csp@csperkins.org>
 * MODIFIED: Orion Hodson    <o.hodson@cs.ucl.ac.uk>
 *           Markus Germeier <mager@tzi.de>
 *           Bill Fenner     <fenner@research.att.com>
 *           Timur Friedman  <timur@research.att.com>
 *           Ladan Gharai    <ladan@isi.edu>
 *           Steve Smith     <ssmith@vislab.usyd.edu.au>
 *
 * The routines in this file implement the Real-time Transport Protocol,
 * RTP, as specified in RFC 3550. Portions of the code are derived from 
 * the algorithms published in that specification.
 *
 * Copyright (c) 1998-2001 University College London
 * Copyright (c) 2001-2004 University of Southern California
 * Copyright (c) 2003-2005 University of Glasgow
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions 
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Science
 *      Department at University College London and by the University of
 *      Southern California Information Sciences Institute.
 * 4. Neither the name of the University nor of the Department may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Revision: 1.4 $
 * $Date: 2005/11/08 14:41:52 $
 *
 * Derived from revision 1.305 of rtp.c
 * 
 */

#include "config.h"
#include "config_unix.h"
#include "memory.h"
#include "debug.h"
#include "rtp/cc.h"
#include "rtp/net_udp.h"
#include "rtp/rtp.h"
#include "rtp/rtp_db.h"
#include "rtp/rtp_tfrc.h"
#include "rtp/rtp_priv.h"
#include "tv.h"

struct rtp_db {
	uint32_t	 my_ssrc;
	rtp_source	*db[RTP_DB_SIZE];
        rtcp_rr_wrapper	 rr[RTP_DB_SIZE][RTP_DB_SIZE]; 	/* Indexed by [hash(reporter)][hash(reportee)] */
	unsigned	 ssrc_count;
};

int 
ssrc_hash(uint32_t ssrc)
{
	/* Hash from an ssrc to a position in the source database.   */
	/* Assumes that ssrc values are uniformly distributed, which */
	/* should be true but probably isn't (Rosenberg has reported */
	/* that many implementations generate ssrc values which are  */
	/* not uniformly distributed over the space, and the H.323   */
	/* spec requires that they are non-uniformly distributed).   */
	/* This routine is written as a function rather than inline  */
	/* code to allow it to be made smart in future: probably we  */
	/* should run MD5 on the ssrc and derive a hash value from   */
	/* that, to ensure it's more uniformly distributed?          */
	return ssrc % RTP_DB_SIZE;
}

void 
check_database(struct rtp *session)
{
	/* This routine performs a sanity check on the database. */
	/* This should not call any of the other routines which  */
	/* manipulate the database, to avoid common failures.    */
#ifdef DEBUG
	rtp_source 	*s;
	int	 	 source_count;
	int		 chain;

	assert (session		!= NULL);
	assert (session->magic	== RTP_MAGIC);

	/* Check that we have a database entry for our ssrc... */
	/* We only do this check if ssrc_count > 0 since it is */
	/* performed during initialisation whilst creating the */
	/* source entry for my_ssrc.                           */
	if (session->ssrc_count > 0) {
		for (s = session->db[ssrc_hash(session->my_ssrc)]; s != NULL; s = s->next) {
			if (s->ssrc == session->my_ssrc) {
				break;
			}
		}
		assert(s != NULL);
	}

	source_count = 0;
	for (chain = 0; chain < RTP_DB_SIZE; chain++) {
		/* Check that the linked lists making up the chains in */
		/* the hash table are correctly linked together...     */
		for (s = session->db[chain]; s != NULL; s = s->next) {
			source_count++;
			if (s->prev == NULL) {
				assert(s == session->db[chain]);
			} else {
				assert(s->prev->next == s);
			}
			if (s->next != NULL) {
				assert(s->next->prev == s);
			}
			/* Check that the SR is for this source... */
			if (s->sr != NULL) {
				assert(s->sr->ssrc == s->ssrc);
			}
		}
	}
	/* Check that the number of entries in the hash table  */
	/* matches session->ssrc_count                         */
	assert(source_count == session->ssrc_count);
#else 
	UNUSED(session);
#endif
}

rtp_source *
get_source(struct rtp *session, uint32_t ssrc)
{
	rtp_source *s;

	assert (session		!= NULL);
	assert (session->magic	== RTP_MAGIC);

	check_database (session);
	for (s = session->db[ssrc_hash(ssrc)]; s != NULL; s = s->next) {
		if (s->ssrc == ssrc) {
			return s;
		}
	}
	return NULL;
}

rtp_source *
create_source (struct rtp *session, uint32_t ssrc, int probation)
{
	rtp_source	*s;
	rtp_event	 event;
	int		 h;
	
	assert (session		!= NULL);
	assert (session->magic	== RTP_MAGIC);

	s = get_source (session, ssrc);
	if (s != NULL) {
		/* Source is already in the database.  Mark it */
		/* active and exit. This is the common case... */
		gettimeofday (&(s->last_active), NULL);
		return s;
	}
	assert(s == NULL);

	/* Create a new source entry, and add it to the database.    */
	/* The database is a hash table, using the separate chaining */
	/* algorithm.                                                */

	h = ssrc_hash(ssrc);
	s = malloc(sizeof(rtp_source));
	if (s != NULL) {
		memset(s, 0, sizeof(rtp_source));
		s->magic = 0xc001feed;
		s->next  = session->db[h];
		s->ssrc  = ssrc;
		if (probation) {
			/* This is a probationary source, which only counts as */
			/* valid once several consecutive packets are received */
			s->probation = -1;
		} else {
			s->probation = 0;
		}
		
		gettimeofday(&(s->last_active), NULL);
		/* Now, add it to the database... */
		if (session->db[h] != NULL) {
			session->db[h]->prev = s;
		}
		session->db[ssrc_hash(ssrc)] = s;
		session->ssrc_count++;
		check_database (session);

		if (ssrc != session->my_ssrc) {
			if (!rtp_filter_event (session, ssrc)) {
				assert (session->callback != NULL);

				event.ssrc = ssrc;
				event.type = SOURCE_CREATED;
				event.data = NULL;
				
				session->callback (session, &event);
			}
		}
	}
	return s;
}

static void 
remove_rr(struct rtp *session, uint32_t ssrc)
{
	/* Remove any RRs from "s" which refer to "ssrc" as either   */
        /* reporter or reportee.                                     */
        rtcp_rr_wrapper *start, *cur, *tmp;
        int i;

	assert (session		!= NULL);
	assert (session->magic	== RTP_MAGIC);

        /* Remove rows, i.e. ssrc == reporter_ssrc                   */
        for(i = 0; i < RTP_DB_SIZE; i++) {
                start = &session->rr[ssrc_hash(ssrc)][i];
                cur   = start->next;
                while (cur != start) {
                        if (cur->reporter_ssrc == ssrc) {
                                tmp = cur;
                                cur = cur->prev;
                                tmp->prev->next = tmp->next;
                                tmp->next->prev = tmp->prev;
				free(tmp->ts);
                                free(tmp->rr);
				if (tmp->rx) free (tmp->rx);
                                free(tmp);
                        }
                        cur = cur->next;
                }
        }

        /* Remove columns, i.e.  ssrc == reporter_ssrc */
        for(i = 0; i < RTP_DB_SIZE; i++) {
                start = &session->rr[i][ssrc_hash(ssrc)];
                cur   = start->next;
                while (cur != start) {
                        if (cur->rr->ssrc == ssrc) {
                                tmp = cur;
                                cur = cur->prev;
                                tmp->prev->next = tmp->next;
                                tmp->next->prev = tmp->prev;
				free(tmp->ts);
                                free(tmp->rr);
				if (tmp->rx) free (tmp->rx);
                                free(tmp);
                        }
                        cur = cur->next;
                }
        }
}

void 
delete_source(struct rtp *session, uint32_t ssrc)
{
	/* Remove a source from the RTP database... */
	rtp_source	*s = get_source(session, ssrc);
	int		 h = ssrc_hash(ssrc);
	rtp_event	 event;

	assert (s		!= NULL); /* Deleting a source which doesn't exist is an error... */
	assert (session		!= NULL);
	assert (session->magic	== RTP_MAGIC);

	check_database(session);
	if (session->db[h] == s) {
		/* It's the first entry in this chain... */
		session->db[h] = s->next;
		if (s->next != NULL) {
			s->next->prev = NULL;
		}
	} else {
		assert(s->prev != NULL);	/* Else it would be the first in the chain... */
		s->prev->next = s->next;
		if (s->next != NULL) {
			s->next->prev = s->prev;
		}
	}
	/* Free the memory allocated to a source... */
	if (s->sdes_cname != NULL) free(s->sdes_cname);
	if (s->sdes_name  != NULL) free(s->sdes_name);
	if (s->sdes_email != NULL) free(s->sdes_email);
	if (s->sdes_phone != NULL) free(s->sdes_phone);
	if (s->sdes_loc   != NULL) free(s->sdes_loc);
	if (s->sdes_tool  != NULL) free(s->sdes_tool);
	if (s->sdes_note  != NULL) free(s->sdes_note);
	if (s->sdes_priv  != NULL) free(s->sdes_priv);
	if (s->sr    != NULL) free(s->sr);

        remove_rr(session, ssrc); 
	session->ssrc_count--;

	rtcp_reverse_reconsideration(session);

	/* Signal to the application that this source is dead... */
	if (!rtp_filter_event (session, ssrc)) {
		event.ssrc = ssrc;
		event.type = SOURCE_DELETED;
		event.data = NULL;
		session->callback(session, &event);
	}
        free(s);
	check_database(session);
}

