/*
 * FILE:   rtp/rtp_tfrc.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.8 $
 * $Date: 2006/02/03 16:29:18 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "memory.h"
#include "debug.h"
#include "rtp/cc.h"
#include "rtp/net_udp.h"
#include "rtp/rtp.h"
#include "rtp/rtp_db.h"
#include "rtp/rtp_tfrc.h"
#include "rtp/rtp_priv.h"
#include "tv.h"
#include "trace.h"
#include "trace_ug.h"

/* RTT flags */
#define TFRC_RTT_UNAVAIL	63
#define TFRC_RTT_AVAIL		64

#define TFRC_MAX_PT		TFRC_RTT_UNAVAIL

/* Return TRUE if TFRC is been used in this session */
int
rtp_tfrc_on (struct rtp *session)
{
#ifdef USE_TFRC_EXTENSIONS	
	return (cc_sndr_get_algo (session->sndr_cc) == tfrc_e ? TRUE : FALSE);
#else
	return FALSE;
#endif	
}

/* Reception of tfrc reports */
void
rtp_tfrc_recv_report (struct rtp *session, rtcp_rx *rx)
{
	uint32_t	 now;
	struct timeval	 tv_now;
	cc_rcvr_info_t	 rx_info;
	
	assert (session			!= NULL);
	assert (session->sndr_cc	!= NULL);
	
	rx_info.ts	= rx->ts;
	rx_info.tdelay	= rx->tdelay;
	rx_info.x_recv	= rx->x_recv;
	rx_info.p	= rx->p;
	
	/* Provide the RTT known by RTP, as some CC don't compute it by themselves */
	rx_info.rtt	= session->rtp_rtt;

	now = tv_usecs_since (&(session->base_time), &tv_now);

	cc_sndr_process (session->sndr_cc, now, &rx_info);	
	
	/* Trace some events */
	trace_at     (&tv_now, te_tx_recv);
	
	trace_val_at (&tv_now, te_rtp_tfrc_tx_ts,	rx_info.ts);
	trace_val_at (&tv_now, te_rtp_tfrc_tx_tdelay,	rx_info.tdelay);
	trace_val_at (&tv_now, te_rtp_tfrc_tx_x_recv,	rx_info.x_recv);
	trace_val_at (&tv_now, te_rtp_tfrc_tx_p,	rx_info.p);	
}

/* Reception of tfrc info in data */
void
rtp_tfrc_recv_data (struct rtp *session, rtp_packet *packet)
{
	uint32_t	 now;
	struct timeval	 tv_now;
	cc_sndr_info_t	 tx_info;
	int		 rtt_available;

	assert (session			!= NULL);
	assert (session->rcvr_cc	!= NULL);
	assert (packet			!= NULL);
	
	/* Use data attached for CC */
	rtt_available	= packet->fields.pt & TFRC_RTT_AVAIL;

	tx_info.seq	= packet->fields.seq;
	tx_info.send_ts	= packet->fields.send_ts;

	if (rtt_available) {
		tx_info.rtt		 = packet->fields.rtt;
		packet->fields.pt	&= TFRC_RTT_UNAVAIL;	/* Clean the RTT bit */
	} else {
		tx_info.rtt		 = 0;
	}
		
	now = tv_usecs_since (&(session->base_time), &tv_now);
	
	cc_rcvr_process (session->rcvr_cc, now, &tx_info);

	/* Trace some events */
	trace_at     (&tv_now, te_rx_recv);

	trace_val_at (&tv_now, te_rtp_tfrc_rx_seq,	tx_info.seq);
	trace_val_at (&tv_now, te_rtp_tfrc_rx_send_ts,	tx_info.send_ts);
	trace_val_at (&tv_now, te_rtp_tfrc_rx_rtt,	tx_info.rtt);
}

/* Format some TFRC info to put in a report */
void
rtp_tfrc_format_rrxp (struct rtp *session, rtcp_rr_rx *rrxp)
{
	struct timeval	 tv_now;
	cc_rcvr_info_t	 info;
	uint32_t	 now;
	
	assert (session != NULL);
	assert (rrxp	!= NULL);

	now = tv_usecs_since (&(session->base_time), &tv_now);

	cc_rcvr_format (session->rcvr_cc, now, &info);

	/* Trace some events */
	trace_val_at (&tv_now, te_rtp_tfrc_rx_ts,	info.ts);
	trace_val_at (&tv_now, te_rtp_tfrc_rx_tdelay,	info.tdelay);
	trace_val_at (&tv_now, te_rtp_tfrc_rx_x_recv,	info.x_recv);
	trace_val_at (&tv_now, te_rtp_tfrc_rx_p,	info.p);	
	
	rrxp->tdelay	= htonl (info.tdelay);
	rrxp->x_recv	= htonl (info.x_recv); 
	rrxp->ts	= htonl (info.ts);
	rrxp->p		= htonl (info.p);
}

/* Format TFRC info to attach in a data packet */
void
rtp_tfrc_format_data (struct rtp *session, rtp_packet *packet)
{
	uint32_t	 now;
	struct timeval	 tv_now;
	uint32_t	 current_rtt;

	assert (packet != NULL);	
	assert (packet->fields.pt <= TFRC_MAX_PT);	/* PT must be betweeen [0..63] */
	
	/* If the RTT has changed... */
	current_rtt = cc_sndr_get_avg_rtt (session->sndr_cc); 

	/* Set the sending time */
	now = tv_usecs_since (&(session->base_time), &tv_now);

	packet->fields.send_ts = htonl (now);

	/* Trace some events */
	trace_val_at (&tv_now, te_rtp_tfrc_tx_seq,	packet->fields.seq);
	trace_val_at (&tv_now, te_rtp_tfrc_tx_send_ts,	now);
	trace_val_at (&tv_now, te_rtp_tfrc_tx_rtt,	current_rtt);
	
	/* Set the RTT-available flag depending on any change in the RTT */
	if (current_rtt != session->sndr_cc_rtt) { 
		packet->fields.rtt	 = htonl (current_rtt);
		packet->fields.pt	|= TFRC_RTT_AVAIL;
		
		/* Save the RTT, to compare */
		session->sndr_cc_rtt	 = current_rtt;
	} else {
		packet->fields.pt	&= TFRC_RTT_UNAVAIL;	/* Clean the RTT bit */
	}
	
}


/* Procesing needed for TFRC info to attach in data packets */
unsigned int
rtp_tfrc_convert_tx (rtp_packet *packet)
{
	unsigned int	vlen;
	
	/* figure out header lenght based on TFRC	*/
	/* might as well extract rtt and send_ts	*/
	
	vlen = 4;
	
	packet->fields.send_ts = htonl (packet->fields.send_ts);

	if (packet->fields.pt & TFRC_RTT_AVAIL) { 
		/* rtt is present in RTP packet */
		vlen += 4;
		packet->fields.rtt	 = htonl (packet->fields.rtt);
	} else {
		packet->fields.rtt = 0;
	}	
	
	return vlen;
}

/* Procesing needed for TFRC info attached feedbacks */
void
rtp_tfrc_convert_rx (rtcp_rx *rx, rtcp_rr_rx *rrxp)
{	
	assert (rrxp != NULL);
	assert (rx != NULL);

	rx->ts		= ntohl (rrxp->ts);
	rx->tdelay	= ntohl (rrxp->tdelay);
	rx->x_recv	= ntohl (rrxp->x_recv);
	rx->p		= ntohl (rrxp->p);	
}
