/*
 * FILE:   sbuf.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.34 $
 * $Date: 2007/01/08 22:22:22 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace_ug.h"
#include "trace.h"

#include "tv.h"
#include "video_types.h"
#include "video_convert.h"
#include "video_display.h"
#include "cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "rtp/sbuf.h"
#include "transmit.h"

#define SBUF_MAGIC		0xaf5abcbe
#define SBUF_NODE_MAGIC		0xaf5cbabe

/* Weighting factors. Higer factors give more weigth to old values... */
#define SBUF_IFI_FACTOR		0.9	/* ... for IFI & OFI */
#define SBUF_BR_FACTOR		0.9	/* ... for the bitrate */
#define AVG_FL_FACTOR		0.8	/* ... for average frame length */

/*
 * Initial values for the sending buffer context. This should be meaningless,
 * as the context should be set with the correct values before the first frame
 * is compressed...
 */
#define SBUF_CTX_INIT_MTU		1470
#define SBUF_CTX_INIT_MAX_OUT_IFI	1000
#define SBUF_CTX_INIT_MAX_FRAMES	100
#define SBUF_CTX_INIT_MAX_FRAMELEN	200000
#define SBUF_CTX_INIT_MAX_FRAGS		(SBUF_CTX_INIT_MAX_FRAMELEN * \
					 SBUF_CTX_INIT_MAX_FRAMES) / SBUF_CTX_INIT_MTU

/* Should we check the sending buffer to see if we can insert? */
#define USE_ALWAYS_INSERT

/*
 * Should we check if the frame will arrive in time? This should be enabled
 * if we always insert frames (USE_ALWAYS_INSERT). Otherwise, we can have a huge
 * sending buffer...
 */
#define USE_CHECK_TRANS_TIME

/* Time error allowed, for the transmission of the frame */
#ifdef USE_CHECK_TRANS_TIME
#define MAX_FRAME_DELAY_ERR_ALLOW	50000	/* usecs */
#endif

/*
 * Sending buffer locking.
 * As long as we are inserting and removing frames in the same thread (and
 * nowhere else), we don't need locking the global state.
 */
#undef USE_LOCK_SBUF

/* Sending buffer locking: functions */
#define sbuf_lock_init(x)	pthread_mutex_init(&(x->lock), NULL)
#define sbuf_lock_done(x)	pthread_mutex_destroy(&(x->lock))

#ifdef USE_LOCK_SBUF
#define sbuf_lock(x)		pthread_mutex_lock(&(x->lock))
#define sbuf_unlock(x)		pthread_mutex_unlock(&(x->lock))
#else
#define sbuf_lock(x)
#define sbuf_unlock(x)
#endif

/* Node of the sending buffer (containing an encoded frame) */
struct sbuf_node {
	uint32_t		 magic;		/* For debugging		*/

	struct sbuf_node	*next;
	struct sbuf_node	*prev;
	
	t_usec			 ts;		/* Timestamp of the frame	*/
	t_usec			 rtp_ts;

	struct encoded_frame	*enc_frame;
	unsigned long		 bytes_sent;
	unsigned int		 frags_sent;	/* Num. of fragments sent in  */
						/* this encoded frame [0-n] */

	unsigned int		 frame_num;	/* Number of frame		*/

	int			 sent;		/* True when it has been sent	*/
};

/*
 * Internal representation of the sending buffer.
 *
 * The sending buffer keeps both sent and unsent frames. Sent frames must be
 * periodically cleaned, by calling sbuf_remove().
 *
 * The sending buffer can be divided in two zones: one for sent frames and other
 * for unsent frames. The first zone starts with the 'first' pointer, and the
 * second with 'first_out'. The former pointer is the limit between both zones,
 * in such a way that 'first <= first_out <= last'.
 *
 * Nodes in the zone [first, first_out] have the 'sent=TRUE' property, and must
 * be periodically removed using 'sbuf_remove()'.
 */
struct sbuf {
	uint32_t		 magic;		/* For debugging		*/

	struct sbuf_node	*first;		/* First node to remove		*/
	struct sbuf_node	*first_out;	/* Next node to be transmitted */
	struct sbuf_node	*last;		/* Last node inserted */
	
	struct sbuf_ctx_t	*ctx;		/* Buffer context		*/
	
	unsigned int		 frames_count;	/* Number of frames inserted	*/
	
	/* Counters */
	unsigned int		 unsent_frames;	/* Number of frames		*/
	unsigned int		 unsent_frags;	/* Number of fragments		*/
	unsigned long		 unsent_bytes;
	
	unsigned int		 total_frames;	/* Total number of frames	*/
	unsigned int		 total_frags;	/* Total number of fragments	*/
	unsigned long		 total_bytes;

	/* Statistics */
	struct timeval		 last_in_time;	/* Last insertion time		*/
	struct timeval		 last_out_time;	/* Last removal time		*/
	t_usec			 in_ifi;	/* Interval between insertions	*/
	t_usec			 out_ifi;	/* Interval between removals	*/

	/* Average input/input compressed/output rates */
	unsigned int		 in_rate;
	unsigned int		 comp_rate;
	unsigned int		 out_rate;
	
	/* Average lengths of uncompressed/compressed frames inserted */
	unsigned int		 avg_in_frame_len;
	unsigned int		 avg_out_frame_len;

	/* Encoder for inserted frames */
	void			*encoder_state;	/* Encoder state	*/
	unsigned int		 encoder_id;	/* ... id		*/
	struct vcodec_ctx_t     *encoder_ctx;	/* and encoding context */
	
	struct timeval		*base_time;
	
	struct video_tx		*tx;
	
	pthread_mutex_t		 lock;
};

/*********************************************************************************/

/* Validates the structure of the sending buffer */
static void
sbuf_validate (struct sbuf * send_buf)
{
#ifdef DEBUG_SBUF
	struct sbuf_node	*cpb, *ppb;
	unsigned int		 cdc,  pbc;

	cpb = send_buf->frst;
	ppb = NULL;
	pbc = 0;
	
	while (cpb != NULL) {
		assert (cpb->magic == SBUF_NODE_MAGIC);
		assert (cpb->prev == ppb);

		if (cpb->prev != NULL) {
			assert (cpb->prev->next == cpb);
			assert (cpb->ts > ppb->ts);
			
			/* FIXME: previous assert will fail when timestamp wraps */
			
			assert (tv_gt (cpb->send_time, ppb->send_time));
		}
		
		if (cpb->next != NULL) {
			assert(cpb->next->prev == cpb);
		} else {
			assert(cpb = send_buf->last);
		}

		if (cpb->edata != NULL) {
			/*
			 * We have coded data... check all the pointers on
			 * that list too...
			 */
			cdc = 0;
			ccd = cpb->edata;
			pcd = NULL;
			
			while (ccd != NULL) {
				cdc++;
				assert(ccd->data != NULL);
				assert(ccd->prev == pcd);
				if (ccd->prev != NULL) {
					assert(ccd->prev->next == ccd);
					assert(seq_lt(ccd->seqno, pcd->seqno));	/* Coded data stored in
										 * descending order */
				}
				if (ccd->next != NULL) {
					assert(ccd->next->prev == ccd);
				}
				pcd = ccd;
				ccd = ccd->next;
			}
			assert(cpb->edata_count == cdc);
		}
		
		/* Advance to next node in the buffer... */
		ppb = cpb;
		cpb = cpb->next;
		pbc++;
	}
	
	assert(pbc == send_buf->size);
#else
	UNUSED(send_buf);
#endif
}


/*******************************************************************************
 * Insertion and removal
 ******************************************************************************/

/*
 * Move forward the first_out pointer, removing the first frame from the counters
 * of unsent frames. 
 */
static void
sbuf_go_forward (struct sbuf *sbuf)
{
	struct sbuf_node	*fo;
	
	assert (sbuf != NULL);
	
	fo = sbuf->first_out;
	if (fo != NULL) {
		
		fo->sent = TRUE;
		
		assert (sbuf->unsent_frames	> 0);
		assert (sbuf->unsent_frags	> 0);
		assert (sbuf->unsent_bytes	> 0);
		
		sbuf->unsent_frames	--;
		sbuf->unsent_bytes	-= fo->bytes_sent;
		sbuf->unsent_frags	-= fo->frags_sent;

		sbuf->first_out		 = fo->next;
		
		//trace (te_sbuf_go_forward);
	}
}

/* Insert a new encoded fram at the end of the sending buffer */
static void
sbuf_add (struct sbuf		*sbuf,
	  struct encoded_frame	*eframe,
	  t_usec		 insertion_ts,
	  t_usec		 rtp_ts)
{
	struct sbuf_node	*new_node		= NULL;
	
	assert (sbuf != NULL);
	
	//trace (te_sbuf_add);
	
	new_node = malloc (sizeof (struct sbuf_node));
	if (new_node != NULL) {

		/* Set up the node */
		new_node->magic		= SBUF_NODE_MAGIC;
		new_node->enc_frame	= eframe;
		new_node->frame_num	= sbuf->frames_count++;

		new_node->frags_sent	= 0;
		new_node->bytes_sent	= 0;

		new_node->sent		= FALSE;

		/* Times */
		new_node->ts		= insertion_ts;
		new_node->rtp_ts        = rtp_ts;

		/* Put the node in the buffer */
		new_node->prev		= sbuf->last;
		new_node->next		= NULL;
		
		/* Update the sbuf: first node */
		if (sbuf->first_out == NULL) {
			sbuf->first_out	= new_node;
		}

		if (sbuf->first == NULL) {
			sbuf->first	= new_node;
		}
		
		/* Update the sbuf: last node */
		if (sbuf->last != NULL) {
			assert (sbuf->last->next == NULL);
			sbuf->last->next = new_node;
		}
		sbuf->last = new_node;
		
		/* Update counters */
		sbuf->unsent_frames	++;
		sbuf->unsent_frags	+= eframe->len;
		sbuf->unsent_bytes	+= eframe->datalen;

		sbuf->total_frames	++;
		sbuf->total_frags	+= eframe->len;
		sbuf->total_bytes	+= eframe->datalen;
		
		assert (sbuf->total_frames >= sbuf->unsent_frames);
		
		assert ((sbuf->first == NULL) ||
			((sbuf->first->frame_num     <= sbuf->first_out->frame_num) &&
			 (sbuf->first_out->frame_num <= sbuf->last->frame_num)));
		
		assert ((sbuf->first == NULL && sbuf->first_out == NULL) ||
			(sbuf->first != NULL && sbuf->first_out != NULL));

	} else {
		fprintf (stderr, "Error allocating memory for new sbuf_node node.\n");
	}
}

/* Remove a frame from the sending buffer (controlling pointers and other tasks...) */
static void
sbuf_delete (struct sbuf * sbuf, struct sbuf_node *node)
{
	assert (node		!= NULL);
	assert (sbuf		!= NULL);
	assert (sbuf->first	!= NULL);
	assert (sbuf->last	!= NULL);

	//trace (te_sbuf_delete);

	if (node == sbuf->first_out) {
		sbuf_go_forward (sbuf);
	}
	
	if (node == sbuf->first)
		sbuf->first		= node->next;
	
	if (node == sbuf->last)
		sbuf->last		= node->prev;
	
	if (node->next != NULL)
		node->next->prev	= node->prev;
	
	if (node->prev != NULL)
		node->prev->next	= node->next;

	/* Update counters */
	assert (sbuf->total_frames	> 0);
	assert (sbuf->total_frags	> 0);
	assert (sbuf->total_bytes	> 0);

	sbuf->total_frames	--;
	sbuf->total_frags	-= node->enc_frame->len;
	sbuf->total_bytes	-= node->enc_frame->datalen;
	
	assert (sbuf->total_frames	>= sbuf->unsent_frames);
	assert (sbuf->total_frags	>= sbuf->unsent_frags);
	//assert (sbuf->total_bytes	>= sbuf->unsent_bytes);
	
	assert (sbuf->first == NULL || sbuf->first->prev == NULL);
	assert (sbuf->last  == NULL || sbuf->last->next  == NULL);
	
	/* Free the encoded_frame and the node */
	vcodec_eframe_free (node->enc_frame->id, node->enc_frame);
	free (node);
}

/*******************************************************************************
 * Sending buffer policy
 ******************************************************************************/

/* Return TRUE if the sending buffer has a length limit and it is full */
static bool
sbuf_full (struct sbuf * send_buf)
{
	if ((send_buf->ctx->max_frames > 0) &&
	    (send_buf->unsent_frames >= send_buf->ctx->max_frames)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

/*
 * Check of the policy.
 * Checks that the given amount of data (corresponsing to an encoded frame) can
 * be inserted in the sending buffer. If it is false, the frame will probably
 * be discarded (or encoded and marked as sent).
 */
static bool
sbuf_can_insert (struct sbuf * send_buf, unsigned int bytes)
{
	int		res		= TRUE;
	
#ifndef USE_ALWAYS_INSERT
	unsigned int	max_frames,
			unsent_frames,
			frame_len,
			new_frames;
	unsigned int	encoder_id;

	if (bytes > 0) {
		max_frames = send_buf->ctx->max_frames;
		
		/* If there is a limit in the number of frames, check it. */
		if (max_frames > 0) {
			
			if (sbuf_full (send_buf)) {
				res = FALSE;
			} else {
				unsent_frames	= send_buf->unsent_frames;
				encoder_id	= send_buf->encoder_id;
				frame_len	= vcodec_encoder_get_framelen (encoder_id);
				
				if (frame_len > 0) {
					new_frames = bytes / frame_len;
				
					if (unsent_frames + new_frames > max_frames) {
						res = FALSE;
					}
				}
			}
		}
	}
#endif	
	UNUSED (send_buf);
	UNUSED (bytes);
	
	return res;
}

/*
 * Calculate if the frame should be marked as old.
 * If the time needed for transmitting the frame, given the current situation
 * (RTT, delay...), results in the frame arriving too late, then just discard it.
 */ 
static void
sbuf_check_arrival_time (struct sbuf * sbuf, struct sbuf_node *node)
{
#ifdef USE_CHECK_TRANS_TIME
	t_usec		exp_display_t	= 0,
			max_display_t	= 0,
			trans_delay	= 0,
			usecs_now	= 0;
	struct timeval	t_now;
	unsigned long	datalen;
#endif	
	
	assert (node		!= NULL);
	assert (sbuf		!= NULL);
	assert (node->magic	== SBUF_NODE_MAGIC);

#ifdef USE_CHECK_TRANS_TIME
	/* Check it, but never discard frames that we already started to send... */
	if (node->bytes_sent == 0) {
		datalen		= node->enc_frame->datalen - node->bytes_sent;
		
		usecs_now	= tv_usecs_since (sbuf->base_time, &t_now);
		
		trans_delay	= tx_trans_delay (sbuf->tx, datalen);
		exp_display_t	= usecs_now + trans_delay;
		max_display_t	= node->ts  + MAX_FRAME_DELAY + MAX_FRAME_DELAY_ERR_ALLOW;
		
		if (exp_display_t > max_display_t) {
			node->bytes_sent	= node->enc_frame->datalen;
			node->frags_sent	= node->enc_frame->len;
			node->sent		= TRUE;
			
			//trace (te_sbuf_discard);
		}

		//trace_val_at (&t_now, te_sbuf_max_arrival, max_display_t);
		//trace_val_at (&t_now, te_sbuf_exp_arrival, exp_display_t);
	}
#endif	
}


/*
 * Encoding policy.
 * Adjust the enconding context (used by the codec) based on the current sbuf
 * context (that specifies some limits...) and the state of the sending buffer.
 */
static bool
sbuf_adjust_encoding_ctx (struct sbuf *send_buf)
{
	struct vcodec_ctx_t     *encoder_ctx;
	unsigned int		 old_bitrate,
				 new_bitrate;
	
	assert (send_buf		!= NULL);
	assert (send_buf->encoder_ctx	!= NULL);
	
	encoder_ctx = send_buf->encoder_ctx;

	new_bitrate	= (send_buf->ctx->max_framelen * 8.0)
		/ usec2sec (send_buf->ctx->max_out_ifi);
	old_bitrate	= encoder_ctx->bitrate;

	encoder_ctx->bitrate = (SBUF_BR_FACTOR * old_bitrate + 
			(1.0 - SBUF_BR_FACTOR) * new_bitrate);
	
	encoder_ctx->framesize	= send_buf->ctx->max_framelen;
	
	assert (encoder_ctx->bitrate	> 0);
	assert (encoder_ctx->framesize	> 0);
	
	//trace_val (te_sbuf_ectx_brate, encoder_ctx->bitrate);
	//trace_val (te_sbuf_ectx_fsize, encoder_ctx->framesize);
	
	return TRUE;
}


/*******************************************************************************
 * Encoding
 ******************************************************************************/

/* Check that we have the needed encoder. Otherwise, initialize a new one... */
static int
check_encoder (struct sbuf * send_buf, unsigned int encoder_id)
{
	void	*encoder	= NULL;

	assert (send_buf		!= NULL);
	assert (encoder_id		!= VCODEC_INVALID_ID);
	assert (send_buf->ctx->mtu	>  0);
		
	if (send_buf->encoder_id != encoder_id) {

		/* Ok, the encoder is different to what we have... create a new one */
		if (vcodec_can_encode (encoder_id) || vcodec_can_packetize (encoder_id)) {

			/* Destroy previous encoder if neeeded */
			if (send_buf->encoder_id != VCODEC_INVALID_ID) {
				debug_msg ("Destroying previous encoder...\n");
				vcodec_encoder_destroy (send_buf->encoder_id,
							send_buf->encoder_state);
			}
			
			/* Create a new encoder */
			encoder = vcodec_encoder_create (encoder_id,
							 send_buf->ctx->mtu);
			if (encoder == NULL) {
				fprintf (stderr, "Error: creating new '%s' encoder for outgoing frames.\n", vcodec_get_name (encoder_id));
				return FALSE;
			}
			
			/* Switch to the new encoder */
			send_buf->encoder_id    = encoder_id;
			send_buf->encoder_state = encoder;			
		} else {
			fprintf (stderr, "Error: new encoder %s can not encode or packetize outgoing frames!.\n", vcodec_get_name (encoder_id));
			return FALSE;
		}
	}
	
	return TRUE;
}


/**********************************************************************************
 * Public API
 *********************************************************************************/

/* Initialize the sending buffer. Use a NULL context for automatic allocation. */
struct sbuf*
sbuf_init (struct sbuf_ctx_t *ctx, struct video_tx *tx, struct timeval *base_time)
{
	struct sbuf    *send_buf = NULL;
	
	send_buf = malloc(sizeof(struct sbuf));
	if (send_buf != NULL) {
		send_buf->magic		= SBUF_MAGIC;
		
		send_buf->first		= NULL;
		send_buf->first_out	= NULL;
		send_buf->last		= NULL;

		send_buf->frames_count	= 0;
		
		send_buf->unsent_frames	= 0;
		send_buf->unsent_frags	= 0;
		send_buf->unsent_bytes	= 0;

		send_buf->total_frames	= 0;
		send_buf->total_frags	= 0;
		send_buf->total_bytes	= 0;
				
		send_buf->in_rate	= 0;
		send_buf->comp_rate	= 0;
		send_buf->out_rate	= 0;
				
		send_buf->encoder_state	= NULL;
		send_buf->encoder_id	= VCODEC_INVALID_ID;

		send_buf->base_time	= base_time;

		send_buf->tx		= tx;
		
		/* Encoder context */
		send_buf->encoder_ctx = (struct vcodec_ctx_t*) malloc (sizeof (struct vcodec_ctx_t));
		if (send_buf->encoder_ctx == NULL) {
			fprintf (stderr, "Error creating the encoder context.\n");
			free (send_buf);
			return NULL;
		}

		/* Sending buffer context */
		if (ctx == NULL) {
			/* Create an empty context */
			send_buf->ctx = (struct sbuf_ctx_t*) malloc (sizeof (struct sbuf_ctx_t));
			if (send_buf->ctx != NULL) {
				send_buf->ctx->mtu		= SBUF_CTX_INIT_MTU;
				send_buf->ctx->max_out_ifi	= SBUF_CTX_INIT_MAX_OUT_IFI;
				send_buf->ctx->max_framelen	= SBUF_CTX_INIT_MAX_FRAMELEN;
				send_buf->ctx->max_frags	= SBUF_CTX_INIT_MAX_FRAGS;
				send_buf->ctx->max_frames	= SBUF_CTX_INIT_MAX_FRAMES;
			} else {
				fprintf (stderr, "Error creating the sending buffer context.\n");
				free (send_buf);
				return NULL;
			}
			sbuf_adjust_encoding_ctx (send_buf);
		} else {
			assert (ctx->max_frames	> 0);
			assert (ctx->max_frags	> 0);

			sbuf_set_context (send_buf, ctx);
		}
		
		/* IFIs */
		gettimeofday (&send_buf->last_in_time, NULL);
		gettimeofday (&send_buf->last_out_time, NULL);
		send_buf->in_ifi	= 0;
		send_buf->out_ifi	= 0;
		
		send_buf->avg_in_frame_len	= 0;
		send_buf->avg_out_frame_len	= 0;
		
		/* Create the sending buffer lock */
		if (sbuf_lock_init (send_buf) != 0) {
			fprintf (stderr, "Error creating the sending buffer mutex.\n");
			free (send_buf);
			return NULL;
		}	
	} else {
		fprintf (stderr, "Failed to allocate memory for sending buffer\n");
	}
	
	return send_buf;
}

/*
 * Terminates the sending buffer
 * Note: this is an entry point to the sbuf, so it is mutex-protected
 */
void
sbuf_done (struct sbuf *send_buf)
{
	struct sbuf_node	*node;
	struct timeval		ts;
	
	assert (send_buf	!= NULL);
	assert (send_buf->magic	== SBUF_MAGIC);
	
	sbuf_lock (send_buf);
	
	if (send_buf->unsent_frames != 0) {
		fprintf (stderr, "Error: destroying non-empty sending buffer.\n");
	}
	
	/* Mark all frames as sent and clean up */
	node = send_buf->first;
	while (node != NULL){
		node->sent = TRUE;
		node = node->next;
	}
	
	gettimeofday (&ts, NULL);
	
	sbuf_unlock (send_buf);		/* Unlock before calling mutex-protected */
	sbuf_remove (send_buf, ts);	
	sbuf_lock (send_buf);

	/* Free memory of temporal frames */	
	if (send_buf->encoder_state != NULL) {
		vcodec_encoder_destroy (send_buf->encoder_id,
					send_buf->encoder_state);
	}
	
	free (send_buf->encoder_ctx);

	sbuf_unlock (send_buf);
	sbuf_lock_done (send_buf);

	free (send_buf);
}

/*
 * Insert a frame in the sending buffer.
 * Note: this is an entry point to the sbuf, so it is mutex-protected
 */
bool
sbuf_insert (struct sbuf	*send_buf,
	     struct video_frame	*frame,
	     unsigned int	 frame_enc_id,
	     unsigned int	 encoder_id)
{
	struct encoded_frame	*enc_frame	= NULL;
	unsigned int		 enc_frame_len	= 0;
	void			*encoder_state	= NULL;
	struct vcodec_ctx_t	*ctx		= NULL;
	struct timeval		 t_now;
	unsigned int		 t_dif;
	t_usec			 ins_ts;
	bool			 res		= TRUE;
	
	assert (send_buf	!= NULL);
	assert (send_buf->magic	== SBUF_MAGIC);
	assert (frame		!= NULL);

	//trace (te_sbuf_insertion);
	
	sbuf_lock (send_buf);
	sbuf_validate (send_buf);

	/* Time difference since last insertion */
	t_dif = tv_usecs_since (&send_buf->last_in_time, &t_now);
	tv_cp (&send_buf->last_in_time, &t_now);

	if (frame->data_len > 0)  {

		ctx = send_buf->encoder_ctx;
		
		/*
		 * If there is enough space, encode or packetize the frame,
		 * depending on the match between both formats...
		 */
		enc_frame_len = vcodec_encoder_get_framelen (encoder_id);
		if (sbuf_can_insert (send_buf, enc_frame_len) == TRUE) {
			
			if (check_encoder (send_buf, encoder_id)) {
				
				/* Get a timestamp for the insertion */
				ins_ts = tv_usecs_since (send_buf->base_time, NULL);
				
				/* Encode the frame */
				enc_frame = vcodec_eframe_alloc (send_buf->encoder_id);
				if (enc_frame == NULL) {
					fprintf (stderr, "Error: allocating memory for encoding outgoing frames.\n");
					res = FALSE;
				} else {
					encoder_state = send_buf->encoder_state;
					
					/*
					 * Note: unlock before encoding, as these operations can
					 * take some time...
					 */
					sbuf_unlock (send_buf);
					if (frame_enc_id == encoder_id) {
						/* Just packetize the encoded frame */
						vcodec_packetize (encoder_id, encoder_state, frame, enc_frame);
					} else {
						/* Encode */
						vcodec_encode (encoder_id, encoder_state, frame, enc_frame, ctx);
					}
					sbuf_lock (send_buf);						

					/*
					 * Calculate the average length of uncompressed and
					 * compressed frames inserted in the sbuf.
					 */
					send_buf->avg_in_frame_len =
						AVG_FL_FACTOR * send_buf->avg_in_frame_len +
						(1.0 - AVG_FL_FACTOR) * frame->data_len;
					send_buf->avg_out_frame_len =
						AVG_FL_FACTOR * send_buf->avg_out_frame_len +
						(1.0 - AVG_FL_FACTOR) * enc_frame->datalen;

					/* Update insertion rates */
					send_buf->in_rate = SBUF_BR_FACTOR * send_buf->in_rate +
						(1.0 - SBUF_BR_FACTOR) * ((double)frame->data_len / usec2sec (t_dif));

					send_buf->comp_rate = SBUF_BR_FACTOR * send_buf->comp_rate +
						(1.0 - SBUF_BR_FACTOR) * ((double) enc_frame->datalen / usec2sec (t_dif));					

					/* Add this new encoded frame */
					sbuf_add (send_buf, enc_frame, ins_ts, frame->ts);
				}
			} else {
				debug_msg ("Error adjusting sbuf encoder.\n");
				res = FALSE;
			}
		}

		/*
		 * Update the inter-frames interval for incoming frames.
		 *
		 * It is computed even if the frame is never inserted, as it will provide
		 * a measure of the insertion process.
		 */
		send_buf->in_ifi = SBUF_IFI_FACTOR * send_buf->in_ifi + (1.0 - SBUF_IFI_FACTOR) * t_dif;		
	} else {
		debug_msg ("WARNING: trying to insert a not null frame, but with framelen=0.\n");
	}
	
	/* Trace info... */
	//trace_val_at (&t_now, te_sbuf_in_ifi,		send_buf->in_ifi);
	//trace_val_at (&t_now, te_sbuf_inrate,		sbuf_get_in_rate (send_buf));		
	//trace_val_at (&t_now, te_sbuf_comprate,		sbuf_get_comp_rate (send_buf));
	//trace_val_at (&t_now, te_sbuf_insert_bytes,	frame->data_len);
	//trace_val_at (&t_now, te_sbuf_insert_c_bytes,	enc_frame->datalen);
	//trace_val_at (&t_now, te_sbuf_avg_in_flen,	send_buf->avg_in_frame_len);
	//trace_val_at (&t_now, te_sbuf_avg_out_flen,	send_buf->avg_out_frame_len);
	//trace_val_at (&t_now, te_sbuf_unsent_length,	send_buf->unsent_frags);
	//trace_val_at (&t_now, te_sbuf_unsent_bytes,	send_buf->unsent_bytes);
	//trace_val_at (&t_now, te_sbuf_total_length,	send_buf->total_frags);
	//trace_val_at (&t_now, te_sbuf_sent_length,	send_buf->total_frags - send_buf->unsent_frags);

	sbuf_validate (send_buf);
	sbuf_unlock (send_buf);

	return res;
}


/*
 * Retrieve next fragment to be sent.
 * Note: this is an entry point to the sbuf, so it is mutex-protected
 */
struct encoded_data*
sbuf_get_next (struct sbuf	*send_buf)
{
	struct encoded_data	*edata		= NULL;
	struct sbuf_node	*fo		= NULL;
	unsigned long		 datalen;
	struct timeval		 t_now;
	unsigned int		 t_dif;

	assert (send_buf	 != NULL);
	assert (send_buf->magic	 == SBUF_MAGIC);

	sbuf_lock (send_buf);
beg:
	fo = send_buf->first_out; 
	if ( fo != NULL) {

		sbuf_check_arrival_time (send_buf, fo);
		
		if (fo->sent == TRUE) {
			sbuf_go_forward (send_buf);
			goto beg;		
		} else {
			/* Get the first unsent node... */
			edata = vcodec_eframe_get (fo->enc_frame->id, fo->enc_frame, fo->frags_sent);
			if (edata != NULL) {				
				/* Update pointers */
				fo->frags_sent++;
				fo->bytes_sent	+= edata->data_len;
				
				edata->rtp_ts = fo->rtp_ts;

				//trace_val (te_sbuf_get_bytes,		edata->data_len);
				//trace_val (te_sbuf_unsent_bytes,	send_buf->unsent_bytes);

				/* Some long assertions... */
				//assert (send_buf->unsent_frags >= fo->enc_frame->len - fo->frags_sent);

				assert (fo->enc_frame->datalen	>= fo->bytes_sent);
				assert (fo->enc_frame->len	>= fo->frags_sent);
			} else {
				//assert (fo->enc_frame->datalen	== fo->bytes_sent);
				assert (fo->enc_frame->len	== fo->frags_sent);
								
				/*
				 * If we get NULL, then there were no more fragments in
				 * the current frame. Then move formard and try with the next
				 * frame.
				 *
				 * Update the inter-frames interval and rate for outgoing frames.
				 *
				 * The inter-frames interval is computed using a moving average value,
				 * weighting the current value as well as the previous ones. 
				 */
				t_dif = tv_usecs_since (&send_buf->last_out_time, &t_now);
				tv_cp (&send_buf->last_out_time, &t_now);
				
				datalen = send_buf->first_out->enc_frame->datalen;
				
				send_buf->out_ifi = SBUF_IFI_FACTOR * send_buf->out_ifi +
					(1.0 - SBUF_IFI_FACTOR) * t_dif;
				
				send_buf->out_rate = SBUF_BR_FACTOR * send_buf->out_rate +
					(1.0 - SBUF_BR_FACTOR) * (datalen / usec2sec (t_dif));

				/* Some trace info */
				//trace_val_at (&t_now, te_sbuf_out_ifi,		send_buf->out_ifi);
				//trace_val_at (&t_now, te_sbuf_outrate,		sbuf_get_out_rate (send_buf));
				
				sbuf_go_forward (send_buf);
				goto beg;
			}			
		} /* if sent */
	} /* if fo */
	sbuf_unlock (send_buf);
	
	//trace_val (te_sbuf_unsent_length, send_buf->unsent_frags);
	//trace_val (te_sbuf_total_length, send_buf->total_frags);
	//trace_val (te_sbuf_sent_length, send_buf->total_frags - send_buf->unsent_frags);
	
	return edata;
}

/*
 * Return the length of the sending buffer (in fragments of frames), considering
 * only unsent fragments.
 * Note: this is an entry point to the sbuf, so it is mutex-protected
 */
unsigned int
sbuf_length  (struct sbuf *send_buf)
{
	unsigned int len	= 0;
	
	assert (send_buf	!= NULL);
	assert (send_buf->magic	== SBUF_MAGIC);
	
	sbuf_lock (send_buf);	
	if (send_buf->first_out != NULL) {
		len = send_buf->unsent_frags - send_buf->first_out->frags_sent;
	}
	sbuf_unlock (send_buf);
	
	return len;
}

/*
 * Return the amount of unsent bytes.
 * Note: this is an entry point to the sbuf, so it is mutex-protected.
 */
unsigned int
sbuf_length_bytes  (struct sbuf *send_buf)
{
	unsigned int len	= 0;
	
	assert (send_buf	!= NULL);
	assert (send_buf->magic	== SBUF_MAGIC);
	
	sbuf_lock (send_buf);	
	if (send_buf->first_out != NULL) {
		len = send_buf->unsent_bytes - send_buf->first_out->bytes_sent;
	}
	sbuf_unlock (send_buf);
	
	return len;
}


bool
sbuf_empty (struct sbuf *send_buf)
{
	return (bool)(sbuf_length (send_buf) == 0);
}


/*
 * Get the sending buffer context.
 * Using the sending buffer context we can control parameters such as how much
 * data is stored in the sending buffer...
 */
void
sbuf_get_context (struct sbuf *send_buf, struct sbuf_ctx_t *ctx)
{
	ctx->max_out_ifi	= send_buf->ctx->max_out_ifi;
	ctx->max_framelen	= send_buf->ctx->max_framelen;
	ctx->max_frags		= send_buf->ctx->max_frags;
	ctx->max_frames		= send_buf->ctx->max_frames;
	ctx->mtu		= send_buf->ctx->mtu;
}

/*
 * Set the sending buffer context.
 * This function should be called by other parts of the system for setting
 * the limits of the sending buffer: max number of frames, ideal bitrate...
 */
void
sbuf_set_context (struct sbuf *send_buf, struct sbuf_ctx_t *ctx)
{
	assert (send_buf	!= NULL);
	assert (send_buf->magic	== SBUF_MAGIC);
	assert (ctx		!= NULL);

	assert (ctx->max_out_ifi	> 0);
	assert (ctx->max_framelen	> 0);
	assert (ctx->max_frags		> 0);
	assert (ctx->max_frames		> 0);
	assert (ctx->mtu		> 0);

	send_buf->ctx->max_out_ifi	=	ctx->max_out_ifi;
	send_buf->ctx->max_framelen	=	ctx->max_framelen;
	send_buf->ctx->max_frags	=	ctx->max_frags;
	send_buf->ctx->max_frames	=	ctx->max_frames;
	send_buf->ctx->mtu		=	ctx->mtu;
	
	sbuf_adjust_encoding_ctx (send_buf);
}

/*
 * Remove from the sending buffer previously encoded frames that have been sent
 * or that have passed their sending time .
 * Note: this is an entry point to the sbuf, so it is mutex-protected
 */
bool
sbuf_remove (struct sbuf	*send_buf,
	     struct timeval	 curr_time)
{
	struct sbuf_node	*curr,
				*next;
	unsigned int		 counter	= 0;	/* Assertion variable */

	UNUSED (curr_time);

	assert (send_buf	!= NULL);
	assert (send_buf->magic	== SBUF_MAGIC);
	
	sbuf_lock (send_buf);
	if (send_buf->total_frames > 0) {
		
		sbuf_validate (send_buf);
		
		//trace (te_sbuf_removal);

		counter = send_buf->total_frames;
		
		/* Start in the first node and look for sent frames */
		curr = send_buf->first;
		while (curr != NULL) {

			assert (curr->magic	== SBUF_NODE_MAGIC);
			assert (curr->enc_frame != NULL);

			next = curr->next;				
			
			/* Delete old frames */
			if (curr->sent == TRUE) {
				//trace_val (te_sbuf_remov_bytes, curr->enc_frame->datalen);
				
				sbuf_delete (send_buf, curr);
			}
			
			curr = next;
			counter--;
		}		
		sbuf_validate (send_buf);
	}
	sbuf_unlock (send_buf);		

	assert (counter == 0);
	
//	trace_val (te_sbuf_unsent_length, send_buf->unsent_frags);
//	trace_val (te_sbuf_unsent_bytes, send_buf->unsent_bytes);
//	trace_val (te_sbuf_total_length, send_buf->total_frags);
//	trace_val (te_sbuf_sent_length, send_buf->total_frags - send_buf->unsent_frags);

	return TRUE;
}

t_usec
sbuf_get_in_ifi	(struct sbuf	*send_buf)
{
	assert (send_buf != NULL);

	return send_buf->in_ifi;
}

t_usec
sbuf_get_out_ifi (struct sbuf *send_buf)
{
	assert (send_buf != NULL);
	
	return send_buf->in_ifi;
}

unsigned int
sbuf_get_in_rate (struct sbuf *send_buf)
{
	return send_buf->in_rate;
}

unsigned int
sbuf_get_comp_rate (struct sbuf *send_buf)
{
	return send_buf->comp_rate;
}

unsigned int
sbuf_get_out_rate (struct sbuf *send_buf)
{
	return send_buf->out_rate;
}

unsigned long
sbuf_get_avg_in_frame_len (struct sbuf *send_buf)
{
	assert (send_buf != NULL);
	
	return send_buf->avg_in_frame_len;
}

unsigned long
sbuf_get_avg_out_frame_len (struct sbuf *send_buf)
{
	assert (send_buf != NULL);
	
	return send_buf->avg_out_frame_len;
}

