/*
 * FILE:   trace.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.16 $
 * $Date: 2007/01/08 22:22:22 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace.h"

#ifdef DEBUG_TRACE

#define MAX_TRACE_EVENT_T	255

/* Table entry */
struct trace_table_entry_t {
	trace_output_t	 output_method;
	
	FILE*		 fd;
	trace_fun_t	 fun;
	
	char		*name;
	char		*desc;
	
};


/* Table of events. */
static struct {
	struct trace_table_entry_t	table [MAX_TRACE_EVENT_T];
	struct timeval			start_time;
	int				rel_time;	
} tr_state;

#endif

/*
 * Format and dump (if necessary) the last event(s).
 * 
 * This function could store all the events corresponding to the same
 * moment in time and, when a new time is detected, dump all old data with
 * the appropiate format.
 */ 
#ifdef DEBUG_TRACE
static void
record_event (FILE		*fd,
	      struct timeval	*ts,
	      char		*name,
	      char		*desc,
	      trace_event_t	 ev,
	      trace_value_t	 val,
	      int		 use_rel)
{
	double rel_time;
	
	UNUSED (ev);

	printf ("\n tracing on: 1");
	
	if (tr_state.start_time.tv_sec == 0 && tr_state.start_time.tv_usec == 0) {
		tr_state.start_time.tv_sec	= ts->tv_sec;
		tr_state.start_time.tv_usec	= ts->tv_usec;
	}
	
	if (use_rel) {
		rel_time = (ts->tv_sec - tr_state.start_time.tv_sec) +
				(ts->tv_usec - tr_state.start_time.tv_usec)/1000000.0;
		
#ifdef TRACE_USE_DESC	
		fprintf (fd, "%f %s %f '%s'\n", rel_time, name, (double)val, desc);
#else
		UNUSED (desc);
		fprintf (fd, "%f %s %f\n", rel_time, name, (double)val);
#endif		
	} else {
#ifdef TRACE_USE_DESC	
		fprintf (fd, "%ld.%06ld %s %f '%s'\n", (long)ts->tv_sec,
			 (long)ts->tv_usec, name, (double)val, desc);
#else
		UNUSED (desc);
		fprintf (fd, "%ld.%06ld %s %f\n", (long)ts->tv_sec, (long)ts->tv_usec,
			 name, (double)val);
#endif
	}
}
#endif


/*
 * Public API
 */
void
trace_init (trace_output_t	  out,
	    void		 *def,
	    char		**names,
	    char		**descs,
	    unsigned int	  len,
	    int			  rel_time)
{
#ifdef DEBUG_TRACE	
	unsigned int	 i;
	FILE		*default_fd;
	FILE		*fd	= NULL;
	trace_fun_t	 fun	= NULL;

	 printf ("\n tracing on: 2");
	
	if (def != NULL) {
		/* Initialize the table with default values */
		if(out == t_file) {
			
			/* Open the default log file */
			default_fd = (FILE*) def;
			if (default_fd == NULL) {
				fprintf (stderr, "Error: creating default trace file: %s.\n", strerror(errno));
				return;
			} else {
				fun	= NULL;
				fd	= default_fd;
			}
			
		} else if (out == t_function) {
			fun	= (trace_fun_t)def;
			fd	= NULL;
		}
	}

	for (i=0; i < MAX_TRACE_EVENT_T; i++) {
		tr_state.table[i].output_method		= out;
		tr_state.table[i].fun			= fun;
		tr_state.table[i].fd			= fd;
		
		/* If there is a description available, use it */
		if (names != NULL && descs != NULL && i < len) {
			tr_state.table[i].name = names[i];
			tr_state.table[i].desc = descs[i];
		} else {
			tr_state.table[i].name = "";
			tr_state.table[i].desc = "";
		}
	}
	
	tr_state.start_time.tv_sec	= 0;
	tr_state.start_time.tv_usec	= 0;
	tr_state.rel_time		= rel_time;
#else
	UNUSED (def);
	UNUSED (names);
	UNUSED (descs);
	UNUSED (out);
	UNUSED (len);
	UNUSED (rel_time);
#endif
}

/*
 * Register a new event type definition.
 */
void
trace_reg (trace_event_t	 ev,
	   char			*name,
	   char			*desc,
	   trace_output_t	 out,
	   void			*target)
{
#ifdef DEBUG_TRACE
	unsigned int index	= (unsigned int) ev;
	
	assert(index  <  MAX_TRACE_EVENT_T);
	assert(target != NULL);

	 printf ("\n tracing on: 3");
	
	if (out == t_file) {
		tr_state.table[index].fd = (FILE*) target;
		if (tr_state.table[index].fd == NULL) {
			
			fprintf(stderr, "Error: opening trace file: %s\n", strerror(errno));
			
			tr_state.table[index].output_method	= t_none;
			tr_state.table[index].fd		= NULL;
			
		} else {
			tr_state.table[index].output_method	= t_file;
			tr_state.table[index].name		= name;
			tr_state.table[index].desc		= desc;

		}
	} else if (out == t_function) {
		tr_state.table[index].output_method	= t_function;
		tr_state.table[index].fun		= (trace_fun_t)target;
		tr_state.table[index].name		= name;
		tr_state.table[index].desc		= desc;
	}
#else
	UNUSED (name);
	UNUSED (ev);
	UNUSED (desc);
	UNUSED (out);
	UNUSED (target);
#endif
}

void
trace_unreg (trace_event_t	 ev)
{
#ifdef DEBUG_TRACE
	unsigned int index	= (unsigned int) ev;
	
	assert(index  <  MAX_TRACE_EVENT_T);
	 printf ("\n tracing on: 4");
	
	tr_state.table[index].output_method	= t_none;
	tr_state.table[index].fun		= NULL;
	tr_state.table[index].fd		= NULL;
	tr_state.table[index].name		= NULL;
	tr_state.table[index].desc		= NULL;
#else
	UNUSED (ev);
#endif
}


void
trace (trace_event_t te)
{
#ifdef DEBUG_TRACE	
	trace_val (te, 1);
	 printf ("\n tracing on: 4");
#else
	UNUSED (te);
#endif
}

void
trace_at (struct timeval* ts, trace_event_t ev)
{
#ifdef DEBUG_TRACE
	trace_val_at (ts, ev, 1);
#else
	UNUSED (ev);
	UNUSED (ts);
#endif
}

void
trace_val (trace_event_t ev, trace_value_t val)
{
#ifdef DEBUG_TRACE	
	struct timeval ts;
	
	gettimeofday (&ts, NULL);
	trace_val_at (&ts, ev, val);
#else
	UNUSED (ev);
	UNUSED (val);
#endif
}

void
trace_val_at (struct timeval* ts, trace_event_t ev, trace_value_t val)
{
#ifdef DEBUG_TRACE	
	unsigned int index	= (unsigned int) ev;

	assert (index < MAX_TRACE_EVENT_T);

	if (tr_state.table[index].output_method == t_file && tr_state.table[index].fd != NULL )
		record_event (tr_state.table[index].fd, ts,
			      tr_state.table[index].name,
			      tr_state.table[index].desc, 
			      ev, val, tr_state.rel_time);
	else if (tr_state.table[index].output_method == t_function && tr_state.table[index].fun != NULL )
		(tr_state.table[index].fun)(ts, ev, val);
#else
	UNUSED (ts);
	UNUSED (ev);
	UNUSED (val);
#endif
}

void
trace_done (void)
{
#ifdef DEBUG_TRACE	
	unsigned int i;
	
	for(i=0; i < MAX_TRACE_EVENT_T; i++) {
		if (tr_state.table[i].output_method == t_file)
			fclose (tr_state.table[i].fd);
	}
#else
#endif
}
