/*
 * FILE:    transmit.c
 * AUTHORS: Colin Perkins <csp@csperkins.org>
 *          Ladan Gharai  <ladan@isi.edu>
 *          Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005-2006 University of Glasgow
 * Copyright (c) 2001-2004 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.71 $
 * $Date: 2007/01/09 18:49:08 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace_ug.h"
#include "trace.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"	/* For struct coded_data */
#include "rtp/sbuf.h"
#include "tv.h"
#include "video_capture.h"
#include "video_codec.h"
#include "transmit.h"

/* Switches: Should we... */
//#define USE_DUMMY_PCKS		/* ... send, even if there is no data?
                                        /* This causes problems if enabled with HDTV - see 
					 * commit for for revision 1.69 of this file [csp]*/ 
//#define USE_DYN_MAXFRAMES	/* ... recalculate the max # of frames in the sbuf? */
#define USE_R_FACTOR		/* ... weight the frame length using the current sendrate dif? */ 

/* Constants */
#define TRANSMIT_MAGIC			0xe80ab15f

#define TX_NODATA_FACTOR		0.8
#define TX_MIN_FRAMELEN			8000	/* Minimum frame length (compressed)	*/
#define TX_INIT_RFACTOR			2	/* Inial r_factor			*/

#define DEFAULT_MAX_FRAMES		2	/* Default max # frames in the sbuf	*/
#define DEFAULT_INIT_MAX_FRAMES		1	/* ... and initial value		*/

#define MAX_FRAME_DELAY_PBUF		PBUF_PLAY_DELAY		/* Delay in the playout buffer		*/


#ifdef USE_DUMMY_PCKS
/* Dummy packet used when we don't have data to transmit... */
static struct encoded_data dummy_pck = { 0, 0, NULL, 0, NULL, 0, 0 };
#endif

/*
 * Transmit info
 */
struct video_tx {
	uint32_t		 magic;

	struct tx_ctx		 ctx;			/* TX context		*/

	uint16_t		 seq;
	
	struct timeval		*base_time;
	
	struct rtp		*rtp_session;		/* RTP			*/
	t_usec			 rtp_last_ts;
	
	struct sbuf		*send_buf;		/* Sending buffer	*/
	struct sbuf_ctx_t	 send_buf_ctx;
	
	struct vidcap		*capture_device;
	
	pthread_t		 tx_grab_thread,	/* Threads		*/
				 tx_send_thread;
	
	unsigned int		 encoder_id;
	unsigned int		 decoder_id;

	double			 nodata_rate;
	struct timeval		 nodata_tv;
	
	struct cc_sndr_t	*cc_sndr;		/* Congestion control	*/
	
	unsigned int		*all_working;		/* Working flags	*/
	unsigned int		 tx_working;
};


static void	tx_stop_all	(struct video_tx*);


/*******************************************************************************
 * Utility functions
 ******************************************************************************/

/*
 * Return the expected transmission time for a frame.
 * This function is used for estimating if the receiver will process a frame in time.
 */
t_usec
tx_trans_delay (struct video_tx *tx, unsigned long datalen)
{
	t_usec		usecs_now	= 0,
			current_rtt	= 0,
			net_delay	= 0;
	unsigned long	current_srate	= 0;
	
	usecs_now	= tv_usecs_since (tx->base_time, NULL);
	
	current_rtt	= cc_sndr_get_avg_rtt (tx->cc_sndr);
	current_srate	= MAX (cc_sndr_get_sendrate (tx->cc_sndr, usecs_now), tx->ctx.mtu);

	net_delay	= ((double)current_rtt / 2.0);
	return  net_delay + ((double)datalen / (double)current_srate) + MAX_FRAME_DELAY_PBUF;
}


/*******************************************************************************
 * Threads
 ******************************************************************************/

/*
 * Update the sending buffer context.
 * This is the glue between the CC system and the sending buffer: here we set
 * the constraints that must be enforced in the sbuf.
 * We update the sending buffer context, taking into account the congestion
 * control info. We could set:
 *
 * 1. Max frames.
 *    Using the current RTT, we could set the maximum number of frames in the
 *    sending buffer, in order to keep the global delay < 100-150ms.
 *
 * 2. Max frame length.
 *    Setting a sbuf bitrate equal to the sending rate of the CC system would
 *    produce a constant-length sending buffer.
 */
static void
tx_send_update_sbuf_ctx (struct video_tx	*tx,
			 uint32_t		 usecs_now,
			 uint32_t		*last_known_sendrate,
			 uint32_t		*last_known_rtt)
{
	double			r_factor		= 1;
	double			current_cc_sendrate	= 0,	/* bytse/sec */
				current_out_rate	= 0,
				current_in_rate		= 0;
	unsigned long		current_in_flen_avg	= 0,
				current_out_flen_avg	= 0,
				current_sbuf_len	= 0;
	t_usec			in_ifi			= 0,
				current_rtt		= 0,
				net_delay		= 0,
				max_sbuf_delay		= 0,
				max_out_ifi		= 0;
	unsigned int		max_frames		= DEFAULT_MAX_FRAMES;
	unsigned long		max_framelen		= 0,
				max_framelen_p		= 0;
	struct timeval		tv_now;

	assert (tx->ctx.mtu > 0);

	current_rtt		= cc_sndr_get_avg_rtt (tx->cc_sndr);
	current_cc_sendrate	= MAX (cc_sndr_get_sendrate (tx->cc_sndr, usecs_now), tx->ctx.mtu);
	
	/*
	 * If the RTT or send rates change, we must re-calculate some things...
	 */
	if (current_rtt	!= *last_known_rtt || current_cc_sendrate	!= *last_known_sendrate) {

		net_delay		= current_rtt / 2.0;
		in_ifi			= sbuf_get_in_ifi (tx->send_buf);

		current_sbuf_len	= sbuf_length_bytes (tx->send_buf);

		current_in_flen_avg	= sbuf_get_avg_in_frame_len (tx->send_buf);		
		current_out_flen_avg	= sbuf_get_avg_out_frame_len (tx->send_buf);
		
		current_in_rate		= sbuf_get_comp_rate (tx->send_buf);
		current_out_rate	= sbuf_get_out_rate (tx->send_buf);
		
		assert (current_out_flen_avg > 0);
		

		/*
		 * 1. Update the maximum delay
		 *
		 * The following must be enforced:
		 * pbuf_delay + sbuf_delay + net_delay < MAX_FRAME_DELAY
		 * So we must calculate the maximum amount of time for
		 * transmitting a frame. When it is impossible, just use a minimum
		 * delay...
		 */
		if (MAX_FRAME_DELAY > MAX_FRAME_DELAY_PBUF + net_delay) {
			max_sbuf_delay = MAX_FRAME_DELAY - MAX_FRAME_DELAY_PBUF - net_delay;
		} else {
			max_sbuf_delay = MIN_FRAME_DELAY_SBUF;
		}
			
		assert (max_sbuf_delay > 0);

		/*
		 * 2. Update the maximum queue length
		 */
#ifdef USE_DYN_MAXFRAMES
		if (in_ifi > 0) {
			max_frames = MAX (max_sbuf_delay / in_ifi, DEFAULT_MAX_FRAMES);
		}
#endif
		assert (max_frames > 0);
		
		/*
		 * 3. Update the maximum frame length
		 */
#ifdef USE_R_FACTOR
		if (current_out_rate > 0) {
			/*
			 * We calculate an r_factor: an acceleration factor
			 * obtained as the ratio between the input and the output
			 * of the buffer. In this way, if we are getting out faster
			 * than expected, we will compress less the frames, and
			 * we will have more data for transmitting...
			 */
			r_factor = (double) current_cc_sendrate / (double) current_out_rate;			
		} else{
			r_factor = TX_INIT_RFACTOR;
		}
#else
		r_factor = 1;
#endif
		assert (r_factor > 0);

		max_framelen_p	= r_factor * usec2sec (max_sbuf_delay) * current_cc_sendrate;		
		max_framelen	=  MAX((double)max_framelen_p / (double)max_frames, TX_MIN_FRAMELEN);
		assert (max_framelen > 0);

		/*
		 * 4. Update the maximum output inter-frame interval
		 */
		max_out_ifi = (double)max_framelen / usec2sec (current_cc_sendrate);
		
		assert (max_out_ifi > 0);

		/*
		 * Update the context.
		 */
		tx->send_buf_ctx.max_out_ifi	= max_out_ifi;
		tx->send_buf_ctx.max_framelen	= max_framelen;
		tx->send_buf_ctx.max_frames	= MAX (max_frames, 1);
		tx->send_buf_ctx.max_frags	=
			(tx->send_buf_ctx.max_frames * MAX(current_out_flen_avg, TX_MIN_FRAMELEN))
			/ tx->send_buf_ctx.mtu;
		
		/* Update 'last known' values */
		*last_known_sendrate	= current_cc_sendrate;
		*last_known_rtt		= current_rtt;
		
		/* CC info */
		gettimeofday (&tv_now, NULL);
		//trace_val_at (&tv_now, te_cc_x,		current_cc_sendrate);
		//trace_val_at (&tv_now, te_cc_ipi,	cc_sndr_get_ipi (tx->cc_sndr));
		//trace_val_at (&tv_now, te_cc_p,		cc_sndr_get_p (tx->cc_sndr));
		//trace_val_at (&tv_now, te_cc_rtt,	current_rtt);		

		//trace_val_at (&tv_now, te_tx_rfactor,	r_factor);		
		//trace_val_at (&tv_now, te_tx_maxflen,	max_framelen);		

		assert (tx->send_buf_ctx.max_out_ifi	> 0);
		assert (tx->send_buf_ctx.max_framelen	> 0);
		assert (tx->send_buf_ctx.max_frames	> 0);
		assert (tx->send_buf_ctx.max_frags	> 0);

		sbuf_set_context (tx->send_buf, &tx->send_buf_ctx);
	}
}


static uint16_t
tx_send_pck (struct video_tx *tx, uint8_t pt_p, struct timeval tv_now)
{
	struct encoded_data	*node		= NULL;
	int			 nodata		= FALSE;
	unsigned int		 t_dif;
	uint8_t			 pt		= pt_p;
	
	node = sbuf_get_next (tx->send_buf);

#ifdef USE_DUMMY_PCKS	
	if (node == NULL) {
		assert (sbuf_empty (tx->send_buf));

		nodata = TRUE;
		
		/* Ok, no data, do we use a dummy packet... */
		node = &dummy_pck;
		
		/* Fill the sequence number and the TS */
		node->rtp_ts	= tx->rtp_last_ts;
		node->seqno	= tx->seq;
		
		if (node->data == NULL) {
			node->data	= malloc (tx->ctx.mtu);
			node->data_len	= tx->ctx.mtu;
		}

		/* Use a dummy PT to mark the packet */
		pt = PBUF_DUMMY_PT;

		//trace_at (&tv_now, te_tx_nodata);
		//trace_val_at (&tv_now, te_tx_nodata_bytes, node->data_len);
	}
#endif		
	
	/* Send the packet */
	if (node != NULL) {
		rtp_send_data_hdr (tx->rtp_session, node->rtp_ts, pt,
				   node->mark, 0, 0,
				   (unsigned char*) node->header, node->header_len,
				   (unsigned char*) node->data, node->data_len,
				   NULL, 0, 0);
		
		//trace_at     (&tv_now, te_tx_send);
		//trace_val_at (&tv_now, te_tx_send_bytes,	tx->ctx.mtu);
		//trace_val_at (&tv_now, te_tx_frag_seq,		node->seqno);	
		
		tx->rtp_last_ts = node->rtp_ts;
		tx->seq		= tx->seq + 1;		
	} else {
		nodata = TRUE;

		//trace_at (&tv_now, te_tx_nodata);
		//trace_val_at (&tv_now, te_tx_nodata_bytes, tx->ctx.mtu);		
	}

	if (nodata == TRUE) {
		t_dif = tv_diff (tv_now, tx->nodata_tv);
		tx->nodata_rate	= TX_NODATA_FACTOR * tx->nodata_rate +
				(1.0 - TX_NODATA_FACTOR) * ((double)tx->ctx.mtu/ (double)t_dif);
		
		tv_cp (&tx->nodata_tv, &tv_now);
	}
	
	return tx->seq;
}

/*
 * Spacing and transmission of packets.
 */
static void*
tx_send_thread (void *arg)
{
	struct video_tx		*tx			= (struct video_tx*) arg;
	uint8_t			 pt;
	unsigned int		 frags_transmitted	= 0;
	t_usec			 usecs_now,
				 usecs_prev,
				 usecs_sendtime,
				 usecs_sleep;
	uint32_t		 current_rtt		= CC_NULL_RTT,
				 current_x		= 0;
	struct timeval		 tv_now;

	assert (tx		!= NULL);
	assert (tx->magic	== TRANSMIT_MAGIC);
	assert (tx->rtp_session	!= NULL);
	assert (tx->cc_sndr	!= NULL);


	/* First, wait until we have something to send... */
	while (sbuf_empty (tx->send_buf) && tx_working (tx)) {
		tv_sleep_for (TV_MIN_SLEEP);
	}

	usecs_prev = tv_usecs_since (tx->base_time, NULL);

	/* The PT is not going to change... or it is? */
	pt = vcodec_get_payload (tx->encoder_id);

	while (tx_working (tx)) {
		/*
		 * 1. Congestion control: try to send, otherwise sleep...
		 */
		usecs_now = tv_usecs_since (tx->base_time, &tv_now);

		if (cc_sndr_send_ok (tx->cc_sndr, usecs_now)) {
			/*
			 * 2. Transmit the packet.
			 */
			tx_send_pck (tx, pt, tv_now);

			/*
			 * 3. Record the delivery and update the context.
			 */
			usecs_now = tv_usecs_since (tx->base_time, &tv_now);

			cc_sndr_send_done (tx->cc_sndr, usecs_now);

			tx_send_update_sbuf_ctx (tx, usecs_now, &current_x, &current_rtt);			

			/* Some trace info */
			//trace_val_at (&tv_now, te_tx_ipi, usecs_now - usecs_prev);			
			usecs_prev = usecs_now;

			frags_transmitted++;
		} else {
			/* Don't send yet: sleep for some time */
			usecs_sendtime	= cc_sndr_get_sendtime (tx->cc_sndr);

			if (usecs_sendtime > usecs_now) {
				usecs_sleep = usecs_sendtime - usecs_now;

				//trace_val (te_tx_sleep, usecs_sleep);
				
				tv_sleep_for (usecs_sleep);
			}				
		}
		
	} /* while working */

	tx_stop_all (tx);

	return NULL;
}


/*
 * Grabbing & queueing thread.
 * The transmission thread grabs images from the capture device, decodes it if
 * necessary, and inserts the decoded frame in the sending buffer.
 */
static void *
tx_grab_thread (void *arg)
{
	struct video_frame	*grab_frame	= NULL,
				*dec_frame	= NULL;
	struct video_tx		*tx		= (struct video_tx*) arg;
	void			*decoder	= NULL;
	unsigned int		 data_len;
	struct timeval		 t_now, t_prev;
		
	assert (tx			!= NULL);
	assert (tx->capture_device	!= NULL);
	assert (tx->rtp_session		!= NULL);
	
	debug_msg ("Starting capture thread\n");
		
	/*
	 * 1. Decoding of grabbed frames.
	 *
	 * If they encoder and decoder don't match, we'll need to decode grabbed 
	 * frames and encode again (in the sending buffer).
	 *
	 * Note: null codecs are represented with a VCODEC_INVALID_ID. For instance,
	 * a raw capture device does not need to be decoded, so its decoder_id
	 * should be VCODEC_INVALID_ID...
	 */
	if (tx->decoder_id != tx->encoder_id) {
		
		/* If needed, create the decoder for decoding grabbed frames */
		if (tx->decoder_id != VCODEC_INVALID_ID) {
			
			if (vcodec_can_decode (tx->decoder_id) == FALSE) {
				fprintf (stderr, "Error: decoder can not decode grabbed frames.\n");
				return NULL;	
			}			
			
			decoder = vcodec_decoder_create (tx->decoder_id);
			if (decoder == NULL) {
				fprintf (stderr, "Error: creating decoder for grabbed frames.\n");
				return NULL;	
			}
			
			/* Create the frame for decoding grabbed images */
			data_len  = vcodec_decoder_get_framelen (tx->decoder_id, decoder);
			dec_frame = video_frame_alloc (data_len);
			if (dec_frame == NULL) {
				fprintf (stderr, "Error: allocating memory for decoding grabbed frames.\n");
				return NULL;	
			} else {
				display_size_t	size;
				
				size = vcodec_decoder_get_size (tx->decoder_id, decoder);
				
				dec_frame->width       = display_get_width (size);
				dec_frame->height      = display_get_height (size);
				dec_frame->colour_mode = vcodec_decoder_get_cm (tx->decoder_id, decoder);
			}
		} else {
			debug_msg ("Capture frames do not need to be decoded, or invalid decoder found...\n");
		}
	}

	gettimeofday (&t_prev, NULL);

	/*
	 * 2. Main loop: decode (if needed) and insert in the sending buffer.
	 */
	while (tx_working (tx)) {

		/* Clean old frames in the sending buffer */
		gettimeofday (&t_now, NULL);
		sbuf_remove (tx->send_buf, t_now);

		/* Capture, decode (if needed), encode and transmit a frame... */
		grab_frame = vidcap_grab (tx->capture_device);

		//trace_val_at (&t_now, te_tx_grab_ifi, tv_diff_usec (t_now, t_prev));
		tv_cp (&t_prev, &t_now);

		if (grab_frame != NULL) {

			if (decoder != NULL) {
				/* We need to decode the grabbed frame */
				vcodec_decode_raw (tx->decoder_id, decoder,
						   grab_frame, dec_frame);

				if (dec_frame != NULL) {
					sbuf_insert (tx->send_buf, dec_frame,
						     tx->decoder_id,
						     tx->encoder_id);
				}
			} else {
				/* Just insert the grabbed frame */
				sbuf_insert (tx->send_buf, grab_frame,
					     tx->decoder_id, tx->encoder_id);
			}
			
			free (grab_frame);
		}
	}
	
	/*
	 * 3. Thread exit: cleanup
	 */
	if (decoder != NULL) {
		vcodec_decoder_destroy (tx->decoder_id, decoder);
		
		if (dec_frame->data != NULL)
			free (dec_frame->data);
		
		free (dec_frame);
	}
	
	tx_stop_all (tx);

	return NULL;
}


/*******************************************************************************
 * Public API
 ******************************************************************************/

/* Initialization off the transmission subsytem */
struct video_tx*
tx_init (struct tx_ctx		*ctx,
	 struct rtp		*rtp_session,
	 struct vidcap		*capture_device,
	 unsigned int		 encoder_id,
	 unsigned int		 decoder_id,
	 struct timeval		*base_time,
	 struct cc_sndr_t	*cc_sndr,
	 unsigned int		*working)
{
	struct video_tx		*tx		= NULL;
	
	assert (ctx		!= NULL);
	assert (ctx->mtu	>  0);
	assert (ctx->fps	>  0);
	assert (ctx->max_framelen	>  0);
	assert (base_time	!= NULL);
	assert (capture_device	!= NULL);
	assert (rtp_session	!= NULL);
	assert (cc_sndr		!= NULL);
	assert (working		!= NULL);
	
	tx = (struct video_tx *) malloc(sizeof(struct video_tx));
	if (tx != NULL) {
		tx->magic		= TRANSMIT_MAGIC;
		tx->rtp_session		= rtp_session;
		tx->cc_sndr		= cc_sndr;
		tx->capture_device	= capture_device;

		tx->encoder_id		= encoder_id;
		tx->decoder_id		= decoder_id;		

		tx->tx_working		= FALSE;
		tx->all_working		= working;		

		tx->base_time		= base_time;

		tx->nodata_rate		= 0;
		gettimeofday (&tx->nodata_tv, NULL);

		
		/* Initialize the sending buffer */
		tx->send_buf = sbuf_init (NULL, tx, base_time);
		if (tx->send_buf == NULL) {
			fprintf (stderr, "Error: creating the sending buffer.\n");
			free (tx);
			return NULL;
		}	
		
		/* Copy the context */
		tx->ctx.mtu		= ctx->mtu;
		tx->ctx.fps		= ctx->fps;
		tx->ctx.max_framelen	= ctx->max_framelen;
		
		/*
		 * Set the initial sending buffer context. For a dynamic congestion
		 * control system, these parameters are not really important, as
		 * they will be fixed when the first RTT is received...
		 * However, for a static (null) congestion control, they must be
		 * as accurate as possible.
		 */
		sbuf_get_context (tx->send_buf, &tx->send_buf_ctx);
		
		tx->send_buf_ctx.mtu		= ctx->mtu;
		tx->send_buf_ctx.max_framelen	= ctx->max_framelen;
		tx->send_buf_ctx.max_frames	= DEFAULT_INIT_MAX_FRAMES;
		tx->send_buf_ctx.max_frags	= MAX (ctx->max_framelen / ctx->mtu, 1);
		tx->send_buf_ctx.max_out_ifi	= MAX_FRAME_DELAY;
		
		assert (tx->send_buf_ctx.mtu		> 0);
		assert (tx->send_buf_ctx.max_frames	> 0);
		assert (tx->send_buf_ctx.max_framelen	> 0);
		assert (tx->send_buf_ctx.max_frags	> 0);
		assert (tx->send_buf_ctx.max_out_ifi	> 0);
		
		sbuf_set_context (tx->send_buf, &tx->send_buf_ctx);
	} else {
		fprintf (stderr, "Error: allocating memory for 'video_tx'\n");
	}
	
	return tx;
}

/* Start threads */
int
tx_start (struct video_tx *tx)
{
	assert (tx		!= NULL);
	assert (tx->magic	== TRANSMIT_MAGIC);
	assert (tx->send_buf	!= NULL);
	
	tx->tx_working = TRUE;

	/* Start the grab & enqueue thread */
	if (pthread_create (&tx->tx_grab_thread, NULL, tx_grab_thread, tx) != 0) {
		fprintf (stderr, "Error: creating the tx grab thread.\n");			
		return FALSE;
	}

	/* Start the transmit thread */
	if (pthread_create (&tx->tx_send_thread, NULL, tx_send_thread, tx) != 0) {
		fprintf (stderr, "Error: creating the tx send thread.\n");			
		return FALSE;
	}

	return TRUE;
}

/* Stop threads */
void
tx_stop (struct video_tx *tx)
{
	assert (tx		!= NULL);
	assert (tx->magic	== TRANSMIT_MAGIC);
	
	tx->tx_working = FALSE;
}

static void
tx_stop_all (struct video_tx *tx)
{
	assert (tx		!= NULL);
	assert (tx->magic	== TRANSMIT_MAGIC);
	
	tx_stop (tx);
	*(tx->all_working) = FALSE;
}


/* Stop threads and cleanup everything... */
void
tx_done (struct video_tx *tx)
{
	assert (tx		!= NULL);
	assert (tx->magic	== TRANSMIT_MAGIC);
	
	tx_stop_all (tx);
	
	/* Waits for the transmit thread */
	if (pthread_join (tx->tx_grab_thread, NULL) != 0) {
		fprintf (stderr, "Error: joining the tx grab thread.\n");
	}
	
	if (pthread_join (tx->tx_send_thread, NULL) != 0) {
		fprintf (stderr, "Error: joining the tx send thread.\n");			
	}
	
	sbuf_done (tx->send_buf);
	
	free (tx);
}

unsigned int
tx_working (struct video_tx *tx)
{
	assert (tx		!= NULL);
	assert (tx->magic	== TRANSMIT_MAGIC);
	
	return (*(tx->all_working) && tx->tx_working);
}
