/*
 * FILE:     tv.c
 * AUTHOR:   Colin Perkins <csp@csperkins.org>
 * MODIFIED: Ladan Gharai  <ladan@isi.edu>
 *           Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2001-2005 University of Southern California
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.43 $
 * $Date: 2006/02/17 14:51:00 $
 *
 */

#ifdef __cplusplus
extern "C" {
#endif

#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/time.h>
#include <time.h>

#include "tv.h"

static struct timeval   start_time;		/* The start time of the application           */
static int              start_time_valid = 0;	/* Non-zero if start_time has been initialised */
static t_usec           random_offset;		/* For the RTP media clock only                */


void
tv_cp (struct timeval *dst, struct timeval *src)
{
	dst->tv_sec = src->tv_sec;
	dst->tv_usec = src->tv_usec;
}

double 
tv_diff(struct timeval curr_time, struct timeval prev_time)
{
    /* Return (curr_time - prev_time) in seconds */
    t_sec      ct, pt;

    ct = (double) curr_time.tv_sec + (((double) curr_time.tv_usec) / 1000000.0);
    pt = (double) prev_time.tv_sec + (((double) prev_time.tv_usec) / 1000000.0);
    return (ct - pt);
}

t_usec
tv_diff_usec (struct timeval curr_time, struct timeval prev_time)
{	
	/* Return curr_time - prev_time in usec - i wonder if these numbers will be too big?*/
	t_usec  tmp, tmp1, tmp2;

	/* We return an unsigned, so fail is prev_time is later than curr_time */
	assert(curr_time.tv_sec >= prev_time.tv_sec);
	if (curr_time.tv_sec == prev_time.tv_sec) {
		assert(curr_time.tv_usec >= prev_time.tv_usec);
	}

	tmp1 = (curr_time.tv_sec - prev_time.tv_sec)*((uint32_t)1000000);
        tmp2 = curr_time.tv_usec - prev_time.tv_usec;
        tmp  = tmp1+tmp2;

	return tmp;
}

void 
tv_add (struct timeval *ts, t_sec offset_secs)
{
	unsigned long offset = (unsigned long) (offset_secs * 1000000.0);

        ts->tv_usec += offset;
        while (ts->tv_usec >= 1000000) {
                ts->tv_sec++;
                ts->tv_usec -= 1000000;
        }
}

void
tv_add_usec (struct timeval *ts, t_usec usecs)
{
        unsigned long offset = (unsigned long) (usecs);

        ts->tv_usec += offset;
        while (ts->tv_usec >= 1000000) {
                ts->tv_sec++;
                ts->tv_usec -= 1000000;
        }
}


int 
tv_gt (struct timeval a, struct timeval b)
{
        /* Returns (a>b) */
        if (a.tv_sec > b.tv_sec) {
                return 1;
        }
        if (a.tv_sec < b.tv_sec) {
                return 0;
        }
        assert(a.tv_sec == b.tv_sec);
        return a.tv_usec > b.tv_usec;
}

int
tv_eq (struct timeval a, struct timeval b)
{
        return (a.tv_sec == b.tv_sec && a.tv_usec == b.tv_usec ? 1 : 0);
}

void
tv_sleep_for (t_usec t)
{
	struct timespec	ts;

	ts.tv_sec  = t / 1000000;
	ts.tv_nsec = (t % 1000000) * 1000;
	nanosleep (&ts, NULL);	
}

void
tv_sleep_until (struct timeval a)
{
	struct timeval	curr_time;
	t_usec		delay;
	
	gettimeofday (&curr_time, NULL);
	
	if (tv_gt (curr_time, a)) {
		return;
	} else {
		delay = tv_diff_usec (a, curr_time);
		tv_sleep_for (delay);
	}
}

t_usec
tv_mediatime (void)
{
        struct timeval          curr_time;
	
        if (!start_time_valid) {
                gettimeofday (&start_time, NULL);
                random_offset = lrand48();
                start_time_valid = 1;
        }
	
        gettimeofday (&curr_time, NULL);
        return ((tv_diff(curr_time, start_time) * 90000) + random_offset);
}


uint32_t
tv_usecs_since (struct timeval *base, struct timeval *t_now)
{
	struct timeval	 t_now_local;
	uint32_t	 t_dif		= 0;

	gettimeofday (&t_now_local, NULL);

	if (tv_gt (t_now_local, *base)) {
		t_dif = tv_diff_usec (t_now_local, *base);
	}

	if (t_now != NULL) {
		tv_cp (t_now, &t_now_local);
	}		
	
	return t_dif;
}

void
tv2usec (uint32_t *us, struct timeval tv)
{
	*us = tv.tv_sec * 1000000 + tv.tv_usec;
}

void
usec2tv (struct timeval *tv, uint32_t us)
{
	tv->tv_sec	= 0;
	tv->tv_usec	= 0;

	tv_add_usec (tv, us);	
}


double
usec2sec (t_usec t)
{
	return ((double)t/1000000.0);
}

t_usec
sec2usec (t_sec t)
{
	return (t*1000000);
}


#ifdef __cplusplus
}
#endif

