/*
 * FILE:   UGMain.m
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.16 $
 * $Date: 2005/10/06 10:20:58 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "controller.h"

#import <Cocoa/Cocoa.h>

#import "ui/cocoa/UGRunEnv.h"
#import "ui/cocoa/UGError.h"
#import "ui/cocoa/UGController.h"
#import "ui/cocoa/UGOptionsDialog.h"
#import "ui/cocoa/UGNewConDialog.h"
#import "ui/cocoa/UGMain.h"

static int		  gArgc;
static const char	**gArgv;
static BOOL		  gFinderLaunch;


uint32_t  RTT = 0;


/*
 * UGApplication
 */
@interface UGApplication : NSApplication
@end

@implementation UGApplication
@end


/*
 * A dummy class
 */
@interface DummyClass : NSObject
+ (void)dummyThread:(NSObject *) obj;
@end

@implementation DummyClass

+ (void)dummyThread:(NSObject *) obj
{
	UNUSED( obj );
	
	return;
}
@end


/*
 * The main class of the application, the application's delegate
 */
@implementation UGMain

- (IBAction)helpSlot:(id)sender
{
	UNUSED( sender );
	
	NSRunAlertPanel(@"Help", @"Sorry, but there is no help available.", @"Ok", nil, nil);
}

- (IBAction)resetSlot:(id)sender
{
	UNUSED(sender);
	
	[ugController stop:self];
}

- (void)shutdown
{
	[ugController stop:self];
	
	exit(0);	
}

/* Set the working directory to the .app's parent directory */
- (void) setupWorkingDirectory:(BOOL)shouldChdir
{
	char parentdir[MAXPATHLEN];
	char *c;

	strncpy ( parentdir, gArgv[0], sizeof(parentdir) );
	c = (char*) parentdir;
 
	while (*c != '\0')     /* go to end */
		c++;
 
	while (*c != '/')      /* back up to parent */
		c--;

	*c++ = '\0';             /* cut off last part (binary name) */

	if (shouldChdir) {
		assert(chdir(parentdir) == 0);   /* chdir to the binary app's parent */
		assert(chdir("../../../") == 0); /* chdir to the .app's parent */
	}
}

/*
 * Setups the multithreading environment.
 * 
 * From tha Apple Documentation:
 *
 * "If you are creating secondary threads using the POSIX thread APIs instead 
 * of NSThread, you cannot use Cocoa, including NSAutoreleasePool, unless Cocoa
 * is in multithreading mode. Cocoa enters multithreading mode only after 
 * detaching its first NSThread. To use Cocoa on secondary POSIX threads, your
 * application must first detach at least one NSThread, which can immediately
 * exit. You can test whether Cocoa is in multithreading mode with the NSThread
 * class method isMultiThreaded."
 *
 * Even if we are not going no use Cocoa _from_ secondary theads, to set up
 * a multithreading environment is useful as it activates the thread-safe
 * code checks...
 */
- (void) setupMultiThreading
{
	
	[NSThread detachNewThreadSelector:@selector(dummyThread:)
				 toTarget:[DummyClass class] 
			       withObject:nil];

	if ([NSThread isMultiThreaded]) {
		debug_msg ("Cocoa in multithreaded mode...\n");
	}
	
	return;
}


/* Called when the internal event loop has just started running */
- (void) applicationDidFinishLaunching: (NSNotification *) note
{
	UNUSED (note);
		
	/* Set the working directory to the .app's parent directory */
	[self setupWorkingDirectory:gFinderLaunch];
		
	/* We want a multithread environment... */
	[self setupMultiThreading];
}

/*
 * The application is going to terminate: notify to the controller
 */
- (void)applicationWillTerminate:(NSNotification *)notification
{
	UNUSED (notification);

	[self shutdown];
}

/*
 * Invoked from the Quit menu item
 */
- (void)terminate:(id)sender
{
	UNUSED (sender);

	[self shutdown];
}

@end

/*
 * Main entry point to executable
 */
int
main (int argc, const char *argv[])
{
	/* Copy the arguments into a global variable */
	int i;

	/* This is passed if we are launched by double-clicking */
	if (argc >= 2 && strncmp (argv[1], "-psn", 4) == 0 ) {
		gArgc = 1;
		gFinderLaunch = YES;
	} else {
		gArgc = argc;
		gFinderLaunch = NO;
	}

	gArgv = (const char**) malloc (sizeof(*gArgv) * (gArgc+1));

	assert (gArgv != NULL);

	for (i = 0; i < gArgc; i++)
		gArgv[i] = argv[i];

	gArgv[i] = NULL;

	[UGApplication poseAsClass:[NSApplication class]];
	NSApplicationMain (argc, argv);

	return 0;
}

