/*
 * FILE:   ui.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 * MODIFIED: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2001-2003 University of Southern California
 * Copyright (c) 2004 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2005/09/27 13:54:04 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "ui/ui.h"
#include "ui/ui_tcltk.h"
#include "ui/ui_sdl.h"
#include "ui/ui_none.h"


/*
 * UI framework interface
 * Each UI framework must implement this interface, where
 *   - pref is the preference for this UI
 *   - func_probe returns >0 if the UI accepts the display device given.
 *   - func_init and func_done initializes and sopts the UI, respec.
 */ 
typedef int ui_pref_t;         /* Preference for a UI */ 

struct ui_framework_api {
        ui_pref_t    pref;
        const char*  display_name;
        int          (*func_probe)(const char*);
        int          (*func_init)(int argc, char *argv[]);
        void         (*func_done)(void);
};

/*
 * Table of UI frameworks
 * Each framework must implement the interface defined in "ui_framework_api"
 */ 
struct ui_framework_api ui_table[] = {
#ifdef HAVE_TCL_TK
        {
                50,
                "Tcl/Tk",
                ui_tcltk_probe,
                ui_tcltk_init,
                ui_tcltk_done
        },
#endif
#ifdef HAVE_LIBSDL
        {
                100,
                "SDL",
                ui_sdl_probe,
                ui_sdl_init,
                ui_sdl_done
        },
#endif
        {
                0,
                "none",
                ui_none_probe,
                ui_none_init,
                ui_none_done
        },
};

#define UI_TABLE_SIZE (sizeof(ui_table)/sizeof(struct ui_framework_api))

unsigned int    current_ui;    /* Currently used UI */

/*
 * Initialization of the UI
 * It selects the best UI from the list, based on the preference order and their
 * support for the given display.
 */ 
int
ui_init(int argc, char *argv[], const char* display)
{
        int             accepts      = 0;
        int             current_pref = -1;
        unsigned int    i            = 0;
       
        /*
         * Look for the best UI for this display
         */
        for (i = 0; i < UI_TABLE_SIZE; i++) {
                accepts = ui_table[i].func_probe(display);
                if (accepts && (ui_table[i].pref > current_pref)) {
                        current_pref = ui_table[i].pref;
                        current_ui   = i;                            
                }
        };

        debug_msg("Using \"%s\" user interface.\n", ui_table[current_ui].display_name); 
        
        /*
         * Starts the UI
         */ 
        return ui_table[current_ui].func_init(argc, argv);
}

void 
ui_done(void)
{
        ui_table[current_ui].func_done();
}
