/*
 * FILE:   video_capture.h
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2002 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.17 $
 * $Date: 2006/01/30 12:43:30 $
 *
 */

#ifndef _VIDEO_CAPTURE_H
#define _VIDEO_CAPTURE_H

#include "video_display.h"

struct run_env;

/*
 * API for probing the valid video capture devices. 
 */
typedef uint32_t	vidcap_id_t;

/*
 * Video capture properties.
 */
typedef uint32_t	vidcap_prop_t;

#define VIDCAP_NEEDS_PARAMS	0x0001	/* Needs initialization params */
#define VIDCAP_PROD_ENCODED	0x0010	/* Produces encoded data */
#define VIDCAP_PARAM_FILENAME	0x0100	/* Needs a filename */
#define VIDCAP_PARAM_FPS	0x0200  /* Needs the FPS */
#define VIDCAP_PARAM_SSIZE	0x0400  /* Needs the screen size */

/*
 * Video capture encoding for frames.
 * Some capture devices produce encoded frames: DV, raw...
 */
typedef enum {
	fe_unk,
	fe_raw,
	fe_yuv,
	fe_dv
} vidcap_frame_enc_t;

/*
 * Video capture definition.
 *
 * Every capture device X must report its capabilities using this structure
 * when vidcap_X_probe() is called. There must be a "capture_format_t" 
 * entry for each possible capture configuration.
 */
typedef struct {
	display_size_t		size;
	video_colour_mode_t	colour_mode;
	unsigned int		fps;
	vidcap_frame_enc_t	encoding;
} capture_format_t;

struct vidcap_type {
	vidcap_id_t		 id;
	const char		*name;
	const char		*description;
	vidcap_prop_t		 props;
	capture_format_t	*formats;
	unsigned int		 num_formats;
};

int			 vidcap_init_devices(const struct run_env*);
void			 vidcap_free_devices(void);
unsigned int		 vidcap_get_device_count(void);
struct vidcap_type	*vidcap_get_device_details(unsigned int index);

/*
 * API for video capture. Normal operation is something like:
 *
 * 	v = vidcap_init(id);
 *	...
 *	while (!done) {
 *		...
 *		f = vidcap_grab(v, timeout);
 *		...use the frame "f"
 *	}
 *	vidcap_stop(v);
 *
 * Where the "id" parameter to vidcap_init() is obtained from
 * the probing API. The vidcap_grab() function returns a pointer
 * to the frame, or NULL if no frame is currently available. It
 * does not block.
 *
 */

struct vidcap;

struct vidcap		*vidcap_init(vidcap_id_t	 id,
				     capture_format_t	*format, 
				     const char		*args);

void			 vidcap_done(struct vidcap *state);
struct video_frame	*vidcap_grab(struct vidcap *state);

#endif
