/*
 * FILE:   firewire_dv_freebsd.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.24 $
 * $Date: 2005/10/18 16:26:56 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "video_capture/firewire_dv_freebsd.h"

#ifdef HAVE_FIREWIRE_DV_FREEBSD
#include <dev/firewire/firewire.h>
#include <dev/firewire/iec68113.h>

/*
 * DV internal state
 */
struct vidcap_dvbsd_state {
	int		 fd;
	display_size_t	 format;
	char		 buffer[2][144000];
	unsigned	 bufnum;
	unsigned	 pos;
	int		 syncd;
};

#define NPACKET	256
#define PSIZE	512

void *
vidcap_dvbsd_init(capture_format_t *format, const char *args)
{
	struct vidcap_dvbsd_state	*s;

	UNUSED(format);
	UNUSED(args);
	
	s = malloc(sizeof(struct vidcap_dvbsd_state));
	if (s != NULL) {
		struct fw_isochreq	isoreq;
		struct fw_isobufreq	bufreq;

		s->fd = open("/dev/fw0.0", O_RDWR);
		if (s->fd == -1) {
			perror("Unable to open firewire device");
			free(s);
			return NULL;
		}

		/* FIXME: how are the bufreq.rx.* values chosen? */
		bufreq.rx.nchunk  = 4;
		bufreq.rx.npacket = NPACKET;
		bufreq.rx.psize   = PSIZE;
		bufreq.tx.nchunk  = 0;
		bufreq.tx.npacket = 0;
		bufreq.tx.psize   = 0;
		if (ioctl(s->fd, FW_SSTBUF, &bufreq) < 0) {
			debug_msg("Unable to set firewire buffering\n");
			goto error;
		}

		isoreq.ch  = 63;	/* Channel number for video. Most camcorders use channel 63 */
		isoreq.tag = 1;		
		if (ioctl(s->fd, FW_SRSTREAM, &isoreq) < 0) {
			debug_msg("Unable to set firewire channel\n");
			goto error;
		}

		s->bufnum = 0;
		s->syncd = 0;
		s->pos   = 0;

		/* FIXME: need to set s->format */
	}
	return s;
error:
	close(s->fd);
	free(s);
	return NULL;
}

void
vidcap_dvbsd_done(void *state)
{
	struct vidcap_dvbsd_state	*s = (struct vidcap_dvbsd_state *) state;

	close(s->fd);
	free(s);
}

struct video_frame *
vidcap_dvbsd_grab(void *state)
{
	struct vidcap_dvbsd_state	*s = (struct vidcap_dvbsd_state *) state;
	struct fw_pkt			*p;
	int				 len, i, EoF;
	char				*ptr, *dif;
	unsigned char			 dbn, dseq, sct;
	char				*buffer;
	struct video_frame		*vf = NULL;

	assert(s->pos < 144000);

	buffer = malloc(NPACKET * PSIZE);
	EoF = 0;	/* = 1 when final packet of frame received */
	do {
		len = read(s->fd, buffer, NPACKET * PSIZE);
		if (len < 0) {
			perror("Unable to read from DV camera\n");
			return NULL;
		}

		/* Unwrap the firewire packets. Each firewire packet contains either */
		/* 0 or 6 DV DIF blocks, each 80 bytes long, which we must process.  */
		ptr = buffer;
		while (ptr < buffer + len) {
			p = (struct fw_pkt *) ptr;
			ptr += sizeof(struct fw_isohdr);
			if (p->mode.stream.len > sizeof(struct ciphdr)) {
				for (i = 0; i < 6; i++) {
					/* Process the DIF blocks... */
					dif  = (ptr + sizeof(struct ciphdr) + (80 * i));

					sct  = (unsigned) ((dif[0] & 0xe0) >> 5);
					dseq = (unsigned) ((dif[1] & 0xf0) >> 4);
					dbn  = (unsigned)   dif[2] ;

					if (s->syncd) {
						/*debug_msg("dbn=%u dseq=%d sct=%u\n", dbn, dseq, sct);*/
						assert(s->pos < 144000);
						memcpy(s->buffer[s->bufnum] + s->pos, dif, 80);
						s->pos += 80;
						if ((sct == 4) && (dseq == 11) && (dbn == 134)) {
							EoF = 1;	/* Final DIF block in the frame */
							s->pos = 0;
							/* Create the frame... */
							vf = malloc(sizeof(struct video_frame));
							if (vf != NULL) {
								vf->colour_mode = YUV_422;
								vf->width       = 720;
								vf->height      = 576;
								vf->data        = s->buffer[s->bufnum];
								vf->data_len    = 144000;
							}
							s->bufnum = (s->bufnum + 1) % 2;
						}
						if (s->pos >= 144000) {
							/* Something has gone wrong: we have too much data, but */
							/* we didn't get the last DIF block in the frame.  Drop */
							/* back into the unsynchronised state, discarding data. */
							/* This should never happen in normal operation.        */
							s->syncd = 0;
							s->pos   = 0;
							EoF      = 0;
							debug_msg("Lost video synchronisation\n");
						}
				} else {
						/* When first started, we don't know where in the DIF sequence we are. */
						/* Wait until we see the last packet in a frame, then mark the state   */
						/* as synchronised. This starts the processing of the video with the   */
						/* first DIF block in the next frame.                                  */
						if ((sct == 4) && (dseq == 11) && (dbn == 134)) {
							s->syncd = 1;
						}
					}

				}
			}
			ptr += p->mode.stream.len;
		}
	} while (!EoF);
	free(buffer);

	return vf;
}

struct vidcap_type *
vidcap_dvbsd_probe(const struct run_env* environ)
{
        struct vidcap_type		*vt;
	struct vidcap_dvbsd_state	*s;
	capture_format_t		*cformat;

	cformat = malloc(2 * sizeof(capture_format_t));
	if (cformat != NULL) {
		cformat[0].size        = DS_PAL;
		cformat[0].colour_mode = YUV_422;
		cformat[0].fps         = 25;
		cformat[0].encoding    = fe_dv;
		
		cformat[1].size        = DS_NTSC;
		cformat[1].colour_mode = YUV_422;
		cformat[1].fps         = 25;
		cformat[1].encoding    = fe_dv;
	} else {
		return NULL;
	}	
	
	s = vidcap_dvbsd_init(cformat, environ->capture_args);
	if (s == NULL) {
		return NULL;	/* No Firewire/DV devices */
	}

	vt = (struct vidcap_type *) malloc(sizeof(struct vidcap_type));
	if (vt != NULL) {
		vt->id          = VIDCAP_DVBSD_ID;
		vt->name        = "dv";
		vt->description = "IEEE-1394/DV";
		vt->props       = 0;
		vt->formats	= cformat;
		vt->num_formats = 2;
	}
	
	vidcap_dvbsd_done(s);

        return vt;
}

#endif

