/*
 * FILE:     firewire_dv_linux.c
 * AUTHOR:   Alvaro Saurin <saurin@dcs.gla.ac.uk>
 * MODIFIED: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2005-2006 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.27 $
 * $Date: 2006/02/16 18:24:37 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_FIREWIRE_DV_LINUX

#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_capture.h"
#include "video_display.h"
#include "video_capture/firewire_dv.h"
#include "video_capture/firewire_dv_linux.h"

#include <dv1394.h>
#include <sys/poll.h>

#ifndef DV1394_IOC_INIT
#define DV1394_IOC_INIT DV1394_INIT
#endif /* DV1394_IOC_INIT */

#define RING_FRAMES     4
#define DEFAULT_CHANNEL 63

/*
 * DV internal state
 */
struct vidcap_dvlinux_state {
	int		 fd;			/* File descriptor for the device */
	const char	*filename;		/* The corresponding file name */

	unsigned	 fps;			/* FPS */
	unsigned	 channel;
	display_size_t	 format;		/* Capture format [NTSC|PAL] */
	unsigned	 frame_len;		/* How much data should we read? */
	
	unsigned char	*ring;
	unsigned int	 ring_current;
	unsigned int	 ring_captured;
	unsigned int	 ring_done;
};

/*
 * List of possible DV device names
 */
static char* dvlinux_devices[] = {
	"/dev/dv1394",
	"/dev/dv1394-0",
	"/dev/dv1394-1",
	"/dev/dv1394-2",
	"/dev/dv1394-3",
	"/dev/dv1394/0",
	"/dev/dv1394/1",
	"/dev/dv1394/2",
	"/dev/dv1394/3",
	NULL
};

/* Initialize the internal state from the environment data. */
static struct vidcap_dvlinux_state *
vidcap_dvlinux_state_init (capture_format_t *format, const char *args)
{
	struct vidcap_dvlinux_state *s = NULL;

	s = (struct vidcap_dvlinux_state*) malloc(sizeof(struct vidcap_dvlinux_state));
	if (s != NULL) {
		s->fd  = -1;

		/* capture_args contains the name of the device/filename */
		s->filename = args;
		s->fps      = format->fps;
		s->format   = format->size;
		s->channel  = DEFAULT_CHANNEL;

		/*
		 * In Linux, we must perform reads of 144000 bytes for PAL
		 * and 120000 for NTSC
		 */		
		if (s->format == DS_PAL) {
			s->frame_len = 144000;
		} else if (s->format == DS_NTSC) {
			s->frame_len = 120000;
		}
	}

	return s;
}


/*******************************************************************************
 * Public API
 ******************************************************************************/

/* Initialization of the device */
void *
vidcap_dvlinux_init (capture_format_t *format, const char *args)
{
	struct vidcap_dvlinux_state	*s;
	struct pollfd			 poll_fd;
	int				 poll_result;
	struct dv1394_init		 init_params;
	struct dv1394_status		 status;
	display_size_t			 actual_size;
	int				 i	= 0;

	/*
	 * Get the state from the environment and open the device
	 */
	s = vidcap_dvlinux_state_init (format, args);
	if (s->filename != NULL)  {
		debug_msg("DV capture: trying to open %s\n", s->filename);
		s->fd = open(s->filename, O_RDONLY);
	} else {
		/* Try to autodetect the device looping over the devices list... */
		while ((s->filename = dvlinux_devices[i++]) != NULL) {
			debug_msg ("DV capture: trying to open %s\n", s->filename);
			s->fd = open(s->filename, O_RDONLY);
			if (s->fd != -1) {
				break;
			}
		}
	}

	if (s->fd < 0) {
		fprintf (stderr, "DV capture: unable to open device\n");
		free(s);
		return NULL;
	}
	
	/*
	 * Configure the capture device.
	 */
	init_params.api_version = DV1394_API_VERSION;
	init_params.channel     = s->channel;
	init_params.n_frames    = RING_FRAMES;
	init_params.cip_n       = 0;
	init_params.cip_d       = 0;
	init_params.syt_offset  = 0;
	
	if (s->format == DS_PAL) {
		init_params.format = DV1394_PAL;
	} else if (s->format == DS_NTSC) {
		init_params.format = DV1394_NTSC;
	} else {
		fprintf(stderr, "DV capture: unknown video format\n");
		return NULL;
	}

	if (ioctl(s->fd, DV1394_IOC_INIT, &init_params) < 0) {
		perror ("DV capture: unable to initialise interface");
		close (s->fd);
		free (s);
		return NULL;
	}

	s->ring = mmap (NULL, s->frame_len * RING_FRAMES, PROT_READ, MAP_PRIVATE, s->fd, 0);
	if (s->ring == MAP_FAILED) {
		perror("DV capture: unable to map capture buffer");
		close(s->fd);
		free(s);
		return NULL;
	}	

	if (ioctl (s->fd, DV1394_IOC_START_RECEIVE, 0) < 0) {
		perror("DV capture: unable to start reception: %s\n");
		vidcap_dvlinux_done(s);
		return NULL;
	}

	/* We need to grab one frame to determine if the device is PAL or NTSC... */
	do {
		poll_fd.fd     = s->fd;
		poll_fd.events = POLLIN | POLLERR | POLLHUP;
		poll_result    = poll (&poll_fd, 1, -1);
	} while (poll_result < 0 && (errno == EAGAIN || errno == EINTR));

	if (poll_result < 0) {
		perror ("DV capture: unable to grab taster frame");
		vidcap_dvlinux_done(s);
		return NULL;
	}

	/* Reset the ring counters, querying the status */
	if (ioctl (s->fd, DV1394_IOC_GET_STATUS, &status) < 0) {
		fprintf (stderr, "Error: ioctl on the DV interface: %s\n", strerror(errno));
	} else {
		s->ring_captured  = status.n_clear_frames;
		s->ring_current   = status.first_clear_frame;
		s->ring_done      = 0;
	
		if (status.dropped_frames) {
			debug_msg("WARNING! Dropped frames in the DV device.\n");
		}
	}

	/* Check the format */
	if (dv_frame_is_PAL (s->ring + (s->ring_current * s->frame_len))) {
		actual_size = DS_PAL;
	} else {
		actual_size = DS_NTSC;
	}
	
	if (s->format != actual_size) {
		vidcap_dvlinux_done (s);
		return NULL;
	}

	/* Update pointers */
	s->ring_current = (s->ring_current + 1) % RING_FRAMES;
	s->ring_captured--;
	s->ring_done++;

	return s;
}

/* Free resources */
void
vidcap_dvlinux_done(void *state)
{
	struct vidcap_dvlinux_state *s;
	
	assert(state != NULL);
	
	s = (struct vidcap_dvlinux_state *) state;

	if (s != NULL) {
		if (ioctl(s->fd, DV1394_IOC_SHUTDOWN, 0) < 0) {
			fprintf(stderr, "Error: shutting down the DV device: %s.\n", strerror(errno));
		}

		if (munmap(s->ring, s->frame_len * RING_FRAMES) < 0) {
			fprintf(stderr, "Error: unmapping the DV ring: %s.\n", strerror(errno));
		}

		if (s->fd >= 0) {
			close(s->fd);
		}

		free(s);
	}
}

/* Grab a frame */
struct video_frame *
vidcap_dvlinux_grab (void *state)
{
	struct video_frame		*vf;
	struct vidcap_dvlinux_state	*s;
	struct pollfd			 poll_fd;
	int				 poll_result;
	struct dv1394_status		 status;

	assert(state != NULL);

	s = (struct vidcap_dvlinux_state *) state;

	assert(s->fd        > 0);
	assert(s->frame_len > 0);

	vf = (struct video_frame *) malloc(sizeof(struct video_frame));
	if (vf != NULL) {
		/* If we have no completed frames, request more */
		if(s->ring_captured <= 0) {
			/* Have we treated previous frames? Yes, then ask for more... */
			if (s->ring_done > 0) {
				if (ioctl (s->fd, DV1394_IOC_RECEIVE_FRAMES, s->ring_done) < 0) {
					debug_msg("DV capture ring overflow\n");
				}
			}

			/* Wait for frames in a poll */
			do {
				poll_fd.fd     = s->fd;
				poll_fd.events = POLLIN | POLLERR | POLLHUP;
				poll_result    = poll(&poll_fd, 1, -1);
			} while (poll_result < 0 && (errno == EAGAIN || errno == EINTR));

			if (poll_result < 0) {
				fprintf(stderr, "Error: in poll: %s.\n", strerror(errno));
				free(vf);
				return NULL;
			}

			/* Reset the ring counters, querying the status */
			if (ioctl (s->fd, DV1394_IOC_GET_STATUS, &status) < 0) {
				fprintf (stderr, "Error: ioctl on the DV interface: %s\n", strerror(errno));
			} else {
				s->ring_captured  = status.n_clear_frames;
				s->ring_current   = status.first_clear_frame;
				s->ring_done      = 0;
			
				if (status.dropped_frames) {
					debug_msg ("WARNING! Dropped frames in the DV device.\n");
				}
			}
		}

		/* create the frame */
		vf->colour_mode	= YUV_422;
		vf->width	= display_get_width (s->format);
		vf->height	= display_get_height(s->format);
		vf->data	= s->ring + (s->ring_current * s->frame_len);
		vf->data_len	= s->frame_len;

		/* update pointers */
		s->ring_current = (s->ring_current + 1) % RING_FRAMES;
		s->ring_captured--;
		s->ring_done++;
        } else {
                debug_msg ("Insufficient space to allocate video_frame\n");
	}

	return vf;
}

struct vidcap_type *
vidcap_dvlinux_probe (const struct run_env* environ)
{
        struct vidcap_type		*vt	 = NULL;
	struct vidcap_dvlinux_state	*s	 = NULL;
	capture_format_t		*cformat = NULL;

	debug_msg("Probing for DV capture device...\n");

	/* Create a default format */
	cformat = malloc (sizeof (capture_format_t));
	cformat->size        = DS_PAL;
	cformat->colour_mode = YUV_422;
	cformat->fps         = 25;
	cformat->encoding    = fe_dv;
	
	s = vidcap_dvlinux_init (cformat, environ->capture_args);
	if (s != NULL) {
		debug_msg("Found DV capture device\n");

		vt = (struct vidcap_type *) malloc (sizeof (struct vidcap_type));
		if (vt != NULL) {
			/* If the init function has detected a different format, fix it */
			if (s->format == DS_NTSC) {
				cformat->size        = DS_NTSC;
				cformat->colour_mode = YUV_422;
				cformat->fps         = 25;
				cformat->encoding    = fe_dv;					
			}
			
			vt->id          = VIDCAP_DVLINUX_ID;
			vt->name        = "dv";
			vt->description = "IEEE-1394 DV";
			vt->props       = 0;
			vt->formats	= cformat;
			vt->num_formats = 1;
		}

		vidcap_dvlinux_done (s);	
		return vt;
	} 

	/* Unknown format?!? */
	debug_msg("No DV capture device found\n");
	return NULL;
}

#endif
