/*
 * FILE:   vidcap_null.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * A fake video capture device, used for systems that either have no capture
 * hardware or do not wish to transmit. This fits the interface of the other
 * capture devices, but never produces any video.
 *
 * Copyright (c) 2004-2005 University of Glasgow
 * Copyright (c) 2003 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.18 $
 * $Date: 2006/02/02 13:00:23 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_capture/null.h"
#include "video_codec.h"
#include "tv.h"

struct vcn_state {
	int		fps;
	struct timeval	next_send_time;
};

void *
vidcap_null_init(capture_format_t *format, const char *args)
{
	struct vcn_state *state;

	UNUSED(args);
	
	state = malloc(sizeof(struct vcn_state));
	if (state != NULL) {
		state->fps = format[0].fps;
		
		gettimeofday(&(state->next_send_time), NULL);
		tv_add(&(state->next_send_time), 1.0/state->fps);
	}
	return state;
}

void
vidcap_null_done(void *state)
{
	assert(state != NULL);
}

struct video_frame *
vidcap_null_grab(void *s)
{
	struct vcn_state *state = (struct vcn_state *) s;

	tv_sleep_until (state->next_send_time);
	gettimeofday (&(state->next_send_time), NULL);
	tv_add (&(state->next_send_time), 1.0/state->fps);

	return NULL;
}

struct vidcap_type *
vidcap_null_probe(const struct run_env* environ)
{
	struct vidcap_type	*vt;
        capture_format_t	*cformat;
	
        assert(environ != NULL);

	cformat = malloc(sizeof(capture_format_t));
	if(cformat != NULL) {
		cformat[0].size        = DS_720p;
		cformat[0].colour_mode = YUV_422;
		cformat[0].fps         = 25;
		cformat[0].encoding    = fe_yuv;		
	} else {
		return NULL;
	}		
	
	vt = (struct vidcap_type *) malloc(sizeof(struct vidcap_type));
	if (vt != NULL) {
		vt->id          = VIDCAP_NULL_ID;
		vt->name        = "none";
		vt->description = "No capture";
		vt->props       = 0;
		vt->formats	= cformat;
		vt->num_formats = 1;
	}
	return vt;
}

