/*
 * FILE:   testfile.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.40 $
 * $Date: 2006/05/12 15:04:37 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace_ug.h"
#include "trace.h"

#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_capture/testfile.h"
#include "video_codec.h"

#include "tv.h"

#include <fnmatch.h>

#define TESTFILE_DEAFULT_FPS	25		/* Default FPS for input movies  */
#define TESTFILE_DEFAULT_ENC	fe_raw
#define TESTFILE_DEAFULT_COL	YUV_422
#define TESTFILE_DEAFULT_SIZE	DS_PAL

#define TESTFILE_NOLOOP		"noloop"	/* Cmd-line option for no looping */
#define TESTFILE_NOSLEEP	"nosleep"	/* Cmd-line option for no sleeping between frames */

/* Mapping type */
struct ext_code_map_t {
        char			*ext;		/* Filename extension		*/
        vidcap_frame_enc_t	 decoder;	/* Codec used for loading frames*/
};

/*
 * Mappings betweeen file name extensions and codecs
 */
struct ext_code_map_t     ext_table[] = {
	{".yuv", fe_raw },
	{".dv",  fe_dv  }
};

#define EXT_TABLE_SIZE   (sizeof(ext_table) / sizeof(struct ext_code_map_t))

/*
 * State structure
 */ 
struct vidcap_testfile_state {
	struct testfile_movie_props	*movie_props;

	int		 fd;
        char		*base_ptr;	/* Start point in memory */
        char		*read_ptr;	/* Current read pointer */
        long int	 len;		/* File size (in bytes) */

	int		 loop;		/* Loop the input movie?		*/
	int		 sleep;		/* Sleep until the next capture time?	*/
	
	unsigned int	 period_usecs;
	struct timeval	 next_send_time;
};

/*
 * Some properties of the movie.
 */ 
struct testfile_movie_props {
        char			*filename;	/* The name of the file		*/
        unsigned int		 fps;		/* Frames per second		*/ 
	display_size_t		 size;		/* Size o frames                */
	video_colour_mode_t	 color;		/* Color mode			*/
        unsigned long int	 frames;	/* Movie size (in frames)	*/
        unsigned int		 frame_bytes;	/* Size of each frame (in bytes)*/
	vidcap_frame_enc_t	 encoding;	/* Codec			*/  
};


/* Find a codec given a filename, based on the file extension. */
static vidcap_frame_enc_t
testfile_find_codec (char* filename)
{
        vidcap_frame_enc_t	ret		= TESTFILE_DEFAULT_ENC;
	int			found		= FALSE;
	unsigned int		i		= 0,
				filename_len	= 0,
				ext_len		= 0,
				position	= 0;
	
	assert (filename != NULL);

        while (!found && i < EXT_TABLE_SIZE) {
		position	= strstr (filename, ext_table[i].ext) - filename;
		ext_len		= strlen (ext_table[i].ext);
		filename_len	= strlen (filename);

		if (position + ext_len == filename_len) {
			ret	= ext_table[i].decoder;
			found	= TRUE;
		} else {
			i++;			
		}
        }
	
	return ret;
}

/*
 * Parse the capture_args arguments and set some other properties.
 * Format: FILENAME@SIZE@FPS
 */ 
static void
vidcap_testfile_parsecmd (const char			*str,
			  struct testfile_movie_props	*props,
			  int				*loop,
			  int				*sleep)
{
	display_size_t	 temp_size;
	char		*token = NULL;
	char		 temp_str [1024];
	unsigned int	 temp_int;

	if (str != NULL) {
		/* Copy the string, or it will be overwritten */
		strcpy (temp_str, str);
		
		props->filename = (char*) strtok (temp_str, "@");	
		
		while ((token = strtok (NULL, "@")) != NULL) {
			
			/* Try to parse the 'noloop' */
			if (strcmp (token, TESTFILE_NOLOOP) == 0) {
				*loop	= FALSE;
				continue;
			}

			/* Try to parse the 'noloop' */
			if (strcmp (token, TESTFILE_NOSLEEP) == 0) {
				*sleep	= FALSE;
				continue;
			}
			
			/* Try to parse the FPS */
			if (sscanf (token, "%d", &temp_int) != 0) {
				props->fps = temp_int;
				continue;
			}
			
			/* Try to get a display size from the token... */
			temp_size = display_get_by_name (token);
			if (temp_size != DS_NONE) {
				/* the user has specified the frame size: get some other parameters... */
				props->size		= temp_size;
				props->frame_bytes	= display_get_width (props->size) *
						display_get_height (props->size) *
						cm_get_bpp (props->color);
				continue;
			}
		}
	}
}


/*
 * Obtain some movie properties from the current codec.
 */ 
static void
vidcap_testfile_fillcodec_props (struct testfile_movie_props	*movie_props,
				 unsigned			 codec,
				 unsigned int			 len)
{
	assert (movie_props != NULL);
	assert (len > 0);
	
	if (movie_props->size == DS_NONE) {
		/* FIXME: added NULL as second argument, but not sure it makes sense [csp] */
		movie_props->size	 = vcodec_decoder_get_size (codec, NULL);						
	}

	if (movie_props->color == 0) {
		movie_props->color	 = vcodec_decoder_get_cm (codec, NULL);		
	}
	
	if (movie_props->frame_bytes == 0) {
		movie_props->frame_bytes = vcodec_encoder_get_framelen (codec);
	}

	if (movie_props->frames == 0) {
		movie_props->frames	 = len / movie_props->frame_bytes;		
	}
}


/*
 * Complete the movie properties, in order to make them coherent.
 */ 
static void
vidcap_testfile_check_props (struct testfile_movie_props *movie_props,
			     unsigned int len)
{
	assert (movie_props != NULL);
	
	if (movie_props->encoding == fe_unk) {
		movie_props->encoding = TESTFILE_DEFAULT_ENC;	
		fprintf (stderr, "WARNING! Unknown encoding for input file. Using default...\n");
	}

	if (movie_props->size == DS_NONE)	{
		movie_props->size = TESTFILE_DEAFULT_SIZE;
		fprintf (stderr, "WARNING! Unknown frame size for input file. Using %s...\n",
			display_get_name (movie_props->size));
	}

	if (movie_props->color == NONE)	{
		movie_props->color = TESTFILE_DEAFULT_COL;
		fprintf (stderr, "WARNING! Unknown color mode for input file. Using %s...\n",
			cm_get_name (movie_props->color));

	}

	if (movie_props->fps == 0) {
		movie_props->fps = TESTFILE_DEAFULT_FPS;	
		fprintf (stderr, "WARNING! Unknown FPS for input file. Using %d fps...\n",
			 movie_props->fps);
	}

	if (movie_props->frame_bytes == 0) {
		movie_props->frame_bytes =	display_get_width (movie_props->size) *
						display_get_height (movie_props->size) *
						cm_get_bpp (movie_props->color);
	}	
	if (movie_props->frames == 0)		movie_props->frames = len / movie_props->frame_bytes;
}


/*
 * Obtain the movie properties.
 */ 
static struct testfile_movie_props*
vidcap_testfile_props (capture_format_t	*format,
		       const char	*args,
		       unsigned int	*len,
		       int		*loop,
		       int		*sleep)
{
        struct testfile_movie_props	*movie_props;
	int				 temp_fd;
	unsigned			 temp_codec	= VCODEC_INVALID_ID;
	struct stat			 statbuf;
	
	UNUSED (format);
	UNUSED (args);
	
        movie_props = (struct testfile_movie_props*) malloc (sizeof (struct testfile_movie_props));
	if (movie_props != NULL) {

		/* Default values: it must be a valid raw image... */
		movie_props->filename	 = NULL;
		movie_props->encoding	 = fe_unk;
		movie_props->size	 = DS_NONE;
		movie_props->frame_bytes = 0;
		movie_props->color	 = NONE;
		movie_props->frames	 = 0;
		movie_props->fps	 = (format == NULL ? TESTFILE_DEAFULT_FPS : format->fps);

		*loop	= TRUE;
		*sleep	= TRUE;
		
		/* Parse the command line:  */ 
		vidcap_testfile_parsecmd (args, movie_props, loop, sleep);
		
		/* If a filename has been provided, it must be valid */
		if (movie_props->filename != NULL) {
			if ((temp_fd = open (movie_props->filename, O_RDONLY))) {

				/* Get the input file size */
				if (fstat (temp_fd, &statbuf) < 0) {
					fprintf (stderr, "Error: performing FSTAT on '%s': %s.\n",
						 movie_props->filename, strerror(errno) );
					return NULL;
				}
				*len = statbuf.st_size;
				
				/* Find the codec for the movie format... */
				movie_props->encoding = testfile_find_codec (movie_props->filename);				
				
				/* Fill up some unknown info based on the codec used */
				temp_codec = vcodec_get_by_enc (movie_props->encoding);

				if (temp_codec != VCODEC_INVALID_ID) {
					vidcap_testfile_fillcodec_props (movie_props, temp_codec, *len);
				}
				
				vidcap_testfile_check_props (movie_props, *len);

				close (temp_fd);
			} else {			
				fprintf (stderr, "Error: file [%s] open error: %s.\n",
					 movie_props->filename, strerror(errno));
				free (movie_props);
				return NULL;
			}
		} else {
			debug_msg ("No filename provided.\n");
		}
        } else {
		fprintf (stderr, "Error: no memory for allocating a testfile_movie_props structure.\n");
	}

        return movie_props;
}


/* External API ***********************************************************************************/


/* Initialize the testfile capture device */
void *
vidcap_testfile_init (capture_format_t *format, const char *args)
{
        struct vidcap_testfile_state	*s;
        struct testfile_movie_props	*props;
	unsigned int			 len;
	int				 fd;
	int				 loop;
	int				 sleep;
	
	assert (format != NULL);
	
        props = vidcap_testfile_props (format, args, &len, &loop, &sleep);
	if (props != NULL) {
		if (props->filename == NULL) {
			fprintf(stderr, "Error: testfile capture needs a file name parameter!\n"
				"Error: use the command line options.\n");
			return NULL;
			
		} else if (props->frames == 0 ) {
			fprintf(stderr, "Error: movie shorter than a frame! Aborting...\n");
			return NULL;
			
		} else {
			fd = open (props->filename, O_RDONLY);
			if (fd >= 0) {
				
				/* Allocate space for the state struct */ 
				s = (struct vidcap_testfile_state*)
				malloc (sizeof(struct vidcap_testfile_state));
				if (s != NULL) {
					
					assert (props->frame_bytes > 0);
					
					s->fd		= fd;
					s->movie_props	= props;
					s->base_ptr	= NULL;
					s->read_ptr	= NULL;
					s->len		= len;
					s->loop		= loop;
					s->sleep	= sleep;
					
					s->period_usecs	= 1000000.0/s->movie_props->fps;

					gettimeofday (&(s->next_send_time), NULL);
					
					/*
					 * Print some information
					 */ 
					debug_msg ("  File: %s\n", s->movie_props->filename);
					debug_msg ("   FPS: %d\n", s->movie_props->fps);
					debug_msg ("Length: %ld frames, %ld seconds\n",
					       s->movie_props->frames,
					       (s->movie_props->frames / s->movie_props->fps));
					debug_msg (" Codec: %s\n", 
					       vcodec_get_name(vcodec_get_by_enc(s->movie_props->encoding)) );
					debug_msg ("  Size: %s\n", 
						  display_get_name (s->movie_props->size));
					
					/* mmap the input file */
					if ((s->base_ptr = mmap (0, s->len, PROT_READ, 
								 MAP_SHARED, s->fd, 0)) == (caddr_t) -1) {
						fprintf (stderr, "Error: in MMAP on %s: %s (%d) (length=%ld, fd=%d).\n",
							 s->movie_props->filename, strerror (errno), errno, s->len, s->fd);
						return NULL;
					}
					
					assert (s->base_ptr != NULL);
					
#ifdef HAVE_MADVISE		
					assert (s->len > 0);
					madvise (s->base_ptr,  s->len, MADV_SEQUENTIAL);
#endif
					
					s->read_ptr = s->base_ptr;
					
					return s;
				} else {
					fprintf (stderr, "Error: unable to allocate memory for testfile state.\n");
					return NULL;
				}      
			
			} else {
				fprintf (stderr, "Error: opening testfile '%s': %s.\n",
					 props->filename, strerror (errno));
				return NULL;
			}
		}
	} else {
		fprintf (stderr, "Error: getting file properties: %s.\n", strerror (errno));
		return NULL;
	}
	
	return NULL;        
}

/* Finish the testfile system */
void
vidcap_testfile_done (void *state)
{
        struct vidcap_testfile_state *s = (struct vidcap_testfile_state *) state;

	if (s != NULL) {
		munmap (s->base_ptr, s->len);
		close (s->fd);
		free (s);
	}
}

/* Grab a frame */
struct video_frame *
vidcap_testfile_grab (void *state)
{
        struct vidcap_testfile_state   *s	= (struct vidcap_testfile_state *) state;
        struct video_frame             *vf	= NULL;
        int                             m_frame_bytes;
        int                             m_offset;

	assert (s			!= NULL);
	assert (s->read_ptr		!= NULL);
	assert (s->base_ptr		!= NULL);
	assert (s->movie_props		!= NULL);
	assert (s->movie_props->frame_bytes > 0);
	assert (s->movie_props->fps	> 0);
	
	trace (te_testfile_grab_start);	
	trace_val (te_testfile_grab_i, tv_usecs_since (&s->next_send_time, NULL));
	
	/* Check that we can grab a frame */
	if (s->sleep) {
		tv_sleep_until (s->next_send_time);		
	}

	/* Update the next time */
	gettimeofday (&(s->next_send_time), NULL);
	tv_add_usec (&(s->next_send_time), s->period_usecs);
	
        m_offset      = s->read_ptr - s->base_ptr;  /* Current offset in the movie */
        m_frame_bytes = s->movie_props->frame_bytes;

        /* If we are going to read out of the file, then go to the begining */                   
        if ( (m_offset + m_frame_bytes) > s->len ) {
		if (s->loop) {
			s->read_ptr = s->base_ptr;
		} else {
			return NULL;
		}
	}

	vf = (struct video_frame *) malloc(sizeof(struct video_frame));
	if (vf != NULL) {
		vf->colour_mode = s->movie_props->color;
		vf->width       = display_get_width (s->movie_props->size);
		vf->height      = display_get_height (s->movie_props->size);
		vf->data        = (unsigned char*) s->read_ptr;
		vf->data_len    = m_frame_bytes;
		
		/* Update the read pointer */
		s->read_ptr    += m_frame_bytes; 
	} else {
		fprintf (stderr,"Error: allocating memory for video_frame.\n");
	}

        return vf;
}

/* Probe the availability of the testfile device */
struct vidcap_type *
vidcap_testfile_probe (const struct run_env* environ)
{
        struct vidcap_type		*vt		= NULL;
        struct testfile_movie_props	*props		= NULL;
	capture_format_t		*cformat	= NULL;
	unsigned int			 len;
	int				 loop;
	int				 sleep;
	
	/* Delay the cformat initialization, until more details are available */
	cformat = malloc (sizeof (capture_format_t));
	if (cformat != NULL) {
		props = vidcap_testfile_props (cformat, environ->capture_args, &len, &loop, &sleep);
		if (props != NULL) {
			
			cformat[0].size        = props->size;
			cformat[0].colour_mode = props->color;
			cformat[0].fps         = props->fps;
			cformat[0].encoding    = props->encoding;
			
			vt = (struct vidcap_type *) malloc (sizeof (struct vidcap_type));
			if (vt != NULL) {
				vt->id          = VIDCAP_TESTFILE_ID;
				vt->name        = "testfile";
				vt->description = "Test File";
				vt->props       = VIDCAP_NEEDS_PARAMS |
						VIDCAP_PARAM_FILENAME;
				vt->formats	= cformat;
				vt->num_formats = 1;
			} else {
				fprintf (stderr, "Error: not enough space when trying to allocate a vidcap_type.\n");
				return NULL;
			}
		} else {
			fprintf (stderr, "Error: retrieving the properties of the movie.\n");
			free (cformat);
			return NULL;
		}
		
        } else {
 		fprintf (stderr, "Error: allocating memory for the list of capture formats.\n");
		return NULL;
	} 
	
        return vt;
}
