/*
 * FILE:   video_codec.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.64 $
 * $Date: 2007/01/08 22:22:22 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace_ug.h"
#include "trace.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"		/* Needed by rtp/pbuf.h  */
#include "rtp/pbuf.h"		/* For struct coded_data */
#include "video_codec.h"
#include "video_codec/dv.h"
#include "video_codec/jpeg.h"
#include "video_codec/mpeg1.h"
#include "video_codec/uv_yuv422.h"

struct vcodec_api {
	int                  (*vcx_init)            (void);
	void                 (*vcx_exit)            (void);
	const char *         (*vcx_get_name)        (void);
	const char *         (*vcx_get_desc)        (void);
	int                  (*vcx_can_encode)      (void);
	int                  (*vcx_can_packetize)   (void);
	int                  (*vcx_can_decode)      (void);
	struct encoded_frame*(*vcx_eframe_alloc)    (void);
	struct encoded_data* (*vcx_eframe_get)      (struct encoded_frame *eframe,
						     unsigned int	   pos);
	void                 (*vcx_eframe_free)     (struct encoded_frame *eframe);
	void                *(*vcx_encoder_create)  (unsigned id, unsigned mtu);
	void                 (*vcx_encoder_destroy) (void *state);
	unsigned int         (*vcx_encoder_get_framelen)(void);
	int                  (*vcx_encode)          (void *state, 
                                                     struct video_frame	  *in, 
					             struct encoded_frame *out,
						     struct vcodec_ctx_t  *ctx);
	int                  (*vcx_packetize)       (void *state, 
                                                     struct video_frame *in, 
					             struct encoded_frame *out);
	video_colour_mode_t  (*vcx_decoder_get_cm)  (void *state);
	display_size_t       (*vcx_decoder_get_size)(void *state);
	unsigned int         (*vcx_decoder_get_framelen)(void *state);
	void                *(*vcx_decoder_create)  (unsigned id);
	void                 (*vcx_decoder_destroy) (void *state);
	int                  (*vcx_decode)          (void *state, 
                                                     struct coded_data *in, 
						     struct video_frame *out);
	int                  (*vcx_decode_raw)     (void *state, 
                                                     struct video_frame *in, 
						     struct video_frame *out);
	int                  (*vcx_decode_partial)  (void *state, 
                                                     struct coded_data *in, 
						     struct video_frame *out);
};

/* Table of pointers to codec functions, one for each codec we */
/* support. These are not necessarily all available for use as */
/* some codecs may fail their initialization (for example, if  */
/* they depend on hardware devices that are not present). The  */
/* available_vcodecs[] array contains pointers to those codecs */
/* that are usable.                                            */
static struct vcodec_api vcodec_table[] = {
#ifdef HAVE_LIBDV
	{
		dv_codec_init,
		dv_codec_exit,
		dv_codec_get_name,
		dv_codec_get_desc,
		dv_codec_can_encode,
		dv_codec_can_packetize,
		dv_codec_can_decode,
		dv_eframe_alloc,
		dv_eframe_get,
		dv_eframe_free,
		dv_encoder_create,
		dv_encoder_destroy,
		dv_encoder_get_framelen,
		dv_encode,
		dv_packetize,
		dv_decoder_get_cm,
		dv_decoder_get_size,		
		dv_decoder_get_framelen,
		dv_decoder_create,
		dv_decoder_destroy,
		dv_decode,
		dv_decode_raw,
		dv_decode_partial
	},
#endif
#ifdef HAVE_LIBJPEG
	{
		jpeg_codec_init,
		jpeg_codec_exit,
		jpeg_codec_get_name,
		jpeg_codec_get_desc,
		jpeg_codec_can_encode,
		jpeg_codec_can_packetize,
		jpeg_codec_can_decode,
		jpeg_eframe_alloc,
		jpeg_eframe_get,
		jpeg_eframe_free,
		jpeg_encoder_create,
		jpeg_encoder_destroy,
		jpeg_encoder_get_framelen,
		jpeg_encode,
		jpeg_packetize,
		jpeg_decoder_get_cm,
		jpeg_decoder_get_size,		
		jpeg_decoder_get_framelen,
		jpeg_decoder_create,
		jpeg_decoder_destroy,
		jpeg_decode,
		jpeg_decode_raw,
		jpeg_decode_partial
	},
#endif
#ifdef HAVE_LIBAVCODEC
	{
		mpeg1_codec_init,
		mpeg1_codec_exit,
		mpeg1_codec_get_name,
		mpeg1_codec_get_desc,
		mpeg1_codec_can_encode,
		mpeg1_codec_can_packetize,
		mpeg1_codec_can_decode,
		mpeg1_eframe_alloc,
		mpeg1_eframe_get,
		mpeg1_eframe_free,
		mpeg1_encoder_create,
		mpeg1_encoder_destroy,
		mpeg1_encoder_get_framelen,
		mpeg1_encode,
		mpeg1_packetize,
		mpeg1_decoder_get_cm,
		mpeg1_decoder_get_size,		
		mpeg1_decoder_get_framelen,
		mpeg1_decoder_create,
		mpeg1_decoder_destroy,
		mpeg1_decode,
		mpeg1_decode_raw,
		mpeg1_decode_partial
	},
#endif	
	{
		uv_yuv422_init,
		uv_yuv422_exit,
		uv_yuv422_get_name,
		uv_yuv422_get_desc,
		uv_yuv422_can_encode,
		uv_yuv422_can_packetize,
		uv_yuv422_can_decode,
		uv_yuv422_eframe_alloc,
		uv_yuv422_eframe_get,
		uv_yuv422_eframe_free,
		uv_yuv422_encoder_create,
		uv_yuv422_encoder_destroy,
		uv_yuv422_encoder_get_framelen,
		uv_yuv422_encode,
		uv_yuv422_packetize,
		uv_yuv422_decoder_get_cm,
		uv_yuv422_decoder_get_size,
		uv_yuv422_decoder_get_framelen,
		uv_yuv422_decoder_create,
		uv_yuv422_decoder_destroy,
		uv_yuv422_decode,
		uv_yuv422_decode_raw,
		uv_yuv422_decode_partial
	}
};

#define NUM_VCODEC_INTERFACES (sizeof(vcodec_table)/sizeof(struct vcodec_api))

#define MAX_PAYLOAD 128

/* Pointers to entries in the vcodec_table[], for codecs that  */
/* are usable...                                               */
static struct vcodec_api *available_vcodecs[NUM_VCODEC_INTERFACES];
static unsigned	          num_vcodecs;
static unsigned           pt_map[MAX_PAYLOAD];


/*
 * Mappings between encodings of capture devices and codecs
 */
struct enc_cod_map_entry {
	vidcap_frame_enc_t	 enc;
	char			*codec;
};

static struct enc_cod_map_entry enc_cod_map_table[] = {
	{ fe_unk,	NULL		},
	{ fe_raw,	NULL		},
	{ fe_yuv,	"uv_yuv422"	},
	{ fe_dv,	"dv"		}
};

#define NUM_ENC_COD_MAP (sizeof(enc_cod_map_table)/sizeof(struct enc_cod_map_entry))


/*
 * Public interface functions follow...
 */

void
vcodec_init(void)
{
	unsigned i;

	num_vcodecs = 0;
	
	// debug_msg ("NUM_VCODEC_INTERFACES:%d\n", NUM_VCODEC_INTERFACES);
	
	for (i = 0; i < NUM_VCODEC_INTERFACES; i++) {
		if (vcodec_table[i].vcx_init()) {
			available_vcodecs[num_vcodecs++] = &vcodec_table[i];
			// debug_msg ("Initializing %s\n", vcodec_table[i].vcx_get_name());
		}
		else {
			debug_msg ("Could not find: %s\n", vcodec_table[i].vcx_get_name());			
		}
	}
	
	for (i = 0; i < MAX_PAYLOAD; i++) {
		/* Illegal - will trigger an assert() if unmapped codec is used*/
		pt_map[i] = VCODEC_INVALID_ID;	
	}
}

void
vcodec_done(void)
{
	unsigned i;

	for (i = 0; i < num_vcodecs; i++) {
		available_vcodecs[i]->vcx_exit();
	}
	num_vcodecs = 0;
}

unsigned
vcodec_get_num_codecs(void)
{
	return num_vcodecs;
}

const char *
vcodec_get_name(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_get_name();
	} else {
		return "";
	}
}

const char *
vcodec_get_description(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_get_desc();
	} else {
		return NULL;
	}
}

uint8_t
vcodec_get_payload(unsigned id) {
	uint8_t pt = 0;
		
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		while ((pt_map[pt] != id) && (pt < MAX_PAYLOAD)) {
			pt++;
		}

		return pt;
	} else {
		return 0;
	}
}

int
vcodec_can_encode(unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_can_encode();
	} else {
		return FALSE;
	}
}

int
vcodec_can_packetize (unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_can_packetize();
	} else {
		return FALSE;
	}
}

int
vcodec_can_decode (unsigned id)
{	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_can_decode();
	} else {
		return FALSE;
	}
}


void
vcodec_map_payload(uint8_t pt, unsigned id)
{
	assert(pt < MAX_PAYLOAD);
	assert(pt_map[pt] == VCODEC_INVALID_ID);	/* Fail if already mapped */
	pt_map[pt] = id;
}

void
vcodec_unmap_payload(uint8_t pt)
{
	assert(pt < MAX_PAYLOAD);
	assert(pt_map[pt] < num_vcodecs);		/* Fail if payload not mapped */
	pt_map[pt] = VCODEC_INVALID_ID;
}

unsigned
vcodec_get_by_payload(uint8_t pt)
{
	assert(pt < MAX_PAYLOAD);

	return pt_map[pt];
}

unsigned
vcodec_get_by_name (const char* name)
{
	unsigned int i;
	
	if (name != NULL)
		for (i = 0; i < vcodec_get_num_codecs(); i++)
			if (strcmp (vcodec_get_name(i), name) == 0)
				return i;

	return VCODEC_INVALID_ID;
}

unsigned
vcodec_get_by_enc (vidcap_frame_enc_t enc) {
	unsigned int	 i;
	char		*codec;

	for (i = 0; i < NUM_ENC_COD_MAP; i++)
		if (enc_cod_map_table[i].enc == enc) {
			codec = enc_cod_map_table[i].codec;
			
			/* The table can contain NULLs (ie, when there is no codec needed) */
			if (codec != NULL)
				return vcodec_get_by_name (codec);
			else
				return VCODEC_INVALID_ID;
		}
	
	return VCODEC_INVALID_ID;
}


/*****************************************************************************/
struct encoded_frame*
vcodec_eframe_alloc (unsigned int id)
{
	struct encoded_frame	*res	= NULL;
	
	if (id != VCODEC_INVALID_ID) {
		assert (id < num_vcodecs);
		
		res = available_vcodecs[id]->vcx_eframe_alloc ();
		res->id = id;	/* Maybe this is not necessary */
	}
	
	return res;
}

struct encoded_data*
vcodec_eframe_get (unsigned int id, struct encoded_frame *eframe, int pos)
{
	struct encoded_data	*res	= NULL;

	if (id != VCODEC_INVALID_ID) {
		assert (id < num_vcodecs);
		assert (eframe->id == id);

		res = available_vcodecs[id]->vcx_eframe_get (eframe, pos);
	}
	
	return res;
}

void
vcodec_eframe_free (unsigned int id, struct encoded_frame *eframe)
{
	if (id != VCODEC_INVALID_ID) {
		assert (id < num_vcodecs);
		assert (eframe->id == id);
		
		available_vcodecs[id]->vcx_eframe_free (eframe);
	}
}

/******************************************************************************/
void*
vcodec_encoder_create (unsigned id, unsigned mtu)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_encoder_create (id, mtu);
	} else {
		return NULL;
	}
}

void
vcodec_encoder_destroy (unsigned id, void *state)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_encoder_destroy(state);
	}
}

unsigned
vcodec_encoder_get_framelen (unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_encoder_get_framelen();
	} else {
		return 0;
	}
}


int
vcodec_encode (unsigned              id,
	       void                 *state, 
	       struct video_frame   *in, 
	       struct encoded_frame *out,
	       struct vcodec_ctx_t  *ctx)
{
	int ret = FALSE;
	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		//trace (te_vcodec_enc_init);
		ret = available_vcodecs[id]->vcx_encode(state, in, out, ctx);
		//trace (te_vcodec_enc_done);
	}
	
	return ret;
}

int
vcodec_packetize (unsigned              id,
		  void                 *state, 
		  struct video_frame   *in, 
		  struct encoded_frame *out)
{
	int ret = FALSE;
	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		//trace (te_vcodec_enc_init);
		return available_vcodecs[id]->vcx_packetize(state, in, out);
		//trace (te_vcodec_enc_init);
	}
	return ret;
}

video_colour_mode_t
vcodec_decoder_get_cm(unsigned id, void *state)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_get_cm(state);
	} else {
		return NONE;
	}
}

display_size_t
vcodec_decoder_get_size(unsigned id, void *state)
{	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_get_size(state);
	} else {
		return DS_NONE;
	}
}

unsigned int
vcodec_decoder_get_framelen(unsigned id, void *state)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_get_framelen(state);
	} else {
		return 0;
	}
}

void *
vcodec_decoder_create (unsigned id)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_create(id);
	} else {
		return NULL;
	}
}

void
vcodec_decoder_destroy (unsigned id, void *state)
{
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		return available_vcodecs[id]->vcx_decoder_destroy(state);
	}
}

int
vcodec_decode (unsigned             id,
	       void                *state, 
	       struct coded_data   *in, 
	       struct video_frame  *out)
{
	int	ret	= FALSE;
	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		//trace (te_vcodec_dec_init);
		ret = available_vcodecs[id]->vcx_decode(state, in, out);
		//trace (te_vcodec_dec_done);
	}
	
	return ret;
}

int
vcodec_decode_raw (unsigned   		 id,
		   void			*state,
		   struct video_frame	*in,
		   struct video_frame	*out)
{
	int	ret = FALSE;
	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		//trace (te_vcodec_dec_init);
		ret = available_vcodecs[id]->vcx_decode_raw(state, in, out);
		//trace (te_vcodec_dec_done);
		out->ts = in->ts;
	} 
	
	return ret;
}

int
vcodec_decode_partial (unsigned             id,
		       void                *state, 
		       struct coded_data   *in, 
		       struct video_frame  *out)
{
	int ret	= FALSE;
	
	if (id != VCODEC_INVALID_ID) {
		assert(id < num_vcodecs);
		
		//trace (te_vcodec_dec_init);
		ret = available_vcodecs[id]->vcx_decode_partial(state, in, out);
		//trace (te_vcodec_dec_done);
	}

	return ret;
}

