/*
 * FILE:   video_codec.h
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.33 $
 * $Date: 2005/11/14 14:03:25 $
 *
 */

#ifndef _VIDEO_CODEC_H
#define _VIDEO_CODEC_H

typedef int vcodec_id_t;

#define VCODEC_INVALID_ID	0xffffffff

/*
 * Encoding context.
 * The encoding context specifies properties for the encoding process. Not
 * all of them will be used by codecs, so any modification must try to keep
 * coherent all the properties.
 */
struct vcodec_ctx_t {
	unsigned int	bitrate;	/* Ideal bitrate (bps) */
	unsigned long	framesize;	/* Max frame size wished (in bytes) */
};

/*
 * The coded representation of a frame.
 * Used as a reference for the currently encoded/frame, but never as a
 * direct source for the encoded data. See encoded_data instead, and its
 * associated functions.
 */
struct encoded_frame {
	uint32_t	 magic;
	unsigned	 id;		/* Codec ID			*/
	uint16_t	 len;		/* Number of nodes		*/
	unsigned long	 datalen;	/* Total length of encoded data	*/
	void		*priv;
};

/*
 * The coded representation of a fragment of a frame.
 */
struct encoded_data {
	t_usec		 rtp_ts;
	uint16_t	 seqno;		/* Sequence number in the frame	[0-n]	*/
	unsigned char	*header;	/* Header				*/
	uint16_t	 header_len;	/* Header length			*/
        unsigned char	*data;		/* Encoded data				*/
	uint16_t	 data_len;	/* Data lenght				*/
	uint16_t	 mark;
};

/*
 * Initialisation, probing of available codecs, name and RTP 
 * payload type mapping functions. 
 */
void        vcodec_init           (void);
void        vcodec_done           (void);
unsigned    vcodec_get_num_codecs (void);

const char *vcodec_get_name       (unsigned id);	/* Single word name */
const char *vcodec_get_description(unsigned id);	/* Descriptive text */
uint8_t     vcodec_get_payload    (unsigned id);

int         vcodec_can_encode     (unsigned id);
int         vcodec_can_packetize  (unsigned id);
int         vcodec_can_decode     (unsigned id);

/* Mapping and unmapping of payload types to codecs */
void        vcodec_map_payload    (uint8_t pt, unsigned int id);
void        vcodec_unmap_payload  (uint8_t pt);

/* Search or codecs, given the payload type, name or encoding */
unsigned    vcodec_get_by_payload (uint8_t		 pt);
unsigned    vcodec_get_by_name    (const char		*name);
unsigned    vcodec_get_by_enc     (vidcap_frame_enc_t	 enc);

/*
 * 'encoded_frame's management.
 *
 * Before a encoding function is used, an encoded_frame must be allocated 
 * with vcodec_eframe_alloc(), and it must be released with vcodec_eframe_free(),
 * after we have retrieved its fragments, with vcodec_eframe_get()...   
 */
struct encoded_frame	*vcodec_eframe_alloc	(unsigned int	 id);
struct encoded_data	*vcodec_eframe_get	(unsigned int	 id,
						 struct encoded_frame	*ef,
						 int		 pos);
void			 vcodec_eframe_free	(unsigned int	 id,
						 struct encoded_frame	*ef);

/*
 * Video encoder and decoder functions. These operate on particular
 * instances of a codec, identified by the "state" parameter.
 */

/* Encoder management */
void                *vcodec_encoder_create (unsigned int	  id,
					    unsigned int	  mtu);
void                 vcodec_encoder_destroy(unsigned int	  id, 
                                            void                 *state);

/* Encoder info */
unsigned int         vcodec_encoder_get_framelen (unsigned int	  id);

/* Encoding */
int                  vcodec_encode         (unsigned int	  id, 
                                            void		 *state,
                                            struct video_frame	 *in,
					    struct encoded_frame *out,
					    struct vcodec_ctx_t	 *ctx);
int                  vcodec_packetize      (unsigned		  id, 
                                            void		 *state,
                                            struct video_frame	 *in,
					    struct encoded_frame *out);

/* Decoder info */
video_colour_mode_t  vcodec_decoder_get_cm  (unsigned id, void *state);
display_size_t       vcodec_decoder_get_size(unsigned id, void *state);
unsigned int         vcodec_decoder_get_framelen(unsigned id, void *state);


/* Decoder management */
void                *vcodec_decoder_create  (unsigned id);
void                 vcodec_decoder_destroy (unsigned id, void   *state);

/* Decoding */
int                  vcodec_decode          (unsigned             id, 
                                             void                *state,
					     struct coded_data   *in,
                                             struct video_frame  *out);
int                  vcodec_decode_raw      (unsigned             id, 
                                             void                *state,
					     struct video_frame  *in,
                                             struct video_frame  *out);
int                  vcodec_decode_partial  (unsigned             id,
                                             void                *state,
					     struct coded_data   *in,
                                             struct video_frame  *out);

#endif /* _VCODEC_H */

