/*
 * FILE:   video_codec/dv.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.68 $
 * $Date: 2006/05/25 16:21:37 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBDV
#include "debug.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "video_codec/dv.h"

#include "libdv/dv.h"

#define MAX_ENCODER_FRAMELEN	144000		/* Hardcoded for PAL */


/* Private structure of an encoded_frame */
struct dv_priv_eframe {
	struct encoded_data	*nodes;

	struct video_frame	*frame;	/* Original video frame */
	
	struct dv_priv_eframe	*next,	/* Not currently used */
				*prev;	/* Not currently used */
};

/* Encoder and decoder internal states. */
struct dv_enc_state {
	uint32_t		 magic;		/* For debugging	*/
	
	unsigned int		 id;		/* Codec id		*/

	dv_encoder_t		*encoder;	/* libDV state		*/ 

	display_size_t		 size;		/* Image size		*/
	unsigned int		 mtu;		/* MTU			*/
};

struct dv_dec_state {
	uint32_t		magic;		/* For debugging	*/
	
	unsigned int		id;		/* Codec id		*/
	
	display_size_t		size;		/* Image size		*/
	
	dv_decoder_t		*decoder;	/* libDV state		*/
};


int
dv_codec_init (void)
{
	return TRUE;	/* Codec always available */
}

void
dv_codec_exit(void)
{
	/* Nothing to do */
}

const char *
dv_codec_get_name(void)
{
	return "dv";
}

const char *
dv_codec_get_desc(void)
{
	return "DV (IEC 61834/SMPTE 314M)";
}

int
dv_codec_can_encode(void)
{
	return TRUE;
}

int
dv_codec_can_packetize(void)
{
	return TRUE;
}


int
dv_codec_can_decode(void)
{
	return TRUE;
}

/*******************************************************************************
 * Encoded frames management
 ******************************************************************************/
struct encoded_frame*
dv_eframe_alloc (void)
{
	struct encoded_frame	*res_eframe		= NULL;
	
	res_eframe = (struct encoded_frame*) malloc (sizeof(struct encoded_frame));
	if (res_eframe != NULL) {
		/* Set some info of the encoded_frame */
		res_eframe->magic	= VCODEC_DV_ID;
		res_eframe->len		= 0;
		res_eframe->datalen	= dv_encoder_get_framelen();
		res_eframe->priv	= NULL;
	} else {
		fprintf (stderr, "Error allocating memory for encoded_frame.\n");		
		dv_eframe_free (res_eframe);
	}
	
	return res_eframe;
}

/* Finish the allocation of the encoded_frame, allocating private structures... */
static struct dv_priv_eframe*
dv_priv_alloc (struct dv_enc_state *state, struct encoded_frame *eframe)
{
	unsigned int		 max_nodes;
	struct dv_priv_eframe	*res_priv_eframe	= NULL;
	
	assert (state		!= NULL);
	assert (eframe		!= NULL);
	assert (eframe->magic	== VCODEC_DV_ID);
	
	res_priv_eframe = (struct dv_priv_eframe*) malloc (sizeof (struct dv_priv_eframe));
	if (res_priv_eframe != NULL) {		
		/* 
		 * Allocate memory for the nodes, estimating how many packets
		 * we will generate for a frame. We should have more nodes than
		 * expected: some wasted memory, but no segmentation fault! 
		 */
		max_nodes = (MAX_ENCODER_FRAMELEN / (state->mtu/80)*80) + 1;
		
		res_priv_eframe->nodes = (struct encoded_data*) malloc(max_nodes * sizeof(struct encoded_data));
		if (res_priv_eframe->nodes == NULL) {
			fprintf(stderr, "Error: allocating memory for %d DV encoder nodes.\n", max_nodes);
			dv_encoder_destroy (res_priv_eframe);
			return NULL;
		}

		res_priv_eframe->frame = video_frame_alloc (MAX_ENCODER_FRAMELEN);
		if (res_priv_eframe->frame == NULL) {
			fprintf(stderr, "Error: allocating memory for video_frame in DV encoder.\n");
			dv_encoder_destroy (res_priv_eframe);
			return NULL;
		}
		
		/* Set some info of the encoded_frame */
		eframe->len	= max_nodes;
		eframe->datalen	= MAX_ENCODER_FRAMELEN;
		eframe->priv	= res_priv_eframe;
	} else {
		fprintf (stderr, "Error allocating memory private data of encoded_frame.\n");
	}
	
	return res_priv_eframe;
}

struct encoded_data*
dv_eframe_get (struct encoded_frame *eframe, unsigned int pos)
{
	struct dv_priv_eframe	*ef	= (struct dv_priv_eframe*)(eframe->priv);
	struct encoded_data	*nodes	= NULL,
				*res	= NULL;
		
	assert (eframe		!= NULL);
	assert (eframe->magic	== VCODEC_DV_ID);
	assert ((eframe->len	>  0 && eframe->priv != NULL) || eframe->len == 0);
	assert ((ef		!= NULL && ef->nodes != NULL) || ef == NULL);
	
	if (ef != NULL) {
		nodes = ef->nodes;

		if (pos < eframe->len) {
			res = (nodes + pos);
		}
	}
	
	return res;
}

void
dv_eframe_free (struct encoded_frame *eframe)
{
	struct dv_priv_eframe	*eframe_priv;
	
	assert (eframe        != NULL);
	assert (eframe->magic == VCODEC_DV_ID);
	
	eframe_priv = (struct dv_priv_eframe*) eframe->priv;
	if (eframe_priv != NULL) {
		if (eframe_priv->nodes != NULL)	free (eframe_priv->nodes);
		
		video_frame_free (eframe_priv->frame);
		
		free (eframe_priv);
	}
	
	free (eframe);
}


/*******************************************************************************
 * Encoder
 ******************************************************************************/
void *
dv_encoder_create (unsigned id, unsigned mtu)
{
	struct dv_enc_state	*s	= NULL;
	
	s = (struct dv_enc_state*) malloc(sizeof(struct dv_enc_state));
	if (s != NULL) {
		s->encoder	= dv_encoder_new(0, 0, 0);
		s->id		= id;
		s->mtu		= mtu;
		s->size		= DS_PAL;	/* This sgould be obtained from somewhere... */
	} else {
		fprintf(stderr, "Error: allocating memory for DV encoder state.\n");
	}

	return s;
}

void
dv_encoder_destroy (void *state)
{
	struct dv_enc_state *s = (struct dv_enc_state*) state;

	assert(s          != NULL);
	assert(s->encoder != NULL );

	dv_encoder_free(s->encoder);
	
	free(s);
}

unsigned int
dv_encoder_get_framelen (void)
{
	return MAX_ENCODER_FRAMELEN;		/* Max size: PAL */
}

int
dv_encode(void                 *state,
	  struct video_frame   *in,
	  struct encoded_frame *out,
	  struct vcodec_ctx_t  *ctx)
{
	struct dv_enc_state	*s = (struct dv_enc_state*) state;
	struct video_frame	*encoded_dv_frame;
	unsigned char		*pixels[3];

	UNUSED(ctx);

	/* Validate the frame we have been give to encode... */
	if (in->colour_mode != YUV_422) {
		debug_msg("dv_encode: invalid input colour mode\n");
		return 0;
	}
	if (in->width != 720) {
		debug_msg("dv_encode: invalid frame width (%d != 720)\n", in->width);
		return 0;
	}
	if (in->height != 480 && in->height != 576) {
		debug_msg("dv_encode: invalid frame height (%d != 480 or 576)\n", in->height);
		return 0;
	}
	if (in->data_len != in->width * in->height * 2) {
		debug_msg("dv_encode: invalue data length (%d != %d)\n", in->data_len, in->width * in->height * 2);
		return 0;
	}

	/* Create the encoded_dv_frame... */
	if (in->height == 480) {
		encoded_dv_frame = video_frame_alloc(120000);	/* NTSC */
		s->encoder->isPAL = 0;
	} else {	
		encoded_dv_frame = video_frame_alloc(144000);	/* PAL  */
		s->encoder->isPAL = 1;
	}

	/* Encode... */
	pixels[0] = in->data;
	pixels[1] = NULL;
	pixels[2] = NULL;
	dv_encode_full_frame (s->encoder, pixels, e_dv_color_yuv, encoded_dv_frame->data);

	return dv_packetize (state, encoded_dv_frame, out);
}

int
dv_packetize (void                 *state,
	      struct video_frame   *in,
	      struct encoded_frame *out)
{
	struct dv_enc_state	*s		= (struct dv_enc_state*) state;
	struct dv_priv_eframe	*out_priv;
	struct encoded_data	*nodes;
	unsigned int		 framelen	= 0;
	unsigned int		 len		= 0;
	unsigned int		 i		= 0;
	unsigned int		 data_left	= 0;
	unsigned char		*data_ptr	= NULL;
	unsigned int		 data_this_packet;
	
	assert (in		!= NULL);
	assert (out		!= NULL);
	assert (out->magic	== VCODEC_DV_ID);
	assert (s		!= NULL);
	assert (s->mtu		>  0);
	
	if (in->data_len > 0) {
		dv_priv_alloc (s, out);	 /* Allocate the private part of the eframe */
		
		out_priv	= (struct dv_priv_eframe*)(out->priv);
		video_frame_copy (out_priv->frame, in);	 /* ... and copy the frame */
		
		nodes		= out_priv->nodes;
		data_ptr	= out_priv->frame->data;
		framelen	= out_priv->frame->data_len;
		data_left	= framelen;
		
		/* Frames always contain an integral number of 80 byte DIF blocks */
		assert ((data_left % 80) == 0);
		
		while (data_left > 0) {		
			data_this_packet = (data_left < (s->mtu/80)*80 ? data_left : (s->mtu/80)*80);

			assert(i < out->len);

			nodes[i].seqno		= i;
			nodes[i].header		= NULL;
			nodes[i].header_len	= 0;
			nodes[i].data		= data_ptr;
			nodes[i].data_len	= data_this_packet;
			nodes[i].mark		= FALSE;
			
			data_left -= data_this_packet;
			data_ptr  += data_this_packet;
			
			i++;
		}
		
		nodes[i-1].mark= TRUE;

		len = i;
	} else {
		debug_msg ("WARNING: empty frame in dv_packetize().\n");
	}

	/* Fill up the output */	
	out->id		= s->id;
	out->len	= len;
	out->datalen	= framelen;
	
	assert (out->magic == VCODEC_DV_ID);
	
	return out->len;	/* return the number of nodes generated */
}


/*******************************************************************************
 * Decoder
 ******************************************************************************/

video_colour_mode_t
dv_decoder_get_cm (void *state)
{
	UNUSED (state);
	
	return YUV_422;
}

display_size_t
dv_decoder_get_size (void *state) 
{
	struct dv_dec_state *s = (struct dv_dec_state*) state;

	if (s != NULL) {
		return s->size;
	} else {
		return DS_PAL;
	}
};

unsigned int
dv_decoder_get_framelen (void *state)
{
	return	display_get_width (dv_decoder_get_size (state)) * 
		display_get_height (dv_decoder_get_size (state)) *
		cm_get_bpp (dv_decoder_get_cm (state));	
}

void *
dv_decoder_create (unsigned id)
{
	struct dv_dec_state	*s;
	
	UNUSED(id);
	
	s = (struct dv_dec_state*) malloc(sizeof(struct dv_dec_state));
	if (s == NULL) {
		fprintf (stderr,"Error: allocating memory for the DV decoder state.\n");
	} else {
		s->decoder			= dv_decoder_new(FALSE, FALSE, FALSE);
		s->decoder->quality		= DV_QUALITY_BEST;
		s->decoder->video->quality	= s->decoder->quality;
		
		/*
		 * Create, by default, the biggest size: if we detect it NTSC,
		 * we'll reduce the size later.
		 */
		s->size	= DS_PAL;
	}
		
	return (void*)s;
}

/* Free decoder resources */
void
dv_decoder_destroy (void *state)
{
	struct dv_dec_state *s = (struct dv_dec_state*) state;

	if (s != NULL) {
		assert( s->decoder != NULL );
		dv_decoder_free(s->decoder);
	
		free(s);
	}
}

/* Decode a frame */
int
dv_decode (void *state, struct coded_data *in, struct video_frame *out)
{
	struct dv_dec_state	*s		= (struct dv_dec_state*) state;
	struct coded_data	*curr		= NULL;
	unsigned int		 i		= 0,
				 offset		= 0;
	unsigned char		*decoder_buf	= NULL;
	unsigned int		 decoder_buf_len;
	int			 pitches[3];
	unsigned char		*pixels[3],
				 tmp[4];
	unsigned char		*pos;
	unsigned char		 sct, dseq, dbn;

	assert (s             != NULL);
	assert (s->decoder    != NULL);
	assert (out           != NULL);
	assert (out->data     != NULL);

	/* Create a buffer for storing the fragments of an encoded frame */
	/* NOTE: According to valgrind libdv-0.104 has a buffer overflow */
	/*       which causes it to access beyond the allocated memory.  */
	/*       We allocate an extra 8 bytes here, to ensure it doesn't */
	/*       hit anything important...                         [csp] */
	decoder_buf_len = dv_encoder_get_framelen();
	decoder_buf = (unsigned char *) malloc (decoder_buf_len + 8);
	if (decoder_buf == NULL) {
		debug_msg ("Error: allocating memory for the DV decoder state.\n");
		dv_decoder_destroy(s);
		return 0;
	}
	memset (decoder_buf + decoder_buf_len, 0, 8);	/* Initialise memory allocated for libdv bug */

#ifdef DEBUG
	/* Zero the decoder_buf, so we can check for multiple writes */
	/* at each location (see test below).                        */
	memset(decoder_buf, 0, decoder_buf_len);
#endif

	/* Create the frame joining the fragments. Each RTP packet may	*/
	/* contain several DIF blocks					*/
	
	//for (curr = in; curr->next != NULL; curr = curr->next) printf (".");
	
	for (curr = in; curr != NULL; curr = curr->next) {
		if ((curr->data->meta.data_len % 80) != 0) {
			/* RFC 3189 requires the payload to contain an integral number */
			/* of 80 byte DIF blocks. Don't try to decode packets that are */
			/* the wrong size.                                             */
			debug_msg ("RTP payload does not contain an integral number of DIF blocks (discarded)\n");
			continue;
		}

		pos = curr->data->meta.data;
		while (pos < (curr->data->meta.data + curr->data->meta.data_len)) {
			sct  = (unsigned) ((pos[0] & 0xe0) >> 5);
			dseq = (unsigned) ((pos[1] & 0xf0) >> 4);
			dbn  = (unsigned)  (pos[2]);

			switch (sct) {
			case 0: /* Header */
				offset = dseq * 12000;	
				break;
			case 1: /* Subcode */
				offset = dseq * 12000 + (dbn + 1) * 80;
				break;
			case 2: /* VAUX */
				offset = dseq * 12000 + (dbn + 3) * 80;
				break;
			case 3: /* Audio */
				offset = dseq * 12000 + 480 + (dbn * 16 * 80);
				break;
			case 4: /* Video */
				offset = dseq * 12000 + 480 + (dbn/15 * 16 * 80) + 80 + (dbn%15 * 80);
				break;
			default:
				debug_msg("Unknown DV frame type\n");
			}

			if ((offset + 80) <= decoder_buf_len) {
#ifdef DEBUG
				/* We should never write the same block of memory twice in a frame. */
				if (decoder_buf[offset] != 0) {
					//debug_msg("Warning duplicate DV block received\n");
				}
#endif
				memcpy (decoder_buf + offset, pos, 80);
			} else {
				debug_msg("DV frame offset out of bounds (%u > %u)\n", offset + 80, decoder_buf_len);
			}
			pos += 80;

		} /* while */

	} /* for */

	if (dv_parse_header (s->decoder, decoder_buf) < 0) {
		debug_msg("Unable to parse DV header\n");
		free(decoder_buf);
		return 0;
	}
	dv_parse_packs (s->decoder, decoder_buf);

	if (dv_is_PAL (s->decoder)) {
		s->size = DS_PAL;
	} else {
		s->size = DS_NTSC;
	}

	/* The DV frame size is only known on parsing the bitstream. Check that we've */
	/* been given a frame of matching size to decode into, and refuse to decode   */
	/* otherwise.                                                                 */
	if (out->data_len < dv_decoder_get_framelen (s)) {
		debug_msg ("Error: cannot decode DV image due to frame length mismatch (%d < %d)\n", out->data_len, dv_decoder_get_framelen (s));
		free(decoder_buf);
		return 0;
	} else {
		/* Update the output frame */
		out->width	 = display_get_width (s->size);
		out->height	 = display_get_height (s->size);
		out->colour_mode = dv_decoder_get_cm (s);
		out->data_len	 = dv_decoder_get_framelen (s);

		/* Decode the frame... */
		pixels[0] = out->data;
		pixels[1] = 0;
		pixels[2] = 0;
		
		pitches[0] = 720 * 2;
		pitches[1] = 0;
		pitches[2] = 0;
		
		dv_decode_full_frame (s->decoder, decoder_buf, e_dv_color_yuv, pixels, pitches);
		
		/* The DV decoder gives us Y1UY2V colour component order, but we need UY1VY2. */
		/* Flip the data around to match the HDTV colour format used elsewhere...     */
		for (i = 0; i < out->width * out->height; i += 2) {
			tmp[0] = out->data[(i * 2) + 0];
			tmp[1] = out->data[(i * 2) + 1];
			tmp[2] = out->data[(i * 2) + 2];
			tmp[3] = out->data[(i * 2) + 3];
#ifdef WORDS_BIGENDIAN
			out->data[(i * 2) + 3] = tmp[1];
			out->data[(i * 2) + 2] = tmp[0];
			out->data[(i * 2) + 1] = tmp[3];
			out->data[(i * 2) + 0] = tmp[2];
#else
			out->data[(i * 2) + 0] = tmp[1];
			out->data[(i * 2) + 1] = tmp[0];
			out->data[(i * 2) + 2] = tmp[3];
			out->data[(i * 2) + 3] = tmp[2];
#endif
		}		
	}
	
	free (decoder_buf);
	
	return TRUE;
}

int
dv_decode_raw(void			*state,
	       struct video_frame	*in,
	       struct video_frame	*out)
{
	struct dv_dec_state	*s = (struct dv_dec_state*) state;
	unsigned int		 i;
	int			 pitches[3];
	unsigned char		*pixels[3];

	assert (s             != NULL);
	assert (s->decoder    != NULL);
	assert (out           != NULL);
	assert (out->data     != NULL);
	assert (in            != NULL);
	assert (in->data      != NULL);

	if ((in->data_len % 80) != 0) {
		debug_msg("Cannot decode DV image: frame not a multiple of 80 octets\n");
		return 0;
	}

	if (dv_parse_header (s->decoder, in->data) < 0) {
		debug_msg("Unable to parse DV header\n");
		return 0;
	}
	dv_parse_packs (s->decoder, in->data);

	if (dv_is_PAL (s->decoder)) {
		s->size = DS_PAL;
	} else {
		s->size = DS_NTSC;
	}

	/* The DV frame size is only known on parsing the bitstream. Check that we've */
	/* been given a frame of matching size to decode into, and refuse to decode   */
	/* otherwise.                                                                 */
	if (out->data_len < dv_decoder_get_framelen (s)) {
		debug_msg("Cannot decode DV image: frame length mismatch (%d < %d)\n", out->data_len, dv_decoder_get_framelen (s));
		return 0;
	} else {
		/* Update the output frame */
		out->width	 = display_get_width (s->size);
		out->height	 = display_get_height (s->size);
		out->colour_mode = dv_decoder_get_cm (s);
		out->data_len	 = dv_decoder_get_framelen (s);

		/* Decode the frame... */
		pixels[0] = out->data;
		pixels[1] = 0;
		pixels[2] = 0;

		pitches[0] = 720 * 2;
		pitches[1] = 0;
		pitches[2] = 0;

		dv_decode_full_frame (s->decoder, in->data, e_dv_color_yuv, pixels, pitches);

		/* The DV decoder gives us Y1UY2V colour component order, but we need UY1VY2. */
		/* Flip the data around to match the HDTV colour format used elsewhere...     */
		for (i = 0; i < out->width * out->height; i += 2) {
			unsigned char	tmp[4];

			tmp[0] = out->data[(i * 2) + 0];
			tmp[1] = out->data[(i * 2) + 1];
			tmp[2] = out->data[(i * 2) + 2];
			tmp[3] = out->data[(i * 2) + 3];
#ifdef WORDS_BIGENDIAN
			out->data[(i * 2) + 3] = tmp[1];
			out->data[(i * 2) + 2] = tmp[0];
			out->data[(i * 2) + 1] = tmp[3];
			out->data[(i * 2) + 0] = tmp[2];
#else
			out->data[(i * 2) + 0] = tmp[1];
			out->data[(i * 2) + 1] = tmp[0];
			out->data[(i * 2) + 2] = tmp[3];
			out->data[(i * 2) + 3] = tmp[2];
#endif
		}		
	}

	return 1;	
}

int
dv_decode_partial(void *state, struct coded_data *in, struct video_frame *out)
{
	debug_msg("partial frame discarded\n");
	UNUSED(state);
	UNUSED(in);
	UNUSED(out);
	return 0;
}

#endif /* HAVE_LIBDV */

