/*
 * FILE:    video_codec/jpeg.c
 * AUTHORS: Colin Perkins <csp@csperkins.org>
 *          Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.8 $
 * $Date: 2006/05/12 15:06:56 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBJPEG

#include "debug.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"	/* Needed by video_codec.h      */
#include "rtp/cc.h"		/* Needed by rtp/rtp.h         */
#include "rtp/rtp.h"		/* Needed by rtp/pbuf.h         */
#include "rtp/pbuf.h"		/* Needed for struct coded_data */
#include "video_codec.h"
#include "video_codec/jpeg.h"
#include "video_codec/jpeg_aux.h"

#define MAX_ENCODER_FRAMELEN	display_get_width (display_get_max()) * \
				display_get_height (display_get_max()) * 3


/*
 * Note: this code is 'over-commented', as it comes from a libjpeg example...
 */

/* Encoder and decoder internal states. */
struct jpeg_enc_state {
	uint32_t			 magic;
	unsigned int			 id;
		
	display_size_t			 size;
	video_colour_mode_t		 color;
	unsigned int			 mtu;
	
	unsigned int			 curr_framesize;
		
	struct libjpeg_enc_state	*libjpeg;
};

struct jpeg_dec_state {
	uint32_t			 magic;
	unsigned int			 id;

	display_size_t			 size;
	video_colour_mode_t		 color;

	struct libjpeg_dec_state	*libjpeg;
};


/*******************************************************************************
 * Codec management and info
 ******************************************************************************/
int
jpeg_codec_init (void)
{
	return TRUE;
}

void
jpeg_codec_exit (void)
{
}

const char *
jpeg_codec_get_name (void)
{
	return "jpeg";
}

const char *
jpeg_codec_get_desc (void)
{
	return "Motion JPEG";
}

int
jpeg_codec_can_encode (void)
{
	return TRUE;
}

int
jpeg_codec_can_packetize(void)
{
	return FALSE;
}

int
jpeg_codec_can_decode (void)
{
	return TRUE;
}


/*******************************************************************************
 * Encoded frames management
 ******************************************************************************/
/* Allocate an encoded frame */
struct encoded_frame*
jpeg_eframe_alloc (void)
{
	struct encoded_frame	*res_eframe		= NULL;
	
	res_eframe = (struct encoded_frame*) malloc (sizeof(struct encoded_frame));
	if (res_eframe != NULL) {
		/* Set some info of the encoded_frame */
		res_eframe->magic	= VCODEC_JPEG_ID;
		res_eframe->priv	= NULL;
		res_eframe->len		= 0;		/* Not known yet */
		res_eframe->datalen	= jpeg_encoder_get_framelen ();
	} else {
		fprintf (stderr, "Error allocating memory for encoded_frame.\n");		
		jpeg_eframe_free (res_eframe);
	}
	
	return res_eframe;	
}

/* Get a fragment of an encoded frame */
struct encoded_data*
jpeg_eframe_get	(struct encoded_frame	*eframe,
		 unsigned int		 pos)
{
	struct encoded_data	*ret	= NULL;
	struct jpeg_enc_outdata	*out;
	
	assert (eframe		!= NULL);
	assert (eframe->priv	!= NULL);
	assert (eframe->magic	== VCODEC_JPEG_ID);
	
	out = (struct jpeg_enc_outdata*) eframe->priv;
	if( pos < out->num_frags ) {
		ret = (struct encoded_data*) malloc (sizeof (struct encoded_data));
		if (ret != NULL) {
			ret->seqno	= pos;
			ret->header_len	= 0;
			ret->header	= NULL;
			ret->data_len	= out->frags_len[pos];
			ret->data	= out->frags[pos];
			
			if (pos + 1 == out->num_frags) {
				ret->mark = 1;		/* Mark last fragment */
			} else {
				ret->mark = 0;
			}
		}
	}
	
	return ret;
}

/* Free an encoded frame */
void
jpeg_eframe_free (struct encoded_frame	*eframe)
{
	struct jpeg_enc_outdata	*out;
	unsigned int		 i;
	
	assert (eframe		!= NULL);
	assert (eframe->priv	!= NULL);
	assert (eframe->magic	== VCODEC_JPEG_ID);
	
	out = (struct jpeg_enc_outdata*) eframe->priv;
	if (out != NULL) {
		/* Free all the fragments */
		for (i = 0; i < out->num_frags; i++) {
			if (out->frags[i] != NULL)
				free (out->frags[i]);
		}
		
		/* Free the output */
		free (out);
	}
	
	free (eframe);	
}

/*******************************************************************************
 * Public encoding and packetization functions 
 ******************************************************************************/

/* Create an encoder */
void*
jpeg_encoder_create (unsigned id, unsigned mtu)
{
	struct jpeg_enc_state	*s	= NULL;
	
	s = (struct jpeg_enc_state*) malloc (sizeof (struct jpeg_enc_state));
	if (s != NULL) {
		s->id			= id;
		s->mtu			= mtu;
		s->size			= DS_NONE;
		s->color		= NONE;
		s->curr_framesize	= jpeg_encoder_get_framelen();
		s->libjpeg		= NULL;
	} else {
		fprintf (stderr, "Error: allocating memory for JPEG encoder state.\n");
	}
	
	return s;
}

/* Destroy the encoder */
void
jpeg_encoder_destroy (void *state)
{
	struct jpeg_enc_state	*s = (struct jpeg_enc_state*) state;

	if (s != NULL) {
		libjpeg_compress_done (s->libjpeg);
		free (s);
	}
}

/* Return the max size of an encoded frame */
unsigned int
jpeg_encoder_get_framelen (void)
{
	/* There shouldn't be any compressed frame bigger than this... */
	return MAX_ENCODER_FRAMELEN;
}

/* Encode a frame */
int
jpeg_encode (void                 *state, 
	     struct video_frame   *in, 
	     struct encoded_frame *out,
	     struct vcodec_ctx_t  *ctx)
{
	struct jpeg_enc_state	*s = (struct jpeg_enc_state*) state;
	struct jpeg_enc_outdata	*enc;
	
	assert (in->width	> 0);
	assert (in->height	> 0);
	assert (in->colour_mode != NONE);
	assert (s		!= NULL);
	assert (s->mtu		> 0);
	
	UNUSED (out);
	
	/* Create an encoder if needed */
	if ((in->width	!= display_get_width (s->size))		||
	    (in->height	!= display_get_height (s->size))	||
	    (in->colour_mode != s->color)) {
		
		/* Destroy previous compressor */
		libjpeg_compress_done (s->libjpeg);
		
		s->libjpeg	= libjpeg_compress_init (in->width, in->height,
							  in->colour_mode,
							  s->mtu);
		s->size		= display_get_size (in->width, in->height);
		s->color	= in->colour_mode;		
	}
	
	/* Initialize the target */
	if (ctx != NULL) {
		s->curr_framesize	= ctx->framesize;
	}
	
	/* Compress */
	enc = (void*) libjpeg_compress (s->libjpeg, s->curr_framesize,
					in->data, in->data_len);

	/* Update pointers in the encoded frame with their real values */
	out->priv	= (void*)enc;
	out->len	= enc->num_frags;
	out->datalen	= enc->datalen;

	if (enc != NULL) {
		return enc->num_frags;
	} else {
		return 0;
	}
}

/* Convert a frame in a sequence of fragments */
int
jpeg_packetize (void                 *state, 
		struct video_frame   *in, 
		struct encoded_frame *out)
{
	UNUSED (state);
	UNUSED (in);
	UNUSED (out);
	
	debug_msg ("Packetization not implemented for JPEG.\n");
	
	return 0;
}


/*******************************************************************************
 * Decoder info
 ******************************************************************************/

/* Return the color mode of a decoded frame */
video_colour_mode_t
jpeg_decoder_get_cm (void *state)
{
	UNUSED (state);
	
	return YUV_422;
}

/* Return the display size of a decoded frame */
display_size_t
jpeg_decoder_get_size (void *state)
{
	struct jpeg_dec_state	*s	= (struct jpeg_dec_state*) state;
	display_size_t		 ret	= display_get_max ();
	
	if (s != NULL) {
		if (s->size != DS_NONE)
			ret = s->size;
	}
	
	return ret;
}

/* Return the length of a decoded frame */
unsigned int
jpeg_decoder_get_framelen (void *state)
{
	return	display_get_width	(jpeg_decoder_get_size (state)) * 
		display_get_height	(jpeg_decoder_get_size (state)) *
		cm_get_bpp		(jpeg_decoder_get_cm (state));
}

/*******************************************************************************
 * Public decoding functions
 ******************************************************************************/
/* Create a new decoder */
void*
jpeg_decoder_create (unsigned id)
{
	struct jpeg_dec_state	*s	= NULL;
	
	s = (struct jpeg_dec_state*) malloc (sizeof (struct jpeg_dec_state));
	if (s != NULL) {
		s->id		= id;
		s->size		= DS_NONE;
		s->color	= NONE;
		s->libjpeg	= NULL;		/* Delay libjpeg initialization */
	} else {
		fprintf (stderr, "Error: allocating memory for JPEG decoder state.\n");
	}
	
	return s;
}

/* Destroy the decoder */
void
jpeg_decoder_destroy (void *state)
{
	struct jpeg_dec_state	*s = (struct jpeg_dec_state*) state;
	
	if (s != NULL) {
		libjpeg_decompress_done (s->libjpeg);
		free (s);
	}
}

/* Decode a frame */
int
jpeg_decode (void *state, struct coded_data *in, struct video_frame *out)
{
	struct jpeg_dec_state	*s	= (struct jpeg_dec_state*) state;
	int			 res	= FALSE;
	
	assert (s		!= NULL);
	assert (in		!= NULL);
	assert (out		!= NULL);
	
	if (s->libjpeg == NULL) {
		s->libjpeg = libjpeg_decompress_init ();		
	}
	
	/* Decompress and update the state */
	if (libjpeg_decompress (s->libjpeg, in, out)) {
		s->size	= display_get_size (out->width, out->height);
		s->color= out->colour_mode;
		res	= TRUE;
	}
	
	return res;
}

/* Decode a raw frame, from a capture device */
int
jpeg_decode_raw (void *state, struct video_frame *in, struct video_frame *out)
{
	UNUSED (state);
	UNUSED (in);
	UNUSED (out);
	
	debug_msg ("decode_raw not imeplemented for JPEG\n");
	
	return FALSE;
}

/* Decode a partial frame */
int
jpeg_decode_partial (void *state, struct coded_data *in, struct video_frame *out)
{
	UNUSED (state);
	UNUSED (in);
	UNUSED (out);

	/* Cross your fingers... */
	return jpeg_decode (state, in, out);
	//return FALSE;
}

#endif

