/*
 * FILE:   video_codec/jpeg_aux.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.12 $
 * $Date: 2006/05/12 15:02:48 $
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBJPEG

#include "debug.h"
#include "trace.h"
#include "trace_ug.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "video_codec/jpeg_aux.h"

#include <setjmp.h>		/* For overriding the error mechanism */

#undef HAVE_STDLIB_H
#include <jpeglib.h>


/* Switches */

/* Uncomment one of these: */
/*#define LIBJPEG_IN_DCT	JDCT_ISLOW  		// slow but accurate integer algorithm	*/
/*#define LIBJPEG_IN_DCT	JDCT_IFAST  		// faster, less accurate integer method	*/
/*#define LIBJPEG_IN_DCT	JDCT_FLOAT		// floating-point method		*/
#define LIBJPEG_IN_DCT		JDCT_DEFAULT		// default method (normally JDCT_ISLOW) */
/*#define LIBJPEG_IN_DCT	JDCT_FASTEST		// fastest method (normally JDCT_IFAST) */ 

/* Selects color dithering method for output. Default is JDITHER_FS. */
/* Uncomment one of these:                                           */
/*#define LIBJPEG_OUT_DITHER	JDITHER_FS		// Floyd-Steinberg dither: slow, high quality	*/
#define LIBJPEG_OUT_DITHER	JDITHER_ORDERED		// ordered dither: moderate speed and quality	*/
/*#define LIBJPEG_OUT_DITHER	JDITHER_NONE		// no dithering: fast, very low quality		*/

/*
 * If non-zero, the input image is smoothed; the value should be 1 for minimal
 * smoothing to 100 for maximum smoothing. Consult jcsample.c for details of the
 * smoothing algorithm. The default is zero.
 */
#define LIBJPEG_SMOOTH_FACT	0

/*
 * To emit restart markers in the JPEG file, set one of these nonzero.
 * Set restart_interval to specify the exact interval in MCU blocks.
 * Set restart_in_rows to specify the interval in MCU rows. (If restart_in_rows
 * is not 0, then restart_interval is set after the image width in MCUs is computed.)
 * Defaults are zero (no restarts). One restart marker per MCU row is often a good
 * choice. NOTE: the overhead of restart markers is higher in grayscale JPEG files
 * than in color files, and MUCH higher in progressive JPEGs. If you use restarts,
 * you may want to use larger intervals in those cases.
 */
#define LIBJPEG_RESTART_ROWS	1

/* Use JPEG headers? */
#define LIBJPEG_USE_HEADERS	FALSE

/*
 * Compression history.
 * libjpeg uses a compression factor, q, for setting the compression level. However,
 * we do not know in advance how much compression it is going to apply to a
 * source image. We need to keep a record of last experiences, in order to estimate
 * the compression level that it is going to apply to a new image.
 *
 * Note: there is a lot of redundant information in this structure...
 */
#define LIBJPEG_MAX_Q	100

struct libjpeg_comp_res {
	unsigned int	q;		/* Q value					*/
	double		comp_ratio;	/* Compression ratio (comp/src size) obtained	*/
	unsigned int	comp_size;	/* Size of compressed image			*/
};

#define LIBJPEG_HIST_SIZE	2	/* Length of the results history.		*/

/* Error management */
struct libjpeg_error_mgr {
	struct jpeg_error_mgr	pub;			/* "public" fields */
	jmp_buf			setjmp_buffer;		/* for return to caller */
};

typedef struct libjpeg_error_mgr	*libjpeg_error_ptr;

/* LibJPEG stuff */
struct libjpeg_enc_state {
	struct jpeg_compress_struct	cinfo;	/* Compressor info */
	struct jpeg_error_mgr		jerr;	/* Standard error */
	struct jpeg_destination_mgr	dest;	/* Sestination manager */
	
	struct libjpeg_comp_res		q_hist[LIBJPEG_HIST_SIZE];
};



struct libjpeg_dec_state {
	struct jpeg_decompress_struct	cinfo;
	struct jpeg_source_mgr		src;

	struct libjpeg_error_mgr	jerr;
};


static JOCTET	libjpeg_eof_buffer[2] = {(JOCTET) 0xFF, (JOCTET) JPEG_EOI};


/*******************************************************************************
 * Compression: history
 ******************************************************************************/

/* Initialize the results history */
static void
libjpeg_q_hist_init (struct libjpeg_comp_res* q_hist)
{
	assert (q_hist != NULL);

	q_hist[0].comp_ratio	= 0.1;
	q_hist[0].q		= 10;
	q_hist[0].comp_size	= 0;
	
	q_hist[1].comp_ratio	= 0.2;
	q_hist[1].q		= 20;
	q_hist[1].comp_size	= 0;
}

/* Register a new value in the history */
static void
libjpeg_q_register (struct libjpeg_comp_res	*q_hist,
		    unsigned int		 q,
		    double			 comp_ratio,
		    unsigned int		 comp_size)
{
	assert (q_hist		!= NULL);
	assert (q		<= LIBJPEG_MAX_Q);
	assert (comp_ratio	<= 1.0);

	/* Rotate the history, shifting values */
	q_hist[0].comp_ratio	= q_hist[1].comp_ratio;
	q_hist[0].comp_size	= q_hist[1].comp_size;
	q_hist[0].q		= q_hist[1].q;

	q_hist[1].comp_ratio	= comp_ratio;
	q_hist[1].comp_size	= comp_size;
	q_hist[1].q		= q;
}

/* Calculate the compression factor, q, for a compression ratio */
static unsigned int
libjpeg_q (struct libjpeg_comp_res* q_hist, double ratio)
{
	double		q_0,	q_1,
			r_0,	r_1;
	unsigned int	q;

	assert (q_hist	!= NULL);
	assert (ratio	!= 0);
	
	/*
	 * Rationale: the 'q' factor is obtained using a function 'q = f(ratio)',
	 * where 'ratio' is the compression ratio that we want.
	 *
	 * The function 'f()' is obtained knowing three points obtained from the
	 * history: (ratio_0, q_0), (ratio_1, q_1) and (ratio_2, q_2).
	 *
	 * We use a simple interpolation function. See 'Numerican Recipes', Sec 3.1,
	 * for interpolation functions with more points...
	 */
	q_0	= q_hist[0].q;
	r_0	= q_hist[0].comp_ratio;

	q_1	= q_hist[1].q;
	r_1	= q_hist[1].comp_ratio;

	if (q_0 != q_1 && r_0 != r_1) {
		q =	((ratio - r_1) / (r_0 - r_1)) * q_0 +
			((ratio - r_0) / (r_1 - r_0)) * q_1;
	} else {
		/*
		 * We have a line! In this case we must use a different
		 * interpolation system, as we would get the same 'q' for
		 * any target ratio...
		 */
		if (r_1 != 0 && q_1 != 0) {
			/*
			 * Ok, we will use the previous sample as a reference.
			 * If we have a previous ratio, use a value proportional
			 * to the previous 'q'...
			 */
			q = q_1 * (ratio / r_1);
		} else {
			/* Otherwise, we must use a simpler 'q' calculation */
			q =  LIBJPEG_MAX_Q * ratio;
		}
	}
	
	/*
	 * A small trick: a lower target ratio must produce a lower q, and vice versa.
	 * 
	 */
	if (ratio < r_1) {
		q = MIN (q, q_1);
	} else if (ratio > r_1) {
		q = MAX (q, q_1);		
	}
	
	return MIN (q, LIBJPEG_MAX_Q);
}

/*******************************************************************************
 * Compression: memory management 
 ******************************************************************************/

/*
 * As the libjpeg does not allow user data in call-back functions, we must use
 * a static struct for some info...
 */
static struct  {
	/* Global settings */
	unsigned int			 mtu;
	
	/* Current output for the encoder */
	unsigned int			 curr_frag;	/* Current fragment */
	struct	jpeg_enc_outdata	*curr_out;	/* Current output */
} libjpeg_enc_static;

/*
 * This is called whenever the buffer has filled (free_in_buffer
 * reaches zero).  In typical applications, it should write out the
 * _entire_ buffer (use the saved start address and buffer length;
 * ignore the current state of next_output_byte and free_in_buffer).
 * Then reset the pointer & count to the start of the buffer, and
 * return TRUE indicating that the buffer has been dumped.
 * free_in_buffer must be set to a positive value when TRUE is
 * returned.  A FALSE return should only be used when I/O suspension is
 * desired (this operating mode is discussed in the next section).
 */	 
static int
libjpeg_comp_dest_empty (j_compress_ptr cinfo)
{
	unsigned char	*frag;
	unsigned int	 frag_len;
	unsigned int	 i;
	
	assert (libjpeg_enc_static.mtu > 0);
	
	/* Allocate a buffer with the MTU as size  */
	frag_len = libjpeg_enc_static.mtu;

	/* Allocate a buffer */
	frag = (unsigned char*) malloc (frag_len);
	if (frag != NULL) {
		/* Update pointers for the output */
		i = libjpeg_enc_static.curr_out->num_frags;
		libjpeg_enc_static.curr_out->frags[i]		= frag;
		libjpeg_enc_static.curr_out->frags_len[i]	= frag_len;
		
		/* Update the destination manager for libjpeg */
		cinfo->dest->next_output_byte			= frag;
		cinfo->dest->free_in_buffer			= frag_len;
		
		/* Increase the number of fragments counter */
		libjpeg_enc_static.curr_out->num_frags ++;
		libjpeg_enc_static.curr_out->datalen += frag_len;
	}

	return TRUE;
}


/* 
 * Initialize destination. This is called by jpeg_start_compress()
 * before any data is actually written. It must initialize
 * next_output_byte and free_in_buffer. free_in_buffer must be
 * initialized to a positive value.
 */
static void
libjpeg_comp_dest_init (j_compress_ptr cinfo)
{
	struct	jpeg_enc_outdata	*out;

	out = (struct jpeg_enc_outdata*) malloc (sizeof (struct jpeg_enc_outdata));
	if (out != NULL) {

		/* Initialize the 'current' vars */ 
		libjpeg_enc_static.curr_frag		= 0;
		libjpeg_enc_static.curr_out		= out;

		libjpeg_enc_static.curr_out->num_frags	= 0;
		libjpeg_enc_static.curr_out->datalen	= 0;

		/*
		 * Initialize the first buffer for a fragment. This will set the
		 * right values for next_output_byte and free_in_buffer.
		 */
		libjpeg_comp_dest_empty (cinfo);
	}
}


/*
 * Terminate destination --- called by jpeg_finish_compress() after all
 * data has been written. 
 */
static void
libjpeg_comp_dest_term (j_compress_ptr cinfo)
{
	unsigned int	last,
			remaining;

	last		= libjpeg_enc_static.curr_out->num_frags;
	remaining	= cinfo->dest->free_in_buffer;
	
	/*
	 * Use free_in_buffer to determine how much data is in the buffer and
	 * update the length of the last fragment.
	 * UPDATE: if we do that, the sending buffer will complain...
	 */
	if (libjpeg_enc_static.curr_out->frags_len[last] > remaining) {
		libjpeg_enc_static.curr_out->frags_len[last]	-= remaining;		
	}
	//
	//if (libjpeg_enc_static.curr_out->datalen > remaining) {
	//	libjpeg_enc_static.curr_out->datalen	-= remaining;	
	//}
	/* Done! */
}


/*******************************************************************************
 * Compression external API
 ******************************************************************************/

/*
 * Internal initialization for libjpeg.
 * This function must be called to set up all the internal structures of libjpeg.
 * Note: the compression level should be adjusted before we start encoding.
 */
struct libjpeg_enc_state*
libjpeg_compress_init	(unsigned int			 width,
			 unsigned int			 height,			
			 video_colour_mode_t		 color,
			 unsigned int			 mtu)
{
	struct libjpeg_enc_state	*ret	= NULL;

	ret = (struct libjpeg_enc_state*) malloc (sizeof (struct libjpeg_enc_state));
	if (ret != NULL) {		
		/* Set some static info */
		libjpeg_enc_static.mtu		= mtu;
		libjpeg_enc_static.curr_out	= NULL;
		libjpeg_enc_static.curr_frag	= 0;

		/* Create the error handler */
		ret->cinfo.err = jpeg_std_error (&ret->jerr);

		/* Now we can initialize the JPEG compression object. */
		jpeg_create_compress (&ret->cinfo);

		/* Specify the destination for the compressed data (eg, a file) */
		ret->dest.init_destination	= libjpeg_comp_dest_init;
		ret->dest.empty_output_buffer	= libjpeg_comp_dest_empty;
		ret->dest.term_destination	= libjpeg_comp_dest_term;
		ret->cinfo.dest = &ret->dest;

		/* Colorspace and bytes per pixel of input image */
		switch (color) {
			case YUV_422:
				ret->cinfo.in_color_space	= JCS_YCbCr;
				ret->cinfo.input_components	= 3;	/* 3 bytes per pixel */
				break;
			/*
			 case RGB:
				debug_msg ("RGB color space.\n");
				ret->cinfo.in_color_space	= JCS_RGB;
				ret->cinfo.input_components	= 3;
				break;
			*/
			default:
				fprintf (stderr, "Error: color type not handled in libjpeg.\n");
				trace (te_jpeg_error);

				libjpeg_compress_done (ret);
				return NULL;
		}
		
		/*
		 * Use the library's routine to set default compression parameters.
		 * You must set in_color_space correctly before calling jpeg_set_defaults().
		 */
		jpeg_set_defaults (&ret->cinfo);
		
		//ret->cinfo.input_gamma		= 1.0;
		ret->cinfo.dct_method		= LIBJPEG_IN_DCT;
		ret->cinfo.write_JFIF_header	= LIBJPEG_USE_HEADERS;
		ret->cinfo.smoothing_factor	= LIBJPEG_SMOOTH_FACT;
		ret->cinfo.restart_in_rows	= LIBJPEG_RESTART_ROWS;
		
		jpeg_set_colorspace (&ret->cinfo, ret->cinfo.in_color_space);

		/* Supply a description of the input image. */
		ret->cinfo.image_width		= width;	/* Image width and height */
		ret->cinfo.image_height		= height;

		//jpeg_suppress_tables (&ret->cinfo, TRUE);
		
		/* Initialize the q history */
		libjpeg_q_hist_init (ret->q_hist);
	} else {
		fprintf (stderr, "Error: allocating memory for libjpeg internal state.\n");
		trace (te_jpeg_error);
	}
	
	return ret;
}

/* Destroy the libjpeg encoder */
void
libjpeg_compress_done (struct libjpeg_enc_state	*s)
{
	if (s != NULL) {
		jpeg_destroy_compress (&s->cinfo);
	}
}

/* Compress a frame */
struct jpeg_enc_outdata*
libjpeg_compress (struct libjpeg_enc_state	*s,
		  unsigned int			 target_size,
		  unsigned char			*in_data,
		  unsigned int			 in_data_len)
{
	unsigned int		 q;
	double			 target_ratio,
				 comp_ratio;
	unsigned int		 orig_size, comp_size;
	struct jpeg_enc_outdata	*ret;
	unsigned int		 width, height,
				 width_ptr, height_ptr;
	struct timeval		 tv_now;
	JSAMPROW		 row_ptr[1];
	int			 Cb, Cr, Y0, Y1;
	char			 line [4000];
	unsigned char		*ptr;

	assert (s		!= NULL);
	assert (target_size	> 0);
	assert (in_data		!= NULL);
	assert (in_data_len	> 0);

	orig_size = in_data_len;
	
	/* Get the compresssion level, q, depending on previous experiences */
	target_ratio	= MIN ((double)target_size / (double)orig_size, 1.0);	
	q		= libjpeg_q (s->q_hist, target_ratio);

	assert (q <= LIBJPEG_MAX_Q);
	
	jpeg_set_quality (&s->cinfo, q, FALSE);

	width = s->cinfo.image_width;
	height = s->cinfo.image_height;

	assert (width > 0 && height > 0);
	
	/*
	 * Start compression.
	 *
	 * Note: this is a completely inefficient version, where we do a copy
	 * converting from packet-YUV to unpacked format. The libjpeg supports
	 * a 'raw' mode, where a list of pointers indicating the layout of the
	 * source image must be provided... any help would be appreciated [as]
	 */
	jpeg_start_compress (&s->cinfo, TRUE);

	row_ptr[0]	= (JSAMPROW)line;
	height_ptr	= 0;
	ptr		= in_data;
	while (height_ptr < height) {
		width_ptr	= 0;
		while (width_ptr < 3 * width) {
			Cb = *ptr++;
			Y0 = *ptr++;
			Cr = *ptr++;
			Y1 = *ptr++;

			line [width_ptr++] = Y0;
			line [width_ptr++] = Cb;
			line [width_ptr++] = Cr;
			line [width_ptr++] = Y1;
			line [width_ptr++] = Cb;
			line [width_ptr++] = Cr;
		}
		jpeg_write_scanlines (&s->cinfo, row_ptr, 1);
		height_ptr++;
	}
	jpeg_finish_compress (&s->cinfo);

	/* Register in the compression history */
	orig_size = in_data_len;
	comp_size = libjpeg_enc_static.curr_out->datalen;	
	comp_ratio = (double)comp_size / (double)orig_size;
	
	libjpeg_q_register (s->q_hist, q, comp_ratio, comp_size);
	
	gettimeofday (&tv_now, NULL);
	trace_val_at (&tv_now, te_jpeg_q,		q);
	trace_val_at (&tv_now, te_jpeg_target_size,	target_size);
	trace_val_at (&tv_now, te_jpeg_in_size,		in_data_len);
	trace_val_at (&tv_now, te_jpeg_comp_size,	comp_size);
	trace_val_at (&tv_now, te_jpeg_t_comp_ratio,	target_ratio);
	trace_val_at (&tv_now, te_jpeg_comp_ratio,	(double)comp_size / (double)orig_size);
	
	ret = libjpeg_enc_static.curr_out;
	
	return ret;
}



/*******************************************************************************
 * Decoding: memory management
 ******************************************************************************/
/*
 * As the libjpeg does not allow user data in call back functions, we must use
 * a static struct for some info...
 */
static struct  {
	struct coded_data		*in_ptr;
} libjpeg_dec_static;

/*
 * Replacement for the standard error_exit method.
 */
static void
libjpeg_dec_error_exit (j_common_ptr cinfo)
{
	libjpeg_error_ptr err;
	
	/* cinfo->err really points to a libjpeg_error_mgr struct, so coerce pointer */
	err = (libjpeg_error_ptr) cinfo->err;
	
	/* Always display the message. */
	/* We could postpone this until after returning, if we chose. */
	(*cinfo->err->output_message) (cinfo);
	
	/* Return control to the setjmp point */
	longjmp (err->setjmp_buffer, 1);
}

/*
 * This routine is called only when the decompressor has failed to find
 * a restart (RSTn) marker where one is expected.  Its mission is to
 * find a suitable point for resuming decompression.  For most
 * applications, we recommend that you just use the default resync
 * procedure, jpeg_resync_to_restart().  However, if you are able to back
 * up in the input data stream, or if you have a-priori knowledge about
 * the likely location of restart markers, you may be able to do better.
 * Read the read_restart_marker() and jpeg_resync_to_restart() routines
 * in jdmarker.c if you think you'd like to implement your own resync
 * procedure.
 */ 
static int
libjpeg_dec_src_resync (j_decompress_ptr cinfo, int desired)
{
	int ret = FALSE;
	
	UNUSED (desired);

	if (libjpeg_dec_static.in_ptr != NULL) {
		/*
		 * An 'struct coded_data' has the annoying feature of providing all the
		 * fragments of the frame in _reverse_ order. So we must go to the end,
		 * and follow all the 'prev' pointers when we fill the buffer...
		 */
		while (libjpeg_dec_static.in_ptr->next != NULL) {
			libjpeg_dec_static.in_ptr = libjpeg_dec_static.in_ptr->next;
		}
		
		cinfo->src->next_input_byte = libjpeg_dec_static.in_ptr->data->meta.data;
		cinfo->src->bytes_in_buffer = libjpeg_dec_static.in_ptr->data->meta.data_len;
	
		ret = TRUE;
	} else {
	//	cinfo->src->next_input_byte = NULL;
	//	cinfo->src->bytes_in_buffer = 0;		
	}
	
	return ret;
}

/*
 * Initialize source. This is called by jpeg_read_header() before any
 * data is actually read.  Unlike init_destination(), it may leave
 * bytes_in_buffer set to 0 (in which case a fill_input_buffer() call
 * will occur immediately).
 */ 
static void
libjpeg_dec_src_init (j_decompress_ptr cinfo)
{	
	assert (libjpeg_dec_static.in_ptr != NULL);

	/* Go to the begining of the frame */
	libjpeg_dec_src_resync (cinfo, 0);
}

/*
 * This is called whenever bytes_in_buffer has reached zero and more
 * data is wanted.  In typical applications, it should read fresh data
 * into the buffer (ignoring the current state of next_input_byte and
 * bytes_in_buffer), reset the pointer & count to the start of the
 * buffer, and return TRUE indicating that the buffer has been reloaded.
 * It is not necessary to fill the buffer entirely, only to obtain at
 * least one more byte.  bytes_in_buffer MUST be set to a positive value
 * if TRUE is returned.  A FALSE return should only be used when I/O
 * suspension is desired (this mode is discussed in the next section).
 */
static int
libjpeg_dec_src_fill (j_decompress_ptr cinfo)
{
	struct coded_data	*ptr;
	
	assert (libjpeg_dec_static.in_ptr != NULL);
	
	/* Go forward, to the next node */
	ptr				= libjpeg_dec_static.in_ptr->prev;
	libjpeg_dec_static.in_ptr	= ptr;

	/* Update pointers */
	if (ptr != NULL) {
		assert (ptr->data		!= NULL);
		assert (ptr->data->meta.data	!= NULL);
		assert (ptr->data->meta.data_len > 0);

		cinfo->src->next_input_byte = ptr->data->meta.data;
		cinfo->src->bytes_in_buffer = ptr->data->meta.data_len;
	} else {
		/* No more nodes? Then insert a fake EOI marker */
		cinfo->src->next_input_byte	= &libjpeg_eof_buffer[0];
		cinfo->src->bytes_in_buffer	= 2;
	}
	
	return TRUE;
}

/*
 * Skip num_bytes worth of data.  The buffer pointer and count should
 * be advanced over num_bytes input bytes, refilling the buffer as
 * needed.  This is used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).  In some applications
 * it may be possible to optimize away the reading of the skipped data,
 * but it's not clear that being smart is worth much trouble; large
 * skips are uncommon.  bytes_in_buffer may be zero on return.
 * A zero or negative skip count should be treated as a no-op.
 */
static void
libjpeg_dec_src_skip (j_decompress_ptr cinfo, long num_bytes)
{
	UNUSED (cinfo);
	UNUSED (num_bytes);

	debug_msg ("libjpeg_dec_src_skip not implemented.\n");
}

/*
 * Terminate source. Called by jpeg_finish_decompress() after all
 * data has been read. Often a no-op.
 */
static void
libjpeg_dec_src_term (j_decompress_ptr cinfo)
{
	UNUSED (cinfo);

	cinfo->src->next_input_byte = NULL;
	cinfo->src->bytes_in_buffer = 0;

	libjpeg_dec_static.in_ptr = NULL;
}


/*******************************************************************************
 * Decoding: external API
 ******************************************************************************/

/*
 * Internal initialization for libjpeg.
 * This function must be called to set up all the internal structures of libjpeg.
 * Note: the compression level should be adjusted before we start encoding.
 */
struct libjpeg_dec_state*
libjpeg_decompress_init (void)
{
	struct libjpeg_dec_state	*ret	= NULL;

	ret = (struct libjpeg_dec_state*) malloc (sizeof (struct libjpeg_dec_state));
	if (ret != NULL) {
		/* Create the error handler */
		ret->cinfo.err = jpeg_std_error (&ret->jerr.pub);
		ret->jerr.pub.error_exit = libjpeg_dec_error_exit;

		/* Initialize the JPEG compression object. */
		jpeg_create_decompress (&ret->cinfo);		

		/* Initialize some static data */
		libjpeg_dec_static.in_ptr	= NULL;

		/* Specify the source for compressed data */
		ret->src.init_source		= libjpeg_dec_src_init;
		ret->src.fill_input_buffer	= libjpeg_dec_src_fill;
		ret->src.skip_input_data	= libjpeg_dec_src_skip;
		ret->src.resync_to_restart	= libjpeg_dec_src_resync;
		ret->src.term_source		= libjpeg_dec_src_term;
		ret->cinfo.src = &ret->src;
	} else {
		fprintf (stderr, "Error: allocating memory for libjpeg internal state.\n");
		trace (te_jpeg_error);
	}
	
	return ret;
}

/* Destroy the libjpeg encoder */
void
libjpeg_decompress_done (struct libjpeg_dec_state *s)
{
	if (s != NULL) {
		jpeg_destroy_decompress (&s->cinfo);
	}
}

/* Decompress a frame */
int
libjpeg_decompress (struct libjpeg_dec_state	*s,
		    struct coded_data		*in_ptr,
		    struct video_frame		*out)
{
	unsigned char	*out_data;
	unsigned int	 out_data_len;
	unsigned int	 width, height, width_ptr, height_ptr, bpp;
	JSAMPROW	 row_ptr[1];
	int		 Cb, Cr, Y0, Y1;
	char		 line [4000];
	unsigned char	*ptr;
	unsigned int	 res		= FALSE;
	
	assert (s		!= NULL);
	assert (in_ptr		!= NULL);
	assert (in_ptr->data	!= NULL);

	/* Set the pointer for input data */
	libjpeg_dec_static.in_ptr	= in_ptr;	
	
	/* Establish the setjmp return context for libjpeg_error_exit to use. */
	if (setjmp (s->jerr.setjmp_buffer)) {
		/*
		 * If we get here, the JPEG code has signaled an error.
		 * Print some info...
		 */
		fprintf (stderr, "Wrong packet: seq#: %d length: %d data: %x %x %x %x...\n",
			 in_ptr->seqno,
			 in_ptr->data->meta.data_len,
			 in_ptr->data->meta.data[0],
			 in_ptr->data->meta.data[1],
			 in_ptr->data->meta.data[2],
			 in_ptr->data->meta.data[3]);
		
		trace (te_jpeg_decomp_error);
		
		s->cinfo.src->next_input_byte = NULL;
		s->cinfo.src->bytes_in_buffer = 0;
		
		libjpeg_dec_static.in_ptr = NULL;
		
		jpeg_abort_decompress (&s->cinfo);
		
		return FALSE;
	}

	/* Read file parameters with jpeg_read_header() */
	jpeg_read_header (&s->cinfo, TRUE);
	
	s->cinfo.dither_mode		= LIBJPEG_OUT_DITHER;
	s->cinfo.do_fancy_upsampling	= FALSE;
	s->cinfo.do_block_smoothing	= TRUE;
	s->cinfo.out_color_space	= JCS_YCbCr;	
	
	width	= s->cinfo.image_width;
	height	= s->cinfo.image_height;
	bpp	= s->cinfo.num_components;

	/* FIXME: we should take into account other colour modes... */
	switch (s->cinfo.jpeg_color_space) {
		case JCS_YCbCr:
			out->colour_mode = YUV_422;
			break;
		default:
			fprintf (stderr, "Error: unknown color mode for input JPEG frames.\n");
			trace (te_jpeg_error);
			return FALSE;
	}

	/* Do we have enough space in the output frame? */
	out_data_len = width * height * cm_get_bpp (out->colour_mode);
	if (out_data_len > out->data_len) {
		
		/* Allocate a new buffer for the output */
		out_data = malloc (out_data_len);
		if (out_data != NULL) {

			/* Update the old frame, using the new buffer */
			free (out->data);
			out->data	= out_data;
			out->data_len	= out_data_len;
		} else {
			fprintf (stderr, "Error: allocating memory for frame.\n");
			trace (te_jpeg_error);
			return FALSE;
		}
	} else {
		out_data = out->data;
	}

	/* Start decompression. */
	jpeg_start_decompress (&s->cinfo);
	
	/* Decompress */
	row_ptr[0]	= (JSAMPROW)line;
	height_ptr	= 0;
	ptr		= out_data;

	while (height_ptr < height) {
		jpeg_read_scanlines (&s->cinfo, row_ptr, 1);

		/* Convert unpacked to packet YUV */
		width_ptr	= 0;
		while (width_ptr < 3 * width) {			
			Y0 = line [width_ptr++];
			Cb = line [width_ptr++];
			Cr = line [width_ptr++];
			Y1 = line [width_ptr++];
			Cb = line [width_ptr++];
			Cr = line [width_ptr++];
			
			*ptr++ = Cb;
			*ptr++ = Y0;
			*ptr++ = Cr;
			*ptr++ = Y1;
		}
		height_ptr++;
	}

	jpeg_finish_decompress (&s->cinfo);
	assert (libjpeg_dec_static.in_ptr == NULL);
	
	/* Finish the frame */
	out->width	= s->cinfo.output_width;
	out->height	= s->cinfo.output_height;
	
	res = TRUE;

	return res;
}

#endif

