/*
 * FILE:   video_codec/mpeg1.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.1 $
 * $Date: 2006/05/12 15:00:57 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBAVCODEC

#include "debug.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "video_codec/mpeg1_aux.h"
#include "video_codec/mpeg1.h"

#define MAX_ENCODER_FRAMELEN	(display_get_width (display_get_max()) * \
				display_get_height (display_get_max()) * 3)

/* Encoder internal states. */
struct mpeg1_enc_state {
	uint32_t		 magic;		/* For debugging	*/
	unsigned int		 id;		/* Codec id		*/
	
	struct ffmpeg_state	*ffmpeg;
	
	display_size_t		 size;		/* Image size		*/	
	unsigned int		 mtu;		/* MTU			*/
	video_colour_mode_t	 color;
};

/* Decoder state */
struct mpeg1_dec_state {
	uint32_t		 magic;		/* For debugging	*/
	unsigned int		 id;		/* Codec id		*/

	struct ffmpeg_state	*ffmpeg;	
};


int
mpeg1_codec_init (void)
{
	return TRUE;	/* Codec always available */
}

void
mpeg1_codec_exit(void)
{
	/* Nothing to do */
}

const char *
mpeg1_codec_get_name(void)
{
	return "mpeg1";
}

const char *
mpeg1_codec_get_desc (void)
{
	return "MPEG1";
}

int
mpeg1_codec_can_encode (void)
{
	return TRUE;
}

int
mpeg1_codec_can_packetize (void)
{
	return TRUE;
}

int
mpeg1_codec_can_decode (void)
{
	return TRUE;
}

#pragma mark-

/*******************************************************************************
 * Encoded frames management
 ******************************************************************************/
struct encoded_frame*
mpeg1_eframe_alloc (void)
{
	struct encoded_frame	*res_eframe		= NULL;
	
	res_eframe = (struct encoded_frame*) malloc (sizeof(struct encoded_frame));
	if (res_eframe != NULL) {
		/* Set some info of the encoded_frame */
		res_eframe->magic	= VCODEC_MPEG1_ID;
		res_eframe->len		= 0;			/* Not know yet */
		res_eframe->datalen	= 0;			/* Not know yet */
		res_eframe->priv	= NULL;
	} else {
		fprintf (stderr, "Error allocating memory for encoded_frame.\n");		
		mpeg1_eframe_free (res_eframe);
	}
	
	return res_eframe;
}

struct encoded_data*
mpeg1_eframe_get (struct encoded_frame *eframe, unsigned int pos)
{
	struct ffmpeg_eframe	*ef	= (struct ffmpeg_eframe*)(eframe->priv);
	struct encoded_data	*nodes	= NULL,
				*res	= NULL;
	
	assert (eframe		!= NULL);
	assert (eframe->magic	== VCODEC_MPEG1_ID);
	assert (ef->nodes	!= NULL);
	assert (eframe->priv	!= NULL);
	
	nodes = ef->nodes;
	
	if(pos < eframe->len) {
		res = (nodes + pos);
	}
	
	return res;
}

void
mpeg1_eframe_free (struct encoded_frame *eframe)
{
	struct ffmpeg_eframe	*eframe_priv;
	
	assert (eframe        != NULL);
	assert (eframe->magic == VCODEC_MPEG1_ID);
	
	eframe_priv = (struct ffmpeg_eframe*) eframe->priv;
	if (eframe_priv != NULL) {
		if (eframe_priv->nodes != NULL)	free (eframe_priv->nodes);
		
		free (eframe_priv->buf);		
		free (eframe_priv);
	}
	
	free (eframe);
}


/*******************************************************************************
 * Encoder
 ******************************************************************************/
void*
mpeg1_encoder_create (unsigned id, unsigned mtu)
{
	struct mpeg1_enc_state	*s	= NULL;
	
	s = (struct mpeg1_enc_state*) malloc (sizeof (struct mpeg1_enc_state));
	if (s != NULL) {
		s->id		= id;
		s->mtu		= mtu;
		s->size		= DS_NONE;
		s->color	= NONE;
		s->ffmpeg	= NULL;
	} else {
		fprintf(stderr, "Error: allocating memory for MPEG1 encoder state.\n");
	}
	
	return s;
}

/* Destroy an encoder */
void
mpeg1_encoder_destroy (void *state)
{
	struct mpeg1_enc_state *s = (struct mpeg1_enc_state*) state;
	
	if (s != NULL) {
		ffmpeg_comp_done (s->ffmpeg);
		free (s);		
	}
}

/* Get the length for an encoded frame */
unsigned int
mpeg1_encoder_get_framelen (void)
{
	unsigned int ret = MAX_ENCODER_FRAMELEN;		/* Max size: PAL */
	
	return ret;
}

/* Encode an image */
int
mpeg1_encode (void                 *state,
	      struct video_frame   *in,
	      struct encoded_frame *out,
	      struct vcodec_ctx_t  *ctx)
{
	struct mpeg1_enc_state	*s		= (struct mpeg1_enc_state*) state;
	struct ffmpeg_eframe	*out_priv	= NULL;
	int			 ret		= 0;
	
	assert (s != NULL);
	
	/* Initialize FFmpeg if needed */
	if (in->height		!= display_get_height (s->size) ||
	    in->width		!= display_get_width (s->size) ||
	    in->colour_mode	!= s->color) {

		ffmpeg_comp_done (s->ffmpeg);

		/* Update the state */
		s->size		= display_get_size (in->width, in->height);
		s->color	= in->colour_mode;
		s->ffmpeg	= ffmpeg_comp_init (s->size, in->colour_mode);
	}

	if (s->ffmpeg != NULL) {
		out_priv = ffmpeg_comp (s->ffmpeg, in, s->mtu, ctx);
		
		/* Fill up the output */
		out->id		= s->id;
		out->len	= out_priv->len;
		out->datalen	= out_priv->buf_len;
		out->priv	= out_priv;
		
		ret		= out->len;
	} else {
		fprintf (stderr, "Error: ffmpeg has not been properly initialized.\n");
		mpeg1_encoder_destroy (s);
	}

	return ret;
}

/* Packetize an image */
int
mpeg1_packetize (void                 *state,
		 struct video_frame   *in,
		 struct encoded_frame *out)
{
	struct mpeg1_enc_state	*s		= (struct mpeg1_enc_state*) state;
	struct ffmpeg_eframe	*out_priv;
	unsigned int		 ret		= 0;
	
	assert (in		!= NULL);
	assert (out		!= NULL);
	assert (out->magic	== VCODEC_MPEG1_ID);
	assert (s		!= NULL);
	assert (s->mtu		>  0);

	/* Initialize FFmpeg if needed */
	if (in->height		!= display_get_height (s->size) ||
	    in->width		!= display_get_width (s->size) ||
	    in->colour_mode	!= s->color) {
		
		ffmpeg_comp_done (s->ffmpeg);
		
		/* Update the state */
		s->size		= display_get_size (in->width, in->height);
		s->color	= in->colour_mode;
		s->ffmpeg	= ffmpeg_comp_init (s->size, in->colour_mode);
	}	

	if (s->ffmpeg != NULL) {
		/* Packetize */
		out_priv = ffmpeg_packetize (in->data, in->data_len, s->mtu);
		
		/* Fill up the output */
		out->id		= s->id;
		out->len	= out_priv->len;
		out->datalen	= out_priv->buf_len;
		out->priv	= out_priv;
		
		assert (out->magic == VCODEC_MPEG1_ID);
		
		ret = out->len;	/* return the number of nodes generated */
	} else {
		fprintf (stderr, "Error: no valid ffmpeg state.\n");
	}
	
	return ret;
}


/*******************************************************************************
 * Decoder
 ******************************************************************************/

video_colour_mode_t
mpeg1_decoder_get_cm (void *state)
{
	UNUSED (state);
	
	return YUV_422;
}

/* Get the frame dimensions, or the maximum possible value */
display_size_t
mpeg1_decoder_get_size (void *state) 
{
	display_size_t		ffmpeg_size;
	display_size_t		ret	= display_get_max ();
	struct mpeg1_dec_state *s	= (struct mpeg1_dec_state*) state;

	assert (s	 != NULL);
	assert (s->magic == VCODEC_MPEG1_ID);
	
	if (s != NULL)
		if (s->ffmpeg != NULL) {
			ffmpeg_size = ffmpeg_decomp_get_size (s->ffmpeg);
			if (ffmpeg_size != DS_NONE) {
				ret = ffmpeg_size;				
			}
		}

	return ret;
}

/* Get the frame length, or the maximum frame length possible */
unsigned int
mpeg1_decoder_get_framelen (void *state)
{
	return	display_get_width (mpeg1_decoder_get_size (state)) * 
		display_get_height (mpeg1_decoder_get_size (state)) *
		cm_get_bpp (mpeg1_decoder_get_cm (state));	
}

void*
mpeg1_decoder_create (unsigned id)
{
	struct mpeg1_dec_state	*s;
	
	UNUSED(id);
	
	s = (struct mpeg1_dec_state*) malloc (sizeof (struct mpeg1_dec_state));
	if (s == NULL) {
		fprintf (stderr,"Error: allocating memory for the mpeg1 decoder state.\n");
	} else {
		s->magic	= VCODEC_MPEG1_ID;
		s->ffmpeg	= NULL;
	}
	
	return (void*)s;
}

void
mpeg1_decoder_destroy (void *state)
{
	struct mpeg1_dec_state *s = (struct mpeg1_dec_state*) state;
	
	if (s != NULL) {
		ffmpeg_decomp_done (s->ffmpeg);
		free (s);
	}
}

/* Decode a frame */
int
mpeg1_decode (void *state, struct coded_data *in, struct video_frame *out)
{
	struct mpeg1_dec_state	*s		= (struct mpeg1_dec_state*) state;

	assert (s             != NULL);
	assert (out           != NULL);
	assert (out->data     != NULL);
		
	/* Initialize FFmpeg if needed */
	if (s->ffmpeg == NULL) {		
		s->ffmpeg = ffmpeg_decomp_init ();
	}

	assert (s->ffmpeg	!= NULL);

	return ffmpeg_decomp (s->ffmpeg, in, out);	
}

int
mpeg1_decode_raw (void			*state,
		  struct video_frame	*in,
		  struct video_frame	*out)
{
	UNUSED (state);
	UNUSED (in);
	UNUSED (out);
	
	debug_msg ("mpeg1_decode_raw() not implemented.\n");
	
	return 0;	
}

int
mpeg1_decode_partial (void *state, struct coded_data *in, struct video_frame *out)
{
	return mpeg1_decode (state, in, out);
}

#endif /* HAVE_LIBAVCODEC */

