/*
 * FILE:   video_codec/mpeg1_aux.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.1 $
 * $Date: 2006/05/12 15:00:57 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBAVCODEC

#include "debug.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "video_convert.h"

#include <ffmpeg/avcodec.h>

#include "video_codec/mpeg1_aux.h"

/* Codec ID used in FFmpeg */
#define AVCODEC_ID		CODEC_ID_MPEG1VIDEO	// MPEG-1
//#define AVCODEC_ID		CODEC_ID_MPEG2VIDEO	// MPEG-2
//#define AVCODEC_ID		CODEC_ID_H263		// H-263
//#define AVCODEC_ID		CODEC_ID_H263P		// H-263+
//#define AVCODEC_ID		CODEC_ID_RAWVIDEO	// A testing format, with no compression
//#define AVCODEC_ID		CODEC_ID_H264		// H.264

/* Printable name */
#define AVCODEC_NAME		"mpeg1"

/*
 * Codec color mode.
 * Some codecs only support one color mode. Defining this constant, all frames 
 * will be converted to this mode...
 */
#define AVCODEC_CM		YV_12
//#define AVCODEC_CM		YUV_422


/* The compressor should never return a frame bigger than this... */
#define MAX_FRAMELEN		display_get_width (display_get_max()) * \
				display_get_height (display_get_max()) * 3

#define FFMPEG_INIT_FRAMERATE	25
#define FFMPEG_INIT_BITRATE	MAX_FRAMELEN*8*FFMPEG_INIT_FRAMERATE

struct ffmpeg_state {
	AVCodec			*codec;
	AVCodecContext		*ctx;
};


/* Allocate private structures... */
static struct ffmpeg_eframe*
ffmpeg_priv_alloc (unsigned int		 max_framelen,
		   unsigned int		 mtu)
{
	unsigned int		 max_nodes;
	struct ffmpeg_eframe	*res		= NULL;

	res = (struct ffmpeg_eframe*) malloc (sizeof (struct ffmpeg_eframe));	
	if (res != NULL) {		
		/* 
		 * Allocate memory for the nodes, estimating how many packets
		 * we will generate for a frame. We should have more nodes than
		 * expected: some wasted memory, but no segmentation fault! 
		 */
		max_nodes = max_framelen / mtu;
		
		res->nodes = (struct encoded_data*)
			malloc (max_nodes * sizeof (struct encoded_data));
		if (res->nodes == NULL) {
			fprintf (stderr, "Error: allocating memory for %d %s encoder nodes.\n", max_nodes, AVCODEC_NAME);
			return NULL;
		}
		
		res->buf = malloc (max_framelen);
		if (res->buf == NULL) {
			fprintf (stderr, "Error: allocating memory for %d %s encoder buffer.\n", max_nodes, AVCODEC_NAME);
			return NULL;
		}
		
		res->buf_len = max_framelen;		
	} else {
		fprintf (stderr, "Error allocating memory private data of encoded_frame.\n");
	}
	
	return res;
}


/*******************************************************************************
 * Compression
 ******************************************************************************/
struct ffmpeg_state*
ffmpeg_comp_init (display_size_t size, video_colour_mode_t cm)
{
	struct ffmpeg_state	*ret;
	
	assert (size	!= DS_NONE);

	ret = (struct ffmpeg_state*) malloc (sizeof (struct ffmpeg_state));
	if (ret) {
		ret->codec	= avcodec_find_encoder (AVCODEC_ID);	
		if (!ret->codec) {
			avcodec_init ();
			avcodec_register_all ();
			
			ret->codec = avcodec_find_encoder (AVCODEC_ID);
			if (!ret->codec) {
				fprintf (stderr, "Error: there is no support for this codec in libavcodec.\n");
				return FALSE;
			}
		}
		
		ret->ctx = avcodec_alloc_context ();
		if (ret->ctx == NULL) {
			fprintf (stderr, "Error: allocating context for %s codec.\n", AVCODEC_NAME);
			return FALSE;
		}
		
		/* Initialize the context */
		memset (ret->ctx, 0, sizeof (ret->ctx));
		avcodec_get_context_defaults (ret->ctx);
		
		ret->ctx->width		= display_get_width (size);
		ret->ctx->height	= display_get_height (size);

		switch (cm) {		
			case RGB:	ret->ctx->pix_fmt = PIX_FMT_RGB24;
					break;
			case YUV_422:	ret->ctx->pix_fmt = PIX_FMT_YUV422;
					break;
			case YV_12:	ret->ctx->pix_fmt = PIX_FMT_YUV420P;
					break;
			default:	fprintf (stderr, "Error: colour mode not supported.\n");
					return NULL;
		}
			
		ret->ctx->gop_size	= 0;		/* 0 = intra frames only */
		ret->ctx->me_method	= ME_ZERO;	/* motion estimation type */
		
		debug_msg ("FFmpeg codec  : %s (#%d)\n", ret->codec->name, ret->ctx->pix_fmt);
		debug_msg ("       bitrate: %d\n", ret->ctx->bit_rate);
		
		if (avcodec_open (ret->ctx, ret->codec) < 0) {
			fprintf (stderr, "Error: %s codec can not be opened.\n", AVCODEC_NAME);
			ret = NULL;
		}
	} else {
		fprintf (stderr, "Error: allocating memory for %s codec state.\n", AVCODEC_NAME);
	}
	
	return ret;
}

/* Finalize the FFmpeg system */
void
ffmpeg_comp_done (struct ffmpeg_state* s)
{
	if (s != NULL) {
		if (s->ctx != NULL)
			avcodec_close (s->ctx);
		
		free (s);
	}
}


/* Packetize a frame */
static unsigned int
ffmpeg_packetize_aux (struct ffmpeg_eframe *out_priv, unsigned int mtu)
{
	struct encoded_data	*nodes;
	unsigned int		 i		= 0;
	unsigned int		 data_left	= 0;
	unsigned char		*data_ptr	= NULL;
	unsigned int		 data_this_packet;

	nodes		= out_priv->nodes;
	data_ptr	= out_priv->buf;
	data_left	= out_priv->buf_len;
	
	/* Create the lis of nodes, iterating for the hole image */
	while (data_left > 0) {		
		data_this_packet = MIN (data_left, mtu);
		
		nodes[i].seqno		= i;
		nodes[i].header		= NULL;
		nodes[i].header_len	= 0;
		nodes[i].data		= data_ptr;
		nodes[i].data_len	= data_this_packet;
		nodes[i].mark		= 0;
		
		data_left -= data_this_packet;
		data_ptr  += data_this_packet;
		
		i++;
	}

	out_priv->len = i;
	
	return i;	/* return the number of nodes generated */	
}

/*
 * Converts the input frame to a FFMpeg frame with the correct colour mode.
 */
static AVFrame*
ffmpeg_comp_pict (struct video_frame	*in,
		  video_colour_mode_t	 cm,
		  enum PixelFormat	*ret_pf)
{
	AVFrame			*dest		= NULL,
				*in_conv;
	enum PixelFormat	 dest_pf	= PIX_FMT_NB,
				 in_conv_pf;
	
	if (in->data != NULL && in->data_len > 0) {
		dest	= avcodec_alloc_frame ();
		in_conv	= avcodec_alloc_frame ();
		
		/* Create the pointers for the image */
		switch (in->colour_mode) {
			case RGB:
				/* RGB 24bits */
				in_conv->data[0]	= in->data;
				in_conv->data[1]	= NULL;
				in_conv->data[2]	= NULL;
				in_conv->data[3]	= NULL;
				in_conv->linesize[0]	= in->width*3;
				in_conv->linesize[1]	= 0;
				in_conv->linesize[2]	= 0;
				in_conv->linesize[3]	= 0;
				in_conv_pf		= PIX_FMT_RGB24;
				break;
			case YUV_422:
				/* Packed YUV */
				in_conv->data[0]	= in->data;
				in_conv->data[1]	= NULL;
				in_conv->data[2]	= NULL;
				in_conv->data[3]	= NULL;
				in_conv->linesize[0]	= in->width*2;
				in_conv->linesize[1]	= 0;
				in_conv->linesize[2]	= 0;
				in_conv->linesize[3]	= 0;
				in_conv_pf		= PIX_FMT_YUV422;
				break;
			case YV_12:
				/* Planar YUV */
				in_conv->data[0]	= in->data;
				in_conv->data[1]	= in_conv->data[1] + in->data_len;
				in_conv->data[2]	= in_conv->data[2] + in->data_len/4;
				in_conv->data[3]	= NULL;
				in_conv->linesize[0]	= in->width;
				in_conv->linesize[1]	= in->width/2;
				in_conv->linesize[2]	= in->width/2;
				in_conv->linesize[3]	= 0;
				in_conv_pf		= PIX_FMT_YUV420P;
				break;
			default:
				fprintf (stderr, "Error: colour mode (#%d) not handled.\n",
					 in->colour_mode);
				return NULL;
		}	
		
		/* Find the destination colour mode in FFMpeg terminology */
		switch (cm) {
			case RGB:		dest_pf	= PIX_FMT_RGB24;	break;
			case YUV_422:		dest_pf	= PIX_FMT_YUV422;	break;
			case YV_12:		dest_pf	= PIX_FMT_YUV420P;	break;
			default:
				fprintf (stderr, "Error: colour mode (#%d) not handled.\n", cm);
				return NULL;
		}
		
		/* Convert the image */
		img_convert ((AVPicture*)dest,		dest_pf,
			     (AVPicture*)in_conv,	in_conv_pf,
			     in->width, in->height);		
	}
	
	*ret_pf = dest_pf;	
	return dest;
}


/* Compress a frame */
struct ffmpeg_eframe*
ffmpeg_comp (struct ffmpeg_state	*s,
	     struct video_frame		*in,
	     unsigned int		 mtu,
	     struct vcodec_ctx_t	*ctx)
{
	AVFrame			*frame;
	unsigned int		 real_len,
				 out_len;
	struct ffmpeg_eframe	*ret		= NULL;
	enum PixelFormat	 avcodec_pf;
	
	assert (s	!= NULL);
	assert (s->ctx	!= NULL);

	if (in != NULL) {
		
		out_len = in->data_len;			/* Guess the output len */
		
		/* Get a FFMpeg frame from the input frame */
		frame = ffmpeg_comp_pict (in, AVCODEC_CM, &avcodec_pf);

		ret = ffmpeg_priv_alloc (out_len, mtu);
		if (ret != NULL) {
			
			//frame.pts			= frame_num;
			//frame.display_picture_number	= frame_num++;
			
			/* Ajust some compression parameters, depending on the context provided */
			assert (ctx->framesize	> 0);
			assert (ctx->bitrate	> 0);
			
			s->ctx->bit_rate	= ctx->bitrate;
			s->ctx->frame_rate_base	= 1;
			s->ctx->frame_rate	= (ctx->bitrate*8) / ctx->framesize;
			s->ctx->pix_fmt		= avcodec_pf;
			
			/* Compress  */	
			real_len = avcodec_encode_video (s->ctx,
							 ret->buf, ret->buf_len,	/* Output */
							 frame);			/* Input */							 

			ret->buf_len = real_len;
							 
			/* Packetize */
			ffmpeg_packetize_aux (ret, mtu);
		}
	}

	return ret;
}

/* Packetize a frame */
struct ffmpeg_eframe*
ffmpeg_packetize (unsigned char		*data,
		  unsigned int		 data_len,
		  unsigned int		 mtu)
{
	struct ffmpeg_eframe	*out_priv	= NULL;
	
	out_priv = ffmpeg_priv_alloc (data_len, mtu);
	if (out_priv != NULL) {
		out_priv->buf		= data;
		out_priv->buf_len	= data_len;
		
		ffmpeg_packetize_aux (out_priv, mtu);
	}
	
	return out_priv;
}


/*******************************************************************************
 * Decompression
 ******************************************************************************/
struct ffmpeg_state*
ffmpeg_decomp_init (void)
{
	struct ffmpeg_state	*ret	= NULL;
	
	ret = (struct ffmpeg_state*) malloc (sizeof (struct ffmpeg_state));
	if (ret != NULL) {
		ret->codec	= avcodec_find_encoder (AVCODEC_ID);
		
		if (!ret->codec) {
			debug_msg ("Initializing AVCodec decoder.\n");
			
			avcodec_init ();
			avcodec_register_all ();
			
			ret->codec = avcodec_find_encoder (AVCODEC_ID);
			if (!ret->codec) {
				fprintf (stderr, "Error: %s codec not found in libavcodec.\n", AVCODEC_NAME);
				return FALSE;
			}
		}
		
		ret->ctx = avcodec_alloc_context ();
		if (!ret->ctx) {
			fprintf (stderr, "Error: allocating context for %s codec.\n", AVCODEC_NAME);
			return FALSE;
		}
		
		memset (ret->ctx, 0, sizeof (ret->ctx));
		avcodec_get_context_defaults (ret->ctx);
		
		/* Handle partial frames */
		if(ret->codec->capabilities & CODEC_CAP_TRUNCATED)
			ret->ctx->flags |= CODEC_FLAG_TRUNCATED;
		
		/* Initialize the context */
		ret->ctx->bit_rate		= FFMPEG_INIT_BITRATE;
		ret->ctx->frame_rate_base	= 1;
		ret->ctx->frame_rate		= FFMPEG_INIT_FRAMERATE;
		
		if (avcodec_open (ret->ctx, ret->codec) < 0) {
			fprintf (stderr, "Error: %s codec can not be opened.\n", AVCODEC_NAME);
			return FALSE;
		}
	}
	
	return ret;
}


/* Finalize the FFmpeg system */
void
ffmpeg_decomp_done (struct ffmpeg_state *s)
{
	if (s != NULL) {
		if (s->ctx != NULL)
			avcodec_close (s->ctx);
		
		free (s);
	}
}

/*
 * Decode a frame.
 *
 * Notes on the current implementation:
 *
 * - The libavcodec library provides an nice feature: we call the decoding
 *   function with fragments until a complete frame is returned. However, as some
 *   codecs access memory in 64bytes chunks, RTP packets should have that size too,
 *   So far, we do't use this feature and we copy all the fragments to
 *   a temporal buffer, instead.
 * - The output is provided in a temporal frame. We must copy the data from this
 *   frame to the frame provided.
 *
 * In conclusion, this codec is slightly expensive: it makes a lot of copies.
 */
int
ffmpeg_decomp (struct ffmpeg_state	*s,
	       struct coded_data	*in,
	       struct video_frame	*out)
{
	AVFrame			 frame;
	struct coded_data	*curr		= NULL;
	int			 len		= 0;
	int			 valid_image	= FALSE;
	int			 ret		= FALSE;
	
	unsigned char		*temp_buf;
	unsigned int		 temp_buf_offset, in_len;
	unsigned char		*in_pos;
	
	assert (s		!= NULL);
	assert (in		!= NULL);
	assert (out		!= NULL);
	assert (out->data	!= NULL);

	temp_buf = (unsigned char*) malloc (MAX_FRAMELEN);
	if (temp_buf) {

		/* a 'coded_data' is in reverse order */
		for (curr = in; curr->next != NULL; curr = curr->next);

		assert (curr->next == NULL);
		
		temp_buf_offset	= 0;
		while (curr != NULL) {
			assert (curr->data			!= NULL);
			assert (curr->data->meta.data		!= NULL);
			assert (curr->data->meta.data_len	>  0);

			in_pos = curr->data->meta.data;
			in_len = curr->data->meta.data_len;
			
			if ((temp_buf_offset + in_len) <= MAX_FRAMELEN) {
				memcpy (temp_buf + temp_buf_offset, in_pos, in_len);
			} else {
				debug_msg ("Input frame offset out of bounds (%u > %u)\n",
					   temp_buf_offset + in_len, MAX_FRAMELEN);
			}
		
			temp_buf_offset += in_len;
			curr		 = curr->prev;
		}

		/* Decode */
		debug_msg ("Processing frame of size %d...\n", temp_buf_offset);

		len = avcodec_decode_video (s->ctx, &frame, &valid_image,
					    temp_buf, temp_buf_offset);
		if (len < 0) {
			fprintf(stderr, "Error: while decoding frame\n");
			ret = FALSE;
		} else if (valid_image) {
			debug_msg ("Received %dx%d image\n", s->ctx->width, s->ctx->height);

			memcpy (out->data, frame.data[0], frame.linesize[0] * s->ctx->height);

			/* Update the output frame */
			out->width	= s->ctx->width;
			out->height	= s->ctx->height;
			
			switch (s->ctx->pix_fmt) {
				case PIX_FMT_RGB24:
					out->colour_mode = RGB;
					break;
				case PIX_FMT_YUV420P:
					out->colour_mode = YV_12;
					break;
				case PIX_FMT_YUV422:
					out->colour_mode = YUV_422;
					break;
				default:
					out->colour_mode = NONE;
			}
			
			out->data_len = out->height * out->width * cm_get_bpp (out->colour_mode); 
			
			ret = TRUE;
		}
		
		free (temp_buf);
	}
	
	return ret;
}

/* Get the current frame size */
display_size_t
ffmpeg_decomp_get_size (struct ffmpeg_state *s)
{
	display_size_t	ret	= DS_NONE;

	if (s->ctx != NULL) {
		if (s->ctx->width > 0 && s->ctx->height > 0)
			ret = display_get_size (s->ctx->width, s->ctx->height);
	}
	
	return ret;
}

#endif
