/*
 * FILE:   video_codec/uv_yuv422.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *         Ladan Gharai  <ladan@isi.edu>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 * Copyright (c) 2003-2004 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.3 $
 * $Date: 2006/01/30 14:25:40 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "video_codec/uv_yuv422.h"

#define BPP_8		(2)	/* Bytes per pixel */
#define BPP_10DVS	(8/3)	/* Bytes per pixel */

#define PGROUP_10DVS	16	/* Bytes per PGROUP - per uncomp-draft */
#define PIXELS_10DVS	6	/* Pixels per PGROUP - per uncomp-draft */

/* We limit packets to having up to 10 payload headers... */
#define MAX_PAYLOAD_COUNT 10    

#define END_OF_FRAME  (1<<15)


/* Payload header structure */
typedef struct {
	uint16_t        y_offset;       /* pixels */
	uint16_t        x_offset;       /* pixels */
	uint16_t        length;         /* pixels */
	uint16_t        flags;          
} payload_hdr_t;


/* Private structure of an encoded_frame */
struct yuv_priv_eframe {
	struct encoded_data	*node;	/* Array of nodes */	
	payload_hdr_t		*head;	/* Array of payload headers */
	
	struct video_frame	*frame;	/* Video frame */
	
	struct yuv_priv_eframe	*next,	/* Not currently used */
				*prev;	/* Not currently used */
};

/* Internal state */
struct vcodec_yuv422_state {
	uint32_t		 magic;
	unsigned int		 mtu;
	unsigned int		 id;
	video_colour_mode_t	 color;
	
	struct yuv_priv_eframe	*first,		/* Not currently used */
				*last;		/* Not currently used */
};

/******************************************************************************
 * Public API
 *****************************************************************************/

int
uv_yuv422_init (void)
{
	return 1;	/* Codec is always available */
}

void
uv_yuv422_exit(void)
{
	/* Nothing to do */
}

const char *
uv_yuv422_get_name (void)
{
	return "uv_yuv422";
}

const char *
uv_yuv422_get_desc (void)
{
	return "Uncompressed YUV 4:2:2 (8 bit)";
}

int
uv_yuv422_can_encode (void)
{
	return TRUE;
}

int
uv_yuv422_can_packetize (void)
{
	return FALSE;
}

int
uv_yuv422_can_decode (void)
{
	return TRUE;
}


/*******************************************************************************
 * Encoded frames management
 ******************************************************************************/


struct encoded_frame*
uv_yuv422_eframe_alloc (void)
{
	struct encoded_frame	*res_eframe		= NULL;
	
	res_eframe = (struct encoded_frame*) malloc (sizeof(struct encoded_frame));
	if (res_eframe != NULL) {
		/* Set some info of the encoded_frame */
		res_eframe->magic	= VCODEC_YUV422_ID;
		res_eframe->priv	= NULL;
		res_eframe->datalen	= uv_yuv422_encoder_get_framelen();
	} else {
		fprintf (stderr, "Error allocating memory for encoded_frame.\n");		
		uv_yuv422_eframe_free (res_eframe);
	}
	
	return res_eframe;
}

/*
 * Finish the allocation of the encoded_frame, allocating private structures...
 * A YUV encoded frame has a private part where we keep the headers (with the
 * info for each fragment), the nodes (just pointers to the original frame) and
 * the frame (a copy of the given one). Every 'eframe_get' operation will return
 * a header/node pair in 
 */
static struct yuv_priv_eframe*
uv_yuv422_priv_alloc (struct vcodec_yuv422_state	*state,
		      struct encoded_frame		*eframe)
{
	unsigned int		 num_nodes,
				 max_head_len,
				 max_node_len,
				 node_array_len,
				 head_array_len;
	struct yuv_priv_eframe	*res_priv_eframe	= NULL;

	assert (state		!= NULL);
	assert (eframe		!= NULL);
	assert (eframe->magic	== VCODEC_YUV422_ID);
	
	res_priv_eframe = (struct yuv_priv_eframe*) malloc (sizeof (struct yuv_priv_eframe));
	if (res_priv_eframe != NULL) {
		
		/* Estimate the number of packets to generate (+1). */
		max_head_len	= 8 * 10;
		max_node_len	= uv_yuv422_decoder_get_framelen (NULL);
		num_nodes	= max_node_len / (state->mtu - 40 - max_head_len) + 1;
		
		/* Allocate arrays for nodes */
		node_array_len = num_nodes * sizeof (struct encoded_data);
		res_priv_eframe->node = (struct encoded_data*) malloc (node_array_len);
		if (res_priv_eframe->node != NULL) {
			memset (res_priv_eframe->node, 0, node_array_len);
		} else {
			fprintf (stderr, "Error allocating memory for %d nodes in the encoded_frame.\n", num_nodes);
			free (res_priv_eframe);
			return NULL;
		}
		
		/* Allocate space for the headers */
		head_array_len = num_nodes * MAX_PAYLOAD_COUNT * sizeof(payload_hdr_t); 
		res_priv_eframe->head = (payload_hdr_t*) malloc (head_array_len);
		if (res_priv_eframe->head != NULL) {
			memset (res_priv_eframe->head, 0, head_array_len);			
		} else {
			fprintf (stderr, "Error allocating memory for %d headers in the encoded_frame.\n", num_nodes * MAX_PAYLOAD_COUNT);
			free (res_priv_eframe->node);
			free (res_priv_eframe);
			return NULL;
		}
		
		/* Allocate memory for the video_frame, too */
		res_priv_eframe->frame = video_frame_alloc (max_node_len);
		if (res_priv_eframe->frame == NULL) {
			fprintf (stderr, "Error: allocating new video_frame in yuv422_encode.\n");
			free (res_priv_eframe->node);
			free (res_priv_eframe);
			return NULL;
		}
		
		/* Set some info of the encoded_frame */
		eframe->len	= num_nodes;
		eframe->datalen	= max_node_len;
		eframe->priv	= res_priv_eframe;
	} else {
		fprintf (stderr, "Error allocating memory private data of encoded_frame.\n");
	}
	
	return res_priv_eframe;
}

/* Get a fragment of the eframe */
struct encoded_data*
uv_yuv422_eframe_get (struct encoded_frame *eframe, unsigned int pos)
{
	struct yuv_priv_eframe	*eframe_priv	= NULL;
	struct encoded_data	*nodes		= NULL,
				*res		= NULL;
	
	assert (eframe		!= NULL);
	assert (eframe->priv	!= NULL);
	assert (eframe->magic	== VCODEC_YUV422_ID);
	
	eframe_priv = (struct yuv_priv_eframe*) eframe->priv; 
	nodes = eframe_priv->node;
	
	if( pos < eframe->len ) {
		res = (nodes + pos);

		assert (res->data != NULL);
		assert (res->data_len > 0);
		assert (res->header != NULL);
		assert (res->header_len > 0);
	}
	
	return res;
}

/* Free an eframe */
void
uv_yuv422_eframe_free (struct encoded_frame *eframe)
{
	struct yuv_priv_eframe	*eframe_priv;

	assert (eframe        != NULL);
	assert (eframe->magic == VCODEC_YUV422_ID);
	
	eframe_priv = (struct yuv_priv_eframe*) eframe->priv;
	if (eframe_priv != NULL) {
		if (eframe_priv->node != NULL)	free (eframe_priv->node);
		if (eframe_priv->head != NULL)	free (eframe_priv->head);
		
		video_frame_free (eframe_priv->frame);
			
		free (eframe_priv);
	}
	
	free (eframe);
}



/*******************************************************************************
 * Encoding and packetization
 ******************************************************************************/

void *
uv_yuv422_encoder_create (unsigned id, unsigned mtu)
{
	struct vcodec_yuv422_state	*state;
	
	state = (struct vcodec_yuv422_state*)
		malloc(sizeof(struct vcodec_yuv422_state));
	
	if( state == NULL )
		fprintf(stderr, "Allocation error for the YUV 4:2:2 codec state.");
	else {
		state->magic	= VCODEC_YUV422_ID;
		state->mtu	= mtu;
		state->id	= id;
		state->first	= NULL;
		state->last	= NULL;
	}
	
	return (void*) state;
}

void
uv_yuv422_encoder_destroy (void *state)
{
	struct vcodec_yuv422_state *s = (struct vcodec_yuv422_state*) state;
	
	assert (s != NULL);

	free (s);
}

/* Return the max frame length */
unsigned int
uv_yuv422_encoder_get_framelen (void)
{
	return	display_get_width (DS_720p) *
		display_get_height (DS_720p) * MAX(BPP_10DVS, BPP_8);
}

/* Encode a 8bits frame */
int
uv_yuv422_encode (void                 *state,
		  struct video_frame   *in,
		  struct encoded_frame *out,
		  struct vcodec_ctx_t  *ctx)
{	
	struct vcodec_yuv422_state	*s = (struct vcodec_yuv422_state*) state;
	payload_hdr_t			*curr_head;
	struct encoded_data		*nodes;
	struct yuv_priv_eframe		*eframe_priv;
	
	unsigned int			 mark		= 0,
					 x		= 0,
					 y		= 0,
					 first_x	= 0,
					 first_y	= 0,
					 data_len	= 0,
					 total_len	= 0,	
					 payload_count	= 0,
					 node_count	= 0,
					 len,
					 funny_counter	= 0,
					 ctr,
					 octets_left_this_line,
					 octets_left_this_node,
					 no_pgroups	= 0,
					 pgroups_left_this_line = 0,
					 pgroups_left_this_node;
	video_colour_mode_t		 cm;
	
	UNUSED (ctx);
	
	assert (out		!= NULL);
	assert (out->magic	== VCODEC_YUV422_ID);
	assert (s		!= NULL);
	assert (s->magic	== VCODEC_YUV422_ID);
	assert (in		!= NULL);
	assert (in->data_len	<= uv_yuv422_decoder_get_framelen (NULL));

	uv_yuv422_priv_alloc (s, out);
	
	cm = in->colour_mode;
	
	if (cm == YUV_422_DVS10)	ctr = PGROUP_10DVS;
	else				ctr = BPP_8;
	
	eframe_priv = (struct yuv_priv_eframe*)(out->priv);
	video_frame_copy (eframe_priv->frame, in);	/* Copy the source frame */
	
	curr_head = eframe_priv->head;
	nodes     = eframe_priv->node;

	/* Main loop */
	do {
		if (cm == YUV_422_DVS10) {
			/* figure out how many groups of 6 pixels are left */
			switch (funny_counter) {
				case 0: pgroups_left_this_line = ((eframe_priv->frame->width +4 - x) / PIXELS_10DVS);
					break;
				case 1: pgroups_left_this_line = ((eframe_priv->frame->width +2 - x) / PIXELS_10DVS);
					break;
				case 2: pgroups_left_this_line = ((eframe_priv->frame->width  - x) / PIXELS_10DVS);
					break;
				default: assert (funny_counter > 2);
			}

			/* figure out how many PGROUPS are left */
			pgroups_left_this_node = (s->mtu - 40 - 
						  data_len - (8 * (payload_count + 1))) / PGROUP_10DVS;
		
			if (pgroups_left_this_node < pgroups_left_this_line)
				no_pgroups = pgroups_left_this_node;
			else
				no_pgroups = pgroups_left_this_line;
			
			len = no_pgroups * PGROUP_10DVS;
			
		} else {
			octets_left_this_line = (eframe_priv->frame->width - x) * BPP_8;
			octets_left_this_node = s->mtu - 40 - 
						data_len - (8 * (payload_count + 1));

			/* Calculate the length of this node */
			if (octets_left_this_node < octets_left_this_line)
				len = octets_left_this_node;
			else
				len = octets_left_this_line;			

			while ((len % BPP_8) != 0)   /* Only send complete pixels */
				len--;
		}

		/* Fill in the head */
		curr_head[payload_count].length   = htons(len);
	
		curr_head[payload_count].y_offset = htons(y);
		curr_head[payload_count].x_offset = htons(x);
		curr_head[payload_count].flags    = htons(0);
		payload_count++;
		
		data_len = data_len + len;
		
		if (cm == YUV_422_DVS10) {
			x += no_pgroups * PIXELS_10DVS;
			
			switch (funny_counter) {
				case 0: if (x==in->width+4) {
						x=4; y++; funny_counter++;
					}
					break;
				case 1: if (x==in->width+2) {
						x=2; y++; funny_counter++;
					}
					break;
				case 2: if (x==in->width) {
						x=0; y++; funny_counter=0;
					}
					break;
				default : assert (funny_counter > 2 );
			}
		} else {
			x += (len / BPP_8);

			/* If we finish this line, reset X and increment Y... */
			if (x == eframe_priv->frame->width) {
				x = 0;
				y++;
			}
		}
				
		if (y == eframe_priv->frame->height)
			mark = 1;
		
                /* Is it time to build a packet? */
		if ((y == eframe_priv->frame->height)     || 
		    (payload_count == MAX_PAYLOAD_COUNT) || 
		    ((40u + data_len + (8u * (payload_count + 1)) + ctr) > s->mtu)) {
			
			/* Mark the end */
			curr_head[payload_count-1].flags = htons(END_OF_FRAME);
				
			/* Fill up the node */
			nodes[node_count].header     = (unsigned char*) curr_head;
			nodes[node_count].header_len = 8 * payload_count;

			if (cm == YUV_422_DVS10) {
				nodes[node_count].data =
					eframe_priv->frame->data + 
						((first_y * eframe_priv->frame->width + first_x ) / PIXELS_10DVS) * PGROUP_10DVS;
			} else {
				nodes[node_count].data =
					eframe_priv->frame->data + 
						(first_y * eframe_priv->frame->width + first_x) * BPP_8;
			}
			nodes[node_count].data_len   = data_len;
			nodes[node_count].seqno      = node_count;
			nodes[node_count].mark       = mark;
			
			/* Update pointers */
			curr_head		+= MAX_PAYLOAD_COUNT;
			
			/* Update the total amount of data for this frame */
			total_len		+= data_len + 8*payload_count;			
			
			/* Reset counters */
			payload_count		 = 0;
			node_count		+= 1;
			data_len		 = 0;
			first_x			 = x;
			first_y			 = y;
			
			assert (node_count < out->len);
		}
	} while (y < eframe_priv->frame->height);

	/* Update info of the encoded frame */
	out->len	= node_count;	
	out->datalen	= total_len;
	out->id		= s->id;

	return node_count;
}

int
uv_yuv422_packetize (void                 *state,
		     struct video_frame   *in,
		     struct encoded_frame *out)
{
	UNUSED (state);
	UNUSED (in);
	UNUSED (out);

	debug_msg ("Error! This codec can not packetize.\n");
	
	return 0;
}	


/*******************************************************************************
 * Decoding
 ******************************************************************************/

video_colour_mode_t 
uv_yuv422_decoder_get_cm (void *state)
{
	struct vcodec_yuv422_state	*s	= (struct vcodec_yuv422_state*) state;
	video_colour_mode_t		 ret	= YUV_422;
	
	if (s != NULL) {
		if (s->color != NONE)
			ret = s->color;
	}
	
	return ret;
}

display_size_t 
uv_yuv422_decoder_get_size (void *state)
{
	UNUSED (state);
	return DS_720p;
}

unsigned int         
uv_yuv422_decoder_get_framelen (void *state) {
	display_size_t	size;

	size = uv_yuv422_decoder_get_size (state);

	return display_get_width (size) * display_get_height (size) * MAX(BPP_8, BPP_10DVS);
}


void *
uv_yuv422_decoder_create (unsigned id)
{
	struct vcodec_yuv422_state	*state;
	
	state = (struct vcodec_yuv422_state*) malloc(sizeof(struct vcodec_yuv422_state));
	if( state == NULL )
		fprintf(stderr, "Allocation error for the YUV 4:2:2 codec state.");
	else {
		state->magic	= VCODEC_YUV422_ID;
		state->mtu	= 0;
		state->id	= id;
		state->color	= NONE;
	}	
		
	return state;
}

void
uv_yuv422_decoder_destroy (void *state)
{
	uv_yuv422_encoder_destroy (state);
}


/* Decode a 8bits YUV frame */
int
uv_yuv422_decode (void *state, struct coded_data *in, struct video_frame *out)
{
	payload_hdr_t			*curr_hdr;
	payload_hdr_t			*hdr[MAX_PAYLOAD_COUNT];
	unsigned int			 hdr_count	= 0,
					 frame_offset	= 0,
					 width,
					 height,
					 i;
	unsigned char			*base,
					*offset;
	video_colour_mode_t		 cm		= NONE;
	struct vcodec_yuv422_state	*s = (struct vcodec_yuv422_state*) state;

	assert(in               != NULL);
	assert(out		!= NULL);
	assert(out->data        != NULL);
	assert(s		!= NULL);
	assert(s->magic		== VCODEC_YUV422_ID );

	width  = display_get_width  (uv_yuv422_decoder_get_size (s));
	height = display_get_height (uv_yuv422_decoder_get_size (s));	
	
	if (s->color == NONE) {
		unsigned int		 total_len	= 0;
		struct coded_data	*act;
		
		/*
		 * FIXME! This is a mess: we guess the color mode using the data
		 * length...
		 */
		act = in;
		while (act != NULL) {
			total_len	+= act->data->meta.data_len;
			act		= act->next;
		}
		
		if (total_len < 2000000 ) {
			debug_msg ("YUV_422 (8bits) format detected.\n");
			s->color = YUV_422;
		} else {
			debug_msg ("YUV_422_DVS10 (10bits) format detected.\n");
			s->color = YUV_422_DVS10;
		}
	}

	cm = s->color;
	
	while (in != NULL) {

		assert (in->data            != NULL) ;
		assert (in->data->meta.data != NULL);

		/* Figure out how many headers... */
		curr_hdr = (payload_hdr_t *) in->data->meta.data;
		while (TRUE) {
			hdr[hdr_count++] = curr_hdr;
			if ((ntohs(curr_hdr->flags) & END_OF_FRAME) != 0) {
				/* Last header... */
				break;
			}
			if (hdr_count == 10) {
				/* Out of space... */
				break;
			}
			curr_hdr++;
		}
		
		/* OK, now we can copy the data */
		offset = (in->data->meta.data) + hdr_count * 8;
		for (i = 1; i < hdr_count + 1; i++) { 
			unsigned short x_offset = ntohs(hdr[i - 1]->x_offset);
			unsigned short y_offset = ntohs(hdr[i - 1]->y_offset);
			unsigned short length   = ntohs(hdr[i - 1]->length);
			
#ifdef DEBUG			
			if ((x_offset > width) || (y_offset > height)) {
				debug_msg("WARNING! Coded frame can be out of bounds: %dx%d > %dx%d\n",
					  x_offset, y_offset, width, height);
			}
#endif			
			if (cm == YUV_422)
				frame_offset = (x_offset + (y_offset * width)) * BPP_8;
			else
				frame_offset = ((x_offset + (y_offset * width))/PIXELS_10DVS) * PGROUP_10DVS;

			/* Ok, move the data to the frame */
			base = out->data + frame_offset;
			memcpy (base, offset, length);
			offset += length;

			/* and free it */
			//free(in->data->meta.data);
		}

		in           = in->next;
		hdr_count    = 0;
		frame_offset = 0;
	}

	out->width       = display_get_width	(uv_yuv422_decoder_get_size (s));
	out->height      = display_get_height	(uv_yuv422_decoder_get_size (s));
	out->colour_mode = uv_yuv422_decoder_get_cm	(s);
	out->data_len    = uv_yuv422_decoder_get_framelen (s);	

	return TRUE;
}


int
uv_yuv422_decode_raw (void *state, struct video_frame *in, struct video_frame *out)
{
	struct coded_data	new_in;
	rtp_packet              packet;
	
	assert(state != NULL);
	assert(in    != NULL);
	assert(out   != NULL);
	
	/*
	 * This function has no real meaning for YUV, because the X_decode_full()
	 * functions are for decoding frames grabbed from a capture device, and
	 * there is no such device that provides frames with this format...
	 */
	packet.meta.data	= in->data;
	packet.meta.data_len	= in->data_len;
	
	new_in.seqno = 0;
	new_in.data  = &packet;
	new_in.next  = NULL;
	new_in.prev  = NULL;
	
	return uv_yuv422_decode (state, &new_in, out);	
}

int
uv_yuv422_decode_partial (void *state, struct coded_data *in, struct video_frame *out)
{
	return uv_yuv422_decode (state, in, out);
}

