/*
 * FILE:   video_convert.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.4 $
 * $Date: 2005/11/14 14:05:09 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_convert.h"
#include "video_convert/rgb_to_yuv422.h"
#include "video_convert/yuv422_to_rgb.h"
#include "video_convert/yv12_to_yuv422.h"
#include "video_convert/yuv422_to_yv12.h"

struct video_frame *
vidconv_prepare (struct video_frame *dst_frame, video_colour_mode_t src_cm)
{
	struct video_frame *f	= NULL;
	
	/* Returns a frame with the same size at dst_frame, but with colour */
	/* mode equal to src_cm. Returns the original frame if that is of a */
	/* suitable colour mode.                                            */
	if (dst_frame->colour_mode == src_cm) {
		f = dst_frame;
	} else {
		f = video_frame_alloc (dst_frame->width * dst_frame->height * cm_get_bpp (src_cm));
		if (f != NULL) {
			f->colour_mode	= src_cm;
			f->width	= dst_frame->width;
			f->height	= dst_frame->height;
		}
	}
	
	return f;
}

void
vidconv_convert (struct video_frame *src, struct video_frame *dst)
{
	if (src == dst) {
		/* Nothing to do... */
		return;
	} else {
		assert(src->colour_mode != dst->colour_mode);
		switch(src->colour_mode) {
		case RGB:
			switch (dst->colour_mode) {
			case YUV_422:
				rgb_to_yuv422(src, dst);
				break;
			default:
				abort(); 
			}
			break;
		case YUV_422:
			switch (dst->colour_mode) {
			case RGB:
				yuv422_to_rgb(src, dst);
				break;
			case YV_12:
				yuv422_to_yv12(src, dst);
				break;
			default:
				abort(); 
			}
			break;
		case YV_12:
			switch (dst->colour_mode) {
				case YUV_422:
					yv12_to_yuv422(src, dst);
					break;
				default:
					abort(); 
			}
			break;
		default:
			abort();
		}

		/* Free the memory we allocated in vidconv_prepare() */
		free(src->data);
		free(src);
	}
}

