/*
 * FILE:   video_display.c
 * AUTHOR: Colin Perkins <csp@isi.edu>
 *
 * Copyright (c) 2004-2006 University of Glasgow
 * Copyright (c) 2001-2003 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.35 $
 * $Date: 2006/02/16 18:44:01 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "trace_ug.h"
#include "trace.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/null.h"
#include "video_display/dumpfile.h"
#include "video_display/xv.h"
#include "video_display/x11.h"
#include "video_display/hdstation.h"
#include "video_display/sdl.h"
#include "video_display/gl.h"
#include "video_display/quartz.h"

/*
 * Interface to probing the valid display types. 
 */
typedef struct {
	display_id_t		 id;
	display_type_t 		*(*func_probe)(display_format_t	*format,
					       const char	*display_args);
	void			*(*func_init)(display_format_t	*format,
					      const char	*display_args,
					      void		*data);
	void 			 (*func_done)(void		*state);
	struct video_frame 	*(*func_getf)(void		*state,
					      display_size_t	 size,
					      video_colour_mode_t color);
	display_result_t  	 (*func_putf)(void		*state,
					      struct video_frame *frame);
	video_colour_mode_t	 (*func_colour)(void		*state);
} display_table_t;

static display_table_t display_device_table[] = {
#ifndef X_DISPLAY_MISSING
#ifdef  HAVE_TCL_TK
	{
		0,
		display_xv_probe,
		display_xv_init,
		display_xv_done,
		display_xv_getf,
		display_xv_putf,
		display_xv_colour
	},
	{
		0,
		display_x11_probe,
		display_x11_init,
		display_x11_done,
		display_x11_getf,
		display_x11_putf,
		display_x11_colour
	},
#endif /* HAVE_TCL_TK       */
#endif /* X_DISPLAY_MISSING */
#ifdef HAVE_HDSTATION
	{
		0,
		display_hdstation_probe,
		display_hdstation_init,
		display_hdstation_done,
		display_hdstation_getf,
		display_hdstation_putf,
		display_hdstation_colour
	},
#endif
#ifdef HAVE_LIBSDL
        {
		0,
                display_sdl_probe,
                display_sdl_init,
                display_sdl_done,
                display_sdl_getf,
                display_sdl_putf,
                display_sdl_colour
        },
#endif
#ifdef HAVE_GL
        {
		0,
                display_gl_probe,
                display_gl_init,
                display_gl_done,
                display_gl_getf,
                display_gl_putf,
                display_gl_colour
        },
#endif
#ifdef HAVE_CARBON	
        {
		0,
                display_quartz_probe,
                display_quartz_init,
                display_quartz_done,
                display_quartz_getf,
                display_quartz_putf,
                display_quartz_colour
        },
#endif	
        {
		0,
                display_dumpfile_probe,
                display_dumpfile_init,
                display_dumpfile_done,
                display_dumpfile_getf,
                display_dumpfile_putf,
                display_dumpfile_colour
        },
	{
		0,
		display_null_probe,
		display_null_init,
		display_null_done,
		display_null_getf,
		display_null_putf,
		display_null_colour
	}
};

#define DISPLAY_DEVICE_TABLE_SIZE (sizeof(display_device_table) / sizeof(display_table_t))

static display_type_t	*available_devices[DISPLAY_DEVICE_TABLE_SIZE];
static unsigned int	 available_device_count = 0;

int
display_init_devices (display_format_t *format, const char* display_args)
{
	unsigned int	 i;
	display_type_t	*dt;

	assert(available_device_count == 0);

	for (i = 0; i < DISPLAY_DEVICE_TABLE_SIZE; i++) {
		debug_msg("Probing display device %d\n", i);
		dt = display_device_table[i].func_probe (format, display_args);
		if (dt != NULL) {
			display_device_table[i].id = dt->id;
			available_devices[available_device_count++] = dt;
		}
	}

	return 0;
}

void 
display_free_devices(void)
{
	unsigned int	i;

	for (i = 0; i < available_device_count; i++) {
		free(available_devices[i]);
		available_devices[i] = NULL;
	}
	available_device_count = 0;
}

unsigned int
display_get_device_count(void)
{
	return available_device_count;
}

display_type_t *
display_get_device_details(unsigned int index)
{
	assert(index < available_device_count);
	assert(available_devices[index] != NULL);

	return available_devices[index];
}

/*
 * Display initialisation and playout routines...
 */

#define DISPLAY_MAGIC 0x01ba7ef1

struct display {
	uint32_t	 magic;
	int		 index;
	void		*state;
};

struct display *
display_init (display_id_t	 id,
	      display_format_t	*format,
	      const char	*display_args,
	      void		*data)
{
	unsigned int	i;

	for (i = 0; i < DISPLAY_DEVICE_TABLE_SIZE; i++) {
		if (display_device_table[i].id == id) {
			struct display *d = (struct display *) malloc(sizeof(struct display));
			d->magic = DISPLAY_MAGIC;
			d->state = display_device_table[i].func_init (format,
								      display_args,
								      data);
			d->index = i;
			if (d->state == NULL) {
				debug_msg("Unable to start display 0x%08lx\n", id);
				free(d);
				return NULL;
			} 
			return d;
		}
	}
	debug_msg("Unknown display id: 0x%08x\n", id);
	return NULL;
}

void
display_done(struct display *d)
{
	assert(d->magic == DISPLAY_MAGIC);
	display_device_table[d->index].func_done(d->state);
}

struct video_frame *
display_get_frame(struct display *d, display_size_t size, video_colour_mode_t color)
{
	assert(d->magic == DISPLAY_MAGIC);
	trace (te_display_getf);
	return display_device_table[d->index].func_getf(d->state, size, color);
}

display_result_t
display_put_frame(struct display *d, struct video_frame *frame)
{
	assert(d->magic == DISPLAY_MAGIC);
	trace (te_display_putf_init);
	return display_device_table[d->index].func_putf(d->state, frame);
}

video_colour_mode_t 
display_get_colour_mode(struct display *d)
{
	assert(d->magic == DISPLAY_MAGIC);
	return display_device_table[d->index].func_colour(d->state);
}

typedef struct {
	display_size_t		 id;
	char			*name;
	char			*description;
	unsigned int		 width;
	unsigned int		 height;
} display_size_ext_t;

static display_size_ext_t displays_table[] = {
	{ DS_NONE,   "none",  "(none)",                     0,    0 },
	{ DS_QCIF,   "QCIF",  "Quarter CIF",              176,  144 },
	{ DS_CIF,    "CIF",   "CIF",                      352,  288 },
	{ DS_SCIF,   "SCIF",  "Super CIF",                702,  576 },
	{ DS_VGA,    "VGA",   "VGA",                      640,  480 },
	{ DS_NTSC,   "NTSC",  "NTSC",                     720,  480 },
	{ DS_PAL,    "PAL",   "PAL",                      720,  576 },
	{ DS_720p,   "720p",  "SMPTE 296M - HDTV 720p",  1280,  720 },
	{ DS_1080i,  "1080i", "SMPTE 274M - HDTV 1080i", 1920, 1080 },
};

#define NUM_DISPLAYS (sizeof(displays_table)/sizeof(display_size_ext_t))

display_size_t
display_get_min (void)
{
	return displays_table[1].id;
}

display_size_t
display_get_max (void)
{
	return displays_table[NUM_DISPLAYS-1].id;
}

unsigned int	
display_get_width (display_size_t d)
{
	assert (d < NUM_DISPLAYS);
	return displays_table[d].width;
}

unsigned int	
display_get_height (display_size_t d)
{
	assert (d < NUM_DISPLAYS);
	return displays_table[d].height;
}

char*
display_get_name (display_size_t d)
{
	assert (d < NUM_DISPLAYS);
	return displays_table[d].description;
}

display_size_t
display_get_by_name (char *name)
{
	unsigned int	i;
	display_size_t	ret	= DS_NONE;
	
	for (i = 0; i < NUM_DISPLAYS; i++) {
		if (strcmp (displays_table[i].name, name) == 0) {
			ret = displays_table[i].id;
			break;
		}
	}
	
	return ret;	
}


display_size_t
display_get_size (unsigned int w, unsigned int h)
{
	unsigned int	i;
	display_size_t	ret	= DS_NONE;
	
	for (i = 0; i < NUM_DISPLAYS; i++) {
		if (displays_table[i].width == w && displays_table[i].height == h) {
			ret = displays_table[i].id;
			break;
		}
	}
	
	return ret;
}

