/*
 * FILE:   video_display/dumpfile.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005-2006 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.16 $
 * $Date: 2006/02/22 01:23:13 $
 *
 */


#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"

#include "video_display/dumpfile.h"

#define MAGIC_DUMPFILE   DISPLAY_DUMPFILE_ID

/*
 * Internal state
 */
struct state_dumpfile {
	uint32_t			 magic;		/* For debugging... */

	int				 fd;		/* FD of the file */
	
	unsigned char			*base_ptr;	/* Start pointer in memory */
	unsigned char			*write_ptr;	/* Current write pointer */	
	
	unsigned int			 frame_counter;
	
	struct video_frame		 frame;		/* Internal frame */
	struct dumpfile_movie_props	*movie_props;
};

/*
 * Some properties of the movie
 */ 
struct dumpfile_movie_props {
	char		*filename;	/* The name of the file */
	unsigned long	 len;		/* Max length for the movie (in bytes). */
};

/*
 * Parse the display_args arguments and set some other properties.
 * By now, display_args=="filename@length". This could change. 
 */ 
static struct dumpfile_movie_props*
dumpfile_props (display_format_t	*format,
		const char		*display_args)
{
	struct dumpfile_movie_props	*movie_props	= NULL;
	char				*token		= NULL;

	UNUSED (format);
	
	movie_props = malloc (sizeof (struct dumpfile_movie_props));
	if (movie_props != NULL) {
		movie_props->filename = NULL;

		if (display_args != NULL) {
			/* Parse filename and length... */
			char	*at_pos = strchr(display_args, '@');
			if (at_pos != NULL) {
				int filename_len = at_pos - display_args;
				movie_props->filename = malloc(filename_len+1);
				strncpy(movie_props->filename, display_args, filename_len);
				movie_props->filename[filename_len] = '\0';
				token = at_pos + 1;
			}
		}
		
		if (token) {
			/* Length in the command line is in megabytes */
			movie_props->len = atol (token) * 1024 * 1024;
		} else {
			movie_props->len = 0;
		}
	} else {
		fprintf (stderr, "Error: Unable to allocate memory for parsing Dumpfile options.\n");
	}
	
	return movie_props;
}



/**********************************************************************************
 * Public API
 *********************************************************************************/ 


/* Dump-to-File output initialization. */
void*
display_dumpfile_init (display_format_t	*format,
		       const char	*display_args,
		       void		*data)
{
	int				 fd;
	struct state_dumpfile		*s	= NULL;
	struct dumpfile_movie_props	*props	= NULL;

	UNUSED (data);
	
	assert (format			!= NULL);
	assert (format->size		!= DS_NONE);
	assert (format->colour_mode	!= NONE);
	
	props = dumpfile_props (format, display_args);	 
	if (props == NULL) {
		fprintf (stderr, "Error: parsing 'dumpfile' device options.\n");
		return NULL;
	}        
	
	if (props->filename == NULL) {
		fprintf (stderr, "Error: no filename parameter for 'dumpfile' display.\n"
			 "Error: set the output file using the '-D' flag.\n");
		return NULL;
	}
	
	if (props->len == 0) {
		fprintf (stderr, "Error: no length specified for the 'dumpfile' device.\n"
			 "Error: set the movie length using the '-D filename@length' format (in MBytes).\n");
		return NULL;
	}
	
	/* Opens the file */	
	if ((fd = open (props->filename, O_RDWR | O_CREAT | O_TRUNC, 0644)) >= 0) {
		if (lseek (fd, props->len, SEEK_SET) < 0) {
			fprintf (stderr, "Error: unable to create output file (seek)\n");
			free (props);
			return NULL;
		}
		
		if (write(fd, "", 1) != 1) {
			fprintf (stderr, "Error: unable to create output file (write)\n");
			free (props);
			return NULL;
		}

		/* Create the state */
		s = (struct state_dumpfile *) malloc (sizeof (struct state_dumpfile));
		if (s != NULL) {
			s->magic	= MAGIC_DUMPFILE;
			s->fd		= fd;
			
			assert (s->fd > 0);			
			
			s->movie_props	= props;
			
			assert (s->movie_props->len > 0);

			s->base_ptr	= NULL;
			s->write_ptr	= NULL;
			
			s->frame.colour_mode	= NONE;
			s->frame.height		= 0;
			s->frame.width		= 0;
			s->frame.data_len	= 0;	
			s->frame.data		= NULL;

			s->frame_counter = 0;

			/* Print some info */
			printf ("Dump-to-file: %s\n", s->movie_props->filename);
			printf ("      Length: %lu MBytes\n",
				s->movie_props->len/(1024*1024));
		} else {
			fprintf (stderr, "Error: allocating memory for 'dumpfile' internal state.\n");
			free (props);
		}
	} else {
		fprintf (stderr, "Error: opening 'dumpfile' '%s': %s.\n",
			 props->filename, strerror (errno));
	}

	return s;
}

void 
display_dumpfile_done (void *state)
{
	struct state_dumpfile *s = (struct state_dumpfile *) state;

	if (s != NULL) {
		assert (s->magic == MAGIC_DUMPFILE);

		/* Close the file */
		munmap (s->base_ptr, s->movie_props->len);
		close (s->fd);
		free (s);
	}
}

/* Return the internal frame */
struct video_frame*
display_dumpfile_getf (void			*state,
		       display_size_t		 size,
		       video_colour_mode_t	 color)
{
	struct state_dumpfile	*s = (struct state_dumpfile *) state;
	unsigned int		 w, h, len;
	struct video_frame	*ret	= NULL;

	assert (s		!= NULL);
	assert (s->magic	== MAGIC_DUMPFILE);
	assert (s->movie_props	!= NULL);
	assert (s->fd		> 0);
	
	/* Do the mmap if it was not done before */
	if (s->base_ptr == NULL) {
		/* 'mmap' the output file */
		if ((s->base_ptr = mmap (0, s->movie_props->len, PROT_READ | PROT_WRITE,
					 MAP_SHARED, s->fd, 0)) == MAP_FAILED) {
			fprintf (stderr, "Error: in mmap for dumpfile: %s.\n", strerror (errno));
			return NULL;
		}
#ifdef HAVE_MADVISE
		madvise (s->base_ptr, s->movie_props->len, MADV_SEQUENTIAL | MADV_WILLNEED);
#endif	
		/* Write pointer */
		s->write_ptr = s->base_ptr;
	}	

	if (s->base_ptr == NULL) {
		fprintf (stderr, "Error: mmap'ing memory for dumpfile.\n");
	} else {
		h	= display_get_height (size);
		w	= display_get_width (size);
		len	= h * w * cm_get_bpp (color);
			
		assert (s->write_ptr != NULL);
		assert (len > 0);
		
		/* Check that there is enough space to write the frame */
		if (s->write_ptr + len <= s->base_ptr + s->movie_props->len) {
			
			/* Creates a frame with the expected information */
			s->frame.colour_mode	= color;
			s->frame.height		= h;
			s->frame.width		= w;
			s->frame.data_len	= len;	
			s->frame.data		= s->write_ptr;

			ret = &(s->frame);
		} else {
			s->frame.colour_mode	= NONE;
			s->frame.height		= 0;
			s->frame.width		= 0;
			s->frame.data_len	= 0;	
			s->frame.data		= NULL;
		}
	}	

	return ret;	/* returns the frame */
}

/* Dump the frame to the file */
display_result_t 
display_dumpfile_putf (void *state, struct video_frame *frame)
{
	struct state_dumpfile	*s = (struct state_dumpfile *) state;
	display_result_t	 ret;
	
	assert (s		!= NULL);
	assert (s->magic	== MAGIC_DUMPFILE);
	assert (frame		!= NULL);
	assert (s->base_ptr	!= NULL);
	assert (s->write_ptr	!= NULL);
	assert (frame->data	== s->write_ptr);

	/*
	 * Update the writing pointer if we have enough space,
	 * otherwise, do nothing and request exit.
	 */
	if (s->write_ptr + frame->data_len < s->base_ptr + s->movie_props->len) {
		s->write_ptr += frame->data_len;
		s->frame_counter++;
		ret = DISPLAY_OK;
	} else {
		printf ("Movie completed (%d frames)\n", s->frame_counter);
		ret = DISPLAY_EXIT_REQ;
	}
	
	return ret;
}

/* Color mode suppported by the dump-to-file device */
video_colour_mode_t 
display_dumpfile_colour (void *state)
{
	struct state_dumpfile	*s	= (struct state_dumpfile *) state;
	video_colour_mode_t	 ret	= YUV_422;
	
	assert (s	 != NULL);
	assert (s->magic == MAGIC_DUMPFILE);

	if (s->frame.colour_mode != NONE) {
		ret = s->frame.colour_mode;
	}
	
	return ret;
}

/* Video formats supported by the dump-to-file device */
display_type_t *
display_dumpfile_probe (display_format_t	*format,
			const char		*display_args)
{
	display_type_t		*dt		= NULL;
	display_format_t	*dformat	= NULL;

     	UNUSED (format);
     	UNUSED (display_args);
	
	dformat = malloc (4 * sizeof (display_format_t));
	if (dformat != NULL) {
		dformat[0].size		= DS_QCIF;
		dformat[0].colour_mode	= YUV_422;
		dformat[0].num_images	= 1;
		dformat[1].size		= DS_CIF;
		dformat[1].colour_mode	= YUV_422;
		dformat[1].num_images	= 1;
		dformat[2].size		= DS_SCIF;
		dformat[2].colour_mode	= YUV_422;
		dformat[2].num_images	= 1;
		dformat[3].size		= DS_720p;
		dformat[3].colour_mode	= YUV_422;
		dformat[3].num_images	= 1;
		
		dt = malloc (sizeof (display_type_t));
		if (dt != NULL) {
			dt->id		= DISPLAY_DUMPFILE_ID;
			dt->name	= "dumpfile";
			dt->description = "Dumpfile";
			dt->formats	= dformat;
			dt->num_formats = 4;
		} else {
			fprintf (stderr, "Error: allocating memory for 'dumpfile' probe.\n");
			free (dformat);
		}
		
	} else {
		fprintf (stderr, "Error: allocating memory for 'dumpfile' probe.\n");
	}
	
	return dt;
}

