/*
 * FILE:   video_display/gl.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.1 $
 * $Date: 2005/09/27 13:21:25 $
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_GL

#include "debug.h"
#include "environment.h"
#include "tv.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/gl.h"

#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glext.h>

#define MAGIC_GL		DISPLAY_GL_ID

#define GL_INITIAL_WSIZE	DS_PAL		/* Initial window size */
#define GL_MAX_FRAMESIZE	display_get_width (display_get_max ()) * \
				display_get_height (display_get_max ()) * 2


/* GL properties */
struct gl_props_t {
	const GLubyte	*desc;
	
	const GLubyte	*ext;
	GLboolean	 ext_fence;
	GLboolean	 ext_shade;	
	GLboolean	 ext_apple_yuv422;	
	
	GLint		 max_texture_size;
};

/* Internal state */
struct state_gl {
	uint32_t		 magic;		/* For debugging... */

	/* Input format */
	display_size_t		 size;		/* Current display size */
	video_colour_mode_t	 color;		/* Current display color mode */ 
	
	/* Buffer and internal frame */
	unsigned char		*buffer;
	unsigned int		 buffer_size;
	struct video_frame	 frame;

	int			 request_exit;
	
	/* GL stuff */
	GLuint			 gl_texture;
	struct gl_props_t	*gl_props;
};

static struct state_gl *s_state;

/*******************************************************************************
 * Checks for OpenGL
 ******************************************************************************/

/* Get device properties */
static struct gl_props_t*
get_gl_properties (void)
{
	struct gl_props_t	*ret	= NULL;
	
	ret = (struct gl_props_t*) malloc (sizeof (struct gl_props_t));
	if (ret != NULL) {
		/* Check core version */
		ret->desc = glGetString (GL_VERSION);
	
		/* Verify extension support */
		ret->ext = glGetString (GL_EXTENSIONS);
	
		
		/* Notes on YUV-to-RGB conversion:
		 * "...there's no widely-supported way of doing YUV-to-RGB in
		 * hardware under OpenGL. There are two extensions called
		 * GL_EXT_422_pixels and GL_MESA_ycbcr_texture, but both of these
		 * do only 4:2:2, not 4:2:0, and appear not to be implemented by
		 * Nvidia, anyway. There's also an Nvidia extension called 
		 * GL_NVX_ycrcb, but again, it's only 4:2:2, and apparently
		 * requires a pixel shader to do the actual YUV-to-RGB conversion."
		 */
		ret->ext_fence = gluCheckExtension ((GLubyte*)"GL_APPLE_fence", ret->ext);
		ret->ext_shade = gluCheckExtension ((GLubyte*)"GL_ARB_shading_language_100", ret->ext);
	
		ret->ext_apple_yuv422 = gluCheckExtension ((GLubyte*)"GL_APPLE_ycbcr_422", ret->ext);
		
		/* Determine OpenGL limits */
		glGetIntegerv (GL_MAX_TEXTURE_SIZE, &(ret->max_texture_size)); 
	}
	
	return ret;
}

/* Get device properties */
static int
check_gl_properties (struct gl_props_t *props)
{
	int	ret	= TRUE;
	
	assert (props != NULL);
	
	/* First, check if we have a nice YUV 4:2:2 extension... */
	if (!props->ext_apple_yuv422) {
		fprintf (stderr, "Error: no valid YUV 4:2:2 conversion for GL device.\n");
		ret = FALSE;
	}
	
	return ret;
}

/*******************************************************************************
 * Events handlers
 ******************************************************************************/

/* Keyboard event handler */
static void
gl_keyboard_handler (unsigned char key, int x, int y)
{
	UNUSED (x);
	UNUSED (y);
	
	switch (key) {
                case 'q':
			s_state->request_exit = TRUE;
                        break;
                default:
                        break;
        }
}

/* Resize callback function */
static void
gl_resize_window (int w, int h)
{
	glViewport (0, 0, w, h);
}

/* Draws the internal buffer and swaps buffers */
static void
gl_draw_buffer (void)
{
	unsigned int	width,
			height;
	
	assert (s_state != NULL);
	
	width  = display_get_width (s_state->size);
	height = display_get_height (s_state->size);
	
	glClear (GL_COLOR_BUFFER_BIT);
        glBindTexture (GL_TEXTURE_RECTANGLE_EXT, s_state->gl_texture);
	
	if (s_state->gl_props->ext_apple_yuv422) {
		glTexImage2D (GL_TEXTURE_RECTANGLE_EXT, 0, GL_RGB, width, height, 0,
		      GL_YCBCR_422_APPLE, GL_UNSIGNED_SHORT_8_8_APPLE, s_state->buffer);
	} else {
		fprintf (stderr, "Warning: no YUV 4:2:2 display method available for GL display\n.");
	}
		
        glBegin (GL_QUADS);
        
        glTexCoord2i (0, height);
        glVertex2f (-1.0, -1.0);
        glTexCoord2i (width, height);
        glVertex2f (1.0, -1.0);
	glTexCoord2i (width, 0);
        glVertex2f (1.0, 1.0);
        glTexCoord2i (0, 0);
        glVertex2f (-1.0, 1.0);
        
        glEnd ();
        
        glutSwapBuffers ();	
}

/*******************************************************************************
 * Windows
 ******************************************************************************/
static void
gl_init_glut (struct state_gl *s)
{
	int	  argc	= 0;
	char		**argv	= NULL;
	
	assert (s != NULL);
	
	glutInit (&argc, argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
}

/* Create a window, using Glut */
static int
gl_create_window (struct state_gl	*s,
		  unsigned int		 initial_width,
		  unsigned int		 initial_height)
{	
	assert( s != NULL );
	assert( s->magic == MAGIC_GL );
        
	glutInitWindowPosition (0, 0);
        glutInitWindowSize (initial_width, initial_height);
        glutCreateWindow ("UltraGrid");	
	
	return TRUE;
}

/* Initialize the system for textures */
static int
gl_create_texture (struct state_gl *s)
{
	unsigned int	 width,
			 height;
	
	assert( s != NULL );
	assert( s->magic == MAGIC_GL );
	
	
	width  = display_get_width (s->size);
	height = display_get_height (s->size);
	
	glEnable	(GL_TEXTURE_RECTANGLE_EXT);
        glGenTextures	(1, &s->gl_texture);
        glBindTexture	(GL_TEXTURE_RECTANGLE_EXT, s->gl_texture);

	/* Assert default GL state */
	// glDisable (GL_ALPHA_TEST);
	// glDisable (GL_BLEND);
	// glDisable (GL_DEPTH_TEST);
	// glDepthMask (GL_FALSE);
	// glDisable (GL_CULL_FACE);
	
	if (s->gl_props->ext_apple_yuv422) {
		glPixelStorei	(GL_UNPACK_CLIENT_STORAGE_APPLE, GL_TRUE);
		glTexImage2D	(GL_TEXTURE_RECTANGLE_EXT, 0, GL_RGB, width, height,
				 0, GL_YCBCR_422_APPLE, GL_UNSIGNED_SHORT_8_8_APPLE,
				 s->buffer);
	} else {
		fprintf (stderr, "Warning: no YUV 4:2:2 display method available for GL display\n.");
	}
	
	return TRUE;
}

/* Shows the window */
static int
gl_show_window (struct state_gl *s)
{
	UNUSED (s);
	
	return TRUE;
}

/* Destroy the GL window */
static int
gl_destroy_window (struct state_gl *s)
{
	assert (s != NULL);
	
	/* FIXME: this should do something more */
	
	s->size		= DS_NONE;
	s->color	= NONE;
	
	return TRUE;
}

/* Force a redisplay */
static void
gl_redraw (void)
{
	glutPostRedisplay();
}

/* Reconfigure the system for a new size/color state */
static void
gl_reconfigure (struct state_gl		*state,
		display_size_t		 size,
		video_colour_mode_t	 color)
{
	assert (state != NULL);
	
	if (size != state->size || color != state->color) {
		debug_msg ("Reconfiguring GL display.\n");
		
		/* Destroy previous window */
		gl_destroy_window (state);
		
		assert (state->size == DS_NONE && state->color == NONE);
		
		/* ... and start a new one with new new configuration */
		state->size  = size;
		state->color = color;
		
		gl_create_window (state,
			display_get_width (size), display_get_height (size));
	}	
}

/***************************************************************************
 * Public API
 **************************************************************************/

/*
 * GL display initialization.
 * A view port can be provided by upper levels. For example, from the Cocoa
 * interface, we can get the current view port using a NSRect:
 * 		NSRect bounds = [self bounds];
 *		x = 0;				y = 0;
 *		width = bounds.size.width;	height = bounds.size.height;
 *		min_x = NSMinX(bounds);		min_y = NSMinY(bounds);
 *		max_x = NSMaxX(bounds);		max_y = NSMaxY(bounds);
 */
void*
display_gl_init (display_format_t	*format,
		 const char		*display_args,
		 void			*data)
{
	struct state_gl		*s		= NULL;
	struct gl_view_port	*gl_vp		= (struct gl_view_port*) data;
	unsigned int		 initial_height,
				 initial_width;
	UNUSED (format);
	UNUSED (display_args);
	
	s = malloc (sizeof (struct state_gl));
	if (s != NULL) {
		
		s->gl_props = get_gl_properties ();
		
		if (check_gl_properties (s->gl_props)) {
			
			s_state		= s;	/* Copy the pointer in the static version */
			
			s->magic	= MAGIC_GL;
			
			s->size		= DS_NONE;	/* Initial null display */
			s->color	= NONE;
			
			s->request_exit	= FALSE;
		
			s->buffer_size	= GL_MAX_FRAMESIZE;
			s->buffer	= malloc (s->buffer_size);
			if (s->buffer != NULL) {
				
				gl_init_glut (s);
				
				/* If not provided, we need to create a window */
				if (gl_vp == NULL) {

					/* Set the initial window size */
					initial_height	= display_get_height (GL_INITIAL_WSIZE);
					initial_width	= display_get_width (GL_INITIAL_WSIZE);
					
					gl_create_window (s, initial_width, initial_height);
					gl_show_window (s);
				} else {
					/* FIXME: To be implemented */
				}
				
				/* Register callbacks and handlers */
				glutDisplayFunc	(gl_draw_buffer);
				glutReshapeFunc	(gl_resize_window);
				glutKeyboardFunc(gl_keyboard_handler);
				
				gl_create_texture (s);
				//glutMainLoop();
				
			} else {
				fprintf (stderr, "Error: allocating memory for internal buffer in GL display.\n");
				display_gl_done (s);
			}
		} else {
			fprintf (stderr, "Error: no valid GL display available.\n");
			display_gl_done (s);		
		}
	} else {
		fprintf (stderr, "Error: allocating gl_display\n");
	}
	
	return s;
}

/* Finish the GL display */
void
display_gl_done (void *state)
{
	struct state_gl *s = (struct state_gl *) state;
	
	assert( s != NULL );
	assert( s->magic == MAGIC_GL );
	
	if (s != NULL) {
		gl_destroy_window (s);
		
		if (s->buffer != NULL)
			free (s->buffer);
		
		/* Free the state */
		free (s->gl_props);
		free (s);
	}
}

/* Get a frame */
struct video_frame*
display_gl_getf (void			*state,
		 display_size_t		 size,
		 video_colour_mode_t	 color)
{
	struct state_gl *s = (struct state_gl *) state;
	
	assert( s != NULL );
	assert( s->magic == MAGIC_GL );
	
	if (s->request_exit) {
		return NULL;
	} else {
		/* Check if we need to reconfigure the system */
		if (color != s->color || size != s->size) {
			gl_reconfigure (s, size, color);
		}
		
		if (color == YUV_422) {
			/* Create the frame */
			s->frame.height		= display_get_height (size);
			s->frame.width		= display_get_width (size);
			s->frame.data		= (unsigned char*)s->buffer;
			s->frame.data_len	= s->frame.width * s->frame.height * 2;
			s->frame.colour_mode	= color;
			
			assert (s->frame.data_len <= s->buffer_size);
			
			return &(s->frame);	/* Return the internal frame */
		} else {
			return NULL;
		}
	}
}

display_result_t
display_gl_putf (void			*state,
		 struct video_frame	*frame)
{
	struct state_gl		*s	= (struct state_gl *) state;
	unsigned int		 ret	= DISPLAY_OK;
	
	assert (s			!= NULL);
	assert (s->magic		== MAGIC_GL);
	assert (frame			!= NULL);
	assert (frame->height		== display_get_height (s->size));
	assert (frame->width		== display_get_width (s->size));
	assert (frame->colour_mode	== s->color);
	assert (frame->data_len		<= s->buffer_size);
	
	/* First, check if the user has closed the window or pressed ESC... */
	if (s->request_exit) {
		ret = DISPLAY_EXIT_REQ;
	} else {
		gl_redraw ();
	}
	
	return ret;
}

/* Color mode suppported by the gl */
video_colour_mode_t
display_gl_colour (void *state)
{
	struct state_gl *s = (struct state_gl *) state;
	
	assert (s		!= NULL);
	assert (s->magic	== MAGIC_GL);
	
	return YUV_422;
}

/*
 * Video formats supported by the gl
 */
display_type_t *
display_gl_probe (display_format_t	*format,
		  const char		*display_args)
{
	display_type_t		*dt		= NULL;
	display_format_t	*dformat	= NULL;
	
	UNUSED (display_args);
	UNUSED (format);
	
	dformat = malloc (4 * sizeof (display_format_t));
	if (dformat != NULL) {
		dformat[0].size		= DS_QCIF;
		dformat[0].colour_mode	= YUV_422;
		dformat[0].num_images	= 1;
		dformat[1].size		= DS_CIF;
		dformat[1].colour_mode	= YUV_422;
		dformat[1].num_images	= 1;
		dformat[2].size		= DS_SCIF;
		dformat[2].colour_mode	= YUV_422;
		dformat[2].num_images	= 1;
		dformat[3].size		= DS_720p;
		dformat[3].colour_mode	= YUV_422;
		dformat[3].num_images	= 1;
		
		dt = malloc (sizeof (display_type_t));
		if (dt != NULL) {
			dt->id = DISPLAY_GL_ID;
			dt->name = "gl";
			dt->description = "GL";
			dt->formats = dformat;
			dt->num_formats = 4;
		} else {
			fprintf (stderr, "Error: allocating memory for dt.\n");
			free (dformat);
		}
	} else {
		fprintf (stderr, "Error: allocating memory for dformat.\n");
	}
	
	return dt;
}

#endif

