/*
 * FILE:   video_display/quartz.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005-2006 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2006/02/16 18:48:57 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifdef HAVE_CARBON

#include "debug.h"
#include "environment.h"
#include "tv.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/quartz.h"

#include <Carbon/Carbon.h>
#include <QuickTime/QuickTime.h>

#define MAGIC_QUARTZ		DISPLAY_QUARTZ_ID

#define QUARTZ_INITIAL_WSIZE	DS_PAL		/* Initial window size */
#define QUARTZ_MAX_FRAMESIZE	display_get_width (display_get_max ()) * \
				display_get_height (display_get_max ()) * 2

/* Internal state */
struct state_quartz {
	uint32_t		 magic;		/* For debugging... */

	/* Input properties */
	display_size_t		 size;		/* Current input size */
	video_colour_mode_t	 color;		/* Current display color mode */ 
	
	/* Buffer and Frame */
	char			*buffer;
	unsigned int		 buffer_size;
	struct video_frame	 frame;		/* Internal frame */

	/* Quartz */
	WindowRef		 window;
	MenuRef			 menu;
	Rect			 window_rect;	// size of the window containg the displayed image (include padding)
	Rect			 dest_rect;	// size of the displayed image (after scaling)
	CGContextRef		 context;
	int			 request_exit;	/* TRUE if we must exit */
        unsigned long		 last_time;

	/* QuickTime */
	ImageDescriptionHandle	 qt_descr;
	ImageSequence		 qt_seq;	/* Sequence id (for QuickTime) */
	MatrixRecord		 qt_matrix;	/* Scale matrix */	
};

static void	quartz_window_resized (struct state_quartz*);

/*******************************************************************************
 * Events and event handlers...
 ******************************************************************************/

/* Default handler for mouse and keyboard events */
static OSStatus
quartz_event_handler (EventHandlerCallRef	 nextHandler,
		      EventRef			 event,
		      void			*arg)
{
	struct state_quartz	*quartz_state	= (struct state_quartz *) arg;
	OSStatus		 result		= noErr;
	UInt32			 class		= GetEventClass (event),
				 kind		= GetEventKind (event); 

	assert (quartz_state	!= NULL);

	result = eventNotHandledErr;

	if (class == kEventClassKeyboard) {
		char	char_codes;
		UInt32	key_code;
		UInt32	key_modifiers;

		GetEventParameter (event, kEventParamKeyMacCharCodes, typeChar, NULL,
				   sizeof (char_codes), NULL, &char_codes);

		GetEventParameter (event, kEventParamKeyCode, typeUInt32, NULL,
				   sizeof (key_code), NULL, &key_code);

		GetEventParameter (event, kEventParamKeyModifiers, typeUInt32, NULL, 
				   sizeof (key_modifiers), NULL, &key_modifiers);
		
		if (key_modifiers != 256) {
			if (kind == kEventRawKeyRepeat || kind == kEventRawKeyDown) {
				
				result = CallNextEventHandler (nextHandler, event);

				//if (key_code == QZ_ESCAPE) {
				//	quartz_state->request_exit = TRUE;
				//}
			}
		}
	}

	return result;
}

/* Window command handler, called when the window is resized, closed... */
static OSStatus
quartz_window_handler (EventHandlerCallRef	 nextHandler,
		       EventRef			 event,
		       void			*arg)
{
	struct state_quartz	*quartz_state	= (struct state_quartz *) arg;
	OSStatus		 result		= noErr;
	UInt32			 class		= GetEventClass (event);
	UInt32			 kind		= GetEventKind (event); 
	HICommand		 theHICommand;
	WindowRef		 window;
	
	assert (quartz_state	!= NULL);
	
	if (class == kEventClassCommand) {
		
		GetEventParameter (event, kEventParamDirectObject, typeHICommand,
				   NULL, sizeof (HICommand), NULL, &theHICommand);
		
		switch (theHICommand.commandID) {
			case kHICommandQuit:
				result = CallNextEventHandler (nextHandler, event);
				quartz_state->request_exit = TRUE;
				break;
				
			default:
				result = eventNotHandledErr;
				break;
		}
		
	} else if (class == kEventClassWindow) {
		
		GetEventParameter (event, kEventParamDirectObject, typeWindowRef,
				   NULL, sizeof (WindowRef), NULL, &window);
				
		switch (kind) {
			case kEventWindowClosed:
				result = CallNextEventHandler (nextHandler, event);
				quartz_state->window		= NULL;
				quartz_state->request_exit	= TRUE;
				break;
				
				/* Resize window */
			case kEventWindowBoundsChanged:
				result = CallNextEventHandler (nextHandler, event);
				quartz_window_resized (quartz_state);
				break;

			default:
				result = eventNotHandledErr;
				break;
		}
	}

	return result;
}

/* Updates the state of the window and prevents launching the screensaver */
static void
quartz_update (struct state_quartz *s)
{
        unsigned long		curr_time;
        DateTimeRec		curr_datetime;
        EventRef		event;
        EventTargetRef		target;
        OSStatus		err;
        
	assert (s != NULL);
	
        /* Get an event and process it */
	target	= GetEventDispatcherTarget ();
	err	= ReceiveNextEvent (0, 0, kEventDurationNoWait, true, &event);
	if (err == noErr && event != NULL) {
                SendEventToEventTarget (event, target);
                ReleaseEvent (event);
        }
        
	/* Prevent the system to start the screen saver, updating the system every minute */
        GetTime (&curr_datetime);
        DateToSeconds ( &curr_datetime, &curr_time);
        
        if ((curr_time - s->last_time >= 60) || (s->last_time == 0)) {
                UpdateSystemActivity (UsrActivity);
                s->last_time = curr_time;
        }
}

/*******************************************************************************
 * Carbon stuff
 ******************************************************************************/
/* Enable foreground operation for our window */
static void
quartz_foreground (struct state_quartz *s)
{
	/*
	 * Note: this function uses some dark, internal Carbon stuff. For example,
	 * it needs this hidden function...
	 */
        void CPSEnableForegroundOperation (ProcessSerialNumber* psn);
        
	ProcessSerialNumber	my_proc, fr_proc;
        Boolean			same_proc;
	
	UNUSED (s);
        
        if (GetFrontProcess (&fr_proc) == noErr) {
                if (GetCurrentProcess (&my_proc) == noErr) {
                        if (SameProcess (&fr_proc, &my_proc, &same_proc) == noErr && !same_proc) {
				CPSEnableForegroundOperation (&my_proc);
                        }
			
                        SetFrontProcess (&my_proc);
                }
        }
}

/* Shows the window */
static void
quartz_show_window (struct state_quartz	*state)
{
	SetThemeWindowBackground (state->window,
				  kThemeBrushModelessDialogBackgroundActive, TRUE);
	RepositionWindow (state->window, NULL, kWindowCenterOnMainScreen); 	
	ShowWindow (state->window);	
}


/* Update the scale matrix after a resize of the output window */
static void
quartz_window_resized (struct state_quartz *state)
{
	int		w,		h,
			new_w,		new_h;
	double          scale_h,	scale_w,
			min_scale;
	Fixed		min_scalef;
        CGRect		tmp_bounds;

	assert (state != NULL);

	// Clear Background
	SetThemeWindowBackground ( state->window, kThemeBrushUtilityWindowBackgroundInactive, TRUE);
	tmp_bounds = CGRectMake ( 0, 0, state->window_rect.right, state->window_rect.bottom);
	CreateCGContextForPort (GetWindowPort(state->window), &state->context);
	CGContextClearRect (state->context, tmp_bounds);

	if (state->size != DS_NONE) { 
		assert (display_get_width (state->size) > 0);
		assert (display_get_width (state->size) > 0);
		
		/* Get current window dimensions */
		GetPortBounds (GetWindowPort (state->window), &(state->window_rect) );

		w	= display_get_width (state->size);
		h	= display_get_height (state->size);
		new_w	= state->window_rect.right;
		new_h	= state->window_rect.bottom;
		
		scale_h	= (double) new_h / (double) h;
		scale_w	= (double) new_w / (double) w;

		min_scale	= MIN (scale_h, scale_w);
		min_scalef	= X2Fix (min_scale);

		/* Update the window, keeping the aspect ratio */
		SizeWindow (state->window, min_scale * w, min_scale * h, true);

		SetIdentityMatrix (&state->qt_matrix);
		ScaleMatrix (&state->qt_matrix, min_scalef, min_scalef, Long2Fix (0), Long2Fix (0));

		/* Let QuickTime know that we have updated the matrix */
		if (state->qt_seq != 0) {
			SetDSequenceMatrix (state->qt_seq, &state->qt_matrix);
		}
		
	}

	CGContextSynchronize (state->context);
}

/*
 * Create the window
 * (from "Calling Functions to Create Windows and Controls" in Apple docs)
 */
static void
quartz_create_window (struct state_quartz	*state,
		      uint32_t			 d_width,
		      uint32_t			 d_height) 
{
	CFStringRef	title;
	CFStringRef	window_title; 
	OSStatus	result;

	WindowAttributes	windowAttrs =
		  kWindowStandardDocumentAttributes
                | kWindowStandardHandlerAttribute
                | kWindowCompositingAttribute
                //| kWindowLiveResizeAttribute
                | kWindowMetalAttribute;
	
	const EventTypeSpec commands[] = {
		{ kEventClassWindow,	kEventWindowClosed },
		{ kEventClassWindow,	kEventWindowBoundsChanged },
		{ kEventClassCommand,	kEventCommandProcess }
	};
	
	const EventTypeSpec events[] = {
		{ kEventClassKeyboard,	kEventRawKeyDown },
		{ kEventClassKeyboard,	kEventRawKeyRepeat },
		{ kEventClassMouse,	kEventMouseDown },
		{ kEventClassMouse,	kEventMouseWheelMoved }
	};

	debug_msg ("Creating %dx%d window.\n", d_width, d_height);

	SetRect (&state->window_rect,	0, 0, d_width, d_height);
	SetRect (&state->dest_rect,	0, 0, d_width, d_height);
	
	//ClearMenuBar ();
	
        /* Create menu */
        //CreateStandardWindowMenu (0, &state->menu);
        //InsertMenu (state->menu, 0);
	//DrawMenuBar();
	
	/* Create a new window */
	CreateNewWindow (kDocumentWindowClass, windowAttrs,
			 &state->window_rect, &state->window);		
	
	title		= CFSTR ("Ultragrid");
	window_title	= CFCopyLocalizedString (title, NULL);
	result		= SetWindowTitleWithCFString (state->window, window_title);
	
	CFRelease (title);
	CFRelease (window_title);
		
	/* Install handlers for events */
	QTSetPixMapHandleRequestedGammaLevel (GetPortPixMap (GetWindowPort (state->window)),
					      kQTUseSourceGammaLevel);
	
	InstallApplicationEventHandler (NewEventHandlerUPP (quartz_event_handler),
					GetEventTypeCount (events),
					events, state, NULL);	

	InstallWindowEventHandler (state->window,
				   NewEventHandlerUPP (quartz_window_handler), 
				   GetEventTypeCount (commands),
				   commands, state, NULL);
	
	BringToFront (state->window);
}

/*******************************************************************************
 * QuickTime
 ******************************************************************************/

/* Check that QuickTime is installed */
static int
quartz_check_quicktime (void) 
{
	short   error;
	long    result;
	
	error = Gestalt (gestaltQuickTime, &result);
	return (error == noErr);
}

/*
 * Return a image descriptor for input frames.
 * Note: a new image descriptor must be obtained when the input size changes.
 */
static ImageDescriptionHandle
quartz_image_descriptor (display_size_t		size,
			 video_colour_mode_t	color)
{
	ImageDescriptionHandle			  descr;
	NCLCColorInfoImageDescriptionExtension	**colr;
        FieldInfoImageDescriptionExtension2	**fiel;
        CleanApertureImageDescriptionExtension	**clap;
	CodecType				  codec;
	
	debug_msg ("Creating Quartz image descriptor...\n");

	/* There are other types for other color modes, but we only use YUV 4:2:2... */
	if (color == YUV_422) {
		codec = k422YpCbCr8CodecType;
	} else {
		codec = 0;		
	}
	
	descr = (ImageDescriptionHandle) NewHandle ( sizeof (ImageDescription));
	(*descr)->idSize		= sizeof (ImageDescription);
	(*descr)->cType			= codec; 
	(*descr)->resvd1		= 0;
	(*descr)->version		= 1;
	(*descr)->revisionLevel		= 0;
	(*descr)->vendor		= 'appl';
	(*descr)->spatialQuality	= codecLosslessQuality; 
	(*descr)->temporalQuality	= codecLosslessQuality;
	(*descr)->width			= display_get_width  (size);
	(*descr)->height		= display_get_height (size);
	(*descr)->hRes			= Long2Fix (72);
	(*descr)->vRes			= Long2Fix (72);
	(*descr)->dataSize		= 0;
	(*descr)->frameCount		= 1;
	(*descr)->depth			= 32;
	(*descr)->clutID		= -1;
	
	colr = (NCLCColorInfoImageDescriptionExtension**)
		NewHandle (sizeof (NCLCColorInfoImageDescriptionExtension));
	(*colr)->colorParamType		= 'nclc';	/* Default value */
	(*colr)->primaries		= kQTPrimaries_SMPTE_C;
	(*colr)->transferFunction	= kQTTransferFunction_Unknown;
	(*colr)->matrix			= kQTMatrix_ITU_R_601_4;

        fiel = (FieldInfoImageDescriptionExtension2**)
		NewHandle (sizeof (FieldInfoImageDescriptionExtension2));
	(*fiel)->fields			= kQTFieldsProgressiveScan;
	(*fiel)->detail			= kQTFieldDetailUnknown;

        clap = (CleanApertureImageDescriptionExtension**)
		NewHandle (sizeof (CleanApertureImageDescriptionExtension));
	(*clap)->cleanApertureWidthN	= display_get_width (size);
	(*clap)->cleanApertureWidthD	= 1;
	(*clap)->cleanApertureHeightN	= display_get_height (size);
	(*clap)->cleanApertureHeightD	= 1;
	(*clap)->horizOffN		= 0;
	(*clap)->horizOffD		= 1;
	(*clap)->vertOffN		= 0;
	(*clap)->vertOffD		= 1;
	
	AddImageDescriptionExtension (descr, (Handle) colr, kColorInfoImageDescriptionExtension);
	AddImageDescriptionExtension (descr, (Handle) fiel, kFieldInfoImageDescriptionExtension);
        AddImageDescriptionExtension (descr, (Handle) clap, kCleanApertureImageDescriptionExtension);

	return descr;
}


/* Set up of all the QuickTime stuff */
static display_result_t
quartz_create_quicktime (struct state_quartz	*state)
{
	display_result_t	res	= DISPLAY_OK;
	OSErr			err;

	assert (state != NULL);
	assert (state->buffer != NULL);
	assert (state->buffer_size > 0);
	
	debug_msg ("Creating Quartz QuickTime display...\n");

	if (EnterMovies () != noErr) {
		fprintf (stderr, "Error: starting QuickTime with quartz.\n");
		res = DISPLAY_ERR;
	} else {
		SetIdentityMatrix (&state->qt_matrix);

		/* Initialize the decompresion system */
		err = DecompressSequenceBeginS (&(state->qt_seq),
						state->qt_descr,
						(Ptr)state->buffer,
						state->buffer_size,
						GetWindowPort (state->window),
						NULL,
						NULL,
						&(state->qt_matrix),
						srcCopy,
						NULL,
						0,
						codecLosslessQuality,
						bestSpeedCodec);
		if (err != noErr) {
			fprintf (stderr, "Error: creating QuickTime sequence with quartz.\n");
			res = DISPLAY_ERR;
		}
		
		DisposeHandle ((Handle)state->qt_descr);
	}

	return res;
}

/* Destroy the QuickTime system */
static display_result_t
quartz_destroy_quicktime (struct state_quartz *state)
{
	OSErr			err;
	display_result_t	res	= DISPLAY_OK;
	
	assert (state != NULL);
	
	/* Destroy only if the system was initialized */
	if (state->size != DS_NONE && state->color != NONE) {

		debug_msg ("Finishing Quartz display...\n");
		
		err = CDSequenceEnd (state->qt_seq);	/* Finish the QuickTime code */
		if (err != noErr) {
			fprintf (stderr, "Error: destroying QuickTime system.\n");
			res = DISPLAY_ERR;
		}
		
		DisposeHandle ((Handle)state->qt_descr);
		
		state->size = DS_NONE;
		state->color = NONE;
	}

	return res;
}

/* Reconfigure the system for a new size/color state */
static void
quartz_reconfigure (struct state_quartz	*state,
		    display_size_t	 size,
		    video_colour_mode_t	 color)
{
	assert (state != NULL);
	
	if (size != state->size || color != state->color) {
		debug_msg ("Reconfiguring Quartz display.\n");
		
		/* Destroy previous QuickTime */
		quartz_destroy_quicktime (state);

		assert (state->size == DS_NONE && state->color == NONE);

		/*
		 * ... and start a new one with new new configuration
		 * There is no too much too do, as the scaling system will deal
		 * with the main problem
		 */
		state->size  = size;
		state->color = color;
		
		state->qt_descr = quartz_image_descriptor (size, color);
		quartz_window_resized (state);
		quartz_create_quicktime (state);

		quartz_show_window (state);
	}	
}

/*******************************************************************************
 * Public API
 ******************************************************************************/

/* Quartz display initialization. */
void*
display_quartz_init (display_format_t	*format,
		     const char		*display_args,
		     void		*data)
{
	struct state_quartz	*state		= NULL;
	unsigned int		 initial_height,
				 initial_width;
	
	UNUSED (display_args);
	UNUSED (data);
	UNUSED (format);

	if (! quartz_check_quicktime()) {
		fprintf (stderr, "Error: QuickTime not installed.\n");
		return NULL;
	}
	
	state = malloc (sizeof (struct state_quartz));
	if (state != NULL) {

		state->buffer_size	= QUARTZ_MAX_FRAMESIZE;
		state->buffer		= malloc (state->buffer_size);
		if (state->buffer != NULL) {

			state->magic		= MAGIC_QUARTZ;
			state->size		= DS_NONE;
			state->color		= NONE;
			
			state->last_time	= 0;
			state->request_exit	= FALSE;
			
			state->qt_seq		= 0;
		
			/* Set the initial window size */
			initial_height	= display_get_height (QUARTZ_INITIAL_WSIZE);
			initial_width	= display_get_width (QUARTZ_INITIAL_WSIZE);

			/* Set up all the quartz stuff... */
			quartz_foreground (state);
			quartz_create_window (state, initial_width, initial_height);
			quartz_show_window (state);
		} else {
			fprintf (stderr, "Error: allocating memory for quartz buffer.\n");
			free (state);
		}
	} else {
		fprintf (stderr, "Error: allocating state for quartz display.\n");
	}
	
	return state;
}

/* Finish the Quartz driver */
void
display_quartz_done (void *state)
{
	struct state_quartz *s = (struct state_quartz *) state;
	
	/* Free the frame and state */
	if (s != NULL) {
		assert (s->magic == MAGIC_QUARTZ);

		quartz_destroy_quicktime (s);
		
		if (s->frame.data != NULL) {
			free (s->frame.data);
		}
	
		free (s);
	}
}

/* Get a frame */
struct video_frame*
display_quartz_getf (void			*state,
		     display_size_t		 size,
		     video_colour_mode_t	 color)
{
	struct state_quartz	*s = (struct state_quartz *) state;
	struct video_frame	*ret;
	
	assert (s        != NULL);
	assert (s->magic == MAGIC_QUARTZ);

	UNUSED (size);
	UNUSED (color);
	
	if (s->request_exit) {
		ret = NULL;
	} else {
		/* Check if we need to reconfigure the system */
		if (color != s->color || size != s->size) {
			quartz_reconfigure (s, size, color);
		}
		
		if (color == YUV_422) {
			/* Create the frame */
			s->frame.height		= display_get_height (size);
			s->frame.width		= display_get_width (size);
			s->frame.data		= (unsigned char*)s->buffer;
			s->frame.data_len	= s->frame.width * s->frame.height * 2;
			s->frame.colour_mode	= color;
		
			assert (s->frame.data_len <= s->buffer_size);

			ret = &(s->frame);	/* Return the internal frame */
		} else {
			ret = NULL;
		}
	}
	
	quartz_update (s);
	
	return ret;
}

/* Display a frame */
display_result_t
display_quartz_putf (void		*state,
		     struct video_frame	*frame)
{
	struct state_quartz	*s		= (struct state_quartz *) state;
	CodecFlags		 in_flags	= 0,
				 out_flags	= 0;
	OSErr			 err;
	unsigned int		 ret	= DISPLAY_OK;

	assert (s			!= NULL);
	assert (s->magic		== MAGIC_QUARTZ);
	assert (frame			!= NULL);
	assert (frame->height		<= display_get_height (s->size));
	assert (frame->width		<= display_get_width (s->size));
	assert (frame->colour_mode	== s->color);
	assert (frame->data_len		<= s->buffer_size);
	
	in_flags = codecFlagUseImageBuffer;
	
	/* First, check if the user has closed the window or pressed ESC... */
	if (s->request_exit) {
		debug_msg ("Exit requested in Quartz display...");
		ret = DISPLAY_EXIT_REQ;
	} else if (s->frame.data != NULL) {
		/* Ok, we can show the frame */
		err = DecompressSequenceFrameWhen (s->qt_seq,
						   (char*)s->frame.data,
						   s->frame.data_len,
						   in_flags,
						   &out_flags,
						   (ICMCompletionProcRecordPtr) -1,
						   NULL);
		if (err != noErr) {
			fprintf (stderr, "Error: displaying frame with Quartz.\n");
			ret = DISPLAY_ERR;
		}		
	}

	quartz_update (s);

	return ret;
}

/* Color mode suppported by the quartz */
video_colour_mode_t
display_quartz_colour (void *state)
{
	struct state_quartz *s = (struct state_quartz *) state;
	
	assert (s		!= NULL);
	assert (s->magic	== MAGIC_QUARTZ);
	
	return YUV_422;
}

/* Video formats supported by the quartz */
display_type_t *
display_quartz_probe (display_format_t	*format,
		      const char	*display_args)
{
	display_type_t		*dt		= NULL;
	display_format_t	*dformat	= NULL;
	
	UNUSED (display_args);
	UNUSED (format);
	
	dformat = malloc (4 * sizeof (display_format_t));
	if (dformat != NULL) {
		dformat[0].size		= DS_QCIF;
		dformat[0].colour_mode	= YUV_422;
		dformat[0].num_images	= 1;
		dformat[1].size		= DS_CIF;
		dformat[1].colour_mode	= YUV_422;
		dformat[1].num_images	= 1;
		dformat[2].size		= DS_SCIF;
		dformat[2].colour_mode	= YUV_422;
		dformat[2].num_images	= 1;
		dformat[3].size		= DS_720p;
		dformat[3].colour_mode	= YUV_422;
		dformat[3].num_images	= 1;
		
		dt = malloc (sizeof (display_type_t));
		if (dt != NULL) {
			dt->id		= DISPLAY_QUARTZ_ID;
			dt->name	= "quartz";
			dt->description = "Quartz";
			dt->formats	= dformat;
			dt->num_formats = 4;
		} else {
			fprintf (stderr, "Error: allocating memory for dt.\n");
			free (dformat);
		}
	} else {
		fprintf (stderr, "Error: allocating memory for dformat.\n");
	}
	
	return dt;
}


#endif
