/*
 * FILE:   video_display/sdl.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.32 $
 * $Date: 2006/01/31 12:21:25 $
 *
 */


#include "config.h"
#include "config_unix.h"

#ifdef HAVE_LIBSDL

#include <SDL.h>  

#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/sdl.h"
#include "tv.h"

#define MAGIC_SDL   DISPLAY_SDL_ID

/* Internal state */
struct state_sdl {    
	SDL_Surface		*screen;   
	display_size_t		 screen_size;		/* Current display size */
	video_colour_mode_t	 screen_color;		/* Current display color mode */ 

	SDL_Overlay		*overlay;
	
	struct video_frame	 frame;
	
	int			 request_exit;		/* Should we exit? */
	
	uint32_t		 magic;			/* For debugging... */
};



/*
 * Set the SDL display.
 * The size and/or the color are expected to be different to the current values.
 */
static display_result_t
sdl_reconfigure_screen (struct state_sdl	*state,
			display_size_t		 size, 
			video_colour_mode_t	 color)
{
	const SDL_VideoInfo	*vi;
	unsigned int		 w, h;
	char			 driver_name [128];
	int			 flags	= SDL_SWSURFACE
					| SDL_ANYFORMAT;

	//int			 flags	= SDL_DOUBLEBUF	// This should improve performance
	//				| SDL_HWSURFACE
	//				| SDL_ANYFORMAT;
	display_result_t	 ret	= DISPLAY_OK;

	
	assert (state != NULL);
	
	if (size != DS_NONE) {

		assert (state->screen == NULL);
		assert (!SDL_WasInit (SDL_INIT_VIDEO));

		if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) == -1) {
			fprintf(stderr, "Error: initializing the SDL.\n");
			display_sdl_done (state);
			return DISPLAY_ERR;
		}

		w = display_get_width (size);
		h = display_get_height (size);
		
		debug_msg ("Trying %dx%d display size.\n", w, h);
		
		/* Check if the video mode is supported */
		if (!SDL_VideoModeOK (w, h, 0, flags)) {
			fprintf (stderr, "WARNING!!! Video size could not be supported...\n");
		}
		
		state->screen = SDL_SetVideoMode (w, h, 0, flags);
		if(state->screen == NULL) {
			fprintf (stderr, "SDL display error: Unable to set video size %dx%d: %s\n", 
				 w, h, SDL_GetError());
			ret = DISPLAY_ERR;
		} else {
			state->screen_size	= size;
			state->screen_color	= color;
			
			/* Some cosmetic stuff... */ 
			SDL_ShowCursor (1);
			SDL_WM_SetCaption ("UltraGrid", NULL);

			/* SDL video info */
			vi = SDL_GetVideoInfo ();
			SDL_VideoDriverName (driver_name, sizeof (driver_name));
			debug_msg ("Video output:\n");
			debug_msg (" Driver: %s SDL [%dx%d, %s]\n",
				   driver_name, state->screen->w, state->screen->h,
				   cm_get_name (state->screen_color));
			debug_msg (" Hardware surface? %s\n", (vi->hw_available ? "yes" : "no"));
			debug_msg (" Window Manager? %s\n", (vi->wm_available ? "yes" : "no"));
			debug_msg (" BLIT accelaration? %s\n", (vi->blit_hw ? "yes" : "no"));
			debug_msg (" Video Memory: %dKB\n", vi->video_mem);
			debug_msg (" Preferred Depth: %d bits/pixel\n", vi->vfmt->BitsPerPixel);
			debug_msg (" Must lock? %s\n", (SDL_MUSTLOCK(state->screen) ? "yes" : "no"));

			/*
			 * Creates a frame with the expected size, and the data
			 * section is the overlay structure...
			 */ 
			state->frame.colour_mode = display_sdl_colour (state);
			state->frame.height      = h;
			state->frame.width       = w;
			state->frame.data_len    = h * w * 2;
		}
	}
	
	return ret;
}

/* Close the screen */
static display_result_t
sdl_destroy_screen (struct state_sdl *state)
{
	assert (state != NULL);
	
	if (state->screen != NULL) {
		
		assert (state->overlay == NULL);	/* there must be no overlay */
		
		/* Clean up the SDL library */
		if (SDL_WasInit (SDL_INIT_VIDEO))
			SDL_QuitSubSystem (SDL_INIT_VIDEO);
		
		state->screen	= NULL;
	}
	
	return DISPLAY_OK;
}

/* Reconfigures the screen overlay */
static display_result_t
sdl_reconfigure_overlay (struct state_sdl	*state,
			 display_size_t		 size, 
			 video_colour_mode_t	 color)
{
	display_result_t ret = DISPLAY_OK;
	
	assert (state != NULL);
		
	if (color == YUV_422) {

		assert (state->screen != NULL);
		
		debug_msg("Creating YUV overlay.\n");

		/* Creates a YUV overlay for the display */ 
		state->overlay = 
			SDL_CreateYUVOverlay (display_get_width (size),
					      display_get_height (size),
					      SDL_UYVY_OVERLAY,
					      state->screen);
		
		if ( state->overlay == NULL ) {
			fprintf (stderr, "SDL display error: can't create SDL YUV overlay: %s\n", SDL_GetError());
			return DISPLAY_ERR;
		}
		
		/* Some debug info */ 
		debug_msg (" YUV Overlay size: %dx%d \n",
			  state->overlay->w, state->overlay->h);
		debug_msg (" YUV hardware acceleration? %s\n", 
			  ( state->overlay->hw_overlay == 1 ? "yes" : "no"));
		
		/* Displays a initial black frame */ 
		SDL_DisplayYUVOverlay (state->overlay, &(state->screen->clip_rect) );
		
		/* Set the overlay data area as the frame data buffer */
		state->frame.data = state->overlay->pixels[0];
	} else {
		fprintf (stderr, "SDL display error: can't handle %s color mode\n.",
			 cm_get_name (state->screen_color));
		ret = DISPLAY_ERR;
	}
	
	return ret;
}

/* Close the overlay */
static display_result_t
sdl_destroy_overlay (struct state_sdl		*state)
{
	assert (state != NULL);
	
	if (state->overlay != NULL) {
		SDL_FreeYUVOverlay (state->overlay);
		state->frame.data	= NULL;
		state->overlay		= NULL;
	}
	
	return DISPLAY_OK;
}

/* Reconfigure the screen if needed */
static display_result_t
sdl_reconfigure (struct state_sdl	*state,
		 display_size_t		 size,
		 video_colour_mode_t	 color)
{
	
	assert (size         != DS_NONE);
	assert (state        != NULL);
	assert (state->magic == MAGIC_SDL);

	/* Do we need to resize the display? */
	if (state->screen_size != size || state->screen_color != color) {
		sdl_destroy_overlay	(state);
		sdl_destroy_screen	(state);

		sdl_reconfigure_screen	(state, size, color);
		sdl_reconfigure_overlay	(state, size, color);
	}
	
	return DISPLAY_OK;
}

/* Process any SDL event */
static void
sdl_update (struct state_sdl  *s)
{
	SDL_Event          event;

	/* Control the exit request */
	SDL_PollEvent (&event);
	if (event.type == SDL_QUIT ||
	    (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE)) {
		s->request_exit = TRUE;
	}
}

/***************************************************************************
 * Public API
 **************************************************************************/

/* SDL output initialization. */
void*
display_sdl_init (display_format_t	*format,
		  const char		*display_args,
		  void			*data)
{
	struct state_sdl    *sdl_display	= NULL;

	UNUSED (display_args);
	UNUSED (format);
	UNUSED (data);
	
	sdl_display = malloc (sizeof (struct state_sdl));
	if (sdl_display != NULL) {
		sdl_display->magic		= MAGIC_SDL;
		
		sdl_display->screen		= NULL;
		sdl_display->screen_size	= DS_NONE;	/* Set the initial display size */
		sdl_display->screen_color	= 0;
		
		sdl_display->overlay		= NULL;

		sdl_display->request_exit	= FALSE;
		
#ifdef HAVE_TCL_TK
		/* We can force XV output setting this... */ 
		setenv ("SDL_VIDEO_X11_NODIRECTCOLOR", "1", 1);  
#endif
		/*
		 * Setting YUV acceleration improves the performance, but is
		 * dodgy in some systems...
		 */
		if (getenv ("SDL_VIDEO_YUV_HWACCEL") != NULL) {
			fprintf (stderr, "WARNING! SDL_VIDEO_YUV_HWACCEL environment variable can result in errors with the SDL display.\n");
		} else {			
			setenv ("SDL_VIDEO_YUV_HWACCEL", "0", 0);			
		}
		
		if (getenv ("SDL_VIDEO_YUV_DIRECT") != NULL) {
			fprintf (stderr, "WARNING! SDL_VIDEO_YUV_DIRECT environment variable can result in errors with the SDL display.\n");
		} else {			
			setenv ("SDL_VIDEO_YUV_DIRECT", "1", 0);
		}
        }
	
	return sdl_display;
}

void
display_sdl_done (void *state)
{
	struct state_sdl *s = (struct state_sdl *) state;
	
	if (s != NULL) {
		assert (s->magic == MAGIC_SDL );

		sdl_destroy_overlay (s);
		sdl_destroy_screen (s);

		assert (!SDL_WasInit (SDL_INIT_VIDEO));
		SDL_Quit ();
		
		/* Free the state */
		free (s);
	}
}

struct video_frame*
display_sdl_getf (void			*state,
		  display_size_t	 size,
		  video_colour_mode_t	 color)
{
	struct state_sdl	*sdl_state	= (struct state_sdl *) state;
	struct video_frame	*ret		= NULL;
	
	assert (sdl_state        != NULL);
	assert (sdl_state->magic == MAGIC_SDL);
	
	/*
	 * Reconfigure the display for the frame that we will return.
	 * If it is the first time we give a frame, we can not delay the SDL 
	 * initialization after this point...
	 */
	if (! sdl_state->request_exit) {
		sdl_reconfigure (sdl_state, size, color);
			
		if (sdl_state->screen != NULL) {
		
			assert (sdl_state->overlay != NULL);
			
			/* If we need it, lock the screen */
			if (SDL_MUSTLOCK (sdl_state->screen)) {
				if (SDL_LockSurface (sdl_state->screen) < 0) {
					fprintf (stderr, "Error locking output screen: %s.\n", SDL_GetError());
					return NULL;
				}
			}
			
			if (SDL_LockYUVOverlay (sdl_state->overlay) < 0) {
				fprintf (stderr, "Error locking overlay: %s.\n", SDL_GetError());
				return NULL;
			}

			ret = &(sdl_state->frame);   /* returns the frame */
		} else {
			fprintf (stderr, "Error getting frame from null SDL screen.\n");
		}
	}
	
	return ret;
}

display_result_t
display_sdl_putf (void *state, struct video_frame *frame)
{
	struct state_sdl	*s	= (struct state_sdl *) state;
	display_result_t	 ret	= DISPLAY_OK;

	assert (s			!= NULL);
	assert (s->magic		== MAGIC_SDL);
	assert (frame			!= NULL);
	assert (frame			== &(s->frame));
	assert (frame->height		<= display_get_height (s->screen_size));
	assert (frame->width		<= display_get_width (s->screen_size));
	assert (frame->colour_mode	== s->screen_color);

	if (s->screen != NULL) {
		assert (s->overlay != NULL);

		if (SDL_MUSTLOCK (s->screen))
			SDL_UnlockSurface (s->screen);
		
		SDL_UnlockYUVOverlay (s->overlay);   
		
		SDL_DisplayYUVOverlay (s->overlay, &(s->screen->clip_rect));
		
		/* Control the exit request */
		sdl_update (s);
		
		if (s->request_exit == TRUE) {
			ret = DISPLAY_EXIT_REQ;
		}		
	} else {
		fprintf (stderr, "Error putting frame in invalid SDL screen.\n");
		ret = DISPLAY_ERR;
	}	
	
	return ret;
}

/*
 * Color modes suppported by the SDL
 */
video_colour_mode_t
display_sdl_colour (void *state)
{
	struct state_sdl *s = (struct state_sdl *) state;
	
	assert (s != NULL);
	assert (s->magic == MAGIC_SDL);
	
	return YUV_422;
}

/*
 * Video formats supported by the SDL
 */
display_type_t *
display_sdl_probe (display_format_t *format, const char* display_args)
{
	display_type_t		*dt		= NULL;
	display_format_t	*dformat	= NULL;
	
	UNUSED (display_args);
	UNUSED (format);
	
	dformat = malloc(4 * sizeof(display_format_t));
	if (dformat != NULL) {
		dformat[0].size		= DS_QCIF;
		dformat[0].colour_mode	= YUV_422;
		dformat[0].num_images	= 1;
		dformat[1].size		= DS_CIF;
		dformat[1].colour_mode	= YUV_422;
		dformat[1].num_images	= 1;
		dformat[2].size		= DS_SCIF;
		dformat[2].colour_mode	= YUV_422;
		dformat[2].num_images	= 1;
		dformat[3].size		= DS_720p;
		dformat[3].colour_mode	= YUV_422;
		dformat[3].num_images	= 1;
		
		dt = malloc (sizeof (display_type_t));
		if (dt != NULL) {
			dt->id = DISPLAY_SDL_ID;
			dt->name = "sdl";
			dt->description = "SDL";
			dt->formats = dformat;
			dt->num_formats = 4;
		}
	}
	
	return dt;
}

#endif
