/*
 * FILE:   video_display/x11.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2004-2005 University of Glasgow
 * Copyright (c) 2001-2003 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.33 $
 * $Date: 2006/02/22 01:05:44 $
 *
 */

#include "config.h"
#include "config_unix.h"

#ifndef X_DISPLAY_MISSING /* Don't try to compile if X is not present */
#ifdef HAVE_TCL_TK        /* Need Tcl/Tk currently                    */

#include "debug.h"
#include "environment.h"
#include "video_types.h"
#include "video_display.h"
#include "video_display/x11.h"
#include "tcl.h"
#include "tk.h"
#include "ui/ui_vwin.h"

/* For X shared memory... */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

extern Tcl_Interp      *interp;		/* FIXME */

#define HD_WIDTH 	1280
#define HD_HEIGHT 	720
#define MAGIC_X11	0xcafebabe

struct state_x11 {
	 Tk_Window		 vw;
	 Display		*display;
	 Window			 window;
	 GC			 gc;
	 int			 vw_depth;
	 Visual			*vw_visual;
	 XImage			*vw_image;
	 XShmSegmentInfo	 vw_shm_segment;
	 int			 xv_port;
	 uint32_t		 magic;			/* For debugging */
	 struct video_frame	*frame;
};

static void 
event_handler_x11(ClientData cd, XEvent *ep)
{
	UNUSED(cd);
	UNUSED(ep);
}

void*
display_x11_init (display_format_t	*format,
		  const char		*display_args,
		  void			*data)
{
	struct state_x11	*s;
	Tk_Window		 m_win;

	UNUSED (display_args);
	UNUSED (format);
	UNUSED (data);

	s = (struct state_x11 *) malloc(sizeof(struct state_x11));
	if (s != NULL) {
		s->magic   = MAGIC_X11;

		/* Create a bare window to draw into... */
		m_win = Tk_MainWindow(interp);
		if (m_win == NULL) {
			printf("X11: Cannot find main Tk window\n");
			return NULL;
		}

		s->vw = Tk_CreateWindowFromPath(interp, m_win, ".vw", 0);
		if (s->vw == 0) {
			printf("X11: Cannot create video window\n");
			return NULL;
		}
		
		Tk_SetClass(s->vw, "hdrecv_video");
		Tk_CreateEventHandler(s->vw, ExposureMask|StructureNotifyMask, event_handler_x11, 0);
		Tk_GeometryRequest(s->vw, HD_WIDTH, HD_HEIGHT);
		Tk_MapWindow(s->vw);
		s->vw_depth  = Tk_Depth(s->vw);
		s->vw_visual = Tk_Visual(s->vw);

		/* For now, we only support 24-bit TrueColor displays... */
		if (s->vw_depth != 24) {
			printf("X11: Display doesn't support 24 bit colour\n");
			return NULL;
		}
		if (s->vw_visual->class != TrueColor) {
			printf("X11: Display doesn't support true colour visual mode\n");
			return NULL;
		}

		/* Do the shared memory magic... */
		s->vw_image = XShmCreateImage(Tk_Display(s->vw), s->vw_visual, s->vw_depth, ZPixmap, NULL, &s->vw_shm_segment, HD_WIDTH, HD_HEIGHT);
		debug_msg ("vw_image                   = %p\n", s->vw_image);
		debug_msg ("vw_image->width            = %d\n", s->vw_image->width);
		debug_msg ("vw_image->height           = %d\n", s->vw_image->height);
		if (s->vw_image->width != HD_WIDTH) {
			printf("X11: Display does not support %d pixel wide shared memory images\n", HD_WIDTH);
			return NULL;
		}
		if (s->vw_image->height != HD_HEIGHT) {
			printf("X11: Display does not support %d pixel tall shared memory images\n", HD_HEIGHT);
			return NULL;
		}

		s->vw_shm_segment.shmid    = shmget (IPC_PRIVATE, s->vw_image->bytes_per_line * s->vw_image->height, IPC_CREAT|0777);
		s->vw_shm_segment.shmaddr  = shmat (s->vw_shm_segment.shmid, 0, 0);
		s->vw_shm_segment.readOnly = False;

		s->vw_image->data = s->vw_shm_segment.shmaddr;

		if (XShmAttach (Tk_Display (s->vw), &s->vw_shm_segment) == 0) {
			printf ("X11: Cannot attach shared memory segment\n");
			XDestroyImage (s->vw_image);
			return NULL;
		}

		/* Get our window onto the screen... */
		Tcl_Eval (interp, "pack .vw");

		s->display = Tk_Display (s->vw);
		s->window  = Tk_WindowId (s->vw);
		s->gc      = Tk_GetGC (s->vw, 0, 0);

		while (Tcl_DoOneEvent (TCL_DONT_WAIT | TCL_ALL_EVENTS)) {
			/* Process Tcl/Tk events */
		}

		s->frame = malloc(sizeof(struct video_frame));
		if (s->frame != NULL) {
			s->frame->colour_mode = RGB;
			s->frame->width       = HD_WIDTH;
			s->frame->height      = HD_HEIGHT;
			s->frame->data        = s->vw_image->data;
			s->frame->data_len    = HD_WIDTH * HD_HEIGHT * 4;
		} else {
			display_x11_done(s);
			return NULL;
		}

	}
	
	return (void *) s;
}

void
display_x11_done(void *state)
{
	struct state_x11 *s = (struct state_x11 *) state;

	assert(s->magic == MAGIC_X11);

	XShmDetach(Tk_Display(s->vw), &(s->vw_shm_segment));
	XDestroyImage(s->vw_image);
	shmdt(s->vw_shm_segment.shmaddr);
	shmctl(s->vw_shm_segment.shmid, IPC_RMID, 0);
	if (s->frame != NULL) {
		free(s->frame);
	}
}

struct video_frame *
display_x11_getf(void *state, display_size_t size, video_colour_mode_t color)
{
	struct state_x11 *s = (struct state_x11 *) state;

	UNUSED(size);
	UNUSED(color);

	assert(s->magic == MAGIC_X11);
	return s->frame;
}

display_result_t
display_x11_putf(void *state, struct video_frame *frame)
{
	struct state_x11 *s = (struct state_x11 *) state;

	assert(s->magic == MAGIC_X11);
	assert(frame == s->frame);
	XShmPutImage(s->display, s->window, s->gc, s->vw_image, 0, 0, 0, 0, s->vw_image->width, s->vw_image->height, True);
	XFlush(s->display);
	while (Tcl_DoOneEvent(TCL_DONT_WAIT | TCL_ALL_EVENTS)) {
		/* Process Tcl/Tk events */
	}
	return DISPLAY_OK;
}

video_colour_mode_t
display_x11_colour(void *state)
{
	struct state_x11 *s = (struct state_x11 *) state;
	assert(s->magic == MAGIC_X11);
	return RGB;
}

display_type_t *
display_x11_probe (display_format_t *format, const char* display_args)
{
	display_type_t		*dt;
	display_format_t	*dformat;
	
	UNUSED( display_args );
	UNUSED( format );

	ui_create_vwin(DS_QCIF, RGB);

	dformat = malloc(3 * sizeof(display_format_t));
	dformat[0].size        = DS_QCIF;
	dformat[0].colour_mode = RGB;
	dformat[0].num_images  = -1;
	dformat[1].size        = DS_CIF;
	dformat[1].colour_mode = RGB;
	dformat[1].num_images  = -1;
	dformat[2].size        = DS_SCIF;
	dformat[2].colour_mode = RGB;
	dformat[2].num_images  = -1;

	dt = malloc(sizeof(display_type_t));
	if (dt != NULL) {
		dt->id	        = DISPLAY_X11_ID;
		dt->name        = "x11";
		dt->description = "X Window System";
		dt->formats     = dformat;
		dt->num_formats = 3;
	}
	return dt;
}

#endif /* HAVE_TCL_TK       */
#endif /* X_DISPLAY_MISSING */

