/*
 * FILE:   video_types.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.12 $
 * $Date: 2006/05/12 15:09:54 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"

typedef struct {
	video_colour_mode_t	 id;
	char			*description;
	double			 bpp;
} video_colour_mode_ext_t;

static video_colour_mode_ext_t video_colour_table[] = {
	{ NONE,			"NONE",			0},
	{ RGB,			"RGB",			4.0},
	{ YV_12,		"YV_12",		3.0/2.0},
	{ YUV_422,		"YUV_422",		2.0},
	{ YUV_422_DVS10,	"YUV_422_DVS10",	2.0},
};

#define NUM_COL_MODES (sizeof(video_colour_table)/sizeof(video_colour_mode_ext_t))


char*	
cm_get_name (video_colour_mode_t c)
{
	assert (c < NUM_COL_MODES);
	return video_colour_table[c].description;
}

double
cm_get_bpp (video_colour_mode_t c)
{
	assert (c < NUM_COL_MODES);
	return video_colour_table[c].bpp;
}

/*******************************************************************************
 * video_frames
 ******************************************************************************/

struct video_frame*
video_frame_alloc (unsigned int data_len)
{
	struct video_frame	*res	= NULL;
	
	res = (struct video_frame*) malloc (sizeof(struct video_frame));
	if (res != NULL) {
		res->ts          = tv_mediatime ();
		res->colour_mode = YUV_422;
		res->height	 = 0;
		res->width	 = 0;
		res->data_len    = data_len;
		
		if (data_len > 0) {
			res->data = (unsigned char*) malloc (data_len);
			if (res->data == NULL) {
				fprintf (stderr, "Error: allocating memory for video_frame");
				free (res);
			}
		}
	} else {
		fprintf (stderr, "Error allocating memory for video_frame");
	}
	
	return res;
}

/* Copy a video_frame. dst must have been previously allocated */
void
video_frame_copy (struct video_frame *dst, struct video_frame *src)
{
	assert (src != NULL);
	assert (dst != NULL);
	assert (dst->data != NULL);
	assert (dst->data_len >= src->data_len);
	
	dst->ts          = src->ts;
	dst->colour_mode = src->colour_mode;
	dst->width	 = src->width;
	dst->height	 = src->height;
	dst->data_len	 = src->data_len;
	
	if (src->data != NULL && src->data != dst->data && dst->data_len > 0) {
		memcpy (dst->data, src->data, src->data_len);
	}
}


void
video_frame_free (struct video_frame *frame)
{
	assert (frame != NULL);
	
	if (frame->data != NULL)
		free (frame->data);
	
	free (frame);
}

