/*
 * FILE:    main.c
 * AUTHORS: Colin Perkins <csp@csperkins.org>
 *          Ladan Gharai  <ladan@isi.edu>
 *
 * Copyright (c) 2001-2004 University of Southern California
 * Copyright (c) 2003-2004 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.22 $
 * $Date: 2005/05/27 20:46:57 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "rtp/rtp.h"
#include "tfrc.h"
#include "ntp.h"
#include "tv.h"
#include "version.h"
#include "rtp/net_udp.h"

#define RUN_TFRC "running TFRC code/version 0.0"

uint32_t  	RTT=0;    /* this is computed by handle_rr in rtp_callback */
int		i=0;
uint32_t	tfrc_sender=FALSE,
		tfrc_receiver=TRUE;
static int      should_exit  = FALSE;
 
void my_rtp_callback(struct rtp *session, rtp_event *e);


static void 
print_usage(void) 
{
	printf("Usage: run_tfrc [-s <address>] [-r <address>] [-m <mtu>] [-p <port>]\n");
}

static void
tfrc_process_rr(struct rtp *session, rtp_event *e)
{
	float		fract_lost, tmp;
        uint32_t	ntp_sec, ntp_frac, now; 
	rtcp_rr		*r = (rtcp_rr *) e->data;

	if (e->ssrc == rtp_my_ssrc(session)) {
		/* Filter out loopback reports */
		return;
	}

	if (r->ssrc == rtp_my_ssrc(session)) {
		/* Received a reception quality report for data we are sending.  */
		/*                                                               */
		/* Check for excessive congestion: exit if it occurs, to protect */
		/* other users of the network.                                   */
		fract_lost = (r->fract_lost * 100.0) / 256.0;	/* percentage lost packets */
		//if (fract_lost > 20) {	
		//	printf("Receiver 0x%08x reports excessive congestion, exiting\n", e->ssrc);
		//	exit(1);
		//} /* we want it to look bad! ha ha */

		/* Compute network round-trip time:                              */
		if (r->lsr != 0) {
			ntp64_time(&ntp_sec, &ntp_frac);
			now = ntp64_to_ntp32(ntp_sec, ntp_frac); 

			if ((now < r->lsr) || (now - r->lsr < r->dlsr)) {
				/* Packet arrived before it was sent, ignore. This represents */
				/* a bug in the remote: either our timestamp was incorrectly  */
				/* echoed back to us, or the remote miscalculated in time for */
				/* which it held the packet.                                  */
				debug_msg("Bogus RTT from 0x%08x ignored\n", e->ssrc);
			} else {
				tmp = ((float) (now-r->lsr-r->dlsr))/65536.0; /* RTT in seconds */
				RTT = tmp * 1000000; /* RTT in usec */
			}
			debug_msg("  RTT=%d usec\n", RTT);
		}
	}
}


void
my_rtp_callback(struct rtp *session, rtp_event *e)
{
	double 	tmp;
	rtcp_rx		*tfrc_rx      = (rtcp_rx *) e->data;
	//rtp_packet	*pckt_rtp     = (rtp_packet *) e->data;
	//struct pdb	*participants = (struct pdb *) rtp_get_userdata(session);

	switch (e->type) {
	case RX_RTP: 
		/* if (pckt_rtp->meta.data_len > 0) {
			if ((pckt_rtp->fields.seq%100)==0)
				printf ("seq #: %d\n", pckt_rtp->fields.seq );
		} */
		break;
	case RX_TFRC_RX:
		tfrc_rx = (rtcp_rx *) e->data;
		tmp = (double) tfrc_rx->p/4294967296.0;

	 	printf ("\n*rcv_rrx: %6d %6d %8d %10.8f", 
			//htonl(tfrc_rx->ts), htonl(tfrc_rx->tdelay), htonl(tfrc_rx->x_recv), htonl(tfrc_rx->p));	
			tfrc_rx->ts, tfrc_rx->tdelay, tfrc_rx->x_recv, tmp);
		i=0;
		break;
	case RX_RTCP_START:
		break;
	case RX_RTCP_FINISH:
		break;
	case RX_SR:
		break;
	case RX_RR:
		tfrc_process_rr(session, e);
		break;
	case RX_RR_EMPTY:
		break;
	case RX_SDES:
		//process_sdes(participants, e->ssrc, (rtcp_sdes_item *) e->data);
		break;
	case RX_APP:
		break;
	case RX_BYE:
		break;
	case SOURCE_DELETED:
		break;
	case SOURCE_CREATED:
		//pdb_add(participants, e->ssrc);
		break;
	case RR_TIMEOUT:
		break;
	default:
		debug_msg("Unknown RTP event (type=%d)\n", e->type);
	}
}




int 
main(int argc, char *argv[])
{
	struct rtp		*rtp_session;
	struct timeval		 timeout, 
				 start_time, 
				 curr_time;
	unsigned 		 requested_mtu = 1400;
	uint32_t                 ts;
        double                   rtcp_bw = 5 * 1024 * 1024;     /* FIXME */
	char            	 *data;
const   char			 *address="none";
	int			 port = 5004;
        int                      pt = 96;
	int                      ch;
	int 			 sent;

	data = malloc (requested_mtu);
	while ((ch = getopt(argc, argv, "s:r:m:p:v")) != -1) {
		switch (ch) {
		case 's' : 
			address=optarg; 
			tfrc_sender=TRUE;
			break;
		case 'r' : 
			address=optarg;
			break;
		case 'm' :
			requested_mtu = atoi(optarg);
			break;
		case 'p' :
			port= atoi(optarg);
			break;
		case 'v' :
     			printf("%s\n", RUN_TFRC);
			return EXIT_SUCCESS;
		default :
			print_usage();
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	printf ("%d %s\n", argc, argv[0]);

	if (argc != 0) {
		printf ("%d %s\n", argc, argv[0]);
		print_usage();
		return 1;
	}

	gettimeofday(&start_time, NULL);


        rtp_session = rtp_init(address, port, port, 255, rtcp_bw, TRUE /*tfrc_on*/, my_rtp_callback, NULL);
        if (rtp_session != NULL) {
                rtp_set_option(rtp_session, RTP_OPT_WEAK_VALIDATION,   TRUE);
                rtp_set_sdes(rtp_session, rtp_my_ssrc(rtp_session), RTCP_SDES_TOOL, RUN_TFRC, strlen(RUN_TFRC));
	} else { 
		printf ("Unable to set up RTP session: %s", address); 
		exit(0);
	}

        printf ("start (media) time:%d\n", get_local_time ()); 

	while (!should_exit) {
		/* Housekeeping and RTCP... */
		gettimeofday(&curr_time, NULL);
		ts = tv_diff(curr_time, start_time) * 90000;
		rtp_update(rtp_session, curr_time);
		rtp_send_ctrl(rtp_session, ts, 0, curr_time);

		/* Receive packets from the network... The timeout is adjusted */
		/* to match the video capture rate, so the transmitter works.  */
		timeout.tv_sec  = 0;
		timeout.tv_usec = 100; /* huh? */
		rtp_recv(rtp_session, &timeout, ts);
		
		//receving RTP packets happens via the callback

		if (tfrc_sender) {
			ts = get_local_time();
			sent = rtp_send_data (rtp_session, ts, pt, 1, 0, 0, data, requested_mtu, 0, 0, 0);
			/*if (sent)
				printf (" %d:%d, ", i++,sent );
			else
				printf (".");*/
		}

	}
	debug_msg("Exiting...\n");
	rtp_done(rtp_session);
	return EXIT_SUCCESS;
}

