/*
 * FILE:    test_video_codec_dv.c
 * AUTHORS: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2005 University of Glasgow
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the University must not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.7 $
 * $Date: 2006/05/15 17:28:01 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "video_types.h"
#include "video_capture.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_codec.h"
#include "video_codec/dv.h"
#include "test_video_codec_dv.h"

static struct video_frame *
load_dv_image_pal(void)
{
	int			 fd;
	struct video_frame	*frame;

	frame = video_frame_alloc(829440);
	if (frame != NULL) {
		fd = open("test/image-pal.yuv", O_RDONLY, 0);
		if (fd < 0) {
			return NULL;
		}
		if (read(fd, frame->data, 829440) != 829440) {
			return NULL;
		}
		close(fd);
		frame->width  = 720;
		frame->height = 576;
	}
	return frame;
}

static struct video_frame *
load_dv_image_ntsc(void)
{
	int			 fd;
	struct video_frame	*frame;

	frame = video_frame_alloc(691200);
	if (frame != NULL) {
		fd = open("test/image-ntsc.yuv", O_RDONLY, 0);
		if (fd < 0) {
			return NULL;
		}
		if (read(fd, frame->data, 691200) != 691200) {
			return NULL;
		}
		close(fd);
		frame->width  = 720;
		frame->height = 480;
	}
	return frame;
}

static int
test_dv_capabilities(void)
{
	unsigned  id = vcodec_get_by_name("dv");

	printf("Testing video codec: DV (capabilities) ................................... "); 
	fflush(stdout);

	if (!vcodec_can_encode(id)) {
		printf("FAIL\n   DV codec cannot encode video\n");
		return 1;
	}

	if (!vcodec_can_decode(id)) {
		printf("FAIL\n   DV codec cannot decode video\n");
		return 1;
	}

	if (!vcodec_can_packetize(id)) {
		printf("FAIL\n   DV codec cannot packetize video\n");
		return 1;
	}

	printf("Ok\n");
	return 0;
}

static int
test_dv_pal_encode(void)
{
	unsigned 		 id = vcodec_get_by_name("dv");
	void			*encoder;
	struct encoded_frame	*ef;
	struct encoded_data	*ed;
	struct video_frame	*vf;
	unsigned		 nn, i, dbc;
	uint16_t		 seq_prev;
	t_usec			 ts_prev;

	printf("Testing video codec: DV (PAL encoder) .................................... "); 
	fflush(stdout);

	vf = load_dv_image_pal();	
	if (vf == NULL) {
		printf("FAIL\n   Cannot read test/image-pal.yuv\n");
		return 1;
	}

	encoder = vcodec_encoder_create(id, 1500);
	if (encoder == NULL) {
		printf("FAIL\n   Cannot create DV encoder\n");
		return 1;
	}

	ef = vcodec_eframe_alloc(id);
	nn = vcodec_encode(id, encoder, vf, ef, NULL);
	if (nn == 0) {
		printf("FAIL\n   Cannot encode DV frame\n");
		return 1;
	}

	/* Parse the encoded frames for correctness... */
	dbc = 0;
	seq_prev = 0xffff;
	ts_prev  = 0;
	for (i = 0; i < nn; i++ ) {
		ed = vcodec_eframe_get(id, ef, i);
		assert(ed->header     == NULL);
		assert(ed->header_len == 0);
		assert(ed->data != NULL);
		assert(ed->data_len % 80 == 0);
		assert(ed->seqno == ++seq_prev);	/* Sequence number monotonically increases     */
		dbc += ed->data_len / 80;		/* Count number of 80 byte DIF blocks          */
		if (i > 0) {				/* Timestamp must be the same on all fragments */
			assert(ed->rtp_ts == ts_prev);
		}
		ts_prev = ed->rtp_ts;
	}
	assert(dbc == 1800);	/* Number of DIF blocks in a PAL frame */

	vcodec_eframe_free(id, ef);
	vcodec_encoder_destroy(id, encoder);

	printf("Ok\n");
	return 0;
}

static int
test_dv_pal_decode_correctness(void)
{
	printf("testing video codec: DV (PAL decoder correctness) ........................ "); 
	fflush(stdout);
	printf("--\n");
	return 0;
}

static int
test_dv_pal_decode_speed(void)
{
	printf("Testing video codec: DV (PAL decoder real-time response) ................. "); 
	fflush(stdout);
	printf("--\n");
	return 0;
}

static int
test_dv_pal_packetisation(void)
{
	printf("Testing video codec: DV (PAL packetisation) .............................. "); 
	fflush(stdout);
	printf("--\n");
	return 0;
}

static int
test_dv_ntsc_encode(void)
{
	unsigned 		 id = vcodec_get_by_name("dv");
	void			*encoder;
	struct encoded_frame	*ef;
	struct encoded_data	*ed;
	struct video_frame	*vf;
	unsigned		 nn, i, dbc;
	uint16_t		 seq_prev;
	t_usec			 ts_prev;

	printf("Testing video codec: DV (NTSC encoder) ................................... "); 
	fflush(stdout);

	vf = load_dv_image_ntsc();	
	if (vf == NULL) {
		printf("FAIL\n   Cannot read test/image-pal.yuv\n");
		return 1;
	}

	encoder = vcodec_encoder_create(id, 1500);
	if (encoder == NULL) {
		printf("FAIL\n   Cannot create DV encoder\n");
		return 1;
	}

	ef = vcodec_eframe_alloc(id);
	nn = vcodec_encode(id, encoder, vf, ef, NULL);
	if (nn == 0) {
		printf("FAIL\n   Cannot encode DV frame\n");
		return 1;
	}

	/* Parse the encoded frames for correctness... */
	dbc = 0;
	seq_prev = 0xffff;
	ts_prev  = 0;
	for (i = 0; i < nn; i++ ) {
		ed = vcodec_eframe_get(id, ef, i);
		assert(ed->header     == NULL);
		assert(ed->header_len == 0);
		assert(ed->data != NULL);
		assert(ed->data_len % 80 == 0);
		assert(ed->seqno == ++seq_prev);	/* Sequence number monotonically increases     */
		dbc += ed->data_len / 80;		/* Count number of 80 byte DIF blocks          */
		if (i > 0) {				/* Timestamp must be the same on all fragments */
			assert(ed->rtp_ts == ts_prev);
		}
		ts_prev = ed->rtp_ts;
	}
	assert(dbc == 1500);	/* Number of DIF blocks in an NTSC frame */

	vcodec_eframe_free(id, ef);
	vcodec_encoder_destroy(id, encoder);
	printf("Ok\n");
	return 0;
}

static int
test_dv_ntsc_decode_correctness(void)
{
	printf("testing video codec: DV (NTSC decoder correctness) ....................... "); 
	fflush(stdout);
	printf("--\n");
	return 0;
}

static int
test_dv_ntsc_decode_speed(void)
{
	printf("Testing video codec: DV (NTSC decoder real-time response) ................ "); 
	fflush(stdout);
	printf("--\n");
	return 0;
}

static int
test_dv_ntsc_packetisation(void)
{
	printf("Testing video codec: DV (NTSC packetisation) ............................. "); 
	fflush(stdout);
	printf("--\n");
	return 0;
}

int 
test_video_codec_dv(void)
{
	int	rc = 0;

	vcodec_init();

	rc  = test_dv_capabilities();
	rc |= test_dv_pal_encode();
	rc |= test_dv_pal_decode_correctness();
	rc |= test_dv_pal_decode_speed();
	rc |= test_dv_pal_packetisation();
	rc |= test_dv_ntsc_encode();
	rc |= test_dv_ntsc_decode_correctness();
	rc |= test_dv_ntsc_decode_speed();
	rc |= test_dv_ntsc_packetisation();

	vcodec_done();

	return rc;
}

