/*
 * FILE:    test_video_codec_jpeg.c
 * AUTHORS: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Test JPEG routines
 *
 * Copyright (c) 2005 University of Glasgow
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions 
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of 
 *      Southern California Information Sciences Institute
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "memory.h"
#include "environment.h"
#include "controller.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "video_capture/testcard.h"
#include "video_codec.h"
#include "video_codec/jpeg.h"

#include "test_video_codec_jpeg.h"

/* Input testfile and size */
#define IN_FILE		"test/image-pal.yuv"
#define IN_FILE_W	(720)
#define IN_FILE_H	(576)
#define IN_FILE_SIZE	(IN_FILE_W*IN_FILE_H*2)

#define OUT_FILE	"test/image-pal_test.jpg"

#define JPEG_ITERS	200

int 
test_video_codec_jpeg (void)
{
	unsigned int		 encoder_id	= VCODEC_INVALID_ID;
	struct video_frame	*frame;
	void			*encoder_state	= NULL;
	struct encoded_frame	*eframe	= NULL;
	unsigned int		 encoded_len	= 0;
	struct encoded_data	*edata;
        struct run_env		*environ;
	int			 in_fd, out_fd;
	struct timeval		 tv_before,
				 tv_after;
	uint32_t		 usecs_elapsed;
	unsigned int		 i, j;
	
	printf ("Testing video codec: JPEG ................................................ ");
	fflush (stdout);
	
        environ = env_init ();
	vcodec_init ();
 	
	/* Check if the JPEG can encode and decode */
	if (jpeg_codec_can_encode() != TRUE) {
		printf("FAIL\n");
		printf("    JPEG codec cannot encode!.\n");
		return 1;					
	}
	
	if (jpeg_codec_can_decode () != TRUE) {
		printf("FAIL\n");
		printf("    JPEG codec cannot decode!.\n");
		return 1;					
	}
	
	
	/* Create an encoder */
	encoder_id	= vcodec_get_by_name ("jpeg");
	encoder_state	= vcodec_encoder_create (encoder_id, environ->mtu);
	if (encoder_state == NULL) {
		printf("FAIL\n");
		printf("    Cannot create encoder.\n");
		return 1;				
	}
	
	/* Open and load the input file */
	in_fd = open (IN_FILE, O_RDONLY);
	if (in_fd < 0) {
		fprintf(stderr, "Error: opening input test file %s.\n", IN_FILE);
		return 1;
	}

	frame	= video_frame_alloc (IN_FILE_SIZE);
	j	= read (in_fd, frame->data, IN_FILE_SIZE);
	if (j != IN_FILE_SIZE) {
		fprintf (stderr, "Error: reading test file %s (only read %d bytes, expecting %d).\n",
			 IN_FILE, j, IN_FILE_SIZE);
		return 1;
	}

	frame->width  = IN_FILE_W;
	frame->height = IN_FILE_H;
	
	/* Allocate space for the output */
	eframe = vcodec_eframe_alloc (encoder_id);
	if (eframe == NULL) {
		fprintf(stderr, "Error: allocating memory for encoding outgoing frames.\n");
		return 1;
	}
	assert (eframe->magic == VCODEC_JPEG_ID);

	/* Encode, measuring how much time we sped encoding a frame... */
	encoded_len = vcodec_encode (encoder_id, encoder_state, frame, eframe, NULL);		

	/* Open the output file */
	out_fd = open (OUT_FILE, O_RDWR | O_CREAT, 644);
	if (out_fd < 0) {
		fprintf(stderr, "Error: opening output test file %s.\n", OUT_FILE);
		return 1;
	}
	
	/* Dump the output */
	j = 0;
	while ((edata = vcodec_eframe_get (encoder_id, eframe, j)) != NULL) {
		write (out_fd, edata->data, edata->data_len);		
		j++;
	}
	
	
	printf("Ok\n");

	
	/*
	 * Performance measurement
	 */
	
	/* Encode, measuring how much time we sped encoding a frame... */
	printf("     Measuring encoding speed: ");
	fflush (stdout);
	
	gettimeofday (&tv_before, NULL);
	for (i = 0; i < JPEG_ITERS; i++) {
		encoded_len = vcodec_encode (encoder_id, encoder_state, frame, eframe, NULL);		
	}
	gettimeofday (&tv_after, NULL);
	usecs_elapsed = tv_usecs_since (&tv_before, &tv_after);	
	printf("%f fps\n", (1000000.0 * JPEG_ITERS) / usecs_elapsed);
	
	close (in_fd);
	close (out_fd);
	
	vcodec_eframe_free (encoder_id, eframe);		
	vcodec_encoder_destroy (encoder_id, encoder_state);
	
	vcodec_done();
	env_done(environ);
	
	return 0;
}

