/*
 * FILE:    test_video_codec_uv_yuv422.c
 * AUTHORS: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Test YUV 4:2:2 routines
 *
 * Copyright (c) 2005 University of Glasgow
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions 
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of 
 *      Southern California Information Sciences Institute
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "memory.h"
#include "environment.h"
#include "controller.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/pbuf.h"
#include "video_types.h"
#include "video_display.h"
#include "video_capture.h"
#include "video_capture/testcard.h"
#include "video_codec.h"
#include "video_codec/uv_yuv422.h"

#include "test_video_codec_uv_yuv422.h"

/* Print byte strings */
static void
dump_strs (unsigned char* o, unsigned char* n, int beg, int end) {
	int k;
	
	printf("Original: ");
	for(k=beg;k<end;k++) {
		printf("%2x ", o[k]);
	}
	printf("\n");

	printf("     New: ");
	for(k=beg;k<end;k++) {
		printf("%2x ", n[k]);
	}
	printf("\n");
}

/* Transform from encoded_frame to coded_data */
static struct coded_data*
encoded_to_coded (struct encoded_frame* encoded_frame) {
	
	struct coded_data	*ret		= NULL,
				*curr_out	= NULL,
				*prev_out	= NULL;
	
	struct encoded_data	*curr_in	= NULL;
	
	unsigned int		 pos = 0;
	unsigned int		 len;

	//assert (encoded_frame->id == VCODEC_YUV422_ID);
	
	ret = (struct coded_data*) malloc (sizeof (struct coded_data));
	
	curr_out = ret;
	prev_out = NULL;
	len = encoded_frame->len;
	
	while( (curr_in = vcodec_eframe_get (encoded_frame->id, encoded_frame, pos)) != NULL ) {
		
		if (curr_out == NULL) {
			curr_out = (struct coded_data*) 
				malloc(sizeof(struct coded_data));
		}
		
		if (prev_out != NULL)
			prev_out->next = curr_out;
		
		curr_out->prev  = prev_out;
		curr_out->next  = NULL;
		curr_out->seqno = pos;
		
		curr_out->data                = malloc (sizeof(rtp_packet));
		curr_out->data->meta.data     = malloc (curr_in->header_len + curr_in->data_len);
		curr_out->data->meta.data_len = curr_in->header_len + curr_in->data_len;
		
		memcpy(curr_out->data->meta.data,                       curr_in->header, curr_in->header_len);
		memcpy(curr_out->data->meta.data + curr_in->header_len, curr_in->data,   curr_in->data_len);
		
		prev_out = curr_out;
		curr_out = curr_out->next;
		
		pos++;
	}
	
	if(pos != len) {
		printf("vcodec_encode_get() did not go through all the frame!\n");
		ret = NULL;
	}
	
	return ret;
};


int 
test_video_codec_uv_yuv422 (void)
{
	void			*testcard_state;
	struct vidcap_type	*vt;
	unsigned int		 mtu, j;
	unsigned int		 encoder_id	= VCODEC_INVALID_ID;
	struct video_frame	*frame,
				*decoded_frame	= NULL;
	void			*encoder_state	= NULL;
	struct encoded_frame	*encoded_frame	= NULL;
	unsigned int		 encoded_len	= 0;
	void			*decoder_state	= NULL;
	struct coded_data	*coded_frame	= NULL, *tmp;
        struct run_env		*environ;

	printf("Testing video codec: YUV 4:2:2 (8 bit) ................................... "); 

        environ = env_init ();
	
	mtu = environ->mtu;

	vcodec_init ();
	
	/* Use the testcard for the testing frame */
	vt = vidcap_testcard_probe (environ);
	testcard_state = vidcap_testcard_init (vt->formats, environ->capture_args);
	if(testcard_state == NULL) {
		printf("FAIL\n");
		printf("    Cannot initialize testcard.\n");
		return 1;
	}
 	
	frame = vidcap_testcard_grab (testcard_state);

	/* Check if the YUV can encode and decode */
	if (uv_yuv422_can_encode () != TRUE) {
		printf("FAIL\n");
		printf("    YUV codec cannot encode!.\n");
		return 1;					
	}
	
	if(uv_yuv422_can_decode () != TRUE) {
		printf("FAIL\n");
		printf("    YUV codec cannot decode!.\n");
		return 1;					
	}
	

	/* Check that the encoder and decoder work */
	encoder_id = vcodec_get_by_name ("uv_yuv422");
	encoder_state = uv_yuv422_encoder_create (encoder_id, mtu);
	if(encoder_state == NULL) {
		printf("FAIL\n");
		printf("    Cannot create encoder.\n");
		return 1;				
	}
	
	encoded_frame = uv_yuv422_eframe_alloc ();
	if (encoded_frame == NULL) {
		fprintf(stderr, "Error: allocating memory for encoding outgoing frames.\n");
	}				
	assert(encoded_frame->magic == VCODEC_YUV422_ID);

	encoded_len = uv_yuv422_encode (encoder_state, frame, encoded_frame, NULL);
	
	decoder_state = uv_yuv422_decoder_create (encoder_id);
	if(decoder_state == NULL) {
		printf("FAIL\n");
		printf("    Cannot create decoder.\n");
		return 1;				
	}

	/* Convert the encoded data */
	coded_frame = encoded_to_coded (encoded_frame);
	
	/* Decode the encoded data */
	decoded_frame       = (struct video_frame*) malloc(sizeof(struct video_frame));	
	decoded_frame->data = malloc (uv_yuv422_decoder_get_framelen (decoder_state));
	if (decoded_frame == NULL || decoded_frame->data == NULL) {
		fprintf (stderr, "Error: allocating memory for decoding grabbed frames.\n");
		ctrlr_done ();
		return 1;
	}
	
	uv_yuv422_decode (decoder_state, coded_frame, decoded_frame);

	if(frame->data_len != decoded_frame->data_len) {
		printf("FAIL\n");
		printf("    Frames length doesn't match.\n");
		return 1;						
	}
	
	/* Compare "frame" with "decoded_frame" */
	for (j=0; j<frame->data_len; j++) {
		if( frame->data[j] != decoded_frame->data[j] ) {
			
			printf("    Encoded and decoded data don't match!.\n");
			printf("    Mismatch at offset %d of %d!.\n", j, frame->data_len);
			printf("    This is the sequence:\n");
			
			dump_strs(frame->data + j, decoded_frame->data + j, -3, 12);
			
			return 1;
		}
	}

	while (coded_frame != NULL) {
		free(coded_frame->data);
		tmp  = coded_frame;
		coded_frame = coded_frame->next;
		free(tmp);
	}
	
	uv_yuv422_eframe_free(encoded_frame);
	free(decoded_frame->data);
	free(decoded_frame);
	
	uv_yuv422_decoder_destroy(decoder_state);
	uv_yuv422_encoder_destroy(encoder_state);
	
	vidcap_testcard_done(testcard_state);
	
	vcodec_done();
        
	env_done(environ);
	
	printf("Ok\n");
	
	return 0;
}

