/*
 * FILE:   grabber.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2005-2006 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.7 $
 * $Date: 2006/01/31 12:20:36 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "version.h"
#include "environment.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/rtp_callback.h"
#include "rtp/pbuf.h"
#include "rtp/sbuf.h"
#include "video_capture.h"
#include "video_codec.h"
#include "pdb.h"
#include "tv.h"
#include "transmit.h"
#include "receive.h"
#include "controller.h"
#include "ui/ui.h"


#include "grabber.h"

#ifdef HAVE_LIBSDL
#include <SDL.h>
#endif

/* Defaults */
#define DEFAULT_FPS	5		/* Frames per second */ 

/* Internal state */
struct state {
	struct display		*display_device;

	struct vidcap		*capture_device;
	display_size_t		 capture_device_size;
	video_colour_mode_t	 capture_device_cm;
	
	void			*decoder;
	unsigned int		 decoder_id;	/*... of grabbed frames */
};


/*******************************************************************************
 * Utils
 ******************************************************************************/
static void 
usage (void) 
{
        printf("Usage: grabber [OPTIONS]...\n");
	printf("Options:\n");
	printf(" Display:\n");	
	printf("  -d <display_device>     Display device\n");
	printf("  -D <display_args>       Optional arguments\n");
	printf(" Capture:\n");
	printf("  -t <capture_device>     Capture device\n");
	printf("  -T <capture_args>       Optional arguments for the capture device\n");
	printf(" Other:\n");
	printf("  -f <framerate>          Frames per second\n");
}


static int
init_env (int* argc, char *argv[], struct run_env* env)
{
        int      ch;
        
	env->addr           = NULL;
	env->fps            = DEFAULT_FPS;
	env->mtu            = 1500;
	env->codec          = NULL;
	env->capture        = NULL;
	env->capture_args   = NULL;
	env->display        = NULL;
	env->display_args   = NULL;
	
        while ((ch = getopt(*argc, argv, "d:D:t:T:f:")) != -1) {
                switch (ch) {
			case 'd' :
				env->display      = optarg;
				break;
			case 'D' :
				env->display_args = optarg;
				break;
			case 't' :
				env->capture      = optarg;
				break;
			case 'T' :
				env->capture_args = optarg;
				break;
			case 'f' :
				env->fps          = atoi(optarg);
				break;
			case '?' :
				usage();
				return ERR_USAGE;
                }
        }
        *argc -= optind;
        argv += optind;
	
        if (*argc != 0 || !env->display || !env->capture) {
                usage();
                return ERR_USAGE;
        };
                
        return ERR_OK;
}


/*******************************************************************************
 * Subsystems
 ******************************************************************************/
static int
init_video_codecs (struct run_env *env)
{
	unsigned int		 nc;
	unsigned int		 i;
	const char		*codec_name	= NULL;
	
	assert (env != NULL);
	
	vcodec_init ();
	nc = vcodec_get_num_codecs();
	for (i = 0; i < nc; i++) {
		codec_name = vcodec_get_name(i);		
		if(codec_name != NULL) {
			/* Map static and "well-known" dynamic payload types */
			if (strcmp (codec_name, "uv_yuv422_8") == 0) {
				vcodec_map_payload (30, i);	/* 30 : Uncompressed YUV 4:2:2 8 bit */
			}
			
			if (strcmp (vcodec_get_name (i), "uv_yuv422_10") == 0) {
				vcodec_map_payload (32, i);      /* 32 : Uncompressed YUV 4:2:2 10 bit */
			}
			
			if (strcmp (codec_name, "dv") == 0) {
				vcodec_map_payload (123, i);	/* 123 : DV */
			}
		}
	}

	return TRUE;
}


static struct display*
init_video_display (struct state	*s,
		    display_format_t	*format,
		    struct run_env	*env)
{
	int			 found	= FALSE;
	unsigned int             i;
	display_type_t		*dt	= NULL;
	struct display		*res	= NULL;
	
	assert(env->display != NULL);	
	
	s->display_device = NULL;
	
	/* We can force a specific format */
	display_init_devices (format, env->display_args);
	
	for (i = 0; i < display_get_device_count(); i++) {
		dt = display_get_device_details (i);
		
		assert (dt->name != NULL);
		
		if (strcmp (env->display, dt->name) == 0) {
			found = TRUE;
			res = display_init (dt->id, format,
					    env->display_args, NULL);
			break;
		}
	}
	
	if (found) {
		printf ("Display device: %s.\n", dt->name);
	} else {
		fprintf (stderr, "Error: display device '%s' not found.\n", env->display);
	}
	
	display_free_devices ();
	
	return (s->display_device = res);
}

static struct vidcap*
init_video_capture (struct state *s, struct run_env *env)
{
	int			 found	= FALSE;
	unsigned int		 i;
	struct vidcap_type	*vt	= NULL;
	capture_format_t	*format	= NULL;
	struct vidcap		*res	= NULL;
	
	assert(env->capture != NULL);
	
	s->capture_device = NULL;
	
	vidcap_init_devices (env);
	for (i = 0; i < vidcap_get_device_count(); i++) {
		
		vt = vidcap_get_device_details (i);
		
		/* FIXME: we use the first capture format as the default... */
		format = vt->formats;
		
		assert (vt->name != NULL);
		
		if (strcmp (vt->name, env->capture) == 0) {
			found = TRUE;
			
			res = vidcap_init (vt->id, format, env->capture_args);
			
			s->capture_device_size = format->size;
			s->capture_device_cm   = format->colour_mode;
			
			s->decoder_id = vcodec_get_by_enc (format->encoding);
			break;
		}
	}

	if (found) {
		printf ("Capture device: %s.\n", vt->name);

		/* If needed, create the decoder for decoding grabbed frames */
		if (s->decoder_id != VCODEC_INVALID_ID) {
			
			if (vcodec_can_decode (s->decoder_id) == FALSE) {
				fprintf (stderr, "Error: decoder can not decode grabbed frames.\n");
				return FALSE;	
			}			
			
			s->decoder = vcodec_decoder_create (s->decoder_id);
			if (s->decoder == NULL) {
				fprintf (stderr, "Error: creating decoder for grabbed frames.\n");
				return FALSE;	
			}
			
			printf ("Decoder for input images: %s.\n",
				vcodec_get_name (s->decoder_id));
		}
	} else {
		fprintf (stderr, "Error: capture device '%s' not found.\n", env->capture);
	}
	
	vidcap_free_devices();
	
	return (s->capture_device = res);
}


/*******************************************************************************
 * Main
 ******************************************************************************/
uint32_t  RTT = 0;

int 
main (int argc, char *argv[])
{
	struct state		*s;
	struct video_frame	*grab_frame	= NULL,
				*dec_frame	= NULL,
				*disp_frame	= NULL;
	display_format_t	 format;
	unsigned int		 data_len;
        struct run_env		 environ;
	display_result_t	 ret;
	int			 should_exit	= FALSE;
	
        if (init_env (&argc, argv, &environ) != ERR_OK) {
                return ERR_USAGE;
	}
	
	s = malloc (sizeof (struct state));
	if (s == NULL) {
		fprintf (stderr, "Error allocating memory for state.\n");
		return 1;
	}

	/* Initialize the codecs */
	if (init_video_codecs (&environ) == FALSE) {
		fprintf (stderr, "Video codecs error.\n");
		return ERR_CODEC;
	}	
	
	/*  the capture system */		
	if (init_video_capture (s, &environ) == NULL) {
		fprintf (stderr, "Video capture error.\n");
		return ERR_CAPTURE;
	}
	
	/* ... and create a matching display */
	format.fps		= environ.fps;
	format.colour_mode	= s->capture_device_cm;
	format.size		= s->capture_device_size;
	
	if (ui_init (argc, argv, environ.display) != 0) {
		return ERR_UI;
	}

	if (init_video_display (s, &format, &environ) == NULL) {
		fprintf (stderr, "Video display error.\n");
		return ERR_DISPLAY;
	}

	/* Create the frame for decoding grabbed images */
	data_len  = vcodec_decoder_get_framelen (s->decoder_id, s->decoder);
	dec_frame = video_frame_alloc (data_len);
	if (dec_frame == NULL) {
		fprintf (stderr, "Error: allocating memory for decoding grabbed frames.\n");
		return ERR_CODEC;
	} else {
		display_size_t	size;
		
		size = vcodec_decoder_get_size (s->decoder_id, s->decoder);
		
		dec_frame->width       = display_get_width (size);
		dec_frame->height      = display_get_height (size);
		dec_frame->colour_mode = vcodec_decoder_get_cm (s->decoder_id, s->decoder);
	}
	
	/* Grab-decode-display loop */	
	while (!should_exit) {
		/* Capture, decode (if needed), encode and transmit a frame... */
		grab_frame = vidcap_grab (s->capture_device);
		if (grab_frame != NULL) {
			/* Get a display frame */
			disp_frame = display_get_frame (s->display_device, s->capture_device_size, s->capture_device_cm);
			if (disp_frame == NULL) {
				return DISPLAY_ERR;	/* No working display */
			}

			if (s->decoder != NULL) {
				/* We need to decode the grabbed frame */
				vcodec_decode_raw (s->decoder_id, s->decoder, grab_frame, dec_frame);
				if (dec_frame != NULL) {
					video_frame_copy (disp_frame, dec_frame);
				}
			} else {
				video_frame_copy (disp_frame, grab_frame);
			}
			ret = display_put_frame (s->display_device, disp_frame);
			if (ret == DISPLAY_EXIT_REQ) {
				printf ("Exit request.\n");
				should_exit = TRUE;
			}

			free (grab_frame);
		} else {
			printf ("No capture frame... Exiting.\n");
			should_exit = TRUE;
		}
	}

	return ERR_OK;
}

