/*
 * FILE:   rtplog.c
 * AUTHOR: Colin Perkins <csp@csperkins.org>
 *
 * Copyright (c) 2005 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.1 $
 * $Date: 2005/11/08 14:50:15 $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <netinet/in.h>

struct rtphdr {
	uint16_t	flags;
	uint16_t	seq;
	uint32_t	ts;
	uint32_t	ssrc;
	struct timeval	atime;
};

int
main(int argc, char *argv[])
{
	int			 port, fd, len;
	unsigned		 num_packets, buflen, i;
	struct rtphdr		*buffer;
	char			*filename;
	struct sockaddr_in       s_in;
	char			 sockbuf[1500];
	FILE			*outf;
	int			 reuse = 1;

	if (argc != 4) {
		printf("Usage: rtplog <num_packets> <port> <filename>\n");
		return 1;
	}

	num_packets = atoi(argv[1]);
	port        = atoi(argv[2]);
	filename    = argv[3];

	/*******************************************************************************/
	/* Allocate and lock the capture buffer....                                    */
	/*******************************************************************************/
	printf("rtplog: allocating buffer... "); fflush(stdout);

	buflen = num_packets * sizeof(struct rtphdr);
	buffer = malloc(buflen);
	if (buffer == NULL) {
		printf("fail\n");
		return 2;
	}

	if (mlock(buffer, buflen) != 0) {
		printf("fail\n");
		return 3;
	}
	if (mlock(sockbuf, 1500) != 0) {
		printf("fail\n");
		return 4;
	}

	printf("done\n");
	/*******************************************************************************/
	/* Capture the data...                                                         */
	/*******************************************************************************/
	printf("rtplog: capturing traffic... "); fflush(stdout);

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		printf("fail\n");
		perror("rtplog: unable to create socket");
		return 5;
	}

        if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (char *) &reuse, sizeof(reuse)) != 0) {
		printf("rtplog: unable to reuse port");
		return 6;
	}

	s_in.sin_family      = AF_INET;
	s_in.sin_addr.s_addr = INADDR_ANY;
	s_in.sin_port        = htons(port);
	if (bind(fd, (struct sockaddr *) &s_in, sizeof(s_in)) != 0) {
		printf("fail\n");
		perror("rtplog: unable to bind socket");
		return 7;
	}

	i = 0;
	while (i < num_packets) {
		len = recv(fd, sockbuf, 1500, 0);
		if (len < 0) {
			printf("fail\n");
			perror("rtplog: unable to receive data");
			break;
		}
		if (len > 0) {
			memcpy(buffer + i, sockbuf, sizeof(struct rtphdr));
			gettimeofday(&buffer[i].atime, NULL);
			i++;
		}
	}

	close(fd);

	printf("done\n");
	/*******************************************************************************/
	/* Write to file...                                                            */
	/*******************************************************************************/
	printf("rtplog: writing data file... "); fflush(stdout);

	outf = fopen(filename, "w");
	if (outf == NULL) {
		printf("fail\n");
		perror("rtplog: unable to create output file");
		return 8;
	}

	for (i = 0; i < num_packets; i++) {
		fprintf(outf, "%u.%06u rtp ", (unsigned) buffer[i].atime.tv_sec, 
		                              (unsigned) buffer[i].atime.tv_usec);
		fprintf(outf, "-v %u ",  (ntohs(buffer[i].flags) >> 14) & 0x0003);
		fprintf(outf, "-p %u ",  (ntohs(buffer[i].flags) >> 13) & 0x0001);
		fprintf(outf, "-x %u ",  (ntohs(buffer[i].flags) >> 12) & 0x0001);
		fprintf(outf, "-cc %u ", (ntohs(buffer[i].flags) >>  8) & 0x000f);
		fprintf(outf, "-m %u ",  (ntohs(buffer[i].flags) >>  7) & 0x0001);
		fprintf(outf, "-pt %u ", (ntohs(buffer[i].flags) >>  0) & 0x007f);
		fprintf(outf, "-ssrc 0x%" PRIx32 " ",  ntohl(buffer[i].ssrc));
		fprintf(outf, "-seq %"    PRIu16 " ",  ntohs(buffer[i].seq));
		fprintf(outf, "-ts %"     PRIu32 "\n", ntohl(buffer[i].ts));
	}

	fclose(outf);

	printf("done\n");
	/*******************************************************************************/
	/* Cleanup...                                                                  */
	/*******************************************************************************/
	munlock(buffer, buflen);
	munlock(sockbuf, 1500);
	return 0;
}

