/*
 * FILE:   psnr.c
 * AUTHOR: Alvaro Saurin <saurin@dcs.gla.ac.uk>
 *
 * Copyright (c) 2006 University of Glasgow
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.3 $
 * $Date: 2006/05/12 15:05:00 $
 *
 */


#include "config.h"
#include "config_unix.h"
#include "debug.h"
#include "version.h"
#include "environment.h"

#include "video_types.h"
#include "video_display.h"
#include "rtp/cc.h"
#include "rtp/rtp.h"
#include "rtp/rtp_callback.h"
#include "rtp/pbuf.h"
#include "rtp/sbuf.h"
#include "video_capture.h"
#include "video_capture/testfile.h"
#include "video_codec.h"
#include "pdb.h"
#include "tv.h"
#include "transmit.h"
#include "receive.h"
#include "controller.h"
#include "ui/ui.h"

#include "psnr.h"

/* Constants */
#define DEFAULT_MAXFRAMENO	INT32_MAX

/* Internal state */
struct state {
	
	display_size_t		 size;
	video_colour_mode_t	 cm;
	
	unsigned long int	 max_frameno;

	char			*filename1;
	char			*filename2;
	
	struct vidcap_type	*vt1,
				*vt2;
	
	struct vidcap		*capture1,
				*capture2;
	
	void			*decoder1,
				*decoder2;

	unsigned int		 decoder1_id,
				 decoder2_id;
	
	int			 noloop_src;
};


/*******************************************************************************
 * Utils
 ******************************************************************************/
static void 
usage (void) 
{
        printf ("Usage: psnr [OPTIONS] <source> <transmited>\n");
	printf ("where the <source> file is supposed to be longer than the <transmited>.\n");
	printf ("Options:\n");
	printf ("  -F <frames>    Number of frames to compare.\n");
	printf ("  -s             Finish when the <source> ends.\n");	
}

static int
init_state (int* argc, char *argv[], struct state* s)
{
        int		ch;

	assert (s != NULL);
	
	s->max_frameno	= DEFAULT_MAXFRAMENO;
	s->noloop_src	= FALSE;
	
        while ((ch = getopt(*argc, argv, "sF:")) != -1) {
                switch (ch) {
			case 'F' :
				s->max_frameno	= atoi (optarg);
				break;
			case 's' :
				s->noloop_src = TRUE;
				break;
			case '?' :
				usage ();
				return ERR_USAGE;
                }
        }

        *argc -= optind;
        argv += optind;
	
        if (*argc != 2) {
                usage ();
                return ERR_USAGE;
        };

	s->filename1 = argv[0];
	s->filename2 = argv[1];
	
        return ERR_OK;
}

/*******************************************************************************
 * Subsystems
 ******************************************************************************/
static int
init_video_codecs (void)
{
	unsigned int		 nc;
	unsigned int		 i;
	const char		*codec_name	= NULL;

	vcodec_init ();
	nc = vcodec_get_num_codecs();
	for (i = 0; i < nc; i++) {
		codec_name = vcodec_get_name (i);		
		if(codec_name != NULL) {
			/* Map static and "well-known" dynamic payload types */
			if (strcmp (codec_name, "uv_yuv422") == 0) {
				vcodec_map_payload (30, i);	/* 30 : Uncompressed YUV 4:2:2 8 bit */
			}
			
			if (strcmp (codec_name, "dv") == 0) {
				vcodec_map_payload (123, i);	/* 123 : DV */
			}
		}
	}
	
	return TRUE;
}

/* Initilize the video capture system */
static struct vidcap*
init_video_capture (struct state *s)
{
	char			 capture_str1 [1024],
				 capture_str2 [1024];
	struct run_env		*env;
	display_size_t		 size1,
				 size2;
	
	s->capture1 = NULL;
	s->capture2 = NULL;
	
	/*
	 * Make the string arguments, adding any aditional parameters. For example,
	 * set the FPS to a high value, in order to avoid sleeping between
	 * frames.
	 */
	if (s->noloop_src) {
		sprintf (capture_str1, "%s@noloop@nosleep@200", s->filename1);		
	} else {
		sprintf (capture_str1, "%s@nosleep@200", s->filename1);
	}
	
	sprintf (capture_str2, "%s@noloop@nosleep@200", s->filename2);
	
	env = env_init ();
	
	/* Initialize the capture devices */
	env->capture_args	= capture_str1;
	s->vt1			= vidcap_testfile_probe (env);
	size1			= s->vt1->formats[0].size;
	
	s->capture1		= vidcap_testfile_init (s->vt1->formats, capture_str1);
	s->decoder1_id		= vcodec_get_by_enc (s->vt1->formats[0].encoding);
	
	/* If needed, create the decoder for decoding grabbed frames */
	if (s->decoder1_id != VCODEC_INVALID_ID) {
	
		if (vcodec_can_decode (s->decoder1_id) == FALSE) {
			fprintf (stderr, "Error: decoder can not decode grabbed frames.\n");
			return FALSE;	
		}			
		
		s->decoder1 = vcodec_decoder_create (s->decoder1_id);
		if (s->decoder1 == NULL) {
			fprintf (stderr, "Error: creating decoder for grabbed frames.\n");
			return FALSE;	
		}
	}

	/* 2nd file */
	env->capture_args	= capture_str2;
	s->vt2			= vidcap_testfile_probe (env);
	size2			= s->vt2->formats[0].size;

	s->capture2		= vidcap_testfile_init (s->vt2->formats, capture_str2);
	s->decoder2_id		= vcodec_get_by_enc (s->vt2->formats[0].encoding);

	/* If needed, create the decoder for decoding grabbed frames */
	if (s->decoder2_id != VCODEC_INVALID_ID) {

		if (vcodec_can_decode (s->decoder2_id) == FALSE) {
			fprintf (stderr, "Error: decoder can not decode grabbed frames.\n");
			return FALSE;	
		}			
		
		s->decoder2 = vcodec_decoder_create (s->decoder2_id);
		if (s->decoder2 == NULL) {
			fprintf (stderr, "Error: creating decoder for grabbed frames.\n");
			return FALSE;	
		}
	}
	
	vidcap_free_devices();

	s->size = size1;
	
	env_done (env);
	
	return s->capture1;
}

/*******************************************************************************
* PSNR
******************************************************************************/
static inline unsigned long long int
sqr (unsigned long long int x)
{
	return (x*x);
}

static inline double
psnr_equ (unsigned long long int ssd, unsigned long long int pixels) {
	double mse = (double)ssd / (double)pixels;
	double psnr = 10.0 * log10 (sqr(255.0)/mse);
	
	return psnr;
}

static double
calc_psnr (struct video_frame *frame1, struct video_frame *frame2,
	   double *psnr, double *psnr_y, double *psnr_u, double *psnr_v,
	   unsigned long long int *total_ssd_y,
	   unsigned long long int *total_ssd_u, unsigned long long int *total_ssd_v,
	   unsigned long long int *ssd,         unsigned long long int *ssd_y,
	   unsigned long long int *ssd_u,       unsigned long long int *ssd_v)
{
	unsigned char		*image1,
				*image2;
	unsigned int		 width,
				 height;

	unsigned long long int	 local_ssd_y	= 0,
				 local_ssd_u	= 0,
				 local_ssd_v	= 0;
	
	int			 image1_u0, image1_y0, image1_v0, image1_y1,  
				 image2_u0, image2_y0, image2_v0, image2_y1;

	unsigned long		 s;
	unsigned int		 num_pixels	= 0;

	/* Frame sizes must match */
	assert (frame1->width > 0 && frame2->height > 0);

	assert (frame1->width  == frame2->width);
	assert (frame1->height == frame2->height); 
	
	/* Take the size from the frame1 */
	width	= frame1->width;
	height	= frame1->height;

	image1	= frame1->data;
	image2	= frame2->data;

	assert (image1 != NULL && image2 != NULL);

	assert (frame1->data_len >= width * height * 2);
	assert (frame2->data_len >= width * height * 2);

	/* Compute differences */
	while (num_pixels < width * height) {
		/* packed YUV 422 = U[i] Y[i] V[i] Y[i+1]  */

		image1_u0	= *image1++;
		image2_u0	= *image2++;
		s		= sqr (abs (image1_u0 - image2_u0));
		local_ssd_u	+= s;
		assert (s == 0 || local_ssd_u > 0);

		image1_y0	= *image1++;
		image2_y0	= *image2++;
		s		= sqr (abs (image1_y0 - image2_y0));
		local_ssd_y	+= s;
		assert (s == 0 || local_ssd_y > 0);
		
		image1_v0	= *image1++;
		image2_v0	= *image2++;
		s		= sqr (abs (image1_v0 - image2_v0));
		local_ssd_v	+= s;		
		assert (s == 0 || local_ssd_v > 0);
		
		image1_y1	= *image1++;
		image2_y1	= *image2++;
		s		= sqr (abs (image1_y1 - image2_y1));
		local_ssd_y	+= s;
		assert (s == 0 || local_ssd_y > 0);
		
		num_pixels += 2;
	}
		
	*ssd_y		= local_ssd_y;
	*ssd_u		= local_ssd_u;
	*ssd_v		= local_ssd_v;
	
	*ssd		= local_ssd_y + local_ssd_u + local_ssd_v;
	
	*total_ssd_y	+= local_ssd_y;
	*total_ssd_u	+= local_ssd_u;
	*total_ssd_v	+= local_ssd_v;
	
	*psnr		= psnr_equ (*ssd,        width * height * 2);
	*psnr_y		= psnr_equ (local_ssd_y, width * height);
	*psnr_u		= psnr_equ (local_ssd_u, (width * height) / 2);
	*psnr_v		= psnr_equ (local_ssd_v, (width * height) / 2);

	return *psnr;
}

/*******************************************************************************
 * Main
 ******************************************************************************/
uint32_t  RTT = 0;

int 
main (int argc, char *argv[])
{
	struct state		*s		= NULL;

	struct video_frame	*grab_frame1	= NULL,
				*grab_frame2	= NULL,
				*dec_frame1	= NULL,
				*dec_frame2	= NULL;
	
	display_size_t		 size;
	video_colour_mode_t	 cm;
	
	unsigned int		 data_len;
	
	unsigned int		 frameno	= 0;
	
	double			 avg_psnr	= 0,
				 avg_psnr_y	= 0,
				 avg_psnr_u	= 0,
				 avg_psnr_v	= 0;

	double			 psnr		= 0,
				 psnr_y		= 0,
				 psnr_u		= 0,
				 psnr_v		= 0;
	
	unsigned long long int	 total_ssd	= 0,
				 total_ssd_y	= 0,
				 total_ssd_u	= 0,
				 total_ssd_v	= 0,
				 ssd		= 0,
				 ssd_y		= 0,
				 ssd_u		= 0,
				 ssd_v		= 0;
	
	unsigned long long int	 pixels;
	
	unsigned int		 should_exit	= FALSE;
	
	
	s = malloc (sizeof (struct state));
	if (s == NULL) {
		fprintf (stderr, "Error allocating memory for state.\n");
		return 1;
	}

        if (init_state (&argc, argv, s) != ERR_OK) {
                return ERR_USAGE;
	}

	/* Some info */
	printf ("# Input file 1: %s\n", s->filename1);
	printf ("# Input file 2: %s\n", s->filename2);
	printf ("#       Frames: %ld\n", s->max_frameno);
	
	/* Initialize the codecs */
	if (init_video_codecs () == FALSE) {
		fprintf (stderr, "Video codecs error.\n");
		return ERR_CODEC;
	}	
	
	/* ... and the capture system */		
	if (init_video_capture (s) == NULL) {
		fprintf (stderr, "Video capture error.\n");
		return ERR_CAPTURE;
	}

	printf ("#         Size: %s\n", display_get_name (s->size));

	/* Create the frames for decoding grabbed images */
	data_len  = vcodec_decoder_get_framelen (s->decoder1_id, s->decoder1);
	dec_frame1 = video_frame_alloc (data_len);
	if (dec_frame1 == NULL) {
		fprintf (stderr, "Error: allocating memory for decoding grabbed frames.\n");
		return ERR_CODEC;
	} else {
		size	= vcodec_decoder_get_size (s->decoder1_id, s->decoder1);
		cm	= vcodec_decoder_get_cm (s->decoder1_id, s->decoder1);
		
		dec_frame1->width       = display_get_width (size);
		dec_frame1->height      = display_get_height (size);
		dec_frame1->colour_mode = cm;
	}

	/* ... and the same for the second movie */
	data_len  = vcodec_decoder_get_framelen (s->decoder2_id, s->decoder2);
	dec_frame2 = video_frame_alloc (data_len);
	if (dec_frame2 == NULL) {
		fprintf (stderr, "Error: allocating memory for decoding grabbed frames.\n");
		return ERR_CODEC;
	} else {		
		size	= vcodec_decoder_get_size (s->decoder2_id, s->decoder2);
		cm	= vcodec_decoder_get_cm (s->decoder2_id, s->decoder2);
		
		dec_frame2->width       = display_get_width (size);
		dec_frame2->height      = display_get_height (size);
		dec_frame2->colour_mode = cm;
	}

	printf ("#\n");
	printf ("# %8s %10s %10s %10s  %10s   %7s  %7s  %7s  %7s\n",
		"frameno", "ssd_y", "ssd_u", "ssd_v", "ssd",
		"psnr_y", "psnr_u", "psnr_v", "psnr");

	/* Grab-decode loop */	
	while (frameno < s->max_frameno && !should_exit) {

		/* Capture, decode (if needed) and calculate... */
		grab_frame1 = vidcap_testfile_grab (s->capture1);
		grab_frame2 = vidcap_testfile_grab (s->capture2);

		if (grab_frame1 != NULL && grab_frame2 != NULL) {
			
			/* Maybe we need to decode the grabbed frames */
			if (s->decoder1 != NULL) {
				vcodec_decode_raw (s->decoder1_id, s->decoder1, grab_frame1, dec_frame1);
			} else {
				dec_frame1 = grab_frame1;
			}

			if (s->decoder2 != NULL) {
				vcodec_decode_raw (s->decoder2_id, s->decoder2, grab_frame2, dec_frame2);
			} else {
				dec_frame2 = grab_frame2;
			}

			/* Check that we can compare frames... */
			if (dec_frame1->width != dec_frame2->width ||
			    dec_frame1->height != dec_frame2->height) {
				fprintf (stderr, "Error: frame sizes does not match: %dx%d != %dx%d.\n",
					 dec_frame1->width, dec_frame1->height,
					 dec_frame2->width, dec_frame2->height);
				
				should_exit = TRUE;
			} else {
				/* Calculate the PSNR */
				calc_psnr (dec_frame1, dec_frame2,
					   &psnr, &psnr_y, &psnr_u, &psnr_v,
					   &total_ssd_y, &total_ssd_u, &total_ssd_v,
					   &ssd, &ssd_y, &ssd_u, &ssd_v);

				frameno++;

				printf ("      %4d %10lli %10lli %10lli %10lli   %7.4f  %7.4f  %7.4f  %7.4f\n",
				       frameno,
				       ssd_y, ssd_u, ssd_v, ssd,
				       psnr_y, psnr_u, psnr_v, psnr);			
				
				avg_psnr    += psnr;
				
				avg_psnr_y  += psnr_y;
				avg_psnr_u  += psnr_u;
				avg_psnr_v  += psnr_v;			
			}			
			
			free (grab_frame1);
			free (grab_frame2);
		} else {
			should_exit = TRUE;
		}
	}
	
	total_ssd = total_ssd_y + total_ssd_u + total_ssd_v;
	
	avg_psnr_y /= frameno;
	avg_psnr_u /= frameno;
	avg_psnr_v /= frameno;
	avg_psnr   /= frameno;
	
	pixels = (unsigned long long int) dec_frame1->width * dec_frame1->height * frameno;
	
	printf ("# Total: frames = %i\n"
		"#     psnr_y = %7.4f -     psnr_u = %7.4f       psnr_v = %7.4f - psnr     = %7.4f\n"
		"# avg_psnr_y = %7.4f - avg_psnr_u = %7.4f - avg_psnr_v = %7.4f - avg_psnr = %7.4f\n",
		frameno,
		psnr_equ (total_ssd,   pixels*2),
		psnr_equ (total_ssd_y, pixels),
		psnr_equ (total_ssd_u, pixels/2),
		psnr_equ (total_ssd_v, pixels/2),
		avg_psnr_y,
		avg_psnr_u,
		avg_psnr_v,
		avg_psnr);	
	
	return ERR_OK;
}

